/** \file Numegration.h
  * \brief Integration Rules
  * \author Martin Bernreuther <martin@bernreuther.net>
  * \version 0.1
  * \date 10.8.2000
*/

/** \mainpage Numegration class library
  *
  * \section intro Introduction
  *
  * This class library does numerical integration of a function
  * given at the integration points. The integral is approximated
  * with a sum of weighted function values. There are several
  * quadrature rules implemented:
  * - Trapezoidal
  * - Simpson
  * - Newton
  * - higher Newton-Cotes formulaes
  * - Gauss-Legendre
  * - Gauss-Lobatto
  * - Gauss-Chebyshev
  *
  * The integration point locations and weights are set for
  * given one and two dimensional rectangular regions. These
  * values can be retrieved.
  * For a given function or function object the result is
  * calculated directly.
*/

#ifndef NUMEGRATION_H
#define NUMEGRATION_H

#include <valarray>
#include <vector>
using namespace std;

/// Namespace
namespace Numegration
{
	/// Enum for integration rules
	/**
	 * - Unspecified
	 *
	 * -# NewtonCotes1:	Newton-Cotes Rule of Order 1 (=Trapezoid)\n
	 *			n>=2
	 * -# NewtonCotes2:	Newton-Cotes Rule of Order 2 (=Simpson)\n
	 *			n>=3, (n-3)%2=0 (n odd)
	 * -# NewtonCotes3:	Newton-Cotes Rule of Order 3 (=Newton)\n
	 *			n>=4, (n-4)%3=0
	 * -# NewtonCotes4:	Newton-Cotes Rule of Order 4\n
	 *			n>=5, (n-5)%4=0
	 * -# NewtonCotes5:	Newton-Cotes Rule of Order 5\n
	 *			n>=6, (n-6)%5=0
	 * -# Trapezoid:	Newton-Cotes Rule of Order 1\n
	 *			n>=2, n odd
	 * -# Simpson:	Newton-Cotes Rule of Order 2\n
	 *			n>=3, (n-1)%2=0
	 * -# Newton:	Newton-Cotes Rule of Order 3\n
	 *			n>=4, (n-1)%3=0
	 * -# GaussLegendre:	Gauss-Legendre Rule (Legendre function roots)\n
	 *                      Legendre functions: orthogonal functions\n
	 *			1<=n<=100
	 * -# GaussLobatto:	Gauss-Lobatto Rule (Lobatto function roots)\n
	 *                      Lobatto functions: orthogonal with weight function (1-x^2)\n
	 *                      -> special case of Gegenbauer (ultraspherical) polynomials\n
	 *                         or Jacobi polynomials\n
	 *			2<=n<100
	 * -# GaussChebyshev:	Gauss-Chebyshev Rule (Chebyshev function roots)
	 */
	enum Rule {Unspecified=0
			  , NewtonCotes1, NewtonCotes2, NewtonCotes3, NewtonCotes4, NewtonCotes5
			  , Trapezoid, Simpson, Newton
			  , GaussLegendre, GaussLobatto, GaussChebyshev};

	/// Exception: Error in integration order
	class ErrN
	{
	private:
		unsigned int m_n;
	public:
		ErrN(unsigned int n) { m_n=n; }
		unsigned int Getn() { return m_n; }
	};

	/// Exception: Error in integration rule
	class ErrRule
	{
	private:
		Rule m_rule;
	public:
		ErrRule(Rule r) { m_rule=r; }
		unsigned int GetRule() { return m_rule; }
	};


	/// Class for one dimensional functions
	/**  The ()-operator with one double argument
	 *  has to be overloaded for a derived class
	 *  The return value is the ordinate computed for
	 *  the abscissa-argument.
	 */
	class Function1D
	{
	public:
		//virtual double Compute(double x)=0;
		virtual double operator()(double x)=0;
	};

	/// One dimensional integration points
	class Points1D
	{
		private:
			unsigned int m_n;
			Rule	m_rule;
			double m_lbound,m_ubound;
			valarray<double> m_abscissas,m_weights;

			void Initialize();
			void FitOrd2Bounds();

		public:
			/// Constuctor
			Points1D(unsigned int n, Rule r, double lbound=-1., double ubound=1.);
	
			/// Set number of integration points (integration order)
			void Setn(unsigned int n);
			/// Set integration rule
			void SetRule(Rule r);
			/// Set integration interval
			void SetBounds(double lbound, double ubound);
			/// Set number of integration points, integration rule & bounds
			void SetParameters(unsigned int n, Rule r, double lbound=-1., double ubound=1.);

			/// Get number of integration points (integration order)
			unsigned int Getn() const { return m_n; }
			/// Get integration rule
			Rule GetRule() const { return m_rule; }
			/// Get lower integration bound
			double GetLBound() const { return m_lbound; }
			/// Get upper integration bound
			double GetUBound() const { return m_ubound; }

			/// Get integration points abscissas
			const valarray<double>& GetAbscissas() const { return m_abscissas; }
			/// Get weights of integration points
			const valarray<double>& GetWeights() const { return m_weights; }

			/// Integrate over given function values at integration points
			double Integrate(const valarray<double>& funcval) const;
			/// Integrate with the given function over the interval
			double Integrate(double (*pf)(double)) const;
			/// Integrate with the given function object over the interval
			double Integrate(Function1D& f) const;
	};


	/// Class for two dimensional functions
	/** The ()-operator with two double arguments
	 *  has to be overloaded for a derived class.
	 *  The return value is the ordinate computed for
	 *  the abscissa-arguments.
	 */
	class Function2D
	{
	public:
		//virtual double Compute(double x1,double x2)=0;
		virtual double operator()(double x1, double x2)=0;
	};

	/// Two dimensional integration points
	class Points2D
	{
		private:
			unsigned int m_n1,m_n2;
			Rule	m_rule1,m_rule2;
			double m_lbound1,m_ubound1,m_lbound2,m_ubound2;
			valarray<double> m_abscissas1,m_abscissas2;
			vector<valarray<double> > m_weights;

			void Initialize();

		public:
			/// Constuctor
			Points2D(unsigned int n1, Rule r1, unsigned int n2=0, Rule r2=Numegration::Unspecified, double lbound1=-1., double ubound1=1., double lbound2=-1., double ubound2=1.);

			/// Set number of integration points (integration order) in both directions
			void Setn(unsigned int n1,unsigned int n2=0);
			/// Set integration rules for both directions
			void SetRule(Rule r1,Rule r2=Unspecified);
			/// Set integration region 
			void SetBounds(double lbound1, double ubound1, double lbound2, double ubound2);
			/// Set number of integration points, integration rule & bounds
			void SetParameters(unsigned int n1, Rule r1, unsigned int n2=0, Rule r2=Numegration::Unspecified, double lbound1=-1., double ubound1=1., double lbound2=-1., double ubound2=1.);

			/// Get number of integration points in the first direction
			unsigned int Getn1() const { return m_n1; }
			/// Get number of integration points in the second direction
			unsigned int Getn2() const { return m_n2; }
			/// Get integration rule for the first direction
			Rule GetRule1() const { return m_rule1; }
			/// Get integration rule for the second direction
			Rule GetRule2() const { return m_rule2; }
			/// Get lower integration bound in the first direction
			double GetLBound1() const { return m_lbound1; }
			/// Get upper integration bound in the first direction
			double GetUBound1() const { return m_ubound1; }
			/// Get lower integration bound in the second direction
			double GetLBound2() const { return m_lbound2; }
			/// Get upper integration bound in the second direction
			double GetUBound2() const { return m_ubound2; }

			/// Get integration points abscissas in the first direction
			const valarray<double>& GetAbscissas1() const { return m_abscissas1; }
			/// Get integration points abscissas in the second direction
			const valarray<double>& GetAbscissas2() const { return m_abscissas2; }
			/// Get weights of integration points
			const vector<valarray<double> >& GetWeights() const { return m_weights; }

			/// Integrate over given function values at integration points
			double Integrate(const vector<valarray<double> >& funcval) const;
			/// Integrate with the given function over the region
			double Integrate(double (*pf)(double,double)) const;
			/// Integrate with the given function object over the region
			double Integrate(Function2D& f) const;

			/// Integrate a product f1(x)*f2(y) over given function values at integration points
			double IntegrateProduct(const valarray<double>& funcval1,const valarray<double>& funcval2) const;
			/// Integrate a product f1(x)*f2(y) with the given functions over the region
			double IntegrateProduct(double (*pf1)(double),double (*pf2)(double)) const;
			/// Integrate a product  f1(x)*f2(y) with the given function objects over the region
			double IntegrateProduct(Function1D& f1,Function1D& f2) const;
	};

}

#endif
