/** \file Numegration.cpp
  * \brief Integration Rules
  * \author Martin Bernreuther <martin@bernreuther.net>
  * \version 0.1
  * \date 10.8.2000
  *
  * - GaussLegendre Integration Points:
  *   - S. Holzer, M. Klein
  *   - see Table A.4.1 of "B. Szabo, I. Babuska: Finite Element Analysis"
  *   - see Tables 1 of "A.H.Stroud, Don Secrest: Gaussian Quadrature Formulas"
  * - GaussLobatto Integration Points:
  *   - S. Holzer
  *   - see Table A.4.2 of "B. Szabo, I. Babuska: Finite Element Analysis"
  *   - see Tables 11 of "A.H.Stroud, Don Secrest: Gaussian Quadrature Formulas"
  *
  * - values calculated with Maple
*/

#include "Numegration.h"
#include <cmath>
//using namespace std;

// Points1D ############################################################

/**
 *	\param	n	Number of integration points	(unsigned int)
 *	\param	r	Integration rule	(Numegration::Rule)
 *	\param	lbound	Lower bound of integration interval	(double)\n
 *		lbound not specified:	lbound is set to -1
 *	\param	ubound	Upper bound of integration interval	(double)\n
 *		ubound not specified:	ubound is set to 1
 */
Numegration::Points1D::Points1D(unsigned int n, Rule r, double lbound, double ubound)
{
	SetParameters(n,r,lbound,ubound);
}

/**
 *	\param	n	Number of integration points	(unsigned int)
 */
void Numegration::Points1D::Setn(unsigned int n)
{
	if(n && n!=m_n)
	{
		m_n=n;
		Initialize();
	}
}

/**
 *	\param	r	Integration rule	(Numegration::Rule)
 */
void Numegration::Points1D::SetRule(Numegration::Rule r)
{
	if(r!=m_rule)
	{
		m_rule=r;
		Initialize();
	}
}

/**
 *	\param	lbound	Lower bound of integration interval	(double)
 *	\param	ubound	Upper bound of integration interval	(double)
 */
void Numegration::Points1D::SetBounds(double lbound, double ubound)
{
	if(lbound!=m_lbound || ubound!=m_ubound)
	{
		m_lbound=lbound;
		m_ubound=ubound;
		Initialize();
	}
}

/**
 *	\param	n	Number of integration points	(unsigned int)
 *	\param	r	Integration rule	(Numegration::Rule)
 *	\param	lbound	Lower bound of integration interval	(double)\n
 *		lbound not specified:	lbound is set to -1
 *	\param	ubound	Upper bound of integration interval	(double)\n
 *		ubound not specified:	ubound is set to 1
 */
void Numegration::Points1D::SetParameters(unsigned int n, Rule r, double lbound, double ubound)
{
	m_n=n;

	m_rule=r;

	m_lbound=lbound;
	m_ubound=ubound;

	Initialize();
}


/**
 *	\param	funcval	Ordinates of the function at integration points	(valarray<double>)
 */
double Numegration::Points1D::Integrate(const valarray<double>& funcval) const
{
	unsigned int size=m_weights.size();
	double result=0.;
	if(size && funcval.size()==size)
	{
		for(unsigned int i=0;i<size;i++)
			result+=m_weights[i]*funcval[i];
	}
	return result;
}

/**
 *	\param	pf	pointer to a C-function	(double* (double))
 */
double Numegration::Points1D::Integrate(double (*pf)(double)) const
{
	double result=0.;
	double x;
	for(unsigned int i=0;i<m_n;i++)
	{
		x=m_abscissas[i];
		result+=m_weights[i]*(*pf)(x);
	}
	return result;
}

/**
 *	\param	f	function object 	(Function1D)
 */
double Numegration::Points1D::Integrate(Function1D& f) const
{
	double result=0.;
	double x;
	for(unsigned int i=0;i<m_n;i++)
	{
		x=m_abscissas[i];
		result+=m_weights[i]*f(x);
	}
	return result;
}


// private -------------------------------------------------------------			

void Numegration::Points1D::FitOrd2Bounds()
/// fit integration abscissas to bounds
{
	// assumes m_abscissas are set for [-1,1]
	unsigned int i;
	double d=m_ubound-m_lbound;
	for (i=0;i<m_abscissas.size();i++)
	{
		m_abscissas[i]=.5*(m_abscissas[i]+1.)*d+m_lbound;
		m_weights[i]*=.5*d;
	}
}


void Numegration::Points1D::Initialize()
/// initialize internal data
{
	unsigned int i,j;

	m_abscissas.resize(m_n);
	m_weights.resize(m_n);

	if(!m_n)
	{
		throw ErrN(m_n);
		return;
	}

	switch(m_rule)
	{
	// Trapez **********************************************************
	case NewtonCotes1:
	case Trapezoid:
	{
		if(m_n<2)
		{
			m_n=2;
			//m_n=0
			m_abscissas.resize(m_n);
			m_weights.resize(m_n);
			throw ErrN(m_n);
			// break;
		}
		double h=2./(m_n-1);
		for(i=0;i<m_n;i++)
		{
			m_abscissas[i]=i*h-1.;
			m_weights[i]=h;
		}
		m_weights[0]=m_weights[m_n-1]=.5*h;
		break;
	}
	// Simpson *********************************************************
	case NewtonCotes2:
	case Simpson:
	{
		if(m_n<3 || m_n%2==0)
		{
			if(m_n<3) m_n=3;
			if(m_n%2==0) m_n++;
			//m_n=0
			m_abscissas.resize(m_n);
			m_weights.resize(m_n);
			throw ErrN(m_n);
			// break;
		}
		double h=2./(m_n-1);
		for(i=0;i<m_n;i++)
			m_abscissas[i]=i*h-1.;
		h/=3.;
		m_weights[0]=m_weights[m_n-1]=h;
		m_weights[1]=4.*h;
		for(i=2;i<m_n-1;i+=2)
		{
			m_weights[i]=2.*h;
			m_weights[i+1]=4.*h;
		}
		break;
	}
	// Newton **********************************************************
	case NewtonCotes3:
	case Newton:
	{
		if(m_n<4 || (m_n-1)%3!=0)
		{
			if(m_n<4) m_n=4;
			if((m_n-1)%3!=0) m_n++;
			//m_n=0
			m_abscissas.resize(m_n);
			m_weights.resize(m_n);
			throw ErrN(m_n);
			// break;
		}
		double h=2./(m_n-1);
		for(i=0;i<m_n;i++)
			m_abscissas[i]=i*h-1.;
		h*=3./8.;
		m_weights[0]=m_weights[m_n-1]=h;
		m_weights[1]=3.*h;
		m_weights[2]=3.*h;
		for(i=3;i<m_n-1;i+=3)
		{
			m_weights[i]=2.*m_weights[0];
			m_weights[i+1]=m_weights[1];
			m_weights[i+2]=m_weights[2];
		}
		break;
	}
	// Newton-Cotes 4. Ordnung *****************************************
	case NewtonCotes4:
	{
		if(m_n<5 || (m_n-1)%4!=0)
		{
			if(m_n<5) m_n=5;
			if((m_n-1)%4!=0) m_n++;
			//m_n=0
			m_abscissas.resize(m_n);
			m_weights.resize(m_n);
			throw ErrN(m_n);
			// break;
		}
		double h=2./(m_n-1);
		for(i=0;i<m_n;i++)
			m_abscissas[i]=i*h-1.;
		h*=2./45.;
		m_weights[0]=m_weights[m_n-1]=7.*h;
		m_weights[1]=32.*h;
		m_weights[2]=12.*h;
		m_weights[3]=32.*h;
		for(i=4;i<m_n-1;i+=4)
		{
			m_weights[i]=2.*m_weights[0];
			m_weights[i+1]=m_weights[1];
			m_weights[i+2]=m_weights[2];
			m_weights[i+3]=m_weights[3];
		}
		break;
	}
	// Newton-Cotes 5. Ordnung *****************************************
	case NewtonCotes5:
	{
		if(m_n<6 || (m_n-1)%5!=0)
		{
			if(m_n<6) m_n=6;
			if((m_n-1)%5!=0) m_n++;
			//m_n=0
			m_abscissas.resize(m_n);
			m_weights.resize(m_n);
			throw ErrN(m_n);
			// break;
		}
		double h=2./(m_n-1);
		for(i=0;i<m_n;i++)
			m_abscissas[i]=i*h-1.;
		h*=5./288.;
		m_weights[0]=m_weights[m_n-1]=19.*h;
		m_weights[1]=75.*h;
		m_weights[2]=50.*h;
		m_weights[3]=50.*h;
		m_weights[3]=75.*h;
		for(i=4;i<m_n-1;i+=4)
		{
			m_weights[i]=2.*m_weights[0];
			m_weights[i+1]=m_weights[1];
			m_weights[i+2]=m_weights[2];
			m_weights[i+3]=m_weights[3];
			m_weights[i+4]=m_weights[4];
		}
		break;
	}
	// GaussLegendre ***************************************************
	case GaussLegendre:
		switch (m_n)
		{
		case 1:
			m_abscissas[0]=0.;
			m_weights[0]=2.;
			break;
		case 2:
			m_abscissas[0]=.577350269189625764509149;
			m_weights[0]=1.;
			break;
		case 3:
			m_abscissas[0]=.774596669241483377035853;
			m_abscissas[1]=0.;
			m_weights[0]=.555555555555555555555556;
			m_weights[1]=.888888888888888888888889;
			break;
		case 4:
			m_abscissas[0]=.861136311594052575223946;
			m_abscissas[1]=.339981043584856264802666;
			m_weights[0]=.347854845137453857373064;
			m_weights[1]=.652145154862546142626936;
			break;
		case 5:
			m_abscissas[0]=.906179845938663992797627;
			m_abscissas[1]=.538469310105683091036314;
			m_abscissas[2]=0.;
			m_weights[0]=.236926885056189087514264;
			m_weights[1]=.478628670499366468041292;
			m_weights[2]=.568888888888888888888889;
			break;
		case 6:
			m_abscissas[0]=.932469514203152027812302;
			m_abscissas[1]=.6612093864662645136614;
			m_abscissas[2]=.238619186083196908630502;
			m_weights[0]=.171324492379170345040296;
			m_weights[1]=.360761573048138607569834;
			m_weights[2]=.46791393457269104738987;
			break;
		case 7:
			m_abscissas[0]=.94910791234275852452619;
			m_abscissas[1]=.741531185599394439863865;
			m_abscissas[2]=.405845151377397166906606;
			m_abscissas[3]=0.;
			m_weights[0]=.129484966168869693270611;
			m_weights[1]=.279705391489276667901468;
			m_weights[2]=.38183005050511894495037;
			m_weights[3]=.417959183673469387755102;
			break;
		case 8:
			m_abscissas[0]=.960289856497536231683561;
			m_abscissas[1]=.796666477413626739591554;
			m_abscissas[2]=.525532409916328985817739;
			m_abscissas[3]=.183434642495649804939476;
			m_weights[0]=.101228536290376259152531;
			m_weights[1]=.222381034453374470544356;
			m_weights[2]=.313706645877887287337962;
			m_weights[3]=.36268378337836198296515;
			break;
		case 9:
			m_abscissas[0]=.968160239507626089835576;
			m_abscissas[1]=.83603110732663579429943;
			m_abscissas[2]=.613371432700590397308702;
			m_abscissas[3]=.324253423403808929038538;
			m_abscissas[4]=0.;
			m_weights[0]=.081274388361574411971892;
			m_weights[1]=.180648160694857404058472;
			m_weights[2]=.260610696402935462318743;
			m_weights[3]=.31234707704000284006863;
			m_weights[4]=.330239355001259763164525;
			break;
		case 10:
			m_abscissas[0]=.973906528517171720077964;
			m_abscissas[1]=.865063366688984510732097;
			m_abscissas[2]=.679409568299024406234327;
			m_abscissas[3]=.433395394129247190799266;
			m_abscissas[4]=.148874338981631210884826;
			m_weights[0]=.066671344308688137593569;
			m_weights[1]=.149451349150580593145776;
			m_weights[2]=.219086362515982043995535;
			m_weights[3]=.269266719309996355091227;
			m_weights[4]=.295524224714752870173893;
			break;
		case 11:
			m_abscissas[0]=.978228658146056992803938;
			m_abscissas[1]=.887062599768095299075158;
			m_abscissas[2]=.730152005574049324093416;
			m_abscissas[3]=.519096129206811815925726;
			m_abscissas[4]=.269543155952344972331532;
			m_abscissas[5]=0.;
			m_weights[0]=.055668567116173666482754;
			m_weights[1]=.125580369464904624634694;
			m_weights[2]=.186290210927734251426098;
			m_weights[3]=.233193764591990479918524;
			m_weights[4]=.262804544510246662180689;
			m_weights[5]=.272925086777900630714484;
			break;
		case 12:
			m_abscissas[0]=.981560634246719250690549;
			m_abscissas[1]=.904117256370474856678466;
			m_abscissas[2]=.769902674194304687036894;
			m_abscissas[3]=.587317954286617447296702;
			m_abscissas[4]=.367831498998180193752692;
			m_abscissas[5]=.125233408511468915472441;
			m_weights[0]=.047175336386511827194616;
			m_weights[1]=.106939325995318430960255;
			m_weights[2]=.160078328543346226334653;
			m_weights[3]=.203167426723065921749064;
			m_weights[4]=.23349253653835480876085;
			m_weights[5]=.249147045813402785000562;
			break;
		case 13:
			m_abscissas[0]=.984183054718588149472829;
			m_abscissas[1]=.917598399222977965206548;
			m_abscissas[2]=.801578090733309912794206;
			m_abscissas[3]=.642349339440340220643985;
			m_abscissas[4]=.448492751036446852877913;
			m_abscissas[5]=.230458315955134794065528;
			m_abscissas[6]=0.;
			m_weights[0]=.040484004765315879520022;
			m_weights[1]=.092121499837728447914422;
			m_weights[2]=.138873510219787238463602;
			m_weights[3]=.178145980761945738280047;
			m_weights[4]=.207816047536888502312523;
			m_weights[5]=.22628318026289723841209;
			m_weights[6]=.23255155323087391019459;
			break;
		case 14:
			m_abscissas[0]=.986283808696812338841597;
			m_abscissas[1]=.928434883663573517336391;
			m_abscissas[2]=.827201315069764993189795;
			m_abscissas[3]=.68729290481168547014802;
			m_abscissas[4]=.515248636358154091965291;
			m_abscissas[5]=.319112368927889760435672;
			m_abscissas[6]=.108054948707343662066245;
			m_weights[0]=.035119460331751863031833;
			m_weights[1]=.080158087159760209805633;
			m_weights[2]=.121518570687903184689415;
			m_weights[3]=.157203167158193534569602;
			m_weights[4]=.185538397477937813741717;
			m_weights[5]=.205198463721295603965924;
			m_weights[6]=.215263853463157790195876;
			break;
		case 15:
			m_abscissas[0]=.987992518020485428489566;
			m_abscissas[1]=.937273392400705904307759;
			m_abscissas[2]=.848206583410427216200648;
			m_abscissas[3]=.724417731360170047416186;
			m_abscissas[4]=.570972172608538847537227;
			m_abscissas[5]=.394151347077563369897207;
			m_abscissas[6]=.201194093997434522300628;
			m_abscissas[7]=0.;
			m_weights[0]=.030753241996117268354628;
			m_weights[1]=.070366047488108124709267;
			m_weights[2]=.10715922046717193501187;
			m_weights[3]=.139570677926154314447805;
			m_weights[4]=.166269205816993933553201;
			m_weights[5]=.186161000015562211026801;
			m_weights[6]=.198431485327111576456118;
			m_weights[7]=.20257824192556127288062;
			break;
		case 16:
			m_abscissas[0]=.989400934991649932596154;
			m_abscissas[1]=.944575023073232576077988;
			m_abscissas[2]=.865631202387831743880468;
			m_abscissas[3]=.755404408355003033895101;
			m_abscissas[4]=.617876244402643748446672;
			m_abscissas[5]=.458016777657227386342419;
			m_abscissas[6]=.281603550779258913230461;
			m_abscissas[7]=.095012509837637440185319;
			m_weights[0]=.027152459411754094851781;
			m_weights[1]=.062253523938647892862844;
			m_weights[2]=.095158511682492784809925;
			m_weights[3]=.124628971255533872052476;
			m_weights[4]=.149595988816576732081502;
			m_weights[5]=.169156519395002538189312;
			m_weights[6]=.182603415044923588866764;
			m_weights[7]=.189450610455068496285397;
			break;
		case 17:
			m_abscissas[0]=.990575475314417335675434;
			m_abscissas[1]=.950675521768767761222717;
			m_abscissas[2]=.880239153726985902122956;
			m_abscissas[3]=.78151400389680140692523;
			m_abscissas[4]=.657671159216690765850302;
			m_abscissas[5]=.512690537086476967886247;
			m_abscissas[6]=.351231763453876315297186;
			m_abscissas[7]=.178484181495847855850677;
			m_abscissas[8]=0.;
			m_weights[0]=.02414830286854793196011;
			m_weights[1]=.055459529373987201129440;
			m_weights[2]=.085036148317179180883535;
			m_weights[3]=.111883847193403971094788;
			m_weights[4]=.13513636846852547328632;
			m_weights[5]=.154045761076810288081432;
			m_weights[6]=.168004102156450044509971;
			m_weights[7]=.176562705366992646325271;
			m_weights[8]=.179446470356206525458266;
			break;
		case 18:
			m_abscissas[0]=.991565168420930946730016;
			m_abscissas[1]=.955823949571397755181196;
			m_abscissas[2]=.892602466497555739206061;
			m_abscissas[3]=.803704958972523115682417;
			m_abscissas[4]=.691687043060353207874891;
			m_abscissas[5]=.559770831073947534607872;
			m_abscissas[6]=.411751161462842646035932;
			m_abscissas[7]=.251886225691505509588973;
			m_abscissas[8]=.084775013041735301242262;
			m_weights[0]=.021616013526483310313343;
			m_weights[1]=.049714548894969796453335;
			m_weights[2]=.076425730254889056529130;
			m_weights[3]=.100942044106287165562814;
			m_weights[4]=.122555206711478460184519;
			m_weights[5]=.140642914670650651204731;
			m_weights[6]=.154684675126265244925418;
			m_weights[7]=.164276483745832722986054;
			m_weights[8]=.169142382963143591840656;
			break;
		case 19:
			m_abscissas[0]=.992406843843584403189018;
			m_abscissas[1]=.960208152134830030852779;
			m_abscissas[2]=.903155903614817901642661;
			m_abscissas[3]=.822714656537142824978922;
			m_abscissas[4]=.720966177335229378617096;
			m_abscissas[5]=.600545304661681023469638;
			m_abscissas[6]=.464570741375960945717267;
			m_abscissas[7]=.316564099963629831990117;
			m_abscissas[8]=.160358645640225375868096;
			m_abscissas[9]=0.;
			m_weights[0]=.019461788229726477036312;
			m_weights[1]=.044814226765699600332838;
			m_weights[2]=.069044542737641226580708;
			m_weights[3]=.091490021622449999464462;
			m_weights[4]=.111566645547333994716024;
			m_weights[5]=.128753962539336227675516;
			m_weights[6]=.142606702173606611775746;
			m_weights[7]=.152766042065859666778855;
			m_weights[8]=.158968843393954347649956;
			m_weights[9]=.161054449848783695979164;
			break;
		case 20:
			m_abscissas[0]=.993128599185094924786122;
			m_abscissas[1]=.963971927277913791267666;
			m_abscissas[2]=.912234428251325905867752;
			m_abscissas[3]=.839116971822218823394529;
			m_abscissas[4]=.746331906460150792614305;
			m_abscissas[5]=.636053680726515025452837;
			m_abscissas[6]=.510867001950827098004364;
			m_abscissas[7]=.373706088715419560672548;
			m_abscissas[8]=.227785851141645078080496;
			m_abscissas[9]=.076526521133497333754640;
			m_weights[0]=.017614007139152118311862;
			m_weights[1]=.040601429800386941331040;
			m_weights[2]=.062672048334109063569507;
			m_weights[3]=.083276741576704748724758;
			m_weights[4]=.10193011981724043503675;
			m_weights[5]=.118194531961518417312377;
			m_weights[6]=.131688638449176626898494;
			m_weights[7]=.142096109318382051329298;
			m_weights[8]=.149172986472603746787829;
			m_weights[9]=.152753387130725850698084;
			break;
		case 21:
			m_abscissas[0]=.993752170620389500260242;
			m_abscissas[1]=.967226838566306294316622;
			m_abscissas[2]=.920099334150400828790187;
			m_abscissas[3]=.853363364583317283647251;
			m_abscissas[4]=.768439963475677908615878;
			m_abscissas[5]=.667138804197412319305967;
			m_abscissas[6]=.551618835887219807059019;
			m_abscissas[7]=.424342120207438783573669;
			m_abscissas[8]=.288021316802401096600793;
			m_abscissas[9]=.145561854160895090937031;
			m_abscissas[10]=0.;
			m_weights[0]=.016017228257774333324225;
			m_weights[1]=.036953789770852493799951;
			m_weights[2]=.057134425426857208283636;
			m_weights[3]=.076100113628379302017052;
			m_weights[4]=.093444423456033861553290;
			m_weights[5]=.108797299167148377663475;
			m_weights[6]=.121831416053728534195367;
			m_weights[7]=.132268938633337461781053;
			m_weights[8]=.139887394791073154722133;
			m_weights[9]=.144524403989970059063827;
			m_weights[10]=.146081133649690427191985;
			break;
		case 22:
			m_abscissas[0]=.994294585482399292073031;
			m_abscissas[1]=.970060497835428727123951;
			m_abscissas[2]=.926956772187174000520693;
			m_abscissas[3]=.865812577720300136536426;
			m_abscissas[4]=.787816805979208162004278;
			m_abscissas[5]=.69448726318668278005069;
			m_abscissas[6]=.587640403506911592958877;
			m_abscissas[7]=.469355837986757026406331;
			m_abscissas[8]=.341935820892084225158147;
			m_abscissas[9]=.207860426688221285478847;
			m_abscissas[10]=.069739273319722221213842;
			m_weights[0]=.014627995298272200684991;
			m_weights[1]=.033774901584814154793302;
			m_weights[2]=.052293335152683285940312;
			m_weights[3]=.069796468424520488094961;
			m_weights[4]=.085941606217067727414444;
			m_weights[5]=.100414144442880964932079;
			m_weights[6]=.112932296080539218393401;
			m_weights[7]=.123252376810512424285561;
			m_weights[8]=.131173504787062370732965;
			m_weights[9]=.136541498346015171352574;
			m_weights[10]=.13925187285563199337541;
			break;
		case 23:
			m_abscissas[0]=.994769334997552123523926;
			m_abscissas[1]=.972542471218115231956024;
			m_abscissas[2]=.932971086826016102349197;
			m_abscissas[3]=.876752358270441667378157;
			m_abscissas[4]=.804888401618839892151118;
			m_abscissas[5]=.718661363131950194461624;
			m_abscissas[6]=.619609875763646156385097;
			m_abscissas[7]=.509501477846007549689793;
			m_abscissas[8]=.390301038030290831421489;
			m_abscissas[9]=.26413568097034493053387;
			m_abscissas[10]=.133256824298466110931743;
			m_abscissas[11]=0.;
			m_weights[0]=.013411859487141772081309;
			m_weights[1]=.030988005856979444310694;
			m_weights[2]=.048037671731084668571641;
			m_weights[3]=.064232421408525852127170;
			m_weights[4]=.079281411776718954922893;
			m_weights[5]=.092915766060035147477019;
			m_weights[6]=.104892091464541410074086;
			m_weights[7]=.114996640222411364941644;
			m_weights[8]=.123049084306729530467578;
			m_weights[9]=.128905722188082149978595;
			m_weights[10]=.132462039404696617371642;
			m_weights[11]=.133654572186106175351457;
			break;
		case 24:
			m_abscissas[0]=.995187219997021360179997;
			m_abscissas[1]=.974728555971309498198392;
			m_abscissas[2]=.938274552002732758523649;
			m_abscissas[3]=.886415527004401034213154;
			m_abscissas[4]=.82000198597390292195395;
			m_abscissas[5]=.740124191578554364243828;
			m_abscissas[6]=.648093651936975569252496;
			m_abscissas[7]=.545421471388839535658376;
			m_abscissas[8]=.433793507626045138487084;
			m_abscissas[9]=.315042679696163374386793;
			m_abscissas[10]=.19111886747361630915864;
			m_abscissas[11]=.064056892862605626085043;
			m_weights[0]=.012341229799987199546806;
			m_weights[1]=.028531388628933663181308;
			m_weights[2]=.044277438817419806168603;
			m_weights[3]=.059298584915436780746368;
			m_weights[4]=.073346481411080305734034;
			m_weights[5]=.086190161531953275917185;
			m_weights[6]=.097618652104113888269881;
			m_weights[7]=.107444270115965634782577;
			m_weights[8]=.115505668053725601353344;
			m_weights[9]=.121670472927803391204463;
			m_weights[10]=.125837456346828296121375;
			m_weights[11]=.127938195346752156974056;
			break;
		case 25:
			m_abscissas[0]=.995556969790498097908785;
			m_abscissas[1]=.976663921459517511498315;
			m_abscissas[2]=.942974571228974339414011;
			m_abscissas[3]=.894991997878275368851042;
			m_abscissas[4]=.833442628760834001421021;
			m_abscissas[5]=.759259263037357630577283;
			m_abscissas[6]=.673566368473468364485121;
			m_abscissas[7]=.57766293024122296772369;
			m_abscissas[8]=.473002731445714960522182;
			m_abscissas[9]=.361172305809387837735822;
			m_abscissas[10]=.24386688372098843204519;
			m_abscissas[11]=.12286469261071039638736;
			m_abscissas[12]=0.;
			m_weights[0]=.011393798501026287947903;
			m_weights[1]=.026354986615032137261902;
			m_weights[2]=.040939156701306312655623;
			m_weights[3]=.054904695975835191925937;
			m_weights[4]=.068038333812356917207187;
			m_weights[5]=.080140700335001018013235;
			m_weights[6]=.091028261982963649811497;
			m_weights[7]=.100535949067050644202207;
			m_weights[8]=.108519624474263653116094;
			m_weights[9]=.114858259145711648339326;
			m_weights[10]=.119455763535784772228178;
			m_weights[11]=.12224244299031004168896;
			m_weights[12]=.123176053726715451203903;
			break;
		case 26:
			m_abscissas[0]=.995885701145616929003217;
			m_abscissas[1]=.97838544595647099110058;
			m_abscissas[2]=.947159066661714250135915;
			m_abscissas[3]=.902637861984307074217666;
			m_abscissas[4]=.845445942788498018797507;
			m_abscissas[5]=.776385948820678856192967;
			m_abscissas[6]=.696427260419957264863814;
			m_abscissas[7]=.606692293017618063231979;
			m_abscissas[8]=.508440714824505717695703;
			m_abscissas[9]=.403051755123486306481077;
			m_abscissas[10]=.292004839485956895142835;
			m_abscissas[11]=.176858820356890183969058;
			m_abscissas[12]=.059230093429313207093719;
			m_weights[0]=.010551372617343007155651;
			m_weights[1]=.024417851092631908789616;
			m_weights[2]=.037962383294362763950303;
			m_weights[3]=.050975825297147811998320;
			m_weights[4]=.063274046329574835539454;
			m_weights[5]=.074684149765659745887076;
			m_weights[6]=.085045894313485239210448;
			m_weights[7]=.094213800355914148463665;
			m_weights[8]=.102059161094425423238414;
			m_weights[9]=.108471840528576590656579;
			m_weights[10]=.113361816546319666549441;
			m_weights[11]=.116660443485296582044663;
			m_weights[12]=.118321415279262276516371;
			break;
		case 27:
			m_abscissas[0]=.996179262888988566938887;
			m_abscissas[1]=.979923475961501222855873;
			m_abscissas[2]=.950900557814705006851908;
			m_abscissas[3]=.909482320677491104300645;
			m_abscissas[4]=.856207908018294490302737;
			m_abscissas[5]=.791771639070508227144397;
			m_abscissas[6]=.717013473739423699294816;
			m_abscissas[7]=.632907971946495140927735;
			m_abscissas[8]=.540551564579456894900301;
			m_abscissas[9]=.441148251750026880585974;
			m_abscissas[10]=.335993903638508899730319;
			m_abscissas[11]=.226459365439536858857239;
			m_abscissas[12]=.113972585609529966932895;
			m_abscissas[13]=0.;
			m_weights[0]=.009798996051294360261150;
			m_weights[1]=.022686231596180623196034;
			m_weights[2]=.035297053757419711022578;
			m_weights[3]=.047449412520615062704097;
			m_weights[4]=.058983536859833599110301;
			m_weights[5]=.069748823766245592984323;
			m_weights[6]=.079604867773057771263075;
			m_weights[7]=.088423158543756950194323;
			m_weights[8]=.096088727370028507565653;
			m_weights[9]=.102501637817745798671248;
			m_weights[10]=.107578285788533187212163;
			m_weights[11]=.111252488356845192672163;
			m_weights[12]=.11347634610896514862037;
			m_weights[13]=.114220867378956989045046;
			break;
		case 28:
			m_abscissas[0]=.996442497573954449950436;
			m_abscissas[1]=.98130316537087275369456;
			m_abscissas[2]=.954259280628938197254102;
			m_abscissas[3]=.915633026392132073869689;
			m_abscissas[4]=.865892522574395048942255;
			m_abscissas[5]=.805641370917179171447886;
			m_abscissas[6]=.735610878013631772028145;
			m_abscissas[7]=.656651094038864961219898;
			m_abscissas[8]=.569720471811401719308003;
			m_abscissas[9]=.475874224955118261034412;
			m_abscissas[10]=.376251516089078710221357;
			m_abscissas[11]=.272061627635178077676826;
			m_abscissas[12]=.164569282133380771281472;
			m_abscissas[13]=.055079289884034270426517;
			m_weights[0]=.009124282593094517738816;
			m_weights[1]=.021132112592771259751500;
			m_weights[2]=.032901427782304379977631;
			m_weights[3]=.044272934759004227839588;
			m_weights[4]=.055107345675716745431483;
			m_weights[5]=.065272923966999595793398;
			m_weights[6]=.074646214234568779023932;
			m_weights[7]=.083113417228901218390396;
			m_weights[8]=.090571744393032840942186;
			m_weights[9]=.096930657997929915850489;
			m_weights[10]=.102112967578060769814217;
			m_weights[11]=.106055765922846417910416;
			m_weights[12]=.108711192258294135253572;
			m_weights[13]=.110047013016475196282376;
			break;
		case 29:
			m_abscissas[0]=.996679442260596586163192;
			m_abscissas[1]=.982545505261413174870926;
			m_abscissas[2]=.957285595778087725798208;
			m_abscissas[3]=.921180232953058785093753;
			m_abscissas[4]=.874637804920102790417793;
			m_abscissas[5]=.818185487615252444989572;
			m_abscissas[6]=.75246285173447713391261;
			m_abscissas[7]=.678214537602686515156185;
			m_abscissas[8]=.596281797138227820379586;
			m_abscissas[9]=.507592955124227642102628;
			m_abscissas[10]=.413152888174008663890707;
			m_abscissas[11]=.314031637867639934948196;
			m_abscissas[12]=.211352286166001074506376;
			m_abscissas[13]=.106278230132679230170982;
			m_abscissas[14]=0.;
			m_weights[0]=.008516903878746409654264;
			m_weights[1]=.019732085056122705983860;
			m_weights[2]=.030740492202093622644409;
			m_weights[3]=.04140206251868283610483;
			m_weights[4]=.051594826902497923912594;
			m_weights[5]=.061203090657079138542110;
			m_weights[6]=.070117933255051278569581;
			m_weights[7]=.078238327135763783828145;
			m_weights[8]=.085472257366172527545345;
			m_weights[9]=.091737757139258763347966;
			m_weights[10]=.0969638340944086063019;
			m_weights[11]=.101091273759914966121821;
			m_weights[12]=.104073310077729373913328;
			m_weights[13]=.105876155097320941406591;
			m_weights[14]=.106479381718314244246511;
			break;
		case 30:
			m_abscissas[0]=.99689348407464954027163;
			m_abscissas[1]=.983668123279747209970033;
			m_abscissas[2]=.960021864968307512216871;
			m_abscissas[3]=.926200047429274325879324;
			m_abscissas[4]=.882560535792052681543116;
			m_abscissas[5]=.829565762382768397442898;
			m_abscissas[6]=.767777432104826194917977;
			m_abscissas[7]=.697850494793315796932292;
			m_abscissas[8]=.620526182989242861140478;
			m_abscissas[9]=.53662414814201989926417;
			m_abscissas[10]=.44703376953808917678061;
			m_abscissas[11]=.352704725530878113471037;
			m_abscissas[12]=.254636926167889846439805;
			m_abscissas[13]=.153869913608583546963795;
			m_abscissas[14]=.051471842555317695833025;
			m_weights[0]=.007968192496166605615466;
			m_weights[1]=.018466468311090959142302;
			m_weights[2]=.028784707883323369349719;
			m_weights[3]=.038799192569627049596802;
			m_weights[4]=.048402672830594052902938;
			m_weights[5]=.057493156217619066481722;
			m_weights[6]=.065974229882180495128129;
			m_weights[7]=.073755974737705206268244;
			m_weights[8]=.080755895229420215354695;
			m_weights[9]=.086899787201082979802388;
			m_weights[10]=.092122522237786128717633;
			m_weights[11]=.096368737174644259639469;
			m_weights[12]=.099593420586795267062780;
			m_weights[13]=.101762389748405504596429;
			m_weights[14]=.102852652893558840341286;
			break;
		case 31:
			m_abscissas[0]=.997087481819477074055627;
			m_abscissas[1]=.984685909665152484002465;
			m_abscissas[2]=.962503925092949661789052;
			m_abscissas[3]=.930756997896648164956946;
			m_abscissas[4]=.889760029948271043374192;
			m_abscissas[5]=.839920320146267340086905;
			m_abscissas[6]=.78173314841662494040636;
			m_abscissas[7]=.715776784586853283905971;
			m_abscissas[8]=.642706722924260346184418;
			m_abscissas[9]=.563249161407149262720945;
			m_abscissas[10]=.478193782044902480440594;
			m_abscissas[11]=.388385901608232943061351;
			m_abscissas[12]=.294718069981701616617904;
			m_abscissas[13]=.198121199335570628772413;
			m_abscissas[14]=.099555312152341520325175;
			m_abscissas[15]=0.;
			m_weights[0]=.007470831579248775858697;
			m_weights[1]=.017318620790310582463158;
			m_weights[2]=.027009019184979421800609;
			m_weights[3]=.036432273912385464024392;
			m_weights[4]=.045493707527201102902316;
			m_weights[5]=.054103082424916853711666;
			m_weights[6]=.062174786561028426910344;
			m_weights[7]=.069628583235410366167756;
			m_weights[8]=.076390386598776616426358;
			m_weights[9]=.082392991761589263903823;
			m_weights[10]=.087576740608477876126198;
			m_weights[11]=.091890113893641478215363;
			m_weights[12]=.095290242912319512807204;
			m_weights[13]=.097743335386328725093474;
			m_weights[14]=.099225011226672307874876;
			m_weights[15]=.099720544793426451427534;
			break;
		case 32:
			m_abscissas[0]=.997263861849481563544981;
			m_abscissas[1]=.985611511545268335400175;
			m_abscissas[2]=.964762255587506430773812;
			m_abscissas[3]=.934906075937739689170919;
			m_abscissas[4]=.896321155766052123965307;
			m_abscissas[5]=.849367613732569970133693;
			m_abscissas[6]=.794483795967942406963097;
			m_abscissas[7]=.732182118740289680387427;
			m_abscissas[8]=.663044266930215200975115;
			m_abscissas[9]=.587715757240762329040745;
			m_abscissas[10]=.506899908932229390023747;
			m_abscissas[11]=.421351276130635345364119;
			m_abscissas[12]=.331868602282127649779917;
			m_abscissas[13]=.239287362252137074544603;
			m_abscissas[14]=.144471961582796493485186;
			m_abscissas[15]=.048307665687738316234813;
			m_weights[0]=.007018610009470096600407;
			m_weights[1]=.016274394730905670605171;
			m_weights[2]=.025392065309262059455753;
			m_weights[3]=.034273862913021433102688;
			m_weights[4]=.042835898022226680656879;
			m_weights[5]=.050998059262376176196163;
			m_weights[6]=.058684093478535547145284;
			m_weights[7]=.06582222277636184683765;
			m_weights[8]=.072345794108848506225399;
			m_weights[9]=.078193895787070306471741;
			m_weights[10]=.083311924226946755222199;
			m_weights[11]=.087652093004403811142771;
			m_weights[12]=.091173878695763884712869;
			m_weights[13]=.093844399080804565639180;
			m_weights[14]=.095638720079274859419082;
			m_weights[15]=.096540088514727800566765;
			break;
		case 33:
			m_abscissas[0]=.997424694246455217266168;
			m_abscissas[1]=.986455726230642488110376;
			m_abscissas[2]=.966822909689992768928378;
			m_abscissas[3]=.938694372611168350355835;
			m_abscissas[4]=.902316767743433583040531;
			m_abscissas[5]=.858009652676504064643061;
			m_abscissas[6]=.806162356274166589796201;
			m_abscissas[7]=.747230496449562157859055;
			m_abscissas[8]=.681731959969742786268216;
			m_abscissas[9]=.610242345836379027307288;
			m_abscissas[10]=.533389904786347643548894;
			m_abscissas[11]=.451850017272450695725993;
			m_abscissas[12]=.366339257748073341070221;
			m_abscissas[13]=.277609097152497029403248;
			m_abscissas[14]=.186439298827991572335799;
			m_abscissas[15]=.093631065854733385670743;
			m_abscissas[16]=0.;
			m_weights[0]=.006606227847587378058649;
			m_weights[1]=.015321701512934676127946;
			m_weights[2]=.023915548101749480350533;
			m_weights[3]=.032300358632328953281561;
			m_weights[4]=.040401541331669591563410;
			m_weights[5]=.048147742818711695670147;
			m_weights[6]=.055470846631663561284944;
			m_weights[7]=.062306482530317480031628;
			m_weights[8]=.068594572818656712805955;
			m_weights[9]=.074279854843954149342472;
			m_weights[10]=.079312364794886738363908;
			m_weights[11]=.083647876067038707613928;
			m_weights[12]=.087248287618844337607282;
			m_weights[13]=.090081958660638577239744;
			m_weights[14]=.092123986643316846213241;
			m_weights[15]=.093356426065596116160999;
			m_weights[16]=.093768446160209996567305;
			break;
		case 34:
			m_abscissas[0]=.997571753790841919243372;
			m_abscissas[1]=.98722781640630948504975;
			m_abscissas[2]=.968708262533344281764647;
			m_abscissas[3]=.942162397405107091631676;
			m_abscissas[4]=.9078096777183244688009;
			m_abscissas[5]=.865934638334564469263572;
			m_abscissas[6]=.816884227900933664591579;
			m_abscissas[7]=.761064876629873014187409;
			m_abscissas[8]=.698939113216262907933;
			m_abscissas[9]=.631021727080528545317776;
			m_abscissas[10]=.55787550066974664273646;
			m_abscissas[11]=.480106545190327034194103;
			m_abscissas[12]=.398359277758645940631495;
			m_abscissas[13]=.313311081339463247458317;
			m_abscissas[14]=.225666691616449483868641;
			m_abscissas[15]=.136152357259182975894429;
			m_abscissas[16]=.045509821953102542749076;
			m_weights[0]=.006229140555908684718606;
			m_weights[1]=.014450162748595035415202;
			m_weights[2]=.022563721985494970084094;
			m_weights[3]=.030491380638446131809442;
			m_weights[4]=.038166593796387516321766;
			m_weights[5]=.045525611523353272453823;
			m_weights[6]=.052507414572678106168246;
			m_weights[7]=.059054135827524493193961;
			m_weights[8]=.065111521554076411378544;
			m_weights[9]=.070629375814255724999039;
			m_weights[10]=.075561974660031931270834;
			m_weights[11]=.079868444339771844738819;
			m_weights[12]=.083513099699845655187020;
			m_weights[13]=.086465739747035749784247;
			m_weights[14]=.088701897835693869287076;
			m_weights[15]=.090203044370640729573942;
			m_weights[16]=.090956740330259873615338;
			break;
		case 35:
			m_abscissas[0]=.997706569099600297260163;
			m_abscissas[1]=.987935764443851498035117;
			m_abscissas[2]=.97043761603922983321507;
			m_abscissas[3]=.945345148207827329538726;
			m_abscissas[4]=.912854261359317614464937;
			m_abscissas[5]=.873219125025222331523282;
			m_abscissas[6]=.826749899092225406834051;
			m_abscissas[7]=.773810252286912555267423;
			m_abscissas[8]=.714814501556628783264409;
			m_abscissas[9]=.650224364665890388675793;
			m_abscissas[10]=.580545344749764509934502;
			m_abscissas[11]=.506322773241488615024298;
			m_abscissas[12]=.428137541517814254187621;
			m_abscissas[13]=.34660155443081394587698;
			m_abscissas[14]=.262352941209296057970895;
			m_abscissas[15]=.176051061165989569974304;
			m_abscissas[16]=.088371343275659263600929;
			m_abscissas[17]=0.;
			m_weights[0]=.005883433420443084975754;
			m_weights[1]=.013650828348361492266404;
			m_weights[2]=.021322979911483580883438;
			m_weights[3]=.028829260108894254048716;
			m_weights[4]=.036110115863463380532717;
			m_weights[5]=.043108422326170218782306;
			m_weights[6]=.049769370401353529805200;
			m_weights[7]=.056040816212370128578328;
			m_weights[8]=.061873671966080188887014;
			m_weights[9]=.067222285269086903964306;
			m_weights[10]=.072044794772560064665462;
			m_weights[11]=.076303457155442053538659;
			m_weights[12]=.079964942242324262932662;
			m_weights[13]=.083000593728856588379927;
			m_weights[14]=.085386653392099125225944;
			m_weights[15]=.087104446997183534243322;
			m_weights[16]=.088140530430275462970739;
			m_weights[17]=.088486794907104290638207;
			break;
		case 36:
			m_abscissas[0]=.997830462484085836198828;
			m_abscissas[1]=.988586478902212238073395;
			m_abscissas[2]=.972027691049697949335605;
			m_abscissas[3]=.948272984399507545202474;
			m_abscissas[4]=.917497774515659066075867;
			m_abscissas[5]=.879929800890397131982427;
			m_abscissas[6]=.835847166992475306418821;
			m_abscissas[7]=.785576230132206512827769;
			m_abscissas[8]=.72948917159355658209026;
			m_abscissas[9]=.668001236585521062097191;
			m_abscissas[10]=.601567658135980535079945;
			m_abscissas[11]=.530680285926245161640709;
			m_abscissas[12]=.455863944433420267207218;
			m_abscissas[13]=.377672547119689216322736;
			m_abscissas[14]=.296684995344028270503239;
			m_abscissas[15]=.213500892316865578943226;
			m_abscissas[16]=.128736103809384788651993;
			m_abscissas[17]=.043018198473708607226969;
			m_weights[0]=.005565719664245045361256;
			m_weights[1]=.012915947284065574404503;
			m_weights[2]=.020181515297735471532099;
			m_weights[3]=.027298621498568779094417;
			m_weights[4]=.034213810770307229921245;
			m_weights[5]=.040875750923644895474115;
			m_weights[6]=.047235083490265978416617;
			m_weights[7]=.053244713977759919092026;
			m_weights[8]=.058860144245324817309675;
			m_weights[9]=.064039797355015489556385;
			m_weights[10]=.068745323835736442613690;
			m_weights[11]=.072941885005653061353873;
			m_weights[12]=.076598410645870674528758;
			m_weights[13]=.079687828912071601908725;
			m_weights[14]=.082187266704339709517223;
			m_weights[15]=.084078218979661934933458;
			m_weights[16]=.085346685739338627491851;
			m_weights[17]=.085983275670394747490085;
			break;
		case 37:
			m_abscissas[0]=.997944582477913648940803;
			m_abscissas[1]=.989185963214319186684099;
			m_abscissas[2]=.97349303005648574432889;
			m_abscissas[3]=.950972343262094821328576;
			m_abscissas[4]=.921781437412463742668144;
			m_abscissas[5]=.88612496215548607894615;
			m_abscissas[6]=.844252987340555967987456;
			m_abscissas[7]=.796459200509902293393046;
			m_abscissas[8]=.743078833981965262547047;
			m_abscissas[9]=.684486309130959357445523;
			m_abscissas[10]=.621092608408924483148153;
			m_abscissas[11]=.553342391861581781235098;
			m_abscissas[12]=.481710877803205554147193;
			m_abscissas[13]=.406700509318326110100932;
			m_abscissas[14]=.328837429883706999498472;
			m_abscissas[15]=.248667792791365758805568;
			m_abscissas[16]=.166753930239851976968672;
			m_abscissas[17]=.083670408954769901943048;
			m_abscissas[18]=0.;
			m_weights[0]=.005273057279497939351722;
			m_weights[1]=.012238780100307556526305;
			m_weights[2]=.019129044489083966043502;
			m_weights[3]=.025886036990558933522759;
			m_weights[4]=.032461639847521481067234;
			m_weights[5]=.038809602501934544488962;
			m_weights[6]=.044885364662437166657410;
			m_weights[7]=.050646297654824601603876;
			m_weights[8]=.056051987998274917808539;
			m_weights[9]=.061064516523225986130988;
			m_weights[10]=.065648722872751249484024;
			m_weights[11]=.069772451555700344885081;
			m_weights[12]=.073406777248488172724627;
			m_weights[13]=.076526207570529237885888;
			m_weights[14]=.079108861837529380767212;
			m_weights[15]=.081136624508465030509878;
			m_weights[16]=.082595272236437250891230;
			m_weights[17]=.083474573625862787252253;
			m_weights[18]=.083768360993138904797017;
			break;
		case 38:
			m_abscissas[0]=.998049930535687619812847;
			m_abscissas[1]=.989739454266385571944426;
			m_abscissas[2]=.974846328590153507640886;
			m_abscissas[3]=.953466330933529595670542;
			m_abscissas[4]=.92574133204858439682511;
			m_abscissas[5]=.891855739004632216794937;
			m_abscissas[6]=.85203502193236218885965;
			m_abscissas[7]=.806544167605316815551565;
			m_abscissas[8]=.755685903753970680737738;
			m_abscissas[9]=.699798680379184355912826;
			m_abscissas[10]=.639254415829681707180345;
			m_abscissas[11]=.574456021047807081132928;
			m_abscissas[12]=.505834717927931103240534;
			m_abscissas[13]=.433847169432376484373256;
			m_abscissas[14]=.358972440479435013256718;
			m_abscissas[15]=.281708809790165261360114;
			m_abscissas[16]=.202570453892116703203978;
			m_abscissas[17]=.122084025337867419869612;
			m_abscissas[18]=.040785147904578239913316;
			m_weights[0]=.005002880749639345675900;
			m_weights[1]=.011613444716468674177668;
			m_weights[2]=.018156577709613236898876;
			m_weights[3]=.024579739738232375895201;
			m_weights[4]=.030839500545175054658731;
			m_weights[5]=.036894081594024738164940;
			m_weights[6]=.042703158504674434235878;
			m_weights[7]=.048228061860758683374352;
			m_weights[8]=.053432019910332319973757;
			m_weights[9]=.058280399146997206022306;
			m_weights[10]=.062740933392133054052970;
			m_weights[11]=.066783937979140411935046;
			m_weights[12]=.070382507066898954739283;
			m_weights[13]=.073512692584743457145206;
			m_weights[14]=.076153663548446396065994;
			m_weights[15]=.078287844658210948075375;
			m_weights[16]=.079901033243527821586028;
			m_weights[17]=.080982493770597100623269;
			m_weights[18]=.081525029280385786699219;
			break;
		case 39:
			m_abscissas[0]=.998147383066432906005472;
			m_abscissas[1]=.990251536854685983639775;
			m_abscissas[2]=.97609870933347105384485;
			m_abscissas[3]=.955775212324652277110892;
			m_abscissas[4]=.929409148486738229697817;
			m_abscissas[5]=.897167119292992887848291;
			m_abscissas[6]=.859252937999906153913797;
			m_abscissas[7]=.815906297430143104353233;
			m_abscissas[8]=.767401242931063499832272;
			m_abscissas[9]=.714044435894534679133867;
			m_abscissas[10]=.656173213432010910734426;
			m_abscissas[11]=.59415345495727798869289;
			m_abscissas[12]=.528377268660437473896344;
			m_abscissas[13]=.459260512309136048663247;
			m_abscissas[14]=.387240163971561455853882;
			m_abscissas[15]=.312771559248185922535997;
			m_abscissas[16]=.236325512461835767336006;
			m_abscissas[17]=.158385339997837799922701;
			m_abscissas[18]=.079443804608755477581917;
			m_abscissas[19]=0.;
			m_weights[0]=.004752944691635101370776;
			m_weights[1]=.011034788939164594242677;
			m_weights[2]=.017256229093724919040805;
			m_weights[3]=.023369384832178164594712;
			m_weights[4]=.029334955983903378592156;
			m_weights[5]=.035115111498131330761065;
			m_weights[6]=.040673276847933843939057;
			m_weights[7]=.045974301108916631884177;
			m_weights[8]=.050984665292129405214021;
			m_weights[9]=.055672690340916299907391;
			m_weights[10]=.060008736088596149574942;
			m_weights[11]=.063965388138682388986706;
			m_weights[12]=.067517630966231265363021;
			m_weights[13]=.070643005970608760770115;
			m_weights[14]=.073321753414268617381154;
			m_weights[15]=.075536937322836057704784;
			m_weights[16]=.077274552544682016728512;
			m_weights[17]=.078523613287371176725063;
			m_weights[18]=.079276222568368471010156;
			m_weights[19]=.079527622139442852417418;
			break;
		case 40:
			m_abscissas[0]=.998237709710559200349623;
			m_abscissas[1]=.990726238699457006453054;
			m_abscissas[2]=.97725994998377426266337;
			m_abscissas[3]=.957916819213791655804541;
			m_abscissas[4]=.932812808278676533360852;
			m_abscissas[5]=.902098806968874296728253;
			m_abscissas[6]=.865959503212259503820782;
			m_abscissas[7]=.82461223083331166319632;
			m_abscissas[8]=.778305651426519387694972;
			m_abscissas[9]=.727318255189927103280996;
			m_abscissas[10]=.671956684614179548379355;
			m_abscissas[11]=.612553889667980237952612;
			m_abscissas[12]=.549467125095128202075931;
			m_abscissas[13]=.483075801686178712908567;
			m_abscissas[14]=.41377920437160500152488;
			m_abscissas[15]=.341994090825758473007492;
			m_abscissas[16]=.268152185007253681141184;
			m_abscissas[17]=.192697580701371099715517;
			m_abscissas[18]=.116084070675255208483451;
			m_abscissas[19]=.038772417506050821933193;
			m_weights[0]=.004521277098533191258472;
			m_weights[1]=.010498284531152813614742;
			m_weights[2]=.016421058381907888712863;
			m_weights[3]=.022245849194166957261504;
			m_weights[4]=.027937006980023401098489;
			m_weights[5]=.033460195282547847392678;
			m_weights[6]=.038782167974472017639972;
			m_weights[7]=.043870908185673271991675;
			m_weights[8]=.048695807635072232061434;
			m_weights[9]=.053227846983936824354996;
			m_weights[10]=.057439769099391551366618;
			m_weights[11]=.061306242492928939166538;
			m_weights[12]=.064804013456601038074555;
			m_weights[13]=.067912045815233903825690;
			m_weights[14]=.070611647391286779695484;
			m_weights[15]=.072886582395804059060511;
			m_weights[16]=.074723169057968264200189;
			m_weights[17]=.076110361900626242371558;
			m_weights[18]=.077039818164247965588308;
			m_weights[19]=.077505947978424811263724;
			break;
		case 41:
			m_abscissas[0]=.998321588574771441519189;
			m_abscissas[1]=.991167109699016308250159;
			m_abscissas[2]=.978338673561083384469171;
			m_abscissas[3]=.959906891730346226099442;
			m_abscissas[4]=.935976987497853825682318;
			m_abscissas[5]=.90668594475810117295834;
			m_abscissas[6]=.872201511692441408833671;
			m_abscissas[7]=.832721200401361331244273;
			m_abscissas[8]=.788471145047409372736222;
			m_abscissas[9]=.739704803069926181060169;
			m_abscissas[10]=.686701502034951289584604;
			m_abscissas[11]=.629764839072196320488649;
			m_abscissas[12]=.569220941610215869654747;
			m_abscissas[13]=.505416599199406032708336;
			m_abscissas[14]=.43871727705140708851712;
			m_abscissas[15]=.369505022640481441428367;
			m_abscissas[16]=.298176277341824865922983;
			m_abscissas[17]=.225139605633422775605786;
			m_abscissas[18]=.150813354863992163574378;
			m_abscissas[19]=.075623258989162996923766;
			m_abscissas[20]=0.;
			m_weights[0]=.004306140358164887684004;
			m_weights[1]=.009999938773905945338496;
			m_weights[2]=.015644938407818588530827;
			m_weights[3]=.021201063368779553075697;
			m_weights[4]=.026635899207110445467549;
			m_weights[5]=.031918211731699281787067;
			m_weights[6]=.037017716703507988435261;
			m_weights[7]=.041905195195909689429340;
			m_weights[8]=.046552648369014342060757;
			m_weights[9]=.050933454294617494781170;
			m_weights[10]=.055022519242578741880147;
			m_weights[11]=.058796420949871944991186;
			m_weights[12]=.062233542580966316471573;
			m_weights[13]=.065314196453527410436164;
			m_weights[14]=.068020736760876766735533;
			m_weights[15]=.070337660620817497481659;
			m_weights[16]=.072251696861023073396346;
			m_weights[17]=.073751882027223469939281;
			m_weights[18]=.074829623176221551891305;
			m_weights[19]=.075478747092715824027247;
			m_weights[20]=.075695535647298372318780;
			break;
		case 42:
			m_abscissas[0]=.998399618990062415022868;
			m_abscissas[1]=.991577288340860919792361;
			m_abscissas[2]=.979342508063748193708983;
			m_abscissas[3]=.961759365338204488746927;
			m_abscissas[4]=.938923557354988178533173;
			m_abscissas[5]=.910959724904127452583848;
			m_abscissas[6]=.878020569812172742711985;
			m_abscissas[7]=.84028598326181690092544;
			m_abscissas[8]=.797962053255487413232706;
			m_abscissas[9]=.751279935689480489568472;
			m_abscissas[10]=.70049459055617121374157;
			m_abscissas[11]=.645883388869247833957496;
			m_abscissas[12]=.587744597485109322840711;
			m_abscissas[13]=.526395749931192287592868;
			m_abscissas[14]=.462171912070421929759075;
			m_abscissas[15]=.395423852042975057677094;
			m_abscissas[16]=.326516124465411512197157;
			m_abscissas[17]=.255825079342879083966415;
			m_abscissas[18]=.183736806564854550852756;
			m_abscissas[19]=.110645027208519868349123;
			m_abscissas[20]=.036948943165351775813096;
			m_weights[0]=.004105998604649084610603;
			m_weights[1]=.009536220301748502411820;
			m_weights[2]=.014922443697357494144678;
			m_weights[3]=.020227869569052644757056;
			m_weights[4]=.025422959526113047886742;
			m_weights[5]=.030479240699603468362905;
			m_weights[6]=.035369071097592110832662;
			m_weights[7]=.040065735180692261760596;
			m_weights[8]=.044543577771965877874316;
			m_weights[9]=.048778140792803245027449;
			m_weights[10]=.052746295699174070343943;
			m_weights[11]=.056426369358018381646427;
			m_weights[12]=.059798262227586654312832;
			m_weights[13]=.062843558045002576409318;
			m_weights[14]=.065545624364908978927005;
			m_weights[15]=.067889703376521944855363;
			m_weights[16]=.069862992492594159766155;
			m_weights[17]=.071454714265170982921810;
			m_weights[18]=.072656175243804104887906;
			m_weights[19]=.073460813453467528264028;
			m_weights[20]=.073864234232172879996386;
			break;
		case 43:
			m_abscissas[0]=.998472332242507713517636;
			m_abscissas[1]=.991959557593244146420718;
			m_abscissas[2]=.980278220980255331505529;
			m_abscissas[3]=.963486613014079993410195;
			m_abscissas[4]=.941671956847637861818477;
			m_abscissas[5]=.914947907206138729456103;
			m_abscissas[6]=.883453765218616863337678;
			m_abscissas[7]=.847353716209315048999086;
			m_abscissas[8]=.806835964136938635279075;
			m_abscissas[9]=.762111747194955121460328;
			m_abscissas[10]=.713414235268957054851921;
			m_abscissas[11]=.660997313751498133165135;
			m_abscissas[12]=.605134259639600935725083;
			m_abscissas[13]=.546116316660084719140405;
			m_abscissas[14]=.484251176785734724069633;
			m_abscissas[15]=.419861376029269252487279;
			m_abscissas[16]=.353282612864303806645476;
			m_abscissas[17]=.284861998032913627105723;
			m_abscissas[18]=.214956244860518209014878;
			m_abscissas[19]=.1439298095107133107703;
			m_abscissas[20]=.072152990874586235422327;
			m_abscissas[21]=0.;
			m_weights[0]=.003919490253844127282967;
			m_weights[1]=.009103996637401403318869;
			m_weights[2]=.014248756431576486108542;
			m_weights[3]=.019319901423683900396125;
			m_weights[4]=.024290456613838815902018;
			m_weights[5]=.029134413261498494915940;
			m_weights[6]=.033826492086860292344968;
			m_weights[7]=.038342222194132657572128;
			m_weights[8]=.042658057197982083763807;
			m_weights[9]=.046751494754346580010647;
			m_weights[10]=.050601192784390156523850;
			m_weights[11]=.054187080318881786863373;
			m_weights[12]=.057490461956910519427609;
			m_weights[13]=.060494115249991294519679;
			m_weights[14]=.063182380449396112325630;
			m_weights[15]=.065541242126322797491234;
			m_weights[16]=.067558402229365169192408;
			m_weights[17]=.069223344193656684282299;
			m_weights[18]=.070527387765085028126286;
			m_weights[19]=.071463734252514141297581;
			m_weights[20]=.072027501971421974345308;
			m_weights[21]=.072215751693798987977462;
			break;
		case 44:
			m_abscissas[0]=.998540200636774224936057;
			m_abscissas[1]=.992316392138515808483338;
			m_abscissas[2]=.981151833077913966662749;
			m_abscissas[3]=.965099650422493139394398;
			m_abscissas[4]=.944239509118194099203255;
			m_abscissas[5]=.918675259984175774323001;
			m_abscissas[6]=.888534238286043202338368;
			m_abscissas[7]=.853966595004710378728302;
			m_abscissas[8]=.815144539645135010487438;
			m_abscissas[9]=.772261479248755899017759;
			m_abscissas[10]=.725531053660717002606965;
			m_abscissas[11]=.675186070666122365333699;
			m_abscissas[12]=.621477345903575847802425;
			m_abscissas[13]=.564672453185470768424637;
			m_abscissas[14]=.505054391388202317982798;
			m_abscissas[15]=.442920174525411483834827;
			m_abscissas[16]=.378579352014707132511765;
			m_abscissas[17]=.312352466502785812236545;
			m_abscissas[18]=.244569456928201251507302;
			m_abscissas[19]=.175568014775516785746508;
			m_abscissas[20]=.105691901708653247117306;
			m_abscissas[21]=.035289236964135359058197;
			m_weights[0]=.003745404803112777515174;
			m_weights[1]=.008700481367524844122566;
			m_weights[2]=.013619586755579985520205;
			m_weights[3]=.018471481736814749172042;
			m_weights[4]=.023231481902019210628958;
			m_weights[5]=.027875782821281010081114;
			m_weights[6]=.032381222812069820880847;
			m_weights[7]=.036725347813808873642909;
			m_weights[8]=.040886512310346218908447;
			m_weights[9]=.044843984081970031446243;
			m_weights[10]=.048578046448352037527639;
			m_weights[11]=.052070096091704461881232;
			m_weights[12]=.055302735563728052548747;
			m_weights[13]=.058259859877595495334211;
			m_weights[14]=.060926736701561968038558;
			m_weights[15]=.063290079733203854950139;
			m_weights[16]=.065338114879181434984241;
			m_weights[17]=.067060638906293652395705;
			m_weights[18]=.068449070269366660985459;
			m_weights[19]=.069496491861572578037084;
			m_weights[20]=.070197685473558212587142;
			m_weights[21]=.070549157789354068811338;
			break;
		case 45:
			m_abscissas[0]=.998603645181936638156548;
			m_abscissas[1]=.992649998447203741748617;
			m_abscissas[2]=.981968715034540568239318;
			m_abscissas[3]=.966608310396894604736425;
			m_abscissas[4]=.946641690995629061784721;
			m_abscissas[5]=.922163936719000388097467;
			m_abscissas[6]=.893291671753241738464649;
			m_abscissas[7]=.860162475960664225339079;
			m_abscissas[8]=.822934220502086337035775;
			m_abscissas[9]=.781784312593906291312363;
			m_abscissas[10]=.736908848945490352623739;
			m_abscissas[11]=.68852168077120052523202;
			m_abscissas[12]=.636853394453223359271224;
			m_abscissas[13]=.582150212569353186680967;
			m_abscissas[14]=.524672820462916067091134;
			m_abscissas[15]=.464695123919635098579602;
			m_abscissas[16]=.402502943858541914077975;
			m_abscissas[17]=.338392654250602161643404;
			m_abscissas[18]=.272669769752377560608765;
			m_abscissas[19]=.205647489783263745719787;
			m_abscissas[20]=.13764520598325302875659;
			m_abscissas[21]=.068986980163144172490415;
			m_abscissas[22]=0.;
			m_weights[0]=.003582663155283558931143;
			m_weights[1]=.008323189296218241645736;
			m_weights[2]=.013031104991582784320631;
			m_weights[3]=.017677535257937590617093;
			m_weights[4]=.022239847550578732393951;
			m_weights[5]=.026696213967577664805675;
			m_weights[6]=.031025374934515467162508;
			m_weights[7]=.035206692201609016247700;
			m_weights[8]=.039220236729302447564187;
			m_weights[9]=.043046880709164971151691;
			m_weights[10]=.046668387718373365267768;
			m_weights[11]=.050067499237952029799132;
			m_weights[12]=.053228016731268951945904;
			m_weights[13]=.056134878759786476643924;
			m_weights[14]=.058774232718841738574362;
			m_weights[15]=.061133500831066522501886;
			m_weights[16]=.063201440073819937749964;
			m_weights[17]=.064968195750723430853827;
			m_weights[18]=.066425348449842528082915;
			m_weights[19]=.067565954163607536270910;
			m_weights[20]=.068384577378669674531692;
			m_weights[21]=.068877316977661322882003;
			m_weights[22]=.069041824829232020110799;
			break;
		case 46:
			m_abscissas[0]=.998663042133817981128268;
			m_abscissas[1]=.992962348906174364073095;
			m_abscissas[2]=.98273366980416686347798;
			m_abscissas[3]=.968021391853991942737748;
			m_abscissas[4]=.948892363446089795622214;
			m_abscissas[5]=.925433798806753950977491;
			m_abscissas[6]=.89775271153394196570132;
			m_abscissas[7]=.865975394866858062915889;
			m_abscissas[8]=.830246837066066053032385;
			m_abscissas[9]=.790730057075274255189144;
			m_abscissas[10]=.747605359615666054000337;
			m_abscissas[11]=.701069512020405697512122;
			m_abscissas[12]=.651334846201997715106477;
			m_abscissas[13]=.598628289712715153177173;
			m_abscissas[14]=.543190330261802635270963;
			m_abscissas[15]=.48527391838816466277232;
			m_abscissas[16]=.425143313282828397322147;
			m_abscissas[17]=.363072877020995710123707;
			m_abscissas[18]=.299345822701870015483439;
			m_abscissas[19]=.234252922206269768626056;
			m_abscissas[20]=.168091179467103528606797;
			m_abscissas[21]=.10116247530558423951586;
			m_abscissas[22]=.033772190016052041519558;
			m_weights[0]=.003430300868107048286019;
			m_weights[1]=.007969898229724622451613;
			m_weights[2]=.012479883770988684206735;
			m_weights[3]=.016933514007836238046231;
			m_weights[4]=.021309998754136501054479;
			m_weights[5]=.025589286397130010634700;
			m_weights[6]=.029751829552202755799052;
			m_weights[7]=.033778627999106896520604;
			m_weights[8]=.037651305357386071327661;
			m_weights[9]=.041352190109678729704220;
			m_weights[10]=.044864395277318126767095;
			m_weights[11]=.048171895101712200530469;
			m_weights[12]=.051259598007143021335365;
			m_weights[13]=.054113415385856754491638;
			m_weights[14]=.056720325843991235816874;
			m_weights[15]=.059068434595546314807551;
			m_weights[16]=.061147027724650481015357;
			m_weights[17]=.062946621064394508178952;
			m_weights[18]=.064459003467139069588279;
			m_weights[19]=.065677274267781207378758;
			m_weights[20]=.066595874768454887375762;
			m_weights[21]=.067210613600678175862374;
			m_weights[22]=.067518685849036458820214;
			break;
		case 47:
			m_abscissas[0]=.998718728584212109183948;
			m_abscissas[1]=.99325521098776863469219;
			m_abscissas[2]=.983451003071623708764901;
			m_abscissas[3]=.969346787326564497145744;
			m_abscissas[4]=.951003969257708442589803;
			m_abscissas[5]=.928502693012360648197185;
			m_abscissas[6]=.901941329438525356867171;
			m_abscissas[7]=.871436015796896316940985;
			m_abscissas[8]=.837120139899902121277599;
			m_abscissas[9]=.799143754167741942915562;
			m_abscissas[10]=.757672918445438633573997;
			m_abscissas[11]=.712888973409064301661879;
			m_abscissas[12]=.664987747390332729136703;
			m_abscissas[13]=.614178699956373608595463;
			m_abscissas[14]=.560684005934664194483058;
			m_abscissas[15]=.504737583863577919774289;
			m_abscissas[16]=.446584073104855702725182;
			m_abscissas[17]=.386477764084667139583302;
			m_abscissas[18]=.324681486337735902210761;
			m_abscissas[19]=.261465459214974570306983;
			m_abscissas[20]=.197106110279111807961491;
			m_abscissas[21]=.131884866554514897053824;
			m_abscissas[22]=.066086923916355675160474;
			m_abscissas[23]=0.;
			m_weights[0]=.003287453842528014883250;
			m_weights[1]=.007638616295848833614104;
			m_weights[2]=.011962848464312320963943;
			m_weights[3]=.016235333146433059670726;
			m_weights[4]=.020436938147668427642035;
			m_weights[5]=.024549211659658818537834;
			m_weights[6]=.028554150700643386504740;
			m_weights[7]=.032434235515184756767617;
			m_weights[8]=.036172496584174951613460;
			m_weights[9]=.039752586122531003780901;
			m_weights[10]=.043158848648479538268302;
			m_weights[11]=.046376389086505911204402;
			m_weights[12]=.049391137747361169604570;
			m_weights[13]=.052189911780057144872212;
			m_weights[14]=.054760472781530225957125;
			m_weights[15]=.057091580293231540222017;
			m_weights[16]=.059173040942338875976155;
			m_weights[17]=.060995753008739645330711;
			m_weights[18]=.062551746220921662640564;
			m_weights[19]=.063834216605717030631294;
			m_weights[20]=.064837556238945726702604;
			m_weights[21]=.065557377766549740251143;
			m_weights[22]=.065990533588810474533571;
			m_weights[23]=.066135129623655479653440;
			break;
		case 48:
			m_abscissas[0]=.998771007252426118600541;
			m_abscissas[1]=.993530172266350757547929;
			m_abscissas[2]=.984124583722826857744584;
			m_abscissas[3]=.970591592546247250461412;
			m_abscissas[4]=.952987703160430860722961;
			m_abscissas[5]=.931386690706554333114174;
			m_abscissas[6]=.905879136715569672822075;
			m_abscissas[7]=.876572020274247885905694;
			m_abscissas[8]=.84358826162439353071109;
			m_abscissas[9]=.807066204029442627082553;
			m_abscissas[10]=.767159032515740339253855;
			m_abscissas[11]=.724034130923814654674482;
			m_abscissas[12]=.677872379632663905211851;
			m_abscissas[13]=.628867396776513623995165;
			m_abscissas[14]=.577224726083972703817809;
			m_abscissas[15]=.523160974722233033678226;
			m_abscissas[16]=.466902904750958404544929;
			m_abscissas[17]=.408686481990716729916225;
			m_abscissas[18]=.348755886292160738159818;
			m_abscissas[19]=.287362487355455576735886;
			m_abscissas[20]=.224763790394689061224865;
			m_abscissas[21]=.161222356068891718056437;
			m_abscissas[22]=.097004699209462698930054;
			m_abscissas[23]=.032380170962869362033322;
			m_weights[0]=.003153346052305838632677;
			m_weights[1]=.007327553901276262102384;
			m_weights[2]=.011477234579234539489593;
			m_weights[3]=.015579315722943848728177;
			m_weights[4]=.019616160457355527814461;
			m_weights[5]=.023570760839324379140519;
			m_weights[6]=.027426509708356948200074;
			m_weights[7]=.031167227832798088902066;
			m_weights[8]=.034777222564770438892549;
			m_weights[9]=.038241351065830706317217;
			m_weights[10]=.041545082943464749214059;
			m_weights[11]=.044674560856694280419449;
			m_weights[12]=.047616658492490474825907;
			m_weights[13]=.050359035553854474957808;
			m_weights[14]=.052890189485193667095505;
			m_weights[15]=.055199503699984162868203;
			m_weights[16]=.057277292100403215705150;
			m_weights[17]=.059114839698395635746475;
			m_weights[18]=.060704439165893880052969;
			m_weights[19]=.062039423159892663904198;
			m_weights[20]=.063114192286254025657126;
			m_weights[21]=.063924238584648186623906;
			m_weights[22]=.064466164435950082206504;
			m_weights[23]=.064737696812683922503025;
			break;
		case 49:
			m_abscissas[0]=.998820150606635379361831;
			m_abscissas[1]=.993788661944167790760114;
			m_abscissas[2]=.984757895914213004359299;
			m_abscissas[3]=.971762200901555380139972;
			m_abscissas[4]=.954853658674137233555244;
			m_abscissas[5]=.934100294755810149058982;
			m_abscissas[6]=.90958565582807328521302;
			m_abscissas[7]=.881408445573008910037032;
			m_abscissas[8]=.849682119844165701034882;
			m_abscissas[9]=.814534427359855431539501;
			m_abscissas[10]=.776106894345446635018143;
			m_abscissas[11]=.734554254237402696213674;
			m_abscissas[12]=.690043824425132113504752;
			m_abscissas[13]=.642754832419237664056857;
			m_abscissas[14]=.592877694108900712455864;
			m_abscissas[15]=.540613246991726066558225;
			m_abscissas[16]=.486171941452492042176976;
			m_abscissas[17]=.429772993341576524658584;
			m_abscissas[18]=.371643501262284888863735;
			m_abscissas[19]=.312017532119748762207861;
			m_abscissas[20]=.251135178612577273507156;
			m_abscissas[21]=.18924159246181358648531;
			m_abscissas[22]=.126585997269672051067985;
			m_abscissas[23]=.063420684982686786028835;
			m_abscissas[24]=0.;
			m_weights[0]=.003027278988922905077481;
			m_weights[1]=.007035099590086451473451;
			m_weights[2]=.011020551031593580497508;
			m_weights[3]=.014962144935624651029584;
			m_weights[4]=.018843595853089458444451;
			m_weights[5]=.022649201587446676498771;
			m_weights[6]=.026363618927066016960946;
			m_weights[7]=.029971884620583825350691;
			m_weights[8]=.033459466791622174342487;
			m_weights[9]=.036812320963000689819467;
			m_weights[10]=.040016945766373021368605;
			m_weights[11]=.043060436981259597988345;
			m_weights[12]=.045930539355595853542500;
			m_weights[13]=.048615695887828240277651;
			m_weights[14]=.051105094330144590674623;
			m_weights[15]=.053388710708258968527943;
			m_weights[16]=.055457349674803588690432;
			m_weights[17]=.057302681530187475485165;
			m_weights[18]=.058917275760027266024528;
			m_weights[19]=.060294630953152017303106;
			m_weights[20]=.061429200979192936296827;
			m_weights[21]=.062316417320057267401077;
			m_weights[22]=.062952707465195699474400;
			m_weights[23]=.063335509296491748590837;
			m_weights[24]=.063463281404790597718253;
			break;
		case 50:
			m_abscissas[0]=.998866404420071050185459;
			m_abscissas[1]=.994031969432090712585108;
			m_abscissas[2]=.98535408404800588230901;
			m_abscissas[3]=.972864385106692073713344;
			m_abscissas[4]=.956610955242807942997746;
			m_abscissas[5]=.936656618944877933780875;
			m_abscissas[6]=.913078556655791893089736;
			m_abscissas[7]=.885967979523613048637541;
			m_abscissas[8]=.855429769429946084611363;
			m_abscissas[9]=.821582070859335948356254;
			m_abscissas[10]=.784555832900399263905305;
			m_abscissas[11]=.744494302226068538260536;
			m_abscissas[12]=.701552468706822251089546;
			m_abscissas[13]=.655896465685439360781625;
			m_abscissas[14]=.607702927184950239180382;
			m_abscissas[15]=.557158304514650054315523;
			m_abscissas[16]=.504458144907464201651459;
			m_abscissas[17]=.449806334974038789147131;
			m_abscissas[18]=.393414311897565127394229;
			m_abscissas[19]=.335500245419437356836988;
			m_abscissas[20]=.276288193779531990327645;
			m_abscissas[21]=.216007236876041756847285;
			m_abscissas[22]=.154890589998145902071629;
			m_abscissas[23]=.093174701560086140854450;
			m_abscissas[24]=.031098338327188876112329;
			m_weights[0]=.002908622553155140958401;
			m_weights[1]=.006759799195745401502779;
			m_weights[2]=.010590548383650969263570;
			m_weights[3]=.014380822761485574419379;
			m_weights[4]=.018115560713489390351260;
			m_weights[5]=.021780243170124792981592;
			m_weights[6]=.025360673570012390440195;
			m_weights[7]=.028842993580535198029906;
			m_weights[8]=.032213728223578016648166;
			m_weights[9]=.035459835615146154160735;
			m_weights[10]=.038568756612587675244770;
			m_weights[11]=.041528463090147697422412;
			m_weights[12]=.044327504338803275492022;
			m_weights[13]=.046955051303948432965633;
			m_weights[14]=.049400938449466314921244;
			m_weights[15]=.051655703069581138489905;
			m_weights[16]=.053710621888996246523459;
			m_weights[17]=.055557744806212517623567;
			m_weights[18]=.057189925647728383723029;
			m_weights[19]=.058600849813222445835122;
			m_weights[20]=.059785058704265457509576;
			m_weights[21]=.06073797084177021603175;
			m_weights[22]=.061455899590316663756407;
			m_weights[23]=.061936067420683243384088;
			m_weights[24]=.062176616655347262321033;
			break;
		case 51:
			m_abscissas[0]=.998909990848903495168996;
			m_abscissas[1]=.994261260436752574621085;
			m_abscissas[2]=.985915991735902996583886;
			m_abscissas[3]=.973903368019323867231755;
			m_abscissas[4]=.958267848613908194557707;
			m_abscissas[5]=.939067544002962383435368;
			m_abscissas[6]=.916373862309780230823571;
			m_abscissas[7]=.890271218029527303277795;
			m_abscissas[8]=.860856711182292371473496;
			m_abscissas[9]=.828239763823064832854818;
			m_abscissas[10]=.792541712099381205234411;
			m_abscissas[11]=.75389535448537552576396;
			m_abscissas[12]=.712444457577036644580525;
			m_abscissas[13]=.66834322117537008686446;
			m_abscissas[14]=.621755704600723273755043;
			m_abscissas[15]=.572855216351303836522395;
			m_abscissas[16]=.521823669366185842514088;
			m_abscissas[17]=.468850904286041063610457;
			m_abscissas[18]=.414133983226303877936872;
			m_abscissas[19]=.357876456688409509775201;
			m_abscissas[20]=.300287606335331939530246;
			m_abscissas[21]=.241581666447798703846733;
			m_abscissas[22]=.181977026957077545323999;
			m_abscissas[23]=.12169542101888876696382;
			m_abscissas[24]=.060961100150578724734195;
			m_abscissas[25]=0.;
			m_weights[0]=.002796807171089895575544;
			m_weights[1]=.006500337783252600292109;
			m_weights[2]=.010185191297821729939238;
			m_weights[3]=.013832634006477822296688;
			m_weights[4]=.017428714723401052259504;
			m_weights[5]=.020959988401703210579793;
			m_weights[6]=.024413300573781434273142;
			m_weights[7]=.027775798594162477195996;
			m_weights[8]=.031034971290160008454426;
			m_weights[9]=.034178693204188336236209;
			m_weights[10]=.037195268923260292842908;
			m_weights[11]=.040073476285496453186809;
			m_weights[12]=.042802607997880086653610;
			m_weights[13]=.045372511407650068748167;
			m_weights[14]=.047773626240623101999995;
			m_weights[15]=.049997020150057409779549;
			m_weights[16]=.052034421936697087564136;
			m_weights[17]=.053878252313045561434099;
			m_weights[18]=.055521652095738693016737;
			m_weights[19]=.056958507720258662100078;
			m_weights[20]=.058183473982592140598438;
			m_weights[21]=.059191993922961543783539;
			m_weights[22]=.059980315777503252090064;
			m_weights[23]=.060545506934737795138125;
			m_weights[24]=.060885464844856343881199;
			m_weights[25]=.060998924841205880159798;
			break;
		case 52:
			m_abscissas[0]=.998951111103950278090961;
			m_abscissas[1]=.994477590929216029245043;
			m_abscissas[2]=.986446195651549840645323;
			m_abscissas[3]=.974883884221744503140706;
			m_abscissas[4]=.959831826933086552532025;
			m_abscissas[5]=.941343853641359056843579;
			m_abscissas[6]=.919486128916424539893596;
			m_abscissas[7]=.894336890534495322520844;
			m_abscissas[8]=.865986162846067585244359;
			m_abscissas[9]=.834535432326734534961726;
			m_abscissas[10]=.800097283430468324334759;
			m_abscissas[11]=.762794995193744960279211;
			m_abscissas[12]=.722762099749983193676817;
			m_abscissas[13]=.680141904227167702092239;
			m_abscissas[14]=.635086977695245924298125;
			m_abscissas[15]=.587758604979579069902026;
			m_abscissas[16]=.538326209285827438375951;
			m_abscissas[17]=.486966745698096077782458;
			m_abscissas[18]=.433864067718761670309087;
			m_abscissas[19]=.379208269116093669246682;
			m_abscissas[20]=.323195003434807825501099;
			m_abscissas[21]=.266024783605001827472972;
			m_abscissas[22]=.207902264156366059686466;
			m_abscissas[23]=.149035508606949180488634;
			m_abscissas[24]=.089635244648900565488855;
			m_abscissas[25]=.029914109797338766043681;
			m_weights[0]=.002691316950047111118952;
			m_weights[1]=.006255523962973276899716;
			m_weights[2]=.009802634579462752061951;
			m_weights[3]=.013315114982340960656600;
			m_weights[4]=.016780023396300735677922;
			m_weights[5]=.020184891507980792202989;
			m_weights[6]=.023517513553984461590323;
			m_weights[7]=.026765953746504013449493;
			m_weights[8]=.029918581147143946641282;
			m_weights[9]=.03296410908971879791501;
			m_weights[10]=.035891634835097232941942;
			m_weights[11]=.038690678310423978985101;
			m_weights[12]=.041351219500560271679040;
			m_weights[13]=.043863734259000407995130;
			m_weights[14]=.046219228372784793507646;
			m_weights[15]=.048409269744074896853960;
			m_weights[16]=.050426018566342377218211;
			m_weights[17]=.052262255383906993034394;
			m_weights[18]=.053911406932757264750837;
			m_weights[19]=.055367569669302652549041;
			m_weights[20]=.056625530902368597190808;
			m_weights[21]=.057680787452526827653932;
			m_weights[22]=.058529561771813868550291;
			m_weights[23]=.059168815466042970369332;
			m_weights[24]=.059596260171248158258311;
			m_weights[25]=.059810365745291860247785;
			break;
		case 53:
			m_abscissas[0]=.998989947776328227121054;
			m_abscissas[1]=.994681919308007078636036;
			m_abscissas[2]=.986947035023371521720373;
			m_abscissas[3]=.975810233714984581632984;
			m_abscissas[4]=.961309694623136332369858;
			m_abscissas[5]=.943495353464441879021071;
			m_abscissas[6]=.922428603042812128267819;
			m_abscissas[7]=.89818205787542662592552;
			m_abscissas[8]=.8708392975582413516017;
			m_abscissas[9]=.840494576545801375430128;
			m_abscissas[10]=.807252498416895478220486;
			m_abscissas[11]=.771227654925532307866036;
			m_abscissas[12]=.732544230807510253782136;
			m_abscissas[13]=.691335575601366723540749;
			m_abscissas[14]=.647743743916510068751106;
			m_abscissas[15]=.601919005713769327464294;
			m_abscissas[16]=.554019328277067881014735;
			m_abscissas[17]=.504209831657133437039117;
			m_abscissas[18]=.452662219461845791383145;
			m_abscissas[19]=.399554186953952977392563;
			m_abscissas[20]=.345068808495722356694311;
			m_abscissas[21]=.289393906451626206426547;
			m_abscissas[22]=.232721403724272593643364;
			m_abscissas[23]=.175246662155325750729791;
			m_abscissas[24]=.117167809071955150140257;
			m_abscissas[25]=.058685054300259465022712;
			m_abscissas[26]=0.;
			m_weights[0]=.002591683720567031811599;
			m_weights[1]=.006024276226948673281242;
			m_weights[2]=.009441202284940344386662;
			m_weights[3]=.012826026144240379179150;
			m_weights[4]=.016166725256687463928061;
			m_weights[5]=.019451721107636895388047;
			m_weights[6]=.022669673057070208398789;
			m_weights[7]=.025809482510757517713962;
			m_weights[8]=.028860323617823736262800;
			m_weights[9]=.031811678459019323063236;
			m_weights[10]=.034653372583534237958388;
			m_weights[11]=.037375609803482915674172;
			m_weights[12]=.039969005843540382127094;
			m_weights[13]=.042424620634520013592282;
			m_weights[14]=.044733989103672810212766;
			m_weights[15]=.046889150340750314021873;
			m_weights[16]=.048882675032699140420448;
			m_weights[17]=.050707691069292715296485;
			m_weights[18]=.052357907229872718199702;
			m_weights[19]=.053827634868731029042081;
			m_weights[20]=.055111807523933599002350;
			m_weights[21]=.056205998381739709808655;
			m_weights[22]=.057106435536267191773383;
			m_weights[23]=.057810014991713196319683;
			m_weights[24]=.058314311362256007556276;
			m_weights[25]=.058617586232720263318072;
			m_weights[26]=.058718794151164364525487;
			break;
		case 54:
			m_abscissas[0]=.999026666867340983851071;
			m_abscissas[1]=.99487511701833888491889;
			m_abscissas[2]=.987420637397343558552146;
			m_abscissas[3]=.97668632885790323720001;
			m_abscissas[4]=.962707645785923583256994;
			m_abscissas[5]=.945530975164995853763837;
			m_abscissas[6]=.925213359866651486256276;
			m_abscissas[7]=.901822286284701580757467;
			m_abscissas[8]=.875435454065568939417909;
			m_abscissas[9]=.846140515970772949425888;
			m_abscissas[10]=.814034785913567835469638;
			m_abscissas[11]=.779224915346254021535949;
			m_abscissas[12]=.741826538809184316285766;
			m_abscissas[13]=.701963889719172919385571;
			m_abscissas[14]=.65976938763198312469229;
			m_abscissas[15]=.615383198331127370729938;
			m_abscissas[16]=.568952768195209429731625;
			m_abscissas[17]=.520632334385933073327017;
			m_abscissas[18]=.470582412481382283683212;
			m_abscissas[19]=.418969263255204528036103;
			m_abscissas[20]=.365964340372191181984332;
			m_abscissas[21]=.31174372083446822888255;
			m_abscissas[22]=.256487520069997300077411;
			m_abscissas[23]=.200379293606213569778602;
			m_abscissas[24]=.14360542731625615394709;
			m_abscissas[25]=.086354518263248215285443;
			m_abscissas[26]=.028816748199341777656236;
			m_weights[0]=.002497481835761585775946;
			m_weights[1]=.005805611015239984878826;
			m_weights[2]=.009099369455509396948034;
			m_weights[3]=.012363328128847644166468;
			m_weights[4]=.015586303035924131702969;
			m_weights[5]=.018757527621469377912008;
			m_weights[6]=.021866451422853085945510;
			m_weights[7]=.024902741467208773050055;
			m_weights[8]=.027856309310595870287002;
			m_weights[9]=.030717342497870676054005;
			m_weights[10]=.033476336464372645716040;
			m_weights[11]=.036124125840383552582887;
			m_weights[12]=.038651914782102516836857;
			m_weights[13]=.041051306136644974221718;
			m_weights[14]=.043314329309597015441926;
			m_weights[15]=.045433466728276713974852;
			m_weights[16]=.047401678806444991058576;
			m_weights[17]=.049212427324528886068791;
			m_weights[18]=.050859697146188144319709;
			m_weights[19]=.052338016198298744665589;
			m_weights[20]=.053642473647553611272101;
			m_weights[21]=.054768736213057986306223;
			m_weights[22]=.055713062560589987683370;
			m_weights[23]=.056472315730625965031044;
			m_weights[24]=.057043973558794598567828;
			m_weights[25]=.057426137054112114859290;
			m_weights[26]=.057617536707147024672376;
			break;
		case 55:
			m_abscissas[0]=.999061419564818541479432;
			m_abscissas[1]=.995057977847411875043189;
			m_abscissas[2]=.987868941198889198522386;
			m_abscissas[3]=.977515735503989208859217;
			m_abscissas[4]=.964031328593135198778537;
			m_abscissas[5]=.94745886804121074185998;
			m_abscissas[6]=.927851424720791696816271;
			m_abscissas[7]=.905271800744000025781911;
			m_abscissas[8]=.879792322419895506067591;
			m_abscissas[9]=.851494606617154471460245;
			m_abscissas[10]=.820469298559320912454241;
			m_abscissas[11]=.786815781127622365897636;
			m_abscissas[12]=.750641856348021908674988;
			m_abscissas[13]=.71206339998663783890935;
			m_abscissas[14]=.671203990319826395796327;
			m_abscissas[15]=.628194512249928140090636;
			m_abscissas[16]=.583172738026032102974305;
			m_abscissas[17]=.5362828859083432967206;
			m_abscissas[18]=.487675158187474097208264;
			m_abscissas[19]=.437505260037174591808481;
			m_abscissas[20]=.385933900740979429755849;
			m_abscissas[21]=.333126278890023885189471;
			m_abscissas[22]=.279251553200806538549862;
			m_abscissas[23]=.224482300647845483400189;
			m_abscissas[24]=.168993963646873208282842;
			m_abscissas[25]=.112964288059329266588438;
			m_abscissas[26]=.056572753818336776327338;
			m_abscissas[27]=0.;
			m_weights[0]=.002408323619979788819170;
			m_weights[1]=.005598632266560767354083;
			m_weights[2]=.008775746107058528177390;
			m_weights[3]=.011925160719848612170753;
			m_weights[4]=.015036458333511788213151;
			m_weights[5]=.018099614520729062407968;
			m_weights[6]=.021104801668016454120209;
			m_weights[7]=.024042388009725622007792;
			m_weights[8]=.02690296145639627066712;
			m_weights[9]=.029677357765161041221297;
			m_weights[10]=.032356689226185831684706;
			m_weights[11]=.034932372873589887407268;
			m_weights[12]=.037396157867965545282916;
			m_weights[13]=.039740151874337179609464;
			m_weights[14]=.041956846317718762395207;
			m_weights[15]=.044039140421606589895168;
			m_weights[16]=.045980363946283838103905;
			m_weights[17]=.047774298551200695550037;
			m_weights[18]=.049415197711551739480759;
			m_weights[19]=.050897805124493979224775;
			m_weights[20]=.052217371545632084564394;
			m_weights[21]=.053369670001605472723571;
			m_weights[22]=.054351009329911102070322;
			m_weights[23]=.055158246002508687596651;
			m_weights[24]=.055788794195284087102936;
			m_weights[25]=.056240634071084368028279;
			m_weights[26]=.056512318249772001400658;
			m_weights[27]=.056602976444560425440106;
			break;
		case 56:
			m_abscissas[0]=.999094343801465584353154;
			m_abscissas[1]=.995231226081069747216309;
			m_abscissas[2]=.988293715540161511089925;
			m_abscissas[3]=.978301709140256383376991;
			m_abscissas[4]=.965285901905490183626195;
			m_abscissas[5]=.949286479561962635646738;
			m_abscissas[6]=.930352880247496300547273;
			m_abscissas[7]=.908543620420655490846107;
			m_abscissas[8]=.883926108327827540789012;
			m_abscissas[9]=.856576433762748635402991;
			m_abscissas[10]=.826579132142881651672127;
			m_abscissas[11]=.794026922893866498030008;
			m_abscissas[12]=.759020422705128902202413;
			m_abscissas[13]=.721667834450188083522594;
			m_abscissas[14]=.682084612694470455501561;
			m_abscissas[15]=.640393106807006894267939;
			m_abscissas[16]=.596722182770663320104135;
			m_abscissas[17]=.551206824855534618754364;
			m_abscissas[18]=.503987718384381714195224;
			m_abscissas[19]=.455210814878459578948832;
			m_abscissas[20]=.405026880927091278118867;
			m_abscissas[21]=.353591032174954520969707;
			m_abscissas[22]=.301062253867220669053094;
			m_abscissas[23]=.247602909434337203972967;
			m_abscissas[24]=.193378238635275258240185;
			m_abscissas[25]=.138555846810376242012887;
			m_abscissas[26]=.083305186822435374440287;
			m_abscissas[27]=.027797035287275437094061;
			m_weights[0]=.002323855375773215501102;
			m_weights[1]=.005402522246015337761311;
			m_weights[2]=.008469063163307887661627;
			m_weights[3]=.011509824340383382173773;
			m_weights[4]=.014515089278021471807771;
			m_weights[5]=.017475512911400946504959;
			m_weights[6]=.020381929882402572634806;
			m_weights[7]=.023225351562565316937258;
			m_weights[8]=.025996987058391952191819;
			m_weights[9]=.028688268473822741729886;
			m_weights[10]=.031290876747310447867835;
			m_weights[11]=.033796767115611761295427;
			m_weights[12]=.036198193872315186035884;
			m_weights[13]=.038487734259247662486826;
			m_weights[14]=.040658311384744517880125;
			m_weights[15]=.042703216084667086511039;
			m_weights[16]=.044616127652692283213415;
			m_weights[17]=.046391133373001896762190;
			m_weights[18]=.048022746793600258120736;
			m_weights[19]=.049505924683047578919966;
			m_weights[20]=.050836082617798480560124;
			m_weights[21]=.052009109151741399843052;
			m_weights[22]=.053021378524010763967992;
			m_weights[23]=.053869761865714485708954;
			m_weights[24]=.054551636870889421061751;
			m_weights[25]=.055064895901762425796305;
			m_weights[26]=.055407952503245123217793;
			m_weights[27]=.055579746306514395846273;
			break;
		case 57:
			m_abscissas[0]=.99912556562526285057201;
			m_abscissas[1]=.995395523678430311135125;
			m_abscissas[2]=.988696577650222048849758;
			m_abscissas[3]=.979047226709468713798821;
			m_abscissas[4]=.966476085171886679114859;
			m_abscissas[5]=.951020626447876741911904;
			m_abscissas[6]=.932726961067101696100754;
			m_abscissas[7]=.911649678521391212728431;
			m_abscissas[8]=.887851678882221329513;
			m_abscissas[9]=.861403983262046944722242;
			m_abscissas[10]=.832385521150439120829011;
			m_abscissas[11]=.800882894547218242076009;
			m_abscissas[12]=.766990119359450195491716;
			m_abscissas[13]=.730808344744523322827345;
			m_abscissas[14]=.692445551199517739040484;
			m_abscissas[15]=.652016228280976891249362;
			m_abscissas[16]=.609641032908715365423849;
			m_abscissas[17]=.565446429269236759018986;
			m_abscissas[18]=.519564311391187606315154;
			m_abscissas[19]=.47213160951797570958806;
			m_abscissas[20]=.423289881451563950960024;
			m_abscissas[21]=.373184890086594458552134;
			m_abscissas[22]=.321966168395378640590208;
			m_abscissas[23]=.269786573161838765763346;
			m_abscissas[24]=.216801828796124036413614;
			m_abscissas[25]=.16317006259126425104328;
			m_abscissas[26]=.109051332808787800978527;
			m_abscissas[27]=.054607151001646824219831;
			m_abscissas[28]=0.;
			m_weights[0]=.002243753872250662909735;
			m_weights[1]=.005216533474718779390507;
			m_weights[2]=.008178160067821232626211;
			m_weights[3]=.011115763732335990145675;
			m_weights[4]=.014020270790753556170247;
			m_weights[5]=.016882959023441549035001;
			m_weights[6]=.019695270699488520382422;
			m_weights[7]=.022448807890776438079689;
			m_weights[8]=.025135350990918122647274;
			m_weights[9]=.027746881402180192321258;
			m_weights[10]=.030275604842693999458491;
			m_weights[11]=.032713974366371568542490;
			m_weights[12]=.035054712782312617505751;
			m_weights[13]=.037290834324417317354735;
			m_weights[14]=.039415665475480114089953;
			m_weights[15]=.041422864870801110363197;
			m_weights[16]=.043306442216215196596432;
			m_weights[17]=.045060776161381157797214;
			m_weights[18]=.046680631073641503783841;
			m_weights[19]=.048161172661687751268851;
			m_weights[20]=.049497982402019678993838;
			m_weights[21]=.050687070724927408656640;
			m_weights[22]=.051724888920517824720624;
			m_weights[23]=.052608339729177432440231;
			m_weights[24]=.053334786584819158426577;
			m_weights[25]=.053902061483298574642809;
			m_weights[26]=.054308471452498643138747;
			m_weights[27]=.054552803604761886480139;
			m_weights[28]=.054634328756584024062841;
			break;
		case 58:
			m_abscissas[0]=.999155200407386606442737;
			m_abscissas[1]=.995551476597290902602772;
			m_abscissas[2]=.98907900824844263649991;
			m_abscissas[3]=.979755014694350309107816;
			m_abscissas[4]=.967606202502924090153301;
			m_abscissas[5]=.952667557518869091442846;
			m_abscissas[6]=.934982137588259348480843;
			m_abscissas[7]=.914600928564352540686506;
			m_abscissas[8]=.891582692022030176399883;
			m_abscissas[9]=.865993794074807479275071;
			m_abscissas[10]=.837908013339373316352098;
			m_abscissas[11]=.807406327913088141048997;
			m_abscissas[12]=.77457668174965274526614;
			m_abscissas[13]=.739513731020042267846617;
			m_abscissas[14]=.702318571153908113479757;
			m_abscissas[15]=.663098445332125266433016;
			m_abscissas[16]=.621966435263079111033984;
			m_abscissas[17]=.579041135130225030489977;
			m_abscissas[18]=.534446309648847586399802;
			m_abscissas[19]=.488310537216718463615553;
			m_abscissas[20]=.440766839186839565193703;
			m_abscissas[21]=.391952296330753150371217;
			m_abscissas[22]=.342007653597995261248368;
			m_abscissas[23]=.291076914311109189533025;
			m_abscissas[24]=.239306924966153454428949;
			m_abscissas[25]=.186846951835761321374384;
			m_abscissas[26]=.133848250595466857022377;
			m_abscissas[27]=.080463630214142729309848;
			m_abscissas[28]=.026847012365942355803339;
			m_weights[0]=.002167723249627449943054;
			m_weights[1]=.005039981612650243085018;
			m_weights[2]=.007901973849998674754017;
			m_weights[3]=.010741553532878774116856;
			m_weights[4]=.013550237112988812145179;
			m_weights[5]=.016319874234970965052121;
			m_weights[6]=.019042465461893408655788;
			m_weights[7]=.021710156140146235766917;
			m_weights[8]=.024315252724963952540259;
			m_weights[9]=.026850243181981868475907;
			m_weights[10]=.029307818044160490718394;
			m_weights[11]=.031680891253809327320292;
			m_weights[12]=.033962620493416010797728;
			m_weights[13]=.036146426867087270540781;
			m_weights[14]=.038226013845858433229459;
			m_weights[15]=.040195385409867796888077;
			m_weights[16]=.042048863329582125994570;
			m_weights[17]=.043781103533640251039026;
			m_weights[18]=.045387111514819802503981;
			m_weights[19]=.046862256729026346918418;
			m_weights[20]=.048202285945417748406570;
			m_weights[21]=.049403335508962392866511;
			m_weights[22]=.050461942479953125297660;
			m_weights[23]=.051375054618285725474515;
			m_weights[24]=.052140039183669818971261;
			m_weights[25]=.052754690526370833429646;
			m_weights[26]=.053217236446579014103481;
			m_weights[27]=.053526343304058252100611;
			m_weights[28]=.053681119863334848863906;
			break;
		case 59:
			m_abscissas[0]=.999183353909294683756424;
			m_abscissas[1]=.995699640383245964687202;
			m_abscissas[2]=.989442365133730931782184;
			m_abscissas[3]=.980427573956715688449608;
			m_abscissas[4]=.968680221681781531353633;
			m_abscissas[5]=.954233009376951055863233;
			m_abscissas[6]=.937126190353453859405138;
			m_abscissas[7]=.917407438788155281352541;
			m_abscissas[8]=.895131711743472085364297;
			m_abscissas[9]=.870361094292882260963426;
			m_abscissas[10]=.843164625816872201470351;
			m_abscissas[11]=.81361810728821157143509;
			m_abscissas[12]=.781803889862360905632679;
			m_abscissas[13]=.747810645278640231889258;
			m_abscissas[14]=.711733118677197731595916;
			m_abscissas[15]=.673671864504937227024027;
			m_abscissas[16]=.633732966238850097512581;
			m_abscissas[17]=.592027740704030144464176;
			m_abscissas[18]=.548672427808396384372256;
			m_abscissas[19]=.503787866557717978768087;
			m_abscissas[20]=.457499158253266690226215;
			m_abscissas[21]=.409935317810418966722824;
			m_abscissas[22]=.361228914169794809992312;
			m_abscissas[23]=.31151570080301370031836;
			m_abscissas[24]=.260934237342811711610616;
			m_abscissas[25]=.209625503392036544923367;
			m_abscissas[26]=.157732505587857968115218;
			m_abscissas[27]=.105399879016344143836914;
			m_abscissas[28]=.052773484088310003951680;
			m_abscissas[29]=0.;
			m_weights[0]=.002095492284541223402701;
			m_weights[1]=.004872239168265284768582;
			m_weights[2]=.007639529453487575142701;
			m_weights[3]=.010385885500995862193798;
			m_weights[4]=.013103366306345191018320;
			m_weights[5]=.015784347313081466147320;
			m_weights[6]=.018421342753610029360616;
			m_weights[7]=.021006998288437187350461;
			m_weights[8]=.023534105393713363425275;
			m_weights[9]=.025995619731298500186651;
			m_weights[10]=.028384680200534797905153;
			m_weights[11]=.030694627836111683239750;
			m_weights[12]=.032919024271045277757511;
			m_weights[13]=.035051669636400108783718;
			m_weights[14]=.037086619818870922691838;
			m_weights[15]=.039018203016160009503031;
			m_weights[16]=.040841035538686707660202;
			m_weights[17]=.042550036811067638667308;
			m_weights[18]=.044140443530297380690798;
			m_weights[19]=.045607822940509769831868;
			m_weights[20]=.046948085186962019193160;
			m_weights[21]=.048157494714606440388147;
			m_weights[22]=.049232680679361985779694;
			m_weights[23]=.050170646342996902810720;
			m_weights[24]=.050968777425393916850248;
			m_weights[25]=.051624849390891482146440;
			m_weights[26]=.052137033648375391383987;
			m_weights[27]=.052503902647828739050941;
			m_weights[28]=.052724433859127931961304;
			m_weights[29]=.052798012621990421415512;
			break;
		case 60:
			m_abscissas[0]=.99921012322743602203423;
			m_abscissas[1]=.995840525118838173876747;
			m_abscissas[2]=.989787895222221717367279;
			m_abscissas[3]=.981067201752598185618577;
			m_abscissas[4]=.969701788765052733721544;
			m_abscissas[5]=.955722255839996107397232;
			m_abscissas[6]=.939166276116423249495419;
			m_abscissas[7]=.920078476177627552856657;
			m_abscissas[8]=.898510310810045941937789;
			m_abscissas[9]=.874519922646898315129308;
			m_abscissas[10]=.848171984785929632490515;
			m_abscissas[11]=.819537526162145759368518;
			m_abscissas[12]=.788693739932264054569945;
			m_abscissas[13]=.755723775306585686868842;
			m_abscissas[14]=.720716513355730399436021;
			m_abscissas[15]=.68376632738135543722293;
			m_abscissas[16]=.644972828489477067813448;
			m_abscissas[17]=.604440597048510363444209;
			m_abscissas[18]=.562278900753944539178273;
			m_abscissas[19]=.518601400058569747417889;
			m_abscissas[20]=.473525841761707111108163;
			m_abscissas[21]=.427173741583078389307453;
			m_abscissas[22]=.379670056576797977154953;
			m_abscissas[23]=.331142848268448194252353;
			m_abscissas[24]=.281722937423261691690695;
			m_abscissas[25]=.231543551376029338010345;
			m_abscissas[26]=.180739964873425417240877;
			m_abscissas[27]=.129449135396945003146444;
			m_abscissas[28]=.077809333949536569419286;
			m_abscissas[29]=.025959772301247798589170;
			m_weights[0]=.002026811968873758496432;
			m_weights[1]=.004712729926953568640895;
			m_weights[2]=.007389931163345455531517;
			m_weights[3]=.010047557182287984357886;
			m_weights[4]=.012678166476815960131495;
			m_weights[5]=.015274618596784799306726;
			m_weights[6]=.017829901014207720260396;
			m_weights[7]=.020337120729457286775032;
			m_weights[8]=.022789516943997819863783;
			m_weights[9]=.025180477621521248379571;
			m_weights[10]=.027503556749924791635223;
			m_weights[11]=.029752491500788945240836;
			m_weights[12]=.031921219019296328949459;
			m_weights[13]=.034003892724946422834914;
			m_weights[14]=.035994898051084503066579;
			m_weights[15]=.037888867569243444030941;
			m_weights[16]=.039680695452380799470123;
			m_weights[17]=.041365551235584755613164;
			m_weights[18]=.042938892835935641954231;
			m_weights[19]=.044396478795787113327784;
			m_weights[20]=.045734379716114486647196;
			m_weights[21]=.046948988848912204847013;
			m_weights[22]=.048037031819971180963667;
			m_weights[23]=.048995575455756835389476;
			m_weights[24]=.049822035690550181011159;
			m_weights[25]=.050514184532509374598239;
			m_weights[26]=.051070156069855627404549;
			m_weights[27]=.051488451500980933995044;
			m_weights[28]=.051767943174910187543804;
			m_weights[29]=.051907877631220639732865;
			break;
		case 61:
			m_abscissas[0]=.999235597631363471731862;
			m_abscissas[1]=.995974599815120234268013;
			m_abscissas[2]=.990116745232517050965532;
			m_abscissas[3]=.981676011284037079685173;
			m_abscissas[4]=.970674258833182908247408;
			m_abscissas[5]=.95714015191298409137208;
			m_abscissas[6]=.941108986681361147477544;
			m_abscissas[7]=.922622581382955261257552;
			m_abscissas[8]=.90172916247400117064204;
			m_abscissas[9]=.878483237214881032478948;
			m_abscissas[10]=.852945450847663445564847;
			m_abscissas[11]=.825182428108659950664282;
			m_abscissas[12]=.795266599282359649152049;
			m_abscissas[13]=.763276011172312197145915;
			m_abscissas[14]=.729294123449465109688957;
			m_abscissas[15]=.693409590894491155499184;
			m_abscissas[16]=.655716032095070871699186;
			m_abscissas[17]=.616311785197921724709617;
			m_abscissas[18]=.57529965135083061860037;
			m_abscissas[19]=.532786626502925265638482;
			m_abscissas[20]=.488883622262252118820699;
			m_abscissas[21]=.443705176538531601995589;
			m_abscissas[22]=.397369154725756609178292;
			m_abscissas[23]=.349996442204066834533434;
			m_abscissas[24]=.301710628963030712604487;
			m_abscissas[25]=.252637687169053495833691;
			m_abscissas[26]=.202905642518058499226947;
			m_abscissas[27]=.152644240230815300529507;
			m_abscissas[28]=.101984606562274068957208;
			m_abscissas[29]=.051058906707974349366888;
			m_abscissas[30]=0.;
			m_weights[0]=.001961453361670282671773;
			m_weights[1]=.004560924006012417184539;
			m_weights[2]=.007152354991749089585834;
			m_weights[3]=.009725461830356133736137;
			m_weights[4]=.012273263507812104629278;
			m_weights[5]=.014789065884937914546179;
			m_weights[6]=.017266292987613743594433;
			m_weights[7]=.019698477746101181330518;
			m_weights[8]=.022079273148319044002476;
			m_weights[9]=.024402467187544202915341;
			m_weights[10]=.026661998524150889662811;
			m_weights[11]=.028851972088183401504342;
			m_weights[12]=.030966674368397394824698;
			m_weights[13]=.033000588275907410632724;
			m_weights[14]=.034948407516533351090852;
			m_weights[15]=.036805050423154817384321;
			m_weights[16]=.038565673207008172746152;
			m_weights[17]=.040225682590998247367640;
			m_weights[18]=.041780747790888492066676;
			m_weights[19]=.043226811812496097901044;
			m_weights[20]=.044560102035083488271541;
			m_weights[21]=.045777140053145959371340;
			m_weights[22]=.046874750750809065976429;
			m_weights[23]=.047850070585095607161833;
			m_weights[24]=.04870055505641152608753;
			m_weights[25]=.049423985346735589939969;
			m_weights[26]=.050018474108178253425052;
			m_weights[27]=.050482470386797404648144;
			m_weights[28]=.050814763668818343207701;
			m_weights[29]=.051014487038697263543735;
			m_weights[30]=.051081119440786217977921;
			break;
		case 62:
			m_abscissas[0]=.999259859308777029698408;
			m_abscissas[1]=.996102296316267132885148;
			m_abscissas[2]=.990429971189290352428071;
			m_abscissas[3]=.982255949097236649491103;
			m_abscissas[4]=.971600723371651806447179;
			m_abscissas[5]=.95849117297392709202956;
			m_abscissas[6]=.942960401392328503826365;
			m_abscissas[7]=.9250476356362037552273;
			m_abscissas[8]=.904798122521093465758486;
			m_abscissas[9]=.882263012831897363072687;
			m_abscissas[10]=.857499231512070922818403;
			m_abscissas[11]=.830569333604004851345931;
			m_abscissas[12]=.801541346103976371538506;
			m_abscissas[13]=.770488596055419318990945;
			m_abscissas[14]=.737489525283156749860784;
			m_abscissas[15]=.702627492222297055120973;
			m_abscissas[16]=.665990561335479446997546;
			m_abscissas[17]=.627671280646885180726955;
			m_abscissas[18]=.58776644795308733800215;
			m_abscissas[19]=.5463768663002510958161;
			m_abscissas[20]=.503607089344755955917021;
			m_abscissas[21]=.459565157240113395207144;
			m_abscissas[22]=.414362323717126048129589;
			m_abscissas[23]=.368112775046564529663022;
			m_abscissas[24]=.320933341594194004074102;
			m_abscissas[25]=.272943202696726343188684;
			m_abscissas[26]=.224263585604165531668991;
			m_abscissas[27]=.175017459249015628555687;
			m_abscissas[28]=.125329223615896808619649;
			m_abscissas[29]=.075324395496234332763827;
			m_abscissas[30]=.025129291421820614727088;
			m_weights[0]=.001899205679513690480397;
			m_weights[1]=.004416333456930904813274;
			m_weights[2]=.006926041901830960871703;
			m_weights[3]=.009418579428420387637938;
			m_weights[4]=.011887390117010501944819;
			m_weights[5]=.014326191823806517767403;
			m_weights[6]=.016728811790177316288550;
			m_weights[7]=.019089176658573198732504;
			m_weights[8]=.021401322277669968841179;
			m_weights[9]=.023659407208682792574517;
			m_weights[10]=.025857726954024698027096;
			m_weights[11]=.027990728163314637541238;
			m_weights[12]=.030053022573989870077009;
			m_weights[13]=.032039400581624678106339;
			m_weights[14]=.033944844379410545091117;
			m_weights[15]=.035764540622768141285587;
			m_weights[16]=.037493892582280029985619;
			m_weights[17]=.039128531751963084123311;
			m_weights[18]=.040664328882417440968285;
			m_weights[19]=.042097404410385096643023;
			m_weights[20]=.043424138258047419580069;
			m_weights[21]=.044641178977124414293645;
			m_weights[22]=.045745452214570180777232;
			m_weights[23]=.046734168478415524802207;
			m_weights[24]=.04760483018410123227045;
			m_weights[25]=.048355237963477672834803;
			m_weights[26]=.048983496220517837104851;
			m_weights[27]=.049488017919699292527866;
			m_weights[28]=.049867528594952394244761;
			m_weights[29]=.050121069569043288074804;
			m_weights[30]=.050248000375256281688403;
			break;
		case 63:
			m_abscissas[0]=.999282984029123780378936;
			m_abscissas[1]=.996224012777970108602193;
			m_abscissas[2]=.990728546892189466810895;
			m_abscissas[3]=.982808810593727234862511;
			m_abscissas[4]=.972484034697570022801961;
			m_abscissas[5]=.959779449758941927070354;
			m_abscissas[6]=.944726134041009802966375;
			m_abscissas[7]=.927360920621843205447031;
			m_abscissas[8]=.907726302778531558036953;
			m_abscissas[9]=.885870328507853426290298;
			m_abscissas[10]=.861846482364123719539612;
			m_abscissas[11]=.835713554319502843471808;
			m_abscissas[12]=.807535495773456760051466;
			m_abscissas[13]=.77738126299037233556333;
			m_abscissas[14]=.745324648317847417829322;
			m_abscissas[15]=.711444099584845807851432;
			m_abscissas[16]=.675822528114986090131103;
			m_abscissas[17]=.638547105821365385000307;
			m_abscissas[18]=.599709051877625235739009;
			m_abscissas[19]=.559403409486285013267698;
			m_abscissas[20]=.517728813290033248124478;
			m_abscissas[21]=.474787247994804399922212;
			m_abscissas[22]=.430683798795111600662089;
			m_abscissas[23]=.385526394212247892477615;
			m_abscissas[24]=.339425541974584402468834;
			m_abscissas[25]=.292494058586251440036157;
			m_abscissas[26]=.244846793245953362748405;
			m_abscissas[27]=.196600346791506684557627;
			m_abscissas[28]=.147872786357871968569839;
			m_abscissas[29]=.098783356446945279529704;
			m_abscissas[30]=.049452187116159627234234;
			m_abscissas[31]=0.;
			m_weights[0]=.001839874595577084117092;
			m_weights[1]=.004278508346863761866078;
			m_weights[2]=.006710291765960136251907;
			m_weights[3]=.009125968676326656354059;
			m_weights[4]=.011519376076880041750751;
			m_weights[5]=.013884612616115610824866;
			m_weights[6]=.016215878410338338882284;
			m_weights[7]=.018507464460161270409261;
			m_weights[8]=.020753761258039090775342;
			m_weights[9]=.022949271004889933148942;
			m_weights[10]=.025088620553344986618630;
			m_weights[11]=.027166574359097933225190;
			m_weights[12]=.029178047208280526945552;
			m_weights[13]=.031118116622219817508216;
			m_weights[14]=.032982034883779341765683;
			m_weights[15]=.034765240645355877697181;
			m_weights[16]=.036463370085457289630452;
			m_weights[17]=.038072267584349556763638;
			m_weights[18]=.039587995891544093984808;
			m_weights[19]=.04100684575966639863511;
			m_weights[20]=.042325345020815822982505;
			m_weights[21]=.043540267083027590798964;
			m_weights[22]=.044648638825941395370333;
			m_weights[23]=.045647747876292608685886;
			m_weights[24]=.046535149245383696510395;
			m_weights[25]=.047308671312268919080605;
			m_weights[26]=.047966421137995131411053;
			m_weights[27]=.04850678909788384786409;
			m_weights[28]=.048928452820511989944709;
			m_weights[29]=.049230380423747560785043;
			m_weights[30]=.049411833039918178967040;
			m_weights[31]=.049472366623931020888669;
			break;
		case 64:
			m_abscissas[0]=.999305041735772139456906;
			m_abscissas[1]=.996340116771955279346925;
			m_abscissas[2]=.991013371476744320739382;
			m_abscissas[3]=.983336253884625956931299;
			m_abscissas[4]=.973326827789910963741854;
			m_abscissas[5]=.961008799652053718918614;
			m_abscissas[6]=.946411374858402816062481;
			m_abscissas[7]=.92956917213193957582149;
			m_abscissas[8]=.910522137078502805756381;
			m_abscissas[9]=.889315445995114105853404;
			m_abscissas[10]=.865999398154092819760783;
			m_abscissas[11]=.840629296252580362751692;
			m_abscissas[12]=.813265315122797559741923;
			m_abscissas[13]=.783972358943341407610221;
			m_abscissas[14]=.752819907260531896611864;
			m_abscissas[15]=.71988185017161082684894;
			m_abscissas[16]=.685236313054233242563558;
			m_abscissas[17]=.648965471254657339857761;
			m_abscissas[18]=.611155355172393250248853;
			m_abscissas[19]=.571895646202634034283878;
			m_abscissas[20]=.531279464019894545658014;
			m_abscissas[21]=.489403145707052957478526;
			m_abscissas[22]=.446366017253464087984948;
			m_abscissas[23]=.402270157963991603695767;
			m_abscissas[24]=.357220158337668115950443;
			m_abscissas[25]=.311322871990210956157513;
			m_abscissas[26]=.264687162208767416373964;
			m_abscissas[27]=.217423643740007084149649;
			m_abscissas[28]=.169644420423992818037314;
			m_abscissas[29]=.121462819296120554470376;
			m_abscissas[30]=.072993121787799039449543;
			m_abscissas[31]=.024350292663424432508956;
			m_weights[0]=.001783280721696432947296;
			m_weights[1]=.004147033260562467635288;
			m_weights[2]=.006504457968978362856117;
			m_weights[3]=.008846759826363947723031;
			m_weights[4]=.011168139460131128818590;
			m_weights[5]=.013463047896718642598061;
			m_weights[6]=.015726030476024719321966;
			m_weights[7]=.017951715775697343085045;
			m_weights[8]=.020134823153530209372340;
			m_weights[9]=.022270173808383254159298;
			m_weights[10]=.024352702568710873338178;
			m_weights[11]=.026377469715054658671692;
			m_weights[12]=.028339672614259483227511;
			m_weights[13]=.030234657072402478867974;
			m_weights[14]=.032057928354851553585468;
			m_weights[15]=.033805161837141609391565;
			m_weights[16]=.035472213256882383810693;
			m_weights[17]=.037055128540240046040415;
			m_weights[18]=.038550153178615629128962;
			m_weights[19]=.039953741132720341386657;
			m_weights[20]=.041262563242623528610156;
			m_weights[21]=.042473515123653589007340;
			m_weights[22]=.043583724529323453376828;
			m_weights[23]=.044590558163756563060135;
			m_weights[24]=.045491627927418144479771;
			m_weights[25]=.046284796581314417295953;
			m_weights[26]=.046968182816210017325326;
			m_weights[27]=.047540165714830308662282;
			m_weights[28]=.047999388596458307728126;
			m_weights[29]=.048344762234802957169770;
			m_weights[30]=.048575467441503426934799;
			m_weights[31]=.048690957009139720383365;
			break;
		case 65:
			m_abscissas[0]=.999326097075412877265694;
			m_abscissas[1]=.996450948061849163055795;
			m_abscissas[2]=.991285276176801668721821;
			m_abscissas[3]=.983839812187034941377638;
			m_abscissas[4]=.974131539833551169074968;
			m_abscissas[5]=.962182754718055237711984;
			m_abscissas[6]=.94802092816840750637377;
			m_abscissas[7]=.931678628228749337965677;
			m_abscissas[8]=.913193440542846261736547;
			m_abscissas[9]=.892607880504738931423286;
			m_abscissas[10]=.869969294926407036194132;
			m_abscissas[11]=.84532975289993028394245;
			m_abscissas[12]=.818745925922651453433919;
			m_abscissas[13]=.79027895749212184304738;
			m_abscissas[14]=.759994322441999786873983;
			m_abscissas[15]=.727961676329424679011974;
			m_abscissas[16]=.694254695213991633552622;
			m_abscissas[17]=.658950906193625133040941;
			m_abscissas[18]=.6221315090854002415826;
			m_abscissas[19]=.583881189660487313327154;
			m_abscissas[20]=.544287924862227138545572;
			m_abscissas[21]=.503442780455006882341043;
			m_abscissas[22]=.461439701569145057697834;
			m_abscissas[23]=.418375296623409009264199;
			m_abscissas[24]=.374348615122066012008794;
			m_abscissas[25]=.329460919837486407645287;
			m_abscissas[26]=.283815453902248730617655;
			m_abscissas[27]=.237517203346416806570712;
			m_abscissas[28]=.190672655626142769774912;
			m_abscissas[29]=.14338955469897517113125;
			m_abscissas[30]=.095776653209197505652219;
			m_abscissas[31]=.047943462353171857522530;
			m_abscissas[32]=0.;
			m_weights[0]=.001729258251300250898340;
			m_weights[1]=.004021524172003736347079;
			m_weights[2]=.006307942578971754550189;
			m_weights[3]=.008580148266881459893636;
			m_weights[4]=.010832678789597968621514;
			m_weights[5]=.013060311639994846336168;
			m_weights[6]=.015257912146448310349265;
			m_weights[7]=.017420421997670248495366;
			m_weights[8]=.019542865836750062826837;
			m_weights[9]=.021620361284934062841654;
			m_weights[10]=.023648129691287236698781;
			m_weights[11]=.025621506938037758214085;
			m_weights[12]=.027535954088450343942500;
			m_weights[13]=.029387067789310668062645;
			m_weights[14]=.031170590380189142464432;
			m_weights[15]=.032882419676368574984050;
			m_weights[16]=.034518618398549058625221;
			m_weights[17]=.036075423225565273932166;
			m_weights[18]=.037549253448257709809772;
			m_weights[19]=.038936719204051197616674;
			m_weights[20]=.040234629273005533815446;
			m_weights[21]=.041439998417240293022686;
			m_weights[22]=.042550054246755802719217;
			m_weights[23]=.043562243595800486532285;
			m_weights[24]=.044474238395082974427324;
			m_weights[25]=.045283941026300230657128;
			m_weights[26]=.045989489146651696963893;
			m_weights[27]=.046589259972233498302255;
			m_weights[28]=.047081874010454522246007;
			m_weights[29]=.047466198232885503152644;
			m_weights[30]=.047741348681240621559039;
			m_weights[31]=.047906692500495862031347;
			m_weights[32]=.047961849394466618120708;
			break;
		case 66:
			m_abscissas[0]=.999346209872177882072817;
			m_abscissas[1]=.996556821089552691122168;
			m_abscissas[2]=.991545030383375745531763;
			m_abscissas[3]=.984320904935140132890553;
			m_abscissas[4]=.974900427746285951855898;
			m_abscissas[5]=.963304586859257370964861;
			m_abscissas[6]=.949559246206235339092467;
			m_abscissas[7]=.933695072544792494372632;
			m_abscissas[8]=.915747462945207015950429;
			m_abscissas[9]=.895756464446563938460836;
			m_abscissas[10]=.873766684166944730401005;
			m_abscissas[11]=.849827189536016219462098;
			m_abscissas[12]=.823991398695587611347082;
			m_abscissas[13]=.79631696124582457053273;
			m_abscissas[14]=.766865629575319625869543;
			m_abscissas[15]=.735703121049512799579237;
			m_abscissas[16]=.702898971358874185836259;
			m_abscissas[17]=.668526379350967592604287;
			m_abscissas[18]=.632662043690963169403895;
			m_abscissas[19]=.595385991714138458327934;
			m_abscissas[20]=.556781400851723064234948;
			m_abscissas[21]=.516934413028231513572575;
			m_abscissas[22]=.47593394244423684296193;
			m_abscissas[23]=.433871477173370668357696;
			m_abscissas[24]=.390840875016182949755122;
			m_abscissas[25]=.346938154066337548899755;
			m_abscissas[26]=.302261278456436100016129;
			m_abscissas[27]=.2569099397615302229942;
			m_abscissas[28]=.210985334548078993474541;
			m_abscissas[29]=.164589938564714607253651;
			m_abscissas[30]=.117827278078675997114714;
			m_abscissas[31]=.070801698868141591289364;
			m_abscissas[32]=.023618133385924570536223;
			m_weights[0]=.001677653744007238599332;
			m_weights[1]=.003901625641744248259228;
			m_weights[2]=.006120192018447936365568;
			m_weights[3]=.008325388765990901416726;
			m_weights[4]=.010512065987705754657377;
			m_weights[5]=.012675303981261681876445;
			m_weights[6]=.014810265002733960173643;
			m_weights[7]=.016912181472245217180350;
			m_weights[8]=.018976361722771325934867;
			m_weights[9]=.020998199091864625777331;
			m_weights[10]=.022973181735326655918099;
			m_weights[11]=.024896902514757372637731;
			m_weights[12]=.026765068754250001908793;
			m_weights[13]=.028573511782931871182822;
			m_weights[14]=.030318196218868519193641;
			m_weights[15]=.031995228964046887271282;
			m_weights[16]=.033600867886112232670349;
			m_weights[17]=.035131530165472555900641;
			m_weights[18]=.036583800288139094413690;
			m_weights[19]=.03795443766594162094913;
			m_weights[20]=.039240383866828330187813;
			m_weights[21]=.040438769438954979125868;
			m_weights[22]=.041546920313241881317735;
			m_weights[23]=.042562363770055716318907;
			m_weights[24]=.043482833956667478647575;
			m_weights[25]=.044306276943153161904603;
			m_weights[26]=.045030855305441500215193;
			m_weights[27]=.045654952225273056120439;
			m_weights[28]=.046177175097915975471666;
			m_weights[29]=.046596358639584103625824;
			m_weights[30]=.046911567487620827746254;
			m_weights[31]=.047122098287644732185445;
			m_weights[32]=.047227481262998554845633;
			break;
		case 67:
			m_abscissas[0]=.999365435552976797970416;
			m_abscissas[1]=.996658027206113735364371;
			m_abscissas[2]=.991793347082309609461698;
			m_abscissas[3]=.984780847755951113442593;
			m_abscissas[4]=.975635583922908052394254;
			m_abscissas[5]=.964377330427712611695647;
			m_abscissas[6]=.951030459540145373212348;
			m_abscissas[7]=.935623873270750876866653;
			m_abscissas[8]=.918190936812738581247247;
			m_abscissas[9]=.898769404961320375818489;
			m_abscissas[10]=.877401339831658720400298;
			m_abscissas[11]=.854133019537050650790743;
			m_abscissas[12]=.829014837853301197689074;
			m_abscissas[13]=.802101195021694434417254;
			m_abscissas[14]=.773450379899980327541549;
			m_abscissas[15]=.743124443704366206695961;
			m_abscissas[16]=.711189065610040327854836;
			m_abscissas[17]=.677713410498280828780525;
			m_abscissas[18]=.642769979156615354278134;
			m_abscissas[19]=.606434451255557386200298;
			m_abscissas[20]=.568785521441471021257572;
			m_abscissas[21]=.529904728900236016169801;
			m_abscissas[22]=.489876280760646626095593;
			m_abscissas[23]=.448786869719890883995133;
			m_abscissas[24]=.406725486286013915985299;
			m_abscissas[25]=.363783226043954170826044;
			m_abscissas[26]=.320053092362536331066875;
			m_abscissas[27]=.275629794969689587715555;
			m_abscissas[28]=.23060954483211573792523;
			m_abscissas[29]=.185089845783640168816231;
			m_abscissas[30]=.139169283353523713563891;
			m_abscissas[31]=.092947311252079876631721;
			m_abscissas[32]=.046524035976017219564639;
			m_abscissas[33]=0.;
			m_weights[0]=.001628325035240012866463;
			m_weights[1]=.003787008301825508445964;
			m_weights[2]=.005940693177582235216513;
			m_weights[3]=.008081790299023136215346;
			m_weights[4]=.010205440034102440986662;
			m_weights[5]=.012307003849288150521954;
			m_weights[6]=.014381919827200550930977;
			m_weights[7]=.016425690622530879204727;
			m_weights[8]=.018433888456804573872166;
			m_weights[9]=.020402163371343540449257;
			m_weights[10]=.022326252196452076922797;
			m_weights[11]=.024201987609673164720692;
			m_weights[12]=.026025307086213238803705;
			m_weights[13]=.027792261662436769987200;
			m_weights[14]=.029499024470945669695847;
			m_weights[15]=.031141899019472823937426;
			m_weights[16]=.032717327191531205427122;
			m_weights[17]=.034221896949536646739839;
			m_weights[18]=.035652349722745006661333;
			m_weights[19]=.037005587463492582023135;
			m_weights[20]=.038278679356179480647637;
			m_weights[21]=.039468868164308882642887;
			m_weights[22]=.040573576201744525227252;
			m_weights[23]=.041590410915199243098548;
			m_weights[24]=.042517170065830491471548;
			m_weights[25]=.043351846498699517359156;
			m_weights[26]=.044092632489751018307832;
			m_weights[27]=.044737923660889825472142;
			m_weights[28]=.04528632245466953156805;
			m_weights[29]=.045736641161063690936894;
			m_weights[30]=.046087904489761576197217;
			m_weights[31]=.046339351682415621108447;
			m_weights[32]=.046490438160264628208315;
			m_weights[33]=.046540836703563508249900;
			break;
		case 68:
			m_abscissas[0]=.999383825529704347163934;
			m_abscissas[1]=.996754836676664292536173;
			m_abscissas[2]=.992030887742448574647845;
			m_abscissas[3]=.985220861440244638894495;
			m_abscissas[4]=.976338950400356956942612;
			m_abscissas[5]=.965403802582199803231955;
			m_abscissas[6]=.952438404479602049323359;
			m_abscissas[7]=.937470018312800886485621;
			m_abscissas[8]=.920530120847918166460552;
			m_abscissas[9]=.901654335920107830215709;
			m_abscissas[10]=.880882359017455917697264;
			m_abscissas[11]=.858257873582659845494962;
			m_abscissas[12]=.833828459041670945820711;
			m_abscissas[13]=.807645490689708665233626;
			m_abscissas[14]=.779764031618897658744951;
			m_abscissas[15]=.750242716902916519767578;
			m_abscissas[16]=.719143630276472992623073;
			m_abscissas[17]=.686532173566012944784356;
			m_abscissas[18]=.652476929144676964200555;
			m_abscissas[19]=.617049515699883422716947;
			m_abscissas[20]=.580324437616350740175135;
			m_abscissas[21]=.542378928291003385814491;
			m_abscissas[22]=.503292787709084872635135;
			m_abscissas[23]=.463148214622939039172772;
			m_abscissas[24]=.422029633686313579146939;
			m_abscissas[25]=.380023517907674477497964;
			m_abscissas[26]=.337218206795880762727485;
			m_abscissas[27]=.293703720580638739978423;
			m_abscissas[28]=.249571570898416734511575;
			m_abscissas[29]=.20491456834193929584253;
			m_abscissas[30]=.15982662727797896099522;
			m_abscissas[31]=.114402568343910684430578;
			m_abscissas[32]=.06873791903837715221139;
			m_abscissas[33]=.022928712825422349713996;
			m_weights[0]=.001581140256372912939095;
			m_weights[1]=.003677366595011730633568;
			m_weights[2]=.005768969918729952021468;
			m_weights[3]=.007848711393177167415050;
			m_weights[4]=.009912001251585937209130;
			m_weights[5]=.011954462319769442103223;
			m_weights[6]=.013971789174457655815965;
			m_weights[7]=.015959735909613800072133;
			m_weights[8]=.017914120457923152489405;
			m_weights[9]=.019830832087955498291029;
			m_weights[10]=.021705839610378079801466;
			m_weights[11]=.023535199685876333361293;
			m_weights[12]=.025315065045176398323903;
			m_weights[13]=.027041692545903961557978;
			m_weights[14]=.028711451027484990710804;
			m_weights[15]=.030320828938553980341579;
			m_weights[16]=.031866441716821064939348;
			m_weights[17]=.033345038903980687903143;
			m_weights[18]=.034753510979751513166793;
			m_weights[19]=.036088895900179870714976;
			m_weights[20]=.037348385326186667716079;
			m_weights[21]=.038529330529106714493254;
			m_weights[22]=.039629247960712308025406;
			m_weights[23]=.040645824475954076140882;
			m_weights[24]=.041576922197402916484575;
			m_weights[25]=.042420583011142499300614;
			m_weights[26]=.043175032684644223225844;
			m_weights[27]=.043838684597956052010605;
			m_weights[28]=.044410143080352755909349;
			m_weights[29]=.044888206345426667826352;
			m_weights[30]=.045271869018443777869412;
			m_weights[31]=.045560324250648285980708;
			m_weights[32]=.045752965416067950519006;
			m_weights[33]=.045849387387250974686564;
			break;
		case 69:
			m_abscissas[0]=.999401427543244788208061;
			m_abscissas[1]=.996847500485195100459497;
			m_abscissas[2]=.992258266716853118700434;
			m_abscissas[3]=.985642080022637593237139;
			m_abscissas[4]=.977012331622094263389466;
			m_abscissas[5]=.9663866216428605668558;
			m_abscissas[6]=.95378664780518189753394;
			m_abscissas[7]=.939238147041608472966836;
			m_abscissas[8]=.922770839180444997188818;
			m_abscissas[9]=.904418364985827140066525;
			m_abscissas[10]=.884218216947252436662819;
			m_abscissas[11]=.862211662472744083857172;
			m_abscissas[12]=.838443659480504213261616;
			m_abscissas[13]=.81296276450029561385648;
			m_abscissas[14]=.785821033446759760335369;
			m_abscissas[15]=.757073915255837851594723;
			m_abscissas[16]=.72678013859600808865716;
			m_abscissas[17]=.695001591882928460929744;
			m_abscissas[18]=.661803196841078244024226;
			m_abscissas[19]=.627252775869849401767458;
			m_abscissas[20]=.59142091348455582556931;
			m_abscissas[21]=.554380812115132014537973;
			m_abscissas[22]=.516208142556933709171821;
			m_abscissas[23]=.476980889379045710341441;
			m_abscissas[24]=.436779191605844420869877;
			m_abscissas[25]=.395685178997244447351756;
			m_abscissas[26]=.353782804262066317960127;
			m_abscissas[27]=.311157671547281186585905;
			m_abscissas[28]=.267896861553503428581599;
			m_abscissas[29]=.224088753633999013281464;
			m_abscissas[30]=.179822845240643164359713;
			m_abscissas[31]=.135189569085682992158472;
			m_abscissas[32]=.090280108392828742562120;
			m_abscissas[33]=.045186210615101723531632;
			m_abscissas[34]=0.;
			m_weights[0]=.001535976952792084075129;
			m_weights[1]=.003572416739397372609704;
			m_weights[2]=.005604579927870594828536;
			m_weights[3]=.007625555931201510611458;
			m_weights[4]=.009631006150415575588660;
			m_weights[5]=.011616796610671965548740;
			m_weights[6]=.013578860649075670999812;
			m_weights[7]=.015513186543406164739769;
			m_weights[8]=.017415821231969829132074;
			m_weights[9]=.019282877128849402789244;
			m_weights[10]=.021110539639871894627890;
			m_weights[11]=.022895074790740785655521;
			m_weights[12]=.024632836784542455364336;
			m_weights[13]=.026320275416869483791761;
			m_weights[14]=.027953943312187705990862;
			m_weights[15]=.029530502957906711779811;
			m_weights[16]=.031046733517890539032686;
			m_weights[17]=.032499537409646111244734;
			m_weights[18]=.033885946630832289497810;
			m_weights[19]=.035203128821683486147754;
			m_weights[20]=.036448393050700514056644;
			m_weights[21]=.037619195311640906508158;
			m_weights[22]=.038713143720492513932739;
			m_weights[23]=.039728003401761641206459;
			m_weights[24]=.040661701054061600537526;
			m_weights[25]=.041512329185654502082874;
			m_weights[26]=.042278150011280512851583;
			m_weights[27]=.042957599002305213878420;
			m_weights[28]=.043549288082926741033576;
			m_weights[29]=.044052008465909284380986;
			m_weights[30]=.044464733122047138096231;
			m_weights[31]=.044786618878312557542135;
			m_weights[32]=.045017008140399802198716;
			m_weights[33]=.045155430236145460516517;
			m_weights[34]=.045201602377079954242289;
			break;
		case 70:
			m_abscissas[0]=.999418285973575842056743;
			m_abscissas[1]=.996936251961680156609813;
			m_abscissas[2]=.992476055211689981098942;
			m_abscissas[3]=.986045558070398659927101;
			m_abscissas[4]=.977657405957592400393908;
			m_abscissas[5]=.967328223664986438388618;
			m_abscissas[6]=.955078509114292842640327;
			m_abscissas[7]=.940932579003815355522815;
			m_abscissas[8]=.924918516897934440272266;
			m_abscissas[9]=.907068116260922849436353;
			m_abscissas[10]=.887416816863348171124373;
			m_abscissas[11]=.866003634213858629380035;
			m_abscissas[12]=.842871081998980242318594;
			m_abscissas[13]=.818065087625441189027232;
			m_abscissas[14]=.791634901007892758107676;
			m_abscissas[15]=.763632996771899568925095;
			m_abscissas[16]=.734114970060942641307652;
			m_abscissas[17]=.703139426151528597062947;
			m_abscissas[18]=.670767864094077405646281;
			m_abscissas[19]=.63706455460977809627886;
			m_abscissas[20]=.602096412485355487336767;
			m_abscissas[21]=.565932863718808286372959;
			m_abscissas[22]=.528645707679711127265081;
			m_abscissas[23]=.490308974557636589269778;
			m_abscissas[24]=.45099877838164786573164;
			m_abscissas[25]=.410793165902630589371263;
			m_abscissas[26]=.369771961638461895839405;
			m_abscissas[27]=.328016609389643257846131;
			m_abscissas[28]=.285610010540037861691665;
			m_abscissas[29]=.242636359463740645783579;
			m_abscissas[30]=.199180976364857664151403;
			m_abscissas[31]=.155330137882070247309005;
			m_abscissas[32]=.111170905794298693735752;
			m_abscissas[33]=.066790954167551324003719;
			m_abscissas[34]=.022278395286140309693493;
			m_weights[0]=.001492721288844515731039;
			m_weights[1]=.003471894893078143254999;
			m_weights[2]=.005447111874217218312821;
			m_weights[3]=.007411769363190210362111;
			m_weights[4]=.009361762769699026811500;
			m_weights[5]=.011293184649931537649631;
			m_weights[6]=.013202190814676747625073;
			m_weights[7]=.015084987865443127682295;
			m_weights[8]=.016937836376302932531837;
			m_weights[9]=.018757057093133423415451;
			m_weights[10]=.020539037824326453384493;
			m_weights[11]=.022280240452256595833891;
			m_weights[12]=.023977207889100292278687;
			m_weights[13]=.025626570908468482798985;
			m_weights[14]=.027225054818664417159107;
			m_weights[15]=.028769485955808280661310;
			m_weights[16]=.030256797980154237816537;
			m_weights[17]=.031684037961308481734653;
			m_weights[18]=.033048372239372420470874;
			m_weights[19]=.034347092049906537568545;
			m_weights[20]=.035577618901292380532770;
			m_weights[21]=.036737509693672695348040;
			m_weights[22]=.037824461569222817197272;
			m_weights[23]=.038836316484073403979003;
			m_weights[24]=.039771065492776567477850;
			m_weights[25]=.040626852736789616351226;
			m_weights[26]=.041401979129045208638227;
			m_weights[27]=.042094905727284406020984;
			m_weights[28]=.042704256789449777769966;
			m_weights[29]=.043228822505068699789395;
			m_weights[30]=.043667561397201440252548;
			m_weights[31]=.044019602390183458757356;
			m_weights[32]=.044284246539055406775800;
			m_weights[33]=.044460968417246370823557;
			m_weights[34]=.044549417159754667202167;
			break;
		case 71:
			m_abscissas[0]=.99943444211973692756166;
			m_abscissas[1]=.997021308251255130065673;
			m_abscissas[2]=.992684784870641569941806;
			m_abscissas[3]=.986432277268261368092142;
			m_abscissas[4]=.978275736113601792874824;
			m_abscissas[5]=.968230877426325345680305;
			m_abscissas[6]=.956317081041052874884704;
			m_abscissas[7]=.942557339922271159124396;
			m_abscissas[8]=.926978212250871410768131;
			m_abscissas[9]=.909609768984534359441635;
			m_abscissas[10]=.890485535355206511673912;
			m_abscissas[11]=.869642425956307411147088;
			m_abscissas[12]=.847120673391873968941067;
			m_abscissas[13]=.822963750567552752389365;
			m_abscissas[14]=.797218286749317960639491;
			m_abscissas[15]=.769933977541036647405165;
			m_abscissas[16]=.741163488949376848010344;
			m_abscissas[17]=.710962355718544370974605;
			m_abscissas[18]=.679388874129644576284578;
			m_abscissas[19]=.646503989470787642801403;
			m_abscissas[20]=.612371178394682336310376;
			m_abscissas[21]=.577056326390519810485799;
			m_abscissas[22]=.540627600606487356004911;
			m_abscissas[23]=.503155318268290070833605;
			m_abscissas[24]=.464711810947597822947579;
			m_abscissas[25]=.425371284942369555863251;
			m_abscissas[26]=.38520967803852729598972;
			m_abscissas[27]=.344304512929447024710467;
			m_abscissas[28]=.302734747576191376434942;
			m_abscissas[29]=.260580622797318662023926;
			m_abscissas[30]=.217923507382453335534383;
			m_abscissas[31]=.174845741028584863169688;
			m_abscissas[32]=.13143047540226608701291;
			m_abscissas[33]=.087761513634500682361807;
			m_abscissas[34]=.043923148558135329332217;
			m_abscissas[35]=0.;
			m_weights[0]=.001451267330029397268498;
			m_weights[1]=.003375555496730675865127;
			m_weights[2]=.005296182844025892632680;
			m_weights[3]=.007206835281831493387345;
			m_weights[4]=.009103626461992005851317;
			m_weights[5]=.010982860154298551706275;
			m_weights[6]=.012840899668087806070418;
			m_weights[7]=.014674155334611529200408;
			m_weights[8]=.016479087207462396550592;
			m_weights[9]=.018252210704678670502329;
			m_weights[10]=.019990102932350111287485;
			m_weights[11]=.021689409135985367961832;
			m_weights[12]=.023346849109223252631715;
			m_weights[13]=.024959223494313873055276;
			m_weights[14]=.026523419942157908008105;
			m_weights[15]=.02803641911174149061798;
			m_weights[16]=.029495300493708812464936;
			m_weights[17]=.030897248045090721698606;
			m_weights[18]=.032239555623443526941907;
			m_weights[19]=.033519632209454030834406;
			m_weights[20]=.034735006907682188375365;
			m_weights[21]=.035883333715648910777968;
			m_weights[22]=.036962396051982031856083;
			m_weights[23]=.037970111034831156214418;
			m_weights[24]=.038904533502262947492403;
			m_weights[25]=.039763859766857581674337;
			m_weights[26]=.040546431097246896434925;
			m_weights[27]=.041250736919866024249109;
			m_weights[28]=.041875417734733006189543;
			m_weights[29]=.042419267739624593035339;
			m_weights[30]=.042881237157580435020607;
			m_weights[31]=.043260434263241266598856;
			m_weights[32]=.043556127104108533371134;
			m_weights[33]=.043767744913402144972310;
			m_weights[34]=.043894879211788586321253;
			m_weights[35]=.043937284595852365758226;
			break;
		case 72:
			m_abscissas[0]=.999449934452962642181854;
			m_abscissas[1]=.997102871642729078465967;
			m_abscissas[2]=.992884951016801908228104;
			m_abscissas[3]=.986803152375830465808073;
			m_abscissas[4]=.978868778557233771796187;
			m_abscissas[5]=.969096697998780425922171;
			m_abscissas[6]=.957505247577698225664203;
			m_abscissas[7]=.944116185272537965007181;
			m_abscissas[8]=.928954645880918010572224;
			m_abscissas[9]=.91204909268867146992303;
			m_abscissas[10]=.893431263588091256195897;
			m_abscissas[11]=.873136111298778905281708;
			m_abscissas[12]=.85120173765443785111195;
			m_abscissas[13]=.82766932202275431461656;
			m_abscissas[14]=.802583043969291852027003;
			m_abscissas[15]=.775990000299982562142197;
			m_abscissas[16]=.747940116632832390670806;
			m_abscissas[17]=.718486053662234961362462;
			m_abscissas[18]=.687683108290467807491409;
			m_abscissas[19]=.655589109811201056148897;
			m_abscissas[20]=.622264311339468191007724;
			m_abscissas[21]=.587771276691655599203773;
			m_abscissas[22]=.552174762927714397400258;
			m_abscissas[23]=.515541598776003523860159;
			m_abscissas[24]=.477940559168940457183751;
			m_abscissas[25]=.439442236124960732083876;
			m_abscissas[26]=.400118906219161612702708;
			m_abscissas[27]=.360044394891419237540074;
			m_abscissas[28]=.319293937846712171332982;
			m_abscissas[29]=.277944039807847606980111;
			m_abscissas[30]=.236072330885759925057805;
			m_abscissas[31]=.193757420837026059803133;
			m_abscissas[32]=.151078751482210030326707;
			m_abscissas[33]=.108116447562102814809633;
			m_abscissas[34]=.064951166311857114074676;
			m_abscissas[35]=.021663946035424044670326;
			m_weights[0]=.001411516393973434135719;
			m_weights[1]=.003283169774667495801901;
			m_weights[2]=.005151436018790886908251;
			m_weights[3]=.007010272321861863296081;
			m_weights[4]=.008855996073706153383957;
			m_weights[5]=.010685108165351897158958;
			m_weights[6]=.012494165619873757767783;
			m_weights[7]=.014279769054554193266555;
			m_weights[8]=.016038564950285155218161;
			m_weights[9]=.017767250789200653594360;
			m_weights[10]=.019462580863294278043017;
			m_weights[11]=.021121372216440553509820;
			m_weights[12]=.022740510555035754455931;
			m_weights[13]=.024316956064419164326347;
			m_weights[14]=.025847749100655890283898;
			m_weights[15]=.027330015738950934433796;
			m_weights[16]=.028760973164701761095125;
			m_weights[17]=.030137934895375479292983;
			m_weights[18]=.031458315822561813977776;
			m_weights[19]=.032719637064293846704313;
			m_weights[20]=.033919530618286059497196;
			m_weights[21]=.035055743807217870434138;
			m_weights[22]=.036126143507637992985631;
			m_weights[23]=.037128720154502899460555;
			m_weights[24]=.038061591513802163834376;
			m_weights[25]=.038923006216169663799963;
			m_weights[26]=.039711347044834901782399;
			m_weights[27]=.040425133971733970043329;
			m_weights[28]=.041063026936075061101936;
			m_weights[29]=.041623828360138598207608;
			m_weights[30]=.042106485397586464146587;
			m_weights[31]=.042510091910057720077801;
			m_weights[32]=.042833890168338813666840;
			m_weights[33]=.043077272274913699745250;
			m_weights[34]=.043239781305222617485265;
			m_weights[35]=.043321112165486537076394;
			break;
		case 73:
			m_abscissas[0]=.999464798845881170996471;
			m_abscissas[1]=.997181130771611267253561;
			m_abscissas[2]=.993077015588945151555204;
			m_abscissas[3]=.987159036624932114781028;
			m_abscissas[4]=.979437892056483624562705;
			m_abscissas[5]=.969927659054985912747008;
			m_abscissas[6]=.958645700697974220413851;
			m_abscissas[7]=.945612621689483571931668;
			m_abscissas[8]=.930852227381224559904698;
			m_abscissas[9]=.914391479176256414741684;
			m_abscissas[10]=.896260444846691221048938;
			m_abscissas[11]=.876492243420651499825943;
			m_abscissas[12]=.855122984594299484507991;
			m_abscissas[13]=.832191702724909743755788;
			m_abscissas[14]=.807740285502735924807061;
			m_abscissas[15]=.781813397421644751334091;
			m_abscissas[16]=.754458398183323480238445;
			m_abscissas[17]=.725725256181559818503136;
			m_abscissas[18]=.69566645722326271722022;
			m_abscissas[19]=.664336908652196338882376;
			m_abscissas[20]=.631793839050118456511041;
			m_abscissas[21]=.598096693698270976000399;
			m_abscissas[22]=.563307025990016512736243;
			m_abscissas[23]=.527488384992870936800441;
			m_abscissas[24]=.490706199365251987700222;
			m_abscissas[25]=.453027657839945395191991;
			m_abscissas[26]=.414521586492575763889121;
			m_abscissas[27]=.375258323019251502905739;
			m_abscissas[28]=.335309588253022456115741;
			m_abscissas[29]=.294748355153836768255855;
			m_abscissas[30]=.253648715511301393555805;
			m_abscissas[31]=.212085744603730526527244;
			m_abscissas[32]=.170135364060700758505646;
			m_abscissas[33]=.127874203179614301754679;
			m_abscissas[34]=.085379458949596300995857;
			m_abscissas[35]=.042728755038413991185691;
			m_abscissas[36]=0.;
			m_weights[0]=.001373376462759619223984;
			m_weights[1]=.003194524377289034522078;
			m_weights[2]=.005012538571606190263813;
			m_weights[3]=.006821631349174792362209;
			m_weights[4]=.008618310479532247613912;
			m_weights[5]=.010399260995000532201304;
			m_weights[6]=.012161220929281112727763;
			m_weights[7]=.013900968788314650867520;
			m_weights[8]=.015615325433591422995534;
			m_weights[9]=.017301158762489083005607;
			m_weights[10]=.018955389048670021689736;
			m_weights[11]=.020574994420361169166020;
			m_weights[12]=.022157016317040072056770;
			m_weights[13]=.023698564864218974626609;
			m_weights[14]=.025196824137538312813332;
			m_weights[15]=.026649057298727482952230;
			m_weights[16]=.028052611590572060323802;
			m_weights[17]=.029404923180116560105457;
			m_weights[18]=.030703521840433504938126;
			m_weights[19]=.031946035461976706486501;
			m_weights[20]=.033130194385043840677069;
			m_weights[21]=.034253835545302215414130;
			m_weights[22]=.035314906424728287509063;
			m_weights[23]=.036311468800697784690347;
			m_weights[24]=.037241702286349773156894;
			m_weights[25]=.038103907655739800595508;
			m_weights[26]=.038896509947696739520476;
			m_weights[27]=.039618061342706143316508;
			m_weights[28]=.040267243807560033364942;
			m_weights[29]=.040842871502938861549361;
			m_weights[30]=.041343892949525494526883;
			m_weights[31]=.041769392948692853754102;
			m_weights[32]=.042118594254255636268946;
			m_weights[33]=.042390858992231594405374;
			m_weights[34]=.042585689826018387025763;
			m_weights[35]=.042702730864857222076601;
			m_weights[36]=.042741768313923562411455;
			break;
		case 74:
			m_abscissas[0]=.999479068780329524271045;
			m_abscissas[1]=.997256261711016390838064;
			m_abscissas[2]=.993261409804644972713885;
			m_abscissas[3]=.987500726616238559607124;
			m_abscissas[4]=.979984345431300336098804;
			m_abscissas[5]=.970725604042518726819355;
			m_abscissas[6]=.959740955459528912699917;
			m_abscissas[7]=.947049926428385008272628;
			m_abscissas[8]=.932675079462054566085601;
			m_abscissas[9]=.916641971642906460677148;
			m_abscissas[10]=.898979108761895795583715;
			m_abscissas[11]=.879717894452088809228763;
			m_abscissas[12]=.858892574265983576796486;
			m_abscissas[13]=.836540174742815739168348;
			m_abscissas[14]=.812700437551974373875255;
			m_abscissas[15]=.787415748819575517819987;
			m_abscissas[16]=.760731063758994475968623;
			m_abscissas[17]=.732693826736880277762463;
			m_abscissas[18]=.70335388691544185906032;
			m_abscissas[19]=.672763409620247602718505;
			m_abscissas[20]=.640976783590692388595342;
			m_abscissas[21]=.608050524277779825525163;
			m_abscissas[22]=.574043173360994152501893;
			m_abscissas[23]=.539015194662817083240727;
			m_abscissas[24]=.503028866645885165914073;
			m_abscissas[25]=.46614817168388118342206;
			m_abscissas[26]=.428438682303003279375829;
			m_abscissas[27]=.389967444596250750571072;
			m_abscissas[28]=.350802859017798047492535;
			m_abscissas[29]=.311014558769390637049758;
			m_abscissas[30]=.270673285994980405537672;
			m_abscissas[31]=.229850766003717031018169;
			m_abscissas[32]=.188619579744918578310721;
			m_abscissas[33]=.147053034761753411728851;
			m_abscissas[34]=.105225034853070971863989;
			m_abscissas[35]=.063209948675116291895727;
			m_abscissas[36]=.021082477516748304447535;
			m_weights[0]=.001336761650069883550329;
			m_weights[1]=.003109420149896754678673;
			m_weights[2]=.004879179758594144584290;
			m_weights[3]=.006640492909114357634761;
			m_weights[4]=.008390045433971397064088;
			m_weights[5]=.010124694538287305421120;
			m_weights[6]=.011841347547499667323168;
			m_weights[7]=.013536949411787494341053;
			m_weights[8]=.015208484243401234808875;
			m_weights[9]=.016852979582024853584847;
			m_weights[10]=.018467511308979879782854;
			m_weights[11]=.020049208702794944252735;
			m_weights[12]=.021595259481675888969700;
			m_weights[13]=.023102914774915823030933;
			m_weights[14]=.024569493995942767245649;
			m_weights[15]=.025992389600723787866774;
			m_weights[16]=.027369071719679352302438;
			m_weights[17]=.028697092653269875342095;
			m_weights[18]=.029974091222461187339965;
			m_weights[19]=.031197796965915426033372;
			m_weights[20]=.032366034176216999525280;
			m_weights[21]=.033476725767828766263722;
			m_weights[22]=.034527896969826461003334;
			m_weights[23]=.035517678836800959925854;
			m_weights[24]=.036444311571658564481811;
			m_weights[25]=.037306147654394155733707;
			m_weights[26]=.038101654771263248896352;
			m_weights[27]=.038829418539137707758082;
			m_weights[28]=.039488145020196468323633;
			m_weights[29]=.040076663022476966759151;
			m_weights[30]=.040593926182194728058077;
			m_weights[31]=.041039014824127266847419;
			m_weights[32]=.041411137596753510820068;
			m_weights[33]=.041709632879240754378710;
			m_weights[34]=.041933969957777021469955;
			m_weights[35]=.042083749969156972474896;
			m_weights[36]=.042158706609943422122231;
			break;
		case 75:
			m_abscissas[0]=.999492775536035456483467;
			m_abscissas[1]=.997328428962231803309066;
			m_abscissas[2]=.993438536578892704377047;
			m_abscissas[3]=.987828966767524520141267;
			m_abscissas[4]=.980509324597416605771693;
			m_abscissas[5]=.971492256343063823363814;
			m_abscissas[6]=.960793363741892314155448;
			m_abscissas[7]=.94843116507928737792192;
			m_abscissas[8]=.934427059964107156836194;
			m_abscissas[9]=.918805291228393991271239;
			m_abscissas[10]=.901592902548446334093964;
			m_abscissas[11]=.882819691447895542160092;
			m_abscissas[12]=.862518157626883425461262;
			m_abscissas[13]=.840723446654958050097012;
			m_abscissas[14]=.817473289103541314843395;
			m_abscissas[15]=.792807935213563510360414;
			m_abscissas[16]=.766770085206641925524095;
			m_abscissas[17]=.739404815358032526496514;
			m_abscissas[18]=.710759499958041126769559;
			m_abscissas[19]=.680883729296269598044384;
			m_abscissas[20]=.649829223810262232586887;
			m_abscissas[21]=.617649744546925964450186;
			m_abscissas[22]=.584401000091577880319607;
			m_abscissas[23]=.550140550125645551503015;
			m_abscissas[24]=.514927705779916640527992;
			m_abscissas[25]=.478823426955803157015809;
			m_abscissas[26]=.441890216792348266537335;
			m_abscissas[27]=.404192013461653613190807;
			m_abscissas[28]=.365794079480035925326018;
			m_abscissas[29]=.3267628887265263023584;
			m_abscissas[30]=.287166011364297246942702;
			m_abscissas[31]=.24707199686423475690424;
			m_abscissas[32]=.206550255333159607114474;
			m_abscissas[33]=.16567093735213781838595;
			m_abscissas[34]=.124504812532900261625084;
			m_abscissas[35]=.083123147002610991697497;
			m_abscissas[36]=.041597580029079455974891;
			m_abscissas[37]=0.;
			m_weights[0]=.001301591717375855993898;
			m_weights[1]=.003027671014606041291226;
			m_weights[2]=.004751069185015273965899;
			m_weights[3]=.006466464907037538401963;
			m_weights[4]=.008170710707327826403718;
			m_weights[5]=.009860824916114018392052;
			m_weights[6]=.011533873328304495966814;
			m_weights[7]=.013186956762824802119614;
			m_weights[8]=.014817212289814468520138;
			m_weights[9]=.016421817119024640043599;
			m_weights[10]=.017997993125645050637946;
			m_weights[11]=.019543011520127889379571;
			m_weights[12]=.021054197512282842236446;
			m_weights[13]=.022528934913865776450546;
			m_weights[14]=.023964670653716959174765;
			m_weights[15]=.025358919190216379094208;
			m_weights[16]=.026709266810120851772354;
			m_weights[17]=.028013375804780540825259;
			m_weights[18]=.029268988515725986805033;
			m_weights[19]=.030473931242214539203138;
			m_weights[20]=.031626118003749648056032;
			m_weights[21]=.032723554150934220521523;
			m_weights[22]=.033764339818334092646956;
			m_weights[23]=.034746673213330406535099;
			m_weights[24]=.035668853735240453089118;
			m_weights[25]=.036529284919290339006851;
			m_weights[26]=.037326477200332090167306;
			m_weights[27]=.038059050491513603135631;
			m_weights[28]=.038725736573432575841466;
			m_weights[29]=.039325381289635162520767;
			m_weights[30]=.039856946544656352575965;
			m_weights[31]=.040319512101141577558174;
			m_weights[32]=.040712277172937330298758;
			m_weights[33]=.041034561811392106676223;
			m_weights[34]=.041285808082467189083461;
			m_weights[35]=.041465581032619092135248;
			m_weights[36]=.041573569441781278782999;
			m_weights[37]=.041609586362141409380467;
			break;
		case 76:
			m_abscissas[0]=.999505948362153078106202;
			m_abscissas[1]=.997397786355355436691434;
			m_abscissas[2]=.993608772723526984616191;
			m_abscissas[3]=.988144453359837438498645;
			m_abscissas[4]=.981013938975655895251838;
			m_abscissas[5]=.97222922852037690447808;
			m_abscissas[6]=.961805126758767827168032;
			m_abscissas[7]=.949759207710896385084895;
			m_abscissas[8]=.936111781934810796263223;
			m_abscissas[9]=.920885861252150118491581;
			m_abscissas[10]=.904107119545566828544426;
			m_abscissas[11]=.88580384929208278721491;
			m_abscissas[12]=.866006913771981955227868;
			m_abscissas[13]=.84474969498334239401089;
			m_abscissas[14]=.822068037328974613477771;
			m_abscissas[15]=.79800018716120016975281;
			m_abscissas[16]=.772586728281809761297673;
			m_abscissas[17]=.745870513503610443148639;
			m_abscissas[18]=.717896592387703639373006;
			m_abscissas[19]=.688712135277640596996525;
			m_abscissas[20]=.658366353758142630275233;
			m_abscissas[21]=.626910417672266559145452;
			m_abscissas[22]=.594397368836792918151506;
			m_abscissas[23]=.560882031601237478999162;
			m_abscissas[24]=.526420920401242590955307;
			m_abscissas[25]=.491072144462193864861745;
			m_abscissas[26]=.454895309813726339433841;
			m_abscissas[27]=.41795141878032729879566;
			m_abscissas[28]=.380302767117504443467504;
			m_abscissas[29]=.342012838966962023931201;
			m_abscissas[30]=.303146199807907778099415;
			m_abscissas[31]=.263768387584994276479405;
			m_abscissas[32]=.223945802196474133762638;
			m_abscissas[33]=.183745593528914477419244;
			m_abscissas[34]=.143235548227267521445508;
			m_abscissas[35]=.102483975391227018300051;
			m_abscissas[36]=.061559591390611208428668;
			m_abscissas[37]=.020531403993998635501747;
			m_weights[0]=.001267791634085359663282;
			m_weights[1]=.002949102953642474900393;
			m_weights[2]=.004627935228037421326129;
			m_weights[3]=.006299180497328445866573;
			m_weights[4]=.007959847477239734621119;
			m_weights[5]=.009607105414713754082404;
			m_weights[6]=.011238168569667687239675;
			m_weights[7]=.012850283847510144944925;
			m_weights[8]=.014440731748276679939890;
			m_weights[9]=.016006829912248570888510;
			m_weights[10]=.017545937291474230954200;
			m_weights[11]=.019055458467190582806802;
			m_weights[12]=.020532847967908021092975;
			m_weights[13]=.021975614534416249168013;
			m_weights[14]=.023381325307011186622480;
			m_weights[15]=.024747609920659671643264;
			m_weights[16]=.026072164497985983524275;
			m_weights[17]=.027352755531827521674152;
			m_weights[18]=.028587223650054003773975;
			m_weights[19]=.029773487255905040956707;
			m_weights[20]=.030909546037491595389937;
			m_weights[21]=.031993484340421600068535;
			m_weights[22]=.033023474397791741006542;
			m_weights[23]=.033997779412056380846742;
			m_weights[24]=.034914756483550767444125;
			m_weights[25]=.035772859380713947527779;
			m_weights[26]=.036570641147329617007244;
			m_weights[27]=.037306756542381609827567;
			m_weights[28]=.037979964308405283195235;
			m_weights[29]=.038589129264506738342921;
			m_weights[30]=.039133224220518440767507;
			m_weights[31]=.039611331709062058423147;
			m_weights[32]=.040022645532596826116462;
			m_weights[33]=.040366472122844023154096;
			m_weights[34]=.040642231710294738777455;
			m_weights[35]=.040849459301828492280392;
			m_weights[36]=.040987805464793951541308;
			m_weights[37]=.041057036916229422593260;
			break;
		case 77:
			m_abscissas[0]=.999518614633410009748581;
			m_abscissas[1]=.99746447786921409987487;
			m_abscissas[2]=.993772470949878248308911;
			m_abscissas[3]=.988447838222561307325019;
			m_abscissas[4]=.981499227331138462388345;
			m_abscissas[5]=.972938030749108390346306;
			m_abscissas[6]=.962778306467734311940162;
			m_abscissas[7]=.951036743600591871523429;
			m_abscissas[8]=.937732631959061044190142;
			m_abscissas[9]=.922887829359402199928045;
			m_abscissas[10]=.906526725321257818021091;
			m_abscissas[11]=.888676200825564069894835;
			m_abscissas[12]=.869365584067734059878654;
			m_abscissas[13]=.848626602229624036469578;
			m_abscissas[14]=.826493329329002119435587;
			m_abscissas[15]=.803002130222925227383638;
			m_abscissas[16]=.778191600852543141872868;
			m_abscissas[17]=.752102504825215941384387;
			m_abscissas[18]=.724777706436911728780786;
			m_abscissas[19]=.69626210024424222479102;
			m_abscissas[20]=.666602537301451493590916;
			m_abscissas[21]=.635847748183312367013651;
			m_abscissas[22]=.604048262920257384970148;
			m_abscissas[23]=.571256327977198250806983;
			m_abscissas[24]=.537525820412377739024425;
			m_abscissas[25]=.502912159357251517807726;
			m_abscissas[26]=.467472214962811794644644;
			m_abscissas[27]=.431264214961935682072524;
			m_abscissas[28]=.394347649001263560468467;
			m_abscissas[29]=.356783170899781070578555;
			m_abscissas[30]=.318632498994687308777523;
			m_abscissas[31]=.279958314738276122064705;
			m_abscissas[32]=.240824159712432197609682;
			m_abscissas[33]=.201294331229944356638148;
			m_abscissas[34]=.161433776694160931607184;
			m_abscissas[35]=.121307986890552586641807;
			m_abscissas[36]=.080982888385503131206498;
			m_abscissas[37]=.040524735209115925930070;
			m_abscissas[38]=0.;
			m_weights[0]=.001235291177139409614172;
			m_weights[1]=.002873553083652691657278;
			m_weights[2]=.004509523600205835333238;
			m_weights[3]=.006138296159756341839270;
			m_weights[4]=.007757025950083070731842;
			m_weights[5]=.009363023692386430769262;
			m_weights[6]=.010953642853911354238591;
			m_weights[7]=.012526267369227365187360;
			m_weights[8]=.014078312340027004050168;
			m_weights[9]=.015607227268749131295081;
			m_weights[10]=.017110499906531104176240;
			m_weights[11]=.018585660248341485509180;
			m_weights[12]=.020030284534316176396246;
			m_weights[13]=.021441999205456135505125;
			m_weights[14]=.022818484790129520512910;
			m_weights[15]=.024157479707955844940594;
			m_weights[16]=.025456783981694403752638;
			m_weights[17]=.026714262849557890832003;
			m_weights[18]=.027927850271276968541507;
			m_weights[19]=.029095552321768761348702;
			m_weights[20]=.030215450466622998691399;
			m_weights[21]=.031285704713905433393956;
			m_weights[22]=.032304556637030975593572;
			m_weights[23]=.033270332263698545302839;
			m_weights[24]=.034181444826115679265318;
			m_weights[25]=.035036397367978278454878;
			m_weights[26]=.035833785203911962602643;
			m_weights[27]=.036572298227327454533458;
			m_weights[28]=.037250723062893718878760;
			m_weights[29]=.037867945060089320261667;
			m_weights[30]=.038422950124554523673681;
			m_weights[31]=.038914826384233785621033;
			m_weights[32]=.039342765687570151937132;
			m_weights[33]=.039706064931289310681038;
			m_weights[34]=.040004127215591237410352;
			m_weights[35]=.040236462824851084195265;
			m_weights[36]=.040402690032217756170326;
			m_weights[37]=.040502535726788031955250;
			m_weights[38]=.040535835862325650292052;
			break;
		case 78:
			m_abscissas[0]=.99953079999242375951869;
			m_abscissas[1]=.997528638378724361502298;
			m_abscissas[2]=.99392996169448874318219;
			m_abscissas[3]=.988739732093717434384563;
			m_abscissas[4]=.981966163099550497784518;
			m_abscissas[5]=.97362007850623644205991;
			m_abscissas[6]=.96371483598675590155984;
			m_abscissas[7]=.952266294689863363717212;
			m_abscissas[8]=.939292786914946593070647;
			m_abscissas[9]=.924815087780063776360327;
			m_abscissas[10]=.908856381573188258207326;
			m_abscissas[11]=.891442224458818578808555;
			m_abscissas[12]=.872600503472724504931392;
			m_abscissas[13]=.852361391822557703286002;
			m_abscissas[14]=.830757300545911272737268;
			m_abscissas[15]=.807822826594201686768652;
			m_abscissas[16]=.783594697421142787071318;
			m_abscissas[17]=.758111712162315315070401;
			m_abscissas[18]=.731414679498830397498445;
			m_abscissas[19]=.703546352303923772120565;
			m_abscissas[20]=.674551359176750790323894;
			m_abscissas[21]=.644476132972792854593098;
			m_abscissas[22]=.613368836445184776799409;
			m_abscissas[23]=.58127928511595175522153;
			m_abscissas[24]=.548258867500612608746769;
			m_abscissas[25]=.51436046281386405802189;
			m_abscissas[26]=.47963835628810723968515;
			m_abscissas[27]=.444148152240408603067264;
			m_abscissas[28]=.407946685027098391646849;
			m_abscissas[29]=.371091928028596274325727;
			m_abscissas[30]=.333642900810210560782719;
			m_abscissas[31]=.295659574607580145222958;
			m_abscissas[32]=.257202776288112442988523;
			m_abscissas[33]=.218334090942211978813689;
			m_abscissas[34]=.179115763260289151562423;
			m_abscissas[35]=.139610597853483594909276;
			m_abscissas[36]=.099881858677728410528495;
			m_abscissas[37]=.059993167722217693635875;
			m_abscissas[38]=.020008403124517922952067;
			m_weights[0]=.001204024566067353280330;
			m_weights[1]=.002800868811838630411613;
			m_weights[2]=.004395596039460346742738;
			m_weights[3]=.005983489944440407989649;
			m_weights[4]=.007561843189439718826978;
			m_weights[5]=.009128099227255087276942;
			m_weights[6]=.010679742157481113353516;
			m_weights[7]=.012214284549789886397683;
			m_weights[8]=.013729267920144148393727;
			m_weights[9]=.015222265680178451693316;
			m_weights[10]=.016690886689343891866212;
			m_weights[11]=.018132778954982328644406;
			m_weights[12]=.019545633343399923377918;
			m_weights[13]=.020927187251877726785378;
			m_weights[14]=.022275228219113886763050;
			m_weights[15]=.023587597461457472096451;
			m_weights[16]=.024862193326222450761443;
			m_weights[17]=.026096974655108835029834;
			m_weights[18]=.027289964051624364864564;
			m_weights[19]=.028439251046897516262390;
			m_weights[20]=.02954299515860694641162;
			m_weights[21]=.030599428838013045289433;
			m_weights[22]=.031606860300304797738883;
			m_weights[23]=.032563676233689044408056;
			m_weights[24]=.033468344382858977972981;
			m_weights[25]=.034319416002689090290292;
			m_weights[26]=.03511552817821718947488;
			m_weights[27]=.035855406007191695445006;
			m_weights[28]=.036537864641684700648192;
			m_weights[29]=.037161811185498386850671;
			m_weights[30]=.037726246444324247864290;
			m_weights[31]=.038230266525850987649620;
			m_weights[32]=.038673064287257674003896;
			m_weights[33]=.039053930627773413147311;
			m_weights[34]=.039372255624233121937220;
			m_weights[35]=.039627529507810542956393;
			m_weights[36]=.039819343480364089225032;
			m_weights[37]=.039947390369088024879305;
			m_weights[38]=.040011465118420482988779;
			break;
		case 79:
			m_abscissas[0]=.999542528479569844324152;
			m_abscissas[1]=.997590394336941435176854;
			m_abscissas[2]=.994081554785542622411155;
			m_abscissas[3]=.989020707687784337874758;
			m_abscissas[4]=.982415659251278930098581;
			m_abscissas[5]=.974276699597807118178834;
			m_abscissas[6]=.96461652911486478610745;
			m_abscissas[7]=.953450227889269135963276;
			m_abscissas[8]=.940795229306577250300017;
			m_abscissas[9]=.926671291880903873664523;
			m_abscissas[10]=.911100468034580800245006;
			m_abscissas[11]=.89410706950423510131043;
			m_abscissas[12]=.875717629303529476659517;
			m_abscissas[13]=.855960860255219376473271;
			m_abscissas[14]=.834867610137777882067313;
			m_abscissas[15]=.812470813507806773764808;
			m_abscissas[16]=.788805440269197765565029;
			m_abscissas[17]=.763908441067170108458326;
			m_abscissas[18]=.737818689591278209525322;
			m_abscissas[19]=.710576921876824163758205;
			m_abscissas[20]=.68222567269907105927762;
			m_abscissas[21]=.652809209159343998524893;
			m_abscissas[22]=.62237346156657723888291;
			m_abscissas[23]=.590965951722139982853282;
			m_abscissas[24]=.558635718719858906655853;
			m_abscissas[25]=.525433242377054547896268;
			m_abscissas[26]=.491410364416119821095861;
			m_abscissas[27]=.45662020751968913114871;
			m_abscissas[28]=.421117092385771928732146;
			m_abscissas[29]=.384956452912350883922247;
			m_abscissas[30]=.348194749643867867239137;
			m_abscissas[31]=.310889381614736437208875;
			m_abscissas[32]=.273098596727523543742536;
			m_abscissas[33]=.234881400805731915877249;
			m_abscissas[34]=.196297465463184641659194;
			m_abscissas[35]=.157407034933861605028226;
			m_abscissas[36]=.118270832007660881166912;
			m_abscissas[37]=.078949963218954405448085;
			m_abscissas[38]=.039505823435974064058995;
			m_abscissas[39]=0.;
			m_weights[0]=.001173930129956613021202;
			m_weights[1]=.002730907065754855918533;
			m_weights[2]=.004285929113126531218219;
			m_weights[3]=.005834459868763465589213;
			m_weights[4]=.007373921131330176830391;
			m_weights[5]=.008901880982652486253740;
			m_weights[6]=.010415946204513382579183;
			m_weights[7]=.011913750215116998699600;
			m_weights[8]=.013392953344825676197309;
			m_weights[9]=.014851245526350069318579;
			m_weights[10]=.016286349576191682091837;
			m_weights[11]=.017696024624310417864663;
			m_weights[12]=.019078069558937488584784;
			m_weights[13]=.020430326438140859878443;
			m_weights[14]=.021750683846608079768642;
			m_weights[15]=.023037080185719026276979;
			m_weights[16]=.024287506888799492639422;
			m_weights[17]=.025500011555128773947340;
			m_weights[18]=.026672700997105556537883;
			m_weights[19]=.027803744195447058944436;
			m_weights[20]=.028891375157607266781637;
			m_weights[21]=.029933895674838362895649;
			m_weights[22]=.030929677973524835288294;
			m_weights[23]=.031877167256611170360519;
			m_weights[24]=.032774884131130817853422;
			m_weights[25]=.033621426918030930049923;
			m_weights[26]=.034415473840676600882592;
			m_weights[27]=.035155785088611131128251;
			m_weights[28]=.035841204753345752289207;
			m_weights[29]=.036470662633153427529256;
			m_weights[30]=.037043175904046784159838;
			m_weights[31]=.037557850654329770477907;
			m_weights[32]=.038013883280326049545518;
			m_weights[33]=.038410561741104177405417;
			m_weights[34]=.038747266670239967068185;
			m_weights[35]=.039023472342879796026505;
			m_weights[36]=.039238747496594643554919;
			m_weights[37]=.039392756004743003934264;
			m_weights[38]=.039485257401291164753722;
			m_weights[39]=.039516107256271311709157;
			break;
		case 80:
			m_abscissas[0]=.99955382265163062988008;
			m_abscissas[1]=.997649864398237688899494;
			m_abscissas[2]=.994227540965688277892064;
			m_abscissas[3]=.989291302499755531026503;
			m_abscissas[4]=.982848572738629070418288;
			m_abscissas[5]=.974909140585727793385645;
			m_abscissas[6]=.965485089043799251452273;
			m_abscissas[7]=.954590766343634905493482;
			m_abscissas[8]=.942242761309872674752266;
			m_abscissas[9]=.928459877172445795953046;
			m_abscissas[10]=.913263102571757654164734;
			m_abscissas[11]=.896675579438770683194324;
			m_abscissas[12]=.878722567678213828703773;
			m_abscissas[13]=.859431406663111096977192;
			m_abscissas[14]=.838831473580255275616623;
			m_abscissas[15]=.816954138681463470371125;
			m_abscissas[16]=.793832717504605449948639;
			m_abscissas[17]=.769502420135041373865616;
			m_abscissas[18]=.744000297583597272316541;
			m_abscissas[19]=.717365185362099880254068;
			m_abscissas[20]=.689637644342027600771208;
			m_abscissas[21]=.660859898986119801735967;
			m_abscissas[22]=.631075773046871966247928;
			m_abscissas[23]=.600330622829751743154746;
			m_abscissas[24]=.568671268122709784725486;
			m_abscissas[25]=.536145920897131932019857;
			m_abscissas[26]=.502804111888784987593673;
			m_abscissas[27]=.468696615170544477036078;
			m_abscissas[28]=.433875370831756093062387;
			m_abscissas[29]=.39839340588196922702438;
			m_abscissas[30]=.362304753499487315619043;
			m_abscissas[31]=.325664370747701914619113;
			m_abscissas[32]=.288528054884511853109139;
			m_abscissas[33]=.250952358392272120493159;
			m_abscissas[34]=.212994502857666132572389;
			m_abscissas[35]=.174712291832646812559339;
			m_abscissas[36]=.136164022809143886559241;
			m_abscissas[37]=.097408398441584599063278;
			m_abscissas[38]=.058504437152420668628993;
			m_abscissas[39]=.019511383256793997654351;
			m_weights[0]=.001144950003186941534544;
			m_weights[1]=.002663533589512681669294;
			m_weights[2]=.004180313124694895236739;
			m_weights[3]=.005690922451403198649269;
			m_weights[4]=.007192904768117312752676;
			m_weights[5]=.008683945269260858426409;
			m_weights[6]=.010161766041103064520832;
			m_weights[7]=.011624114120797826916467;
			m_weights[8]=.013068761592401339293787;
			m_weights[9]=.014493508040509076116962;
			m_weights[10]=.015896183583725688044903;
			m_weights[11]=.017274652056269306358584;
			m_weights[12]=.018626814208299031428735;
			m_weights[13]=.019950610878141998928892;
			m_weights[14]=.021244026115782006388711;
			m_weights[15]=.022505090246332461926222;
			m_weights[16]=.023731882865930101293193;
			m_weights[17]=.024922535764115491105118;
			m_weights[18]=.026075235767565117902969;
			m_weights[19]=.027188227500486380674419;
			m_weights[20]=.028259816057276862396753;
			m_weights[21]=.029288369583267847692768;
			m_weights[22]=.03027232175955798066122;
			m_weights[23]=.031210174188114701642443;
			m_weights[24]=.032100498673487773148056;
			m_weights[25]=.032941939397645401382836;
			m_weights[26]=.033733214984611522816675;
			m_weights[27]=.034473120451753928794364;
			m_weights[28]=.035160529044747593495527;
			m_weights[29]=.035794393953416054602862;
			m_weights[30]=.036373749905835978043965;
			m_weights[31]=.036897714638276008839151;
			m_weights[32]=.037365490238730490026705;
			m_weights[33]=.037776364362001397489775;
			m_weights[34]=.038129711314477638344207;
			m_weights[35]=.038424993006959423185212;
			m_weights[36]=.038661759774076463327077;
			m_weights[37]=.038839651059051968931774;
			m_weights[38]=.038958395962769531198626;
			m_weights[39]=.039017813656306654811280;
			break;
		case 81:
			m_abscissas[0]=.999564703690319034476721;
			m_abscissas[1]=.997707159988346463657951;
			m_abscissas[2]=.994368193285219865022677;
			m_abscissas[3]=.989552021371021999127889;
			m_abscissas[4]=.983265708566430227219941;
			m_abscissas[5]=.975518572672361296916254;
			m_abscissas[6]=.966322116338056111070092;
			m_abscissas[7]=.955689999756402515248837;
			m_abscissas[8]=.943638017652813821180582;
			m_abscissas[9]=.930184074915167503022546;
			m_abscissas[10]=.915348159640750717289078;
			m_abscissas[11]=.899152313286156530852666;
			m_abscissas[12]=.881620597846720146168274;
			m_abscissas[13]=.862779060069791894673747;
			m_abscissas[14]=.842655692736467046206713;
			m_abscissas[15]=.821280393060910223458768;
			m_abscissas[16]=.798684918265028282157201;
			m_abscissas[17]=.774902838392423431405638;
			m_abscissas[18]=.749969486430620672222515;
			m_abscissas[19]=.723921905815052614443364;
			m_abscissas[20]=.69679879539243560968186;
			m_abscissas[21]=.668640451925090146604899;
			m_abscissas[22]=.639488710221494554644762;
			m_abscissas[23]=.609386880981936546058808;
			m_abscissas[24]=.578379686451551369482078;
			m_abscissas[25]=.546513193976311273079586;
			m_abscissas[26]=.51383474756065821552202;
			m_abscissas[27]=.480392897528448270142663;
			m_abscissas[28]=.446237328391699056648103;
			m_abscissas[29]=.411418785034297541686006;
			m_abscissas[30]=.375988997320331287920517;
			m_abscissas[31]=.340000603239048339775819;
			m_abscissas[32]=.303507070700626164551632;
			m_abscissas[33]=.266562618098935339145502;
			m_abscissas[34]=.229222133759316105441507;
			m_abscissas[35]=.191541094391042855739847;
			m_abscissas[36]=.153575482665630636856788;
			m_abscissas[37]=.115381704043436711258854;
			m_abscissas[38]=.077016502972127177318291;
			m_abscissas[39]=.038536878581511983836486;
			m_abscissas[40]=0.;
			m_weights[0]=.001117029847124606606121;
			m_weights[1]=.002598622299928953013497;
			m_weights[2]=.004078551113421395586021;
			m_weights[3]=.005552611370256278902273;
			m_weights[4]=.007018460484931625511610;
			m_weights[5]=.008473893785345565449616;
			m_weights[6]=.009916741809595875499750;
			m_weights[7]=.011344858495416255762008;
			m_weights[8]=.012756121115134421000256;
			m_weights[9]=.014148432513236065548252;
			m_weights[10]=.015519723912464362938245;
			m_weights[11]=.016867957867635139474335;
			m_weights[12]=.018191131241255761151763;
			m_weights[13]=.019487278155601918215927;
			m_weights[14]=.020754472901445608539523;
			m_weights[15]=.021990832792751632770501;
			m_weights[16]=.023194520960273919881455;
			m_weights[17]=.024363749078563097332491;
			m_weights[18]=.025496780021665677069480;
			m_weights[19]=.026591930443219921090920;
			m_weights[20]=.027647573276924926911086;
			m_weights[21]=.028662140153560676225792;
			m_weights[22]=.029634123730905597658475;
			m_weights[23]=.030562079933052661895660;
			m_weights[24]=.031444630095774066418037;
			m_weights[25]=.032280463014732688872403;
			m_weights[26]=.033068336893488004420875;
			m_weights[27]=.033807081188396245551196;
			m_weights[28]=.034495598347659795894746;
			m_weights[29]=.035132865441939379415979;
			m_weights[30]=.035717935684104568530726;
			m_weights[31]=.036249939835863412798326;
			m_weights[32]=.036728087499180439516906;
			m_weights[33]=.037151668290563712144743;
			m_weights[34]=.037520052896475837859239;
			m_weights[35]=.037832694008300559247575;
			m_weights[36]=.038089127135475601831023;
			m_weights[37]=.038288971295583524430320;
			m_weights[38]=.038431929580375172100257;
			m_weights[39]=.038517789596884695237838;
			m_weights[40]=.038546423782980703389514;
			break;
		case 82:
			m_abscissas[0]=.999575191501652648874129;
			m_abscissas[1]=.997762385826385862562358;
			m_abscissas[2]=.994503768378075628934;
			m_abscissas[3]=.989803338839908949932625;
			m_abscissas[4]=.983667823522007603367888;
			m_abscissas[5]=.97610609709449882869142;
			m_abscissas[6]=.967129116252590695996077;
			m_abscissas[7]=.956749893861610903121799;
			m_abscissas[8]=.944983477438960489182246;
			m_abscissas[9]=.931846926454622887854733;
			m_abscissas[10]=.917359287253280107145762;
			m_abscissas[11]=.901541565288841258196285;
			m_abscissas[12]=.884416694596724708227991;
			m_abscissas[13]=.866009504504765448570684;
			m_abscissas[14]=.846346683612913701795724;
			m_abscissas[15]=.825456741085767656028635;
			m_abscissas[16]=.803369965310108299301882;
			m_abscissas[17]=.780118379975358245193967;
			m_abscissas[18]=.755735697639560417297554;
			m_abscissas[19]=.730257270847596245051693;
			m_abscissas[20]=.703720040872166678493836;
			m_abscissas[21]=.676162484151647024565951;
			m_abscissas[22]=.647624556502346598566312;
			m_abscissas[23]=.61814763518597868204103;
			m_abscissas[24]=.587774458916284867153709;
			m_abscissas[25]=.556549065891763812347322;
			m_abscissas[26]=.524516729944327767975079;
			m_abscissas[27]=.491723894896449322463459;
			m_abscissas[28]=.458218107221963197822863;
			m_abscissas[29]=.424047947108150722163559;
			m_abscissas[30]=.389262958019054815921047;
			m_abscissas[31]=.353913574862147936405395;
			m_abscissas[32]=.318051050862501505010084;
			m_abscissas[33]=.281727383250480102368264;
			m_abscissas[34]=.244995237870704537839177;
			m_abscissas[35]=.207907872821592350966229;
			m_abscissas[36]=.170519061236190159767099;
			m_abscissas[37]=.132883013316257525764814;
			m_abscissas[38]=.095054297732644873952887;
			m_abscissas[39]=.057087762505926931902960;
			m_abscissas[40]=.019038455482006813761455;
			m_weights[0]=.001090118595275830866107;
			m_weights[1]=.002536054696856106109822;
			m_weights[2]=.003980457937856074619029;
			m_weights[3]=.005419276232446765090705;
			m_weights[4]=.006850274534183526184326;
			m_weights[5]=.008271351818383685604430;
			m_weights[6]=.009680440704371073736965;
			m_weights[7]=.011075495781759896320225;
			m_weights[8]=.012454493401142104679734;
			m_weights[9]=.013815433714126459387728;
			m_weights[10]=.015156343270762561788469;
			m_weights[11]=.016475277763983708891012;
			m_weights[12]=.017770324798498407146982;
			m_weights[13]=.019039606640178925073039;
			m_weights[14]=.020281282926912158901571;
			m_weights[15]=.021493553330774844043490;
			m_weights[16]=.022674660164914103102442;
			m_weights[17]=.023822890930047826342227;
			m_weights[18]=.024936580796240755155772;
			m_weights[19]=.026014115016017023753869;
			m_weights[20]=.027053931265124771519787;
			m_weights[21]=.028054521907454230471714;
			m_weights[22]=.029014436180764403961453;
			m_weights[23]=.029932282300012724635086;
			m_weights[24]=.030806729475215629813668;
			m_weights[25]=.031636509840900245537623;
			m_weights[26]=.032420420294340605077836;
			m_weights[27]=.033157324239907211327758;
			m_weights[28]=.033846153236996858744637;
			m_weights[29]=.034485908549150705507379;
			m_weights[30]=.035075662592112690384780;
			m_weights[31]=.035614560278727472680496;
			m_weights[32]=.036101820258727023075696;
			m_weights[33]=.036536736051607652842198;
			m_weights[34]=.036918677070954456998532;
			m_weights[35]=.03724708953872766418784;
			m_weights[36]=.037521497288185020871574;
			m_weights[37]=.037741502454275869671537;
			m_weights[38]=.037906786050505784779464;
			m_weights[39]=.038017108431435269905303;
			m_weights[40]=.038072309640141871207696;
			break;
		case 83:
			m_abscissas[0]=.999585304807049343812604;
			m_abscissas[1]=.997815640403429961756266;
			m_abscissas[2]=.994634507631780323407966;
			m_abscissas[3]=.990045701297265222772328;
			m_abscissas[4]=.984055629596681471537067;
			m_abscissas[5]=.976672750072842934232188;
			m_abscissas[6]=.967907505450128729547015;
			m_abscissas[7]=.957772299122767105961512;
			m_abscissas[8]=.946281475011790874442295;
			m_abscissas[9]=.93345129640183861923568;
			m_abscissas[10]=.919299922587207943780996;
			m_abscissas[11]=.903847383022899508081721;
			m_abscissas[12]=.887115548905063340740519;
			m_abscissas[13]=.869128102178716983499116;
			m_abscissas[14]=.849910501998933797338429;
			m_abscissas[15]=.829489948684984276401755;
			m_abscissas[16]=.807895345214589632297865;
			m_abscissas[17]=.785157256310815840287636;
			m_abscissas[18]=.761307865178458740719255;
			m_abscissas[19]=.736380927950564267407382;
			m_abscissas[20]=.710411725909217464480458;
			m_abscissas[21]=.683437015548021780255841;
			m_abscissas[22]=.655494976546823717483086;
			m_abscissas[23]=.626625157732239358494923;
			m_abscissas[24]=.596868421100418312391831;
			m_abscissas[25]=.566266883981240803007684;
			m_abscissas[26]=.534863859425785310580243;
			m_abscissas[27]=.502703794901426132021111;
			m_abscissas[28]=.46983220938132032871745;
			m_abscissas[29]=.436295628917319275485497;
			m_abscissas[30]=.402141520787488735564029;
			m_abscissas[31]=.367418226311440367401593;
			m_abscissas[32]=.332174892428564182315162;
			m_abscissas[33]=.296461402136003191110713;
			m_abscissas[34]=.260328303884825917912496;
			m_abscissas[35]=.223826740034327399193192;
			m_abscissas[36]=.187008374465722680789553;
			m_abscissas[37]=.149925319457686817796109;
			m_abscissas[38]=.112630061927240307807978;
			m_abscissas[39]=.075175389140377282326814;
			m_abscissas[40]=.037614313997584382667579;
			m_abscissas[41]=0.;
			m_weights[0]=.001064168219666567756381;
			m_weights[1]=.002475719322545939743329;
			m_weights[2]=.003885859435353202192003;
			m_weights[3]=.005290681445859865555238;
			m_weights[4]=.006688051635243685741357;
			m_weights[5]=.008075966593123452283593;
			m_weights[6]=.009452455092479699888243;
			m_weights[7]=.010815566558037158720361;
			m_weights[8]=.012163370727798612063033;
			m_weights[9]=.013493959512375234980700;
			m_weights[10]=.014805449400717877680849;
			m_weights[11]=.016095984011935370915438;
			m_weights[12]=.017363736673824622350166;
			m_weights[13]=.018606912985478472841668;
			m_weights[14]=.019823753345654939049312;
			m_weights[15]=.021012535437269917874009;
			m_weights[16]=.022171576661803622621990;
			m_weights[17]=.023299236518900549370161;
			m_weights[18]=.024393918927158557497434;
			m_weights[19]=.025454074482489496750818;
			m_weights[20]=.026478202650673762485109;
			m_weights[21]=.027464853890903261238928;
			m_weights[22]=.028412631707247641563751;
			m_weights[23]=.029320194625104527918041;
			m_weights[24]=.030186258089814417054102;
			m_weights[25]=.031009596284739194843067;
			m_weights[26]=.031789043866222150643549;
			m_weights[27]=.032523497612968065991291;
			m_weights[28]=.033211917987505015181173;
			m_weights[29]=.033853330607515198699310;
			m_weights[30]=.034446827624950516832522;
			m_weights[31]=.034991569010979654731525;
			m_weights[32]=.035486783744947102645843;
			m_weights[33]=.035931770905660647347331;
			m_weights[34]=.036325900663462288899896;
			m_weights[35]=.036668615171678090043901;
			m_weights[36]=.036959429356184971079758;
			m_weights[37]=.037197931601976730544001;
			m_weights[38]=.037383784335757404410918;
			m_weights[39]=.037516724503737272715058;
			m_weights[40]=.037596563943955177591969;
			m_weights[41]=.037623189652600123325750;
			break;
		case 84:
			m_abscissas[0]=.999595061226923242204747;
			m_abscissas[1]=.997867016421711961561419;
			m_abscissas[2]=.994760638261284641195837;
			m_abscissas[3]=.990279528965356687334216;
			m_abscissas[4]=.984429797127581651519075;
			m_abscissas[5]=.977219507358315782654301;
			m_abscissas[6]=.968658618173621897690628;
			m_abscissas[7]=.958758958729693472019949;
			m_abscissas[8]=.947534209947443122811276;
			m_abscissas[9]=.934999884763558817752401;
			m_abscissas[10]=.921173306363051217727207;
			m_abscissas[11]=.906073584094008421339976;
			m_abscissas[12]=.889721586988315460803939;
			m_abscissas[13]=.872139914883579032029811;
			m_abscissas[14]=.853352867168918661598741;
			m_abscissas[15]=.833386409190021428846052;
			m_abscissas[16]=.812268136356123039022938;
			m_abscissas[17]=.790027235996596867967738;
			m_abscissas[18]=.7666944470188362826669;
			m_abscissas[19]=.742302017422609920773522;
			m_abscissas[20]=.716883659729274823822882;
			m_abscissas[21]=.690474504387248471220307;
			m_abscissas[22]=.663111051218014634336199;
			m_abscissas[23]=.634831118969691483644014;
			m_abscissas[24]=.605673793047833389684269;
			m_abscissas[25]=.575679371495674156041916;
			m_abscissas[26]=.544889309298449688912647;
			m_abscissas[27]=.513346161088761400048104;
			m_abscissas[28]=.481093522332156176054309;
			m_abscissas[29]=.4481759690742023311585;
			m_abscissas[30]=.41463899633233121933999;
			m_abscissas[31]=.380528955217588700538801;
			m_abscissas[32]=.34589298887319705903313;
			m_abscissas[33]=.310778967318463972627971;
			m_abscissas[34]=.275235421288088560499354;
			m_abscissas[35]=.239311475158303365709293;
			m_abscissas[36]=.203056779052553478128809;
			m_abscissas[37]=.166521440220548158966445;
			m_abscissas[38]=.129755953785524741800538;
			m_abscissas[39]=.092811132955437882746250;
			m_abscissas[40]=.055738038794528216999273;
			m_abscissas[41]=.018587909652332133286830;
			m_weights[0]=.001039133516451971889188;
			m_weights[1]=.002417511265443122855235;
			m_weights[2]=.003794591650452349696393;
			m_weights[3]=.005166605182746808329881;
			m_weights[4]=.006531513687713654601121;
			m_weights[5]=.007887405752648146382106;
			m_weights[6]=.009232400784190247014383;
			m_weights[7]=.010564637623008245264848;
			m_weights[8]=.011882274089801223495052;
			m_weights[9]=.013183488679182345986799;
			m_weights[10]=.014466482789161186242275;
			m_weights[11]=.015729483095583598201600;
			m_weights[12]=.016970743955211611343082;
			m_weights[13]=.018188549796056549927600;
			m_weights[14]=.019381217477318808647807;
			m_weights[15]=.020547098609756278611524;
			m_weights[16]=.021684581830644822989244;
			m_weights[17]=.022792095028942129338889;
			m_weights[18]=.023868107516958239384715;
			m_weights[19]=.024911132145206428884391;
			m_weights[20]=.025919727357334647725160;
			m_weights[21]=.026892499182197637515816;
			m_weights[22]=.027828103160258406035767;
			m_weights[23]=.028725246201621802212665;
			m_weights[24]=.029582688373110845289605;
			m_weights[25]=.030399244611902469773114;
			m_weights[26]=.031173786363345661291967;
			m_weights[27]=.031905243140692727484023;
			m_weights[28]=.032592604004584257183613;
			m_weights[29]=.033234918960240444074716;
			m_weights[30]=.033831300270425984803725;
			m_weights[31]=.034380923682372700621335;
			m_weights[32]=.034883029566963308456417;
			m_weights[33]=.035336923968601276160389;
			m_weights[34]=.035741979564315307277889;
			m_weights[35]=.036097636530772566701753;
			m_weights[36]=.036403403318002122488626;
			m_weights[37]=.036658857328759075636577;
			m_weights[38]=.036863645502590307718452;
			m_weights[39]=.037017484803794520585244;
			m_weights[40]=.037120162612602094273728;
			m_weights[41]=.037171537019034067603284;
			break;
		case 85:
			m_abscissas[0]=.999604477357478454326304;
			m_abscissas[1]=.997916601198116933150309;
			m_abscissas[2]=.994882374295616278453367;
			m_abscissas[3]=.990505217716415963787124;
			m_abscissas[4]=.984790957685580497327164;
			m_abscissas[5]=.977747288412243391933334;
			m_abscissas[6]=.969383711923678328005136;
			m_abscissas[7]=.959711515957188362396347;
			m_abscissas[8]=.948743756254578953404649;
			m_abscissas[9]=.936495238116430699686698;
			m_abscissas[10]=.922982496096090045588769;
			m_abscissas[11]=.908223771539091884896825;
			m_abscissas[12]=.892238987891355901581714;
			m_abscissas[13]=.875049723769097087638653;
			m_abscissas[14]=.856679183809956189095811;
			m_abscissas[15]=.837152167337082650198542;
			m_abscissas[16]=.816495034874791614720573;
			m_abscissas[17]=.794735672559116505447149;
			m_abscissas[18]=.771903454490293078660311;
			m_abscissas[19]=.74802920307743407395627;
			m_abscissas[20]=.723145147428601021384583;
			m_abscissas[21]=.697284879842249547650964;
			m_abscissas[22]=.670483310458662660702672;
			m_abscissas[23]=.642776620132514714843399;
			m_abscissas[24]=.614202211590136993071297;
			m_abscissas[25]=.584798658937388085710292;
			m_abscissas[26]=.554605655586269433158991;
			m_abscissas[27]=.523663960670567874788045;
			m_abscissas[28]=.492015344022851310477142;
			m_abscissas[29]=.459702529787088720761469;
			m_abscissas[30]=.426769138743009701721994;
			m_abscissas[31]=.393259629420059190650215;
			m_abscissas[32]=.359219238080438048756645;
			m_abscissas[33]=.324693917652247557454017;
			m_abscissas[34]=.28973027569517370353871;
			m_abscissas[35]=.254375511482453522497731;
			m_abscissas[36]=.218677352284059015710243;
			m_abscissas[37]=.182683988937112673876281;
			m_abscissas[38]=.146444010790510989210902;
			m_abscissas[39]=.11000634011157723735268;
			m_abscissas[40]=.073420166043291125619508;
			m_abscissas[41]=.036734878201249663520469;
			m_abscissas[42]=0.;
			m_weights[0]=.001014971908967743695373;
			m_weights[1]=.002361331704285020896767;
			m_weights[2]=.003706500125759316706871;
			m_weights[3]=.005046838426924442725452;
			m_weights[4]=.006380398587897515098686;
			m_weights[5]=.007705355960382757079897;
			m_weights[6]=.009019915439993631278967;
			m_weights[7]=.010322300230524245893816;
			m_weights[8]=.011610751286703898009624;
			m_weights[9]=.012883528856498084290506;
			m_weights[10]=.014138914548400832930555;
			m_weights[11]=.015375213542389626874408;
			m_weights[12]=.016590756831154670075205;
			m_weights[13]=.017783903451398170907744;
			m_weights[14]=.018953042688182840446805;
			m_weights[15]=.020096596243575421741795;
			m_weights[16]=.021213020364089379672417;
			m_weights[17]=.022300807922839374189458;
			m_weights[18]=.023358490452989891897698;
			m_weights[19]=.024384640129435683142416;
			m_weights[20]=.025377871695866088477370;
			m_weights[21]=.026336844334514359821732;
			m_weights[22]=.027260263476011164785770;
			m_weights[23]=.028146882546865075846383;
			m_weights[24]=.028995504652190152089866;
			m_weights[25]=.029804984191395887375612;
			m_weights[26]=.030574228404649995723924;
			m_weights[27]=.031302198848020870448397;
			m_weights[28]=.03198791279530467445977;
			m_weights[29]=.032630444564642178189038;
			m_weights[30]=.033228926768132769762529;
			m_weights[31]=.033782551482757530331312;
			m_weights[32]=.034290571341029846708222;
			m_weights[33]=.034752300539900637529247;
			m_weights[34]=.035167115766555788249813;
			m_weights[35]=.035534457039855699081992;
			m_weights[36]=.035853828466280812556915;
			m_weights[37]=.036124798909362460374752;
			m_weights[38]=.036347002571695203766757;
			m_weights[39]=.036520139488744884857477;
			m_weights[40]=.036643975933785702486407;
			m_weights[41]=.036718344733419616222152;
			m_weights[42]=.036743145493252106600207;
			break;
		case 86:
			m_abscissas[0]=.999613568841325857017352;
			m_abscissas[1]=.997964477035243944785198;
			m_abscissas[2]=.994999917485337896224028;
			m_abscissas[3]=.990723140745518705219946;
			m_abscissas[4]=.985139706732502413611135;
			m_abscissas[5]=.978256960253684716622482;
			m_abscissas[6]=.970083972685740131358567;
			m_abscissas[7]=.960631520942898448629415;
			m_abscissas[8]=.949912070852210255712008;
			m_abscissas[9]=.937939759909892614433043;
			m_abscissas[10]=.924730378322861949777046;
			m_abscissas[11]=.910301348046165830210553;
			m_abscissas[12]=.894671699739446334379039;
			m_abscissas[13]=.877862047633365937339916;
			m_abscissas[14]=.859894562322684955300209;
			m_abscissas[15]=.840792941514433051699711;
			m_abscissas[16]=.82058237876616022153705;
			m_abscissas[17]=.799289530253661788136253;
			m_abscissas[18]=.77694247961102337105447;
			m_abscissas[19]=.753570700888808903497941;
			m_abscissas[20]=.729205019678928322898718;
			m_abscissas[21]=.703877572457267694796427;
			m_abscissas[22]=.677621764197588510655952;
			m_abscissas[23]=.650472224312526232850995;
			m_abscissas[24]=.622464760979750811913254;
			m_abscissas[25]=.593636313913498033672087;
			m_abscissas[26]=.564024905643741329015105;
			m_abscissas[27]=.533669591367248683303849;
			m_abscissas[28]=.502610407436657274501913;
			m_abscissas[29]=.470888318555497783577945;
			m_abscissas[30]=.438545163748809094011783;
			m_abscissas[31]=.405623601180600396151174;
			m_abscissas[32]=.372167051890939610324135;
			m_abscissas[33]=.338219642526872666411964;
			m_abscissas[34]=.303826147142705735062772;
			m_abscissas[35]=.26903192814641030510041;
			m_abscissas[36]=.233882876470037463740116;
			m_abscissas[37]=.198425351043051402814531;
			m_abscissas[38]=.162706117648411714337532;
			m_abscissas[39]=.126772287242048252986199;
			m_abscissas[40]=.090671253817080167031377;
			m_abscissas[41]=.054450631894731206029935;
			m_abscissas[42]=.018158193724385815933443;
			m_weights[0]=.000991643266620363525578;
			m_weights[1]=.002307087488809902925968;
			m_weights[2]=.003621439249610901437556;
			m_weights[3]=.004931184096960103696422;
			m_weights[4]=.006234459139140123463887;
			m_weights[5]=.007529521612194562606845;
			m_weights[6]=.008814657101954815703782;
			m_weights[7]=.010088168460386105654673;
			m_weights[8]=.011348375156177703977167;
			m_weights[9]=.012593614678069697810409;
			m_weights[10]=.013822244452766670866648;
			m_weights[11]=.015032643905081378684945;
			m_weights[12]=.016223216549729022588084;
			m_weights[13]=.017392392075690542386720;
			m_weights[14]=.018538628406709859206315;
			m_weights[15]=.019660413729562179807402;
			m_weights[16]=.020756268484909142790582;
			m_weights[17]=.021824747316927627800684;
			m_weights[18]=.022864440978548006445773;
			m_weights[19]=.023873978189479004973217;
			m_weights[20]=.024852027444399835918326;
			m_weights[21]=.025797298768839535407771;
			m_weights[22]=.026708545420372209575307;
			m_weights[23]=.027584565532851248387384;
			m_weights[24]=.028424203701493494757312;
			m_weights[25]=.029226352506709944583661;
			m_weights[26]=.029989953974664932491338;
			m_weights[27]=.03071400097263205318304;
			m_weights[28]=.031397538537302865558533;
			m_weights[29]=.032039665134294016110228;
			m_weights[30]=.032639533847189921956099;
			m_weights[31]=.033196353494551597120090;
			m_weights[32]=.033709389673417554864972;
			m_weights[33]=.034177965727919904634238;
			m_weights[34]=.034601463641737692259934;
			m_weights[35]=.034979324853210099371413;
			m_weights[36]=.035311050992034205080585;
			m_weights[37]=.035596204536575495590691;
			m_weights[38]=.035834409390924055789779;
			m_weights[39]=.036025351380935257710090;
			m_weights[40]=.036168778668600637582745;
			m_weights[41]=.036264502084202387431492;
			m_weights[42]=.036312395375813338282315;
			break;
		case 87:
			m_abscissas[0]=.999622350432484272389096;
			m_abscissas[1]=.998010721562982358081679;
			m_abscissas[2]=.995113458137992005945508;
			m_abscissas[3]=.990933650110964570522868;
			m_abscissas[4]=.985476606068420367963808;
			m_abscissas[5]=.978749341003816272426478;
			m_abscissas[6]=.970760519747284274353547;
			m_abscissas[7]=.96152043693607316926423;
			m_abscissas[8]=.951041001389313573919119;
			m_abscissas[9]=.939335719974205747877988;
			m_abscissas[10]=.926419679891237219951464;
			m_abscissas[11]=.912309529094122884382983;
			m_abscissas[12]=.897023454767569367017326;
			m_abscissas[13]=.880581159852041841057713;
			m_abscissas[14]=.86300383762970660257019;
			m_abscissas[15]=.844314144397035220240557;
			m_abscissas[16]=.824536170255780480133435;
			m_abscissas[17]=.803695408058177778235321;
			m_abscissas[18]=.781818720545437736354678;
			m_abscissas[19]=.758934305721349315735428;
			m_abscissas[20]=.735071660505313794102036;
			m_abscissas[21]=.710261542711472902841968;
			m_abscissas[22]=.684535931402823674119862;
			m_abscissas[23]=.657927985671349391042954;
			m_abscissas[24]=.630472001897250447293695;
			m_abscissas[25]=.602203369542335201687955;
			m_abscissas[26]=.573158525534530411874213;
			m_abscissas[27]=.543374907302293353442616;
			m_abscissas[28]=.51289090451945225914091;
			m_abscissas[29]=.481745809622666805089462;
			m_abscissas[30]=.449979767165284378401573;
			m_abscissas[31]=.417633722072869055029911;
			m_abscissas[32]=.384749366867096863663811;
			m_abscissas[33]=.351369087926041315715694;
			m_abscissas[34]=.317535910850115714331344;
			m_abscissas[35]=.283293445004091873194452;
			m_abscissas[36]=.248685827306677131599538;
			m_abscissas[37]=.213757665340101603593763;
			m_abscissas[38]=.178553979853044214298229;
			m_abscissas[39]=.143120146731008139012399;
			m_abscissas[40]=.107501838508942771605762;
			m_abscissas[41]=.071744965501499429712831;
			m_abscissas[42]=.035895616626800899472717;
			m_abscissas[43]=0.;
			m_weights[0]=.000969109738177075337607;
			m_weights[1]=.002254690753752853092481;
			m_weights[2]=.003539271655388628540182;
			m_weights[3]=.004819456238501885899309;
			m_weights[4]=.006093462047634872130102;
			m_weights[5]=.007359623648818063660771;
			m_weights[6]=.008616302838488951832950;
			m_weights[7]=.009861877713701826716584;
			m_weights[8]=.011094741940560719279721;
			m_weights[9]=.012313306030048036542287;
			m_weights[10]=.013515999118245658081881;
			m_weights[11]=.014701270887239842229894;
			m_weights[12]=.015867593518826319002922;
			m_weights[13]=.017013463643001534433645;
			m_weights[14]=.018137404265354252050218;
			m_weights[15]=.019237966665356558785051;
			m_weights[16]=.020313732260655569526570;
			m_weights[17]=.021363314433802531593616;
			m_weights[18]=.022385360318485478214197;
			m_weights[19]=.023378552542660172257824;
			m_weights[20]=.024341610926167632339214;
			m_weights[21]=.025273294130557073164119;
			m_weights[22]=.026172401258933558949726;
			m_weights[23]=.027037773403735807283977;
			m_weights[24]=.027868295140429205989634;
			m_weights[25]=.028662895965176218388587;
			m_weights[26]=.029420551674623048249225;
			m_weights[27]=.030140285686018824743951;
			m_weights[28]=.030821170295962234153719;
			m_weights[29]=.031462327876150763937962;
			m_weights[30]=.032062932004589667777658;
			m_weights[31]=.032622208530801443925800;
			m_weights[32]=.033139436573662023536289;
			m_weights[33]=.033613949450576935584222;
			m_weights[34]=.034045135536799373455184;
			m_weights[35]=.034432439053782243765938;
			m_weights[36]=.034775360785547829248711;
			m_weights[37]=.035073458722151536556625;
			m_weights[38]=.035326348629410213698431;
			m_weights[39]=.035533704544160593911335;
			m_weights[40]=.035695259194409433776479;
			m_weights[41]=.035810804343833741756626;
			m_weights[42]=.035880191060187015877735;
			m_weights[43]=.035903329907265501442083;
			break;
		case 88:
			m_abscissas[0]=.999630836056270655781323;
			m_abscissas[1]=.998055408053249677606002;
			m_abscissas[2]=.995223175887988870781101;
			m_abscissas[3]=.991137078154015555111612;
			m_abscissas[4]=.985802186087763552283936;
			m_abscissas[5]=.979225203154296703776892;
			m_abscissas[6]=.971414410141322276480368;
			m_abscissas[7]=.962379646063770682989888;
			m_abscissas[8]=.952132293463802112522033;
			m_abscissas[9]=.940685263302643545262005;
			m_abscissas[10]=.928052978394692055580503;
			m_abscissas[11]=.914251355104527191339978;
			m_abscissas[12]=.899297783230057054844673;
			m_abscissas[13]=.883211104059450581623722;
			m_abscissas[14]=.866011586613781334022913;
			m_abscissas[15]=.847720902098209087197015;
			m_abscissas[16]=.828362096590456244850838;
			m_abscissas[17]=.807959561999236491824533;
			m_abscissas[18]=.786539005328287157479979;
			m_abscissas[19]=.764127416284206841967597;
			m_abscissas[20]=.740753033268607773581767;
			m_abscissas[21]=.716445307797250485659942;
			m_abscissas[22]=.691234867390880438020343;
			m_abscissas[23]=.665153476984453172221994;
			m_abscissas[24]=.638233998903326450283541;
			m_abscissas[25]=.610510351456818986239587;
			m_abscissas[26]=.582017466201287357580758;
			m_abscissas[27]=.552791243926555472797176;
			m_abscissas[28]=.522868509421143753727045;
			m_abscissas[29]=.492286965073286722885887;
			m_abscissas[30]=.461085143366196510612751;
			m_abscissas[31]=.429302358327424370211025;
			m_abscissas[32]=.396978655993491050245984;
			m_abscissas[33]=.364154763952198287815301;
			m_abscissas[34]=.330872040026196275109126;
			m_abscissas[35]=.297172420162464305587542;
			m_abscissas[36]=.263098365593362600228498;
			m_abscissas[37]=.228692809335831313896401;
			m_abscissas[38]=.19399910209614678857796;
			m_abscissas[39]=.159060957648394215944091;
			m_abscissas[40]=.123922397755479061647622;
			m_abscissas[41]=.088627696702076058804614;
			m_abscissas[42]=.053221325509403576521964;
			m_abscissas[43]=.017747895902112098288902;
			m_weights[0]=.000947335598161927266771;
			m_weights[1]=.002204058563143696628540;
			m_weights[2]=.003459867667862796423975;
			m_weights[3]=.004711479279598661743022;
			m_weights[4]=.005957186996138046583129;
			m_weights[5]=.007195398459796372059760;
			m_weights[6]=.008424547492702473015099;
			m_weights[7]=.009643083322053204400770;
			m_weights[8]=.010849469775429271259402;
			m_weights[9]=.012042186439581212309740;
			m_weights[10]=.013219730313627911708181;
			m_weights[11]=.014380617601299944235935;
			m_weights[12]=.015523385536933553840165;
			m_weights[13]=.016646594208217656045113;
			m_weights[14]=.017748828360324074556495;
			m_weights[15]=.01882869917375545139471;
			m_weights[16]=.019884846011274113243601;
			m_weights[17]=.020915938130576624232254;
			m_weights[18]=.021920676359989853595634;
			m_weights[19]=.022897794734781142327248;
			m_weights[20]=.023846062091859661263579;
			m_weights[21]=.024764283620768733025321;
			m_weights[22]=.025651302368961947884779;
			m_weights[23]=.026506000699434737721409;
			m_weights[24]=.027327301698855330835624;
			m_weights[25]=.028114170534408613491579;
			m_weights[26]=.028865615757635429246477;
			m_weights[27]=.029580690553619349113352;
			m_weights[28]=.030258493933943525335137;
			m_weights[29]=.030898171871912197633703;
			m_weights[30]=.031498918378604892320042;
			m_weights[31]=.032059976518406388069267;
			m_weights[32]=.032580639362732108686239;
			m_weights[33]=.033060250880746700145281;
			m_weights[34]=.033498206765953092528065;
			m_weights[35]=.033893955197610259239893;
			m_weights[36]=.034246997536020078737370;
			m_weights[37]=.034556888950807084134865;
			m_weights[38]=.034823238981399354993129;
			m_weights[39]=.035045712029004261396586;
			m_weights[40]=.035224027779459108532879;
			m_weights[41]=.035357961556423843793669;
			m_weights[42]=.035447344604470769706143;
			m_weights[43]=.035492064301714545296067;
			break;
		case 89:
			m_abscissas[0]=.999639038864533459581454;
			m_abscissas[1]=.998098605710273929484265;
			m_abscissas[2]=.995329240406749760327228;
			m_abscissas[3]=.991333738808665300673175;
			m_abscissas[4]=.986116947861103674004151;
			m_abscissas[5]=.979685276583876641129617;
			m_abscissas[6]=.972046642748909359704165;
			m_abscissas[7]=.963210454656535441853845;
			m_abscissas[8]=.953187597292840940578782;
			m_abscissas[9]=.941990418170214919067879;
			m_abscissas[10]=.929632711823718603373074;
			m_abscissas[11]=.916129702688823431211062;
			m_abscissas[12]=.901498026283997752087531;
			m_abscissas[13]=.885755708684469948280781;
			m_abscissas[14]=.868922144297083719255345;
			m_abscissas[15]=.851018071956677446875567;
			m_abscissas[16]=.832065549370079127108473;
			m_abscissas[17]=.812087925937486030236493;
			m_abscissas[18]=.791109813983793371456488;
			m_abscissas[19]=.769157058434801100328117;
			m_abscissas[20]=.74625670497536000938089;
			m_abscissas[21]=.7224369667285082433704;
			m_abscissas[22]=.697727189496539905894596;
			m_abscissas[23]=.67215781560675955082267;
			m_abscissas[24]=.645760346406419848716587;
			m_abscissas[25]=.618567303453019120698681;
			m_abscissas[26]=.590612188447752393043095;
			m_abscissas[27]=.561929441961464304933288;
			m_abscissas[28]=.532554401003943956969125;
			m_abscissas[29]=.50252325548882955144576;
			m_abscissas[30]=.471873003647753163790076;
			m_abscissas[31]=.44064140644865181859485;
			m_abscissas[32]=.408866941074398818696415;
			m_abscissas[33]=.376588753519067603880406;
			m_abscissas[34]=.343846610360227952477911;
			m_abscissas[35]=.310680849766689799743267;
			m_abscissas[36]=.277132331802052117947073;
			m_abscissas[37]=.243242388085282049589001;
			m_abscissas[38]=.20905277087034175587291;
			m_abscissas[39]=.174605601607596274075495;
			m_abscissas[40]=.139943319050374196667504;
			m_abscissas[41]=.105108626970613411698691;
			m_abscissas[42]=.070144441548005792132557;
			m_abscissas[43]=.035093838497457001384747;
			m_abscissas[44]=0.;
			m_weights[0]=.000926287105193472815530;
			m_weights[1]=.002155112582219113764636;
			m_weights[2]=.003383104792407455132634;
			m_weights[3]=.004607087343463241433055;
			m_weights[4]=.005825425788770107459645;
			m_weights[5]=.007036596870989114137390;
			m_weights[6]=.008239102525389078730573;
			m_weights[7]=.009431459260797890539711;
			m_weights[8]=.010612197289972188032681;
			m_weights[9]=.011779861580874892176619;
			m_weights[10]=.012933013392602677291586;
			m_weights[11]=.014070231948644482813887;
			m_weights[12]=.015190116144666123397473;
			m_weights[13]=.016291286254792384577541;
			m_weights[14]=.017372385621502401669641;
			m_weights[15]=.018432082321784115675846;
			m_weights[16]=.019469070805151873138675;
			m_weights[17]=.020482073500400270212245;
			m_weights[18]=.021469842388561140843413;
			m_weights[19]=.022431160539836367128359;
			m_weights[20]=.023364843612455445827169;
			m_weights[21]=.024269741311522339273662;
			m_weights[22]=.025144738806002568622816;
			m_weights[23]=.025988758102073836251481;
			m_weights[24]=.026800759371129177712566;
			m_weights[25]=.027579742230784582533477;
			m_weights[26]=.028324746977305207228035;
			m_weights[27]=.029034855767926811830019;
			m_weights[28]=.029709193751612459627302;
			m_weights[29]=.030346930146849129343408;
			m_weights[30]=.030947279265154844789479;
			m_weights[31]=.031509501479034283658799;
			m_weights[32]=.032032904133189585507032;
			m_weights[33]=.032516842397863206967586;
			m_weights[34]=.032960720063261117074292;
			m_weights[35]=.033363990274077320939719;
			m_weights[36]=.033726156203214570706310;
			m_weights[37]=.034046771663871087167356;
			m_weights[38]=.034325441659239087817965;
			m_weights[39]=.034561822869137808136434;
			m_weights[40]=.034755624072981420920812;
			m_weights[41]=.034906606508560709891011;
			m_weights[42]=.035014584166196443369153;
			m_weights[43]=.035079424017902025317168;
			m_weights[44]=.035101046181273077031624;
			break;
		case 90:
			m_abscissas[0]=.999646971286638437463248;
			m_abscissas[1]=.998140379938568153561306;
			m_abscissas[2]=.995431812058344663926755;
			m_abscissas[3]=.991523928811062786129147;
			m_abscissas[4]=.986421365057832848734254;
			m_abscissas[5]=.980130251345148385458953;
			m_abscissas[6]=.972658162090193139997465;
			m_abscissas[7]=.964014098171505483393667;
			m_abscissas[8]=.95420847388150033616072;
			m_abscissas[9]=.943253103645357768153575;
			m_abscissas[10]=.931161187500432007005847;
			m_abscissas[11]=.917947295066586383372356;
			m_abscissas[12]=.903627347931302693869986;
			m_abscissas[13]=.888218600434745981298376;
			m_abscissas[14]=.871739618862903434474028;
			m_abscissas[15]=.854210259067071882286022;
			m_abscissas[16]=.835651642533377045564199;
			m_abscissas[17]=.816086130929481056443754;
			m_abscissas[18]=.795537299158248134863601;
			m_abscissas[19]=.774029906950334246806958;
			m_abscissas[20]=.751589869029638468178415;
			m_abscissas[21]=.728244223887390363625799;
			m_abscissas[22]=.704021101202391143555469;
			m_abscissas[23]=.678949687946597146181269;
			m_abscissas[24]=.653060193216842191961262;
			m_abscissas[25]=.626383811835045126762609;
			m_abscissas[26]=.598952686760742185888769;
			m_abscissas[27]=.570799870361220978705362;
			m_abscissas[28]=.541959284585913426189305;
			m_abscissas[29]=.512465680093027970988462;
			m_abscissas[30]=.482354594377665692520121;
			m_abscissas[31]=.451662308951869367576379;
			m_abscissas[32]=.420425805628197756101395;
			m_abscissas[33]=.388682721959498206776509;
			m_abscissas[34]=.356471305888567846131189;
			m_abscissas[35]=.323830369662345966511395;
			m_abscissas[36]=.290799243066166651549602;
			m_abscissas[37]=.257417726034420129920888;
			m_abscissas[38]=.223726040694722859268958;
			m_abscissas[39]=.189764782903379019020874;
			m_abscissas[40]=.155574873330529119511405;
			m_abscissas[41]=.121197508153924082968749;
			m_abscissas[42]=.086674109420734770085237;
			m_abscissas[43]=.052046275137206949059279;
			m_abscissas[44]=.017355729146299652461298;
			m_weights[0]=.000905932371214833093725;
			m_weights[1]=.002107778774526329891476;
			m_weights[2]=.003308867243336018195432;
			m_weights[3]=.004506123613674977864138;
			m_weights[4]=.005697981560747352600848;
			m_weights[5]=.006882983208463284314730;
			m_weights[6]=.008059694944620015658672;
			m_weights[7]=.009226696957741990940320;
			m_weights[8]=.010382582309893214613808;
			m_weights[9]=.011525957889148058850594;
			m_weights[10]=.012655445837168128868876;
			m_weights[11]=.013769685112337093430751;
			m_weights[12]=.014867333088043324050385;
			m_weights[13]=.015947067151006639013207;
			m_weights[14]=.017007586285222675709397;
			m_weights[15]=.018047612634460236164050;
			m_weights[16]=.019065893039137318425324;
			m_weights[17]=.020061200544639595964532;
			m_weights[18]=.021032335878722563117062;
			m_weights[19]=.021978128895934133838692;
			m_weights[20]=.022897439987163184634989;
			m_weights[21]=.023789161452528723210101;
			m_weights[22]=.024652218835904852935966;
			m_weights[23]=.025485572219443228484467;
			m_weights[24]=.026288217476514587361596;
			m_weights[25]=.027059187481547958521614;
			m_weights[26]=.027797553275302275158039;
			m_weights[27]=.028502425184161416318755;
			m_weights[28]=.029172953892100742486558;
			m_weights[29]=.029808331464031275487148;
			m_weights[30]=.030407792319286952690394;
			m_weights[31]=.030970614154080920945937;
			m_weights[32]=.031496118811818636076958;
			m_weights[33]=.031983673100218576039456;
			m_weights[34]=.032432689554255616911790;
			m_weights[35]=.032842627144007504578630;
			m_weights[36]=.033212991926551316514041;
			m_weights[37]=.033543337641124276682933;
			m_weights[38]=.033833266246831687257928;
			m_weights[39]=.034082428402253995463605;
			m_weights[40]=.034290523886375041931697;
			m_weights[41]=.034457301960324256174596;
			m_weights[42]=.034582561669496891418054;
			m_weights[43]=.034666152085688240188272;
			m_weights[44]=.034707972488950057920460;
			break;
		case 91:
			m_abscissas[0]=.999654645076576130382989;
			m_abscissas[1]=.99818079259053475467322;
			m_abscissas[2]=.995531042505352453689383;
			m_abscissas[3]=.99170792881727710156762;
			m_abscissas[4]=.986715885723469786907936;
			m_abscissas[5]=.980560780241213439036863;
			m_abscissas[6]=.973249861830690734322468;
			m_abscissas[7]=.964791745747276075776495;
			m_abscissas[8]=.955196400732276363874022;
			m_abscissas[9]=.944475136545718668062791;
			m_abscissas[10]=.932640590356264406612214;
			m_abscissas[11]=.919706711723450488141331;
			m_abscissas[12]=.905688746096571865002581;
			m_abscissas[13]=.890603216814397403791499;
			m_abscissas[14]=.874467905612230385893605;
			m_abscissas[15]=.857301831652642579078184;
			m_abscissas[16]=.839125229101387031796118;
			m_abscissas[17]=.819959523273280333976318;
			m_abscissas[18]=.79982730537529224065049;
			m_abscissas[19]=.778752305876122344694653;
			m_abscissas[20]=.756759366533369288592607;
			m_abscissas[21]=.733874411111094782254468;
			m_abscissas[22]=.710124414822193871743145;
			m_abscissas[23]=.685537372531524655146344;
			m_abscissas[24]=.660142265757234790190265;
			m_abscissas[25]=.633969028509153281498804;
			m_abscissas[26]=.607048512004496052778236;
			m_abscissas[27]=.579412448302463132325792;
			m_abscissas[28]=.551093412900583559842208;
			m_abscissas[29]=.522124786336890588473257;
			m_abscissas[30]=.492540714843183417561822;
			m_abscissas[31]=.462376070095751451638247;
			m_abscissas[32]=.431666408111001803186173;
			m_abscissas[33]=.400447927334439305030325;
			m_abscissas[34]=.368757425972399562862388;
			m_abscissas[35]=.336632258616828491012942;
			m_abscissas[36]=.304110292214235318462596;
			m_abscissas[37]=.271229861430719269390085;
			m_abscissas[38]=.23802972346568211899255;
			m_abscissas[39]=.204549012367488773656851;
			m_abscissas[40]=.170827192904925167218356;
			m_abscissas[41]=.136904014048826416020622;
			m_abscissas[42]=.102819462118707721920625;
			m_abscissas[43]=.068613713649625415750365;
			m_abscissas[44]=.034327088034825330846991;
			m_abscissas[45]=0.;
			m_weights[0]=.000886241240669414176583;
			m_weights[1]=.002061987122032229660682;
			m_weights[2]=.003237045507972104977099;
			m_weights[3]=.004408439747302676819063;
			m_weights[4]=.005574668047479788997832;
			m_weights[5]=.006734334432268884665261;
			m_weights[6]=.007886066314628901599631;
			m_weights[7]=.009028504189234487748913;
			m_weights[8]=.010160300654415476728892;
			m_weights[9]=.011280121273769682983409;
			m_weights[10]=.012386645903556743054535;
			m_weights[11]=.013478570153830979194319;
			m_weights[12]=.014554606885208696084841;
			m_weights[13]=.015613487707050059750951;
			m_weights[14]=.016653964463061240172258;
			m_weights[15]=.017674810697521905060372;
			m_weights[16]=.018674823098168125421786;
			m_weights[17]=.019652822912969146604742;
			m_weights[18]=.020607657338598460740459;
			m_weights[19]=.021538200878685666296224;
			m_weights[20]=.022443356670097373373321;
			m_weights[21]=.023322057775598802835996;
			m_weights[22]=.024173268441322879422218;
			m_weights[23]=.024995985317534957432561;
			m_weights[24]=.02578923864123601618879;
			m_weights[25]=.026552093379198908103079;
			m_weights[26]=.027283650330082980278990;
			m_weights[27]=.027983047184323166381186;
			m_weights[28]=.028649459540541024396496;
			m_weights[29]=.029282101877277479718264;
			m_weights[30]=.029880228478900374932771;
			m_weights[31]=.030443134314594394903447;
			m_weights[32]=.030970155869396544215619;
			m_weights[33]=.031460671926297088545190;
			m_weights[34]=.031914104298483697288599;
			m_weights[35]=.032329918510865394484094;
			m_weights[36]=.032707624430072782948420;
			m_weights[37]=.033046776842191791200169;
			m_weights[38]=.033346975977549838636978;
			m_weights[39]=.033607867981935753109824;
			m_weights[40]=.033829145333697935793656;
			m_weights[41]=.034010547206229078665489;
			m_weights[42]=.034151859775410126185675;
			m_weights[43]=.034252916471651060067192;
			m_weights[44]=.034313598176231398572420;
			m_weights[45]=.034333833361708963163405;
			break;
		case 92:
			m_abscissas[0]=.999662071356524554448781;
			m_abscissas[1]=.998219902195450153782215;
			m_abscissas[2]=.995627075269215169124288;
			m_abscissas[3]=.991886004437253478622237;
			m_abscissas[4]=.98700093392401204090004;
			m_abscissas[5]=.980977481210154027373163;
			m_abscissas[6]=.973822588026941934734002;
			m_abscissas[7]=.965544504421595901364174;
			m_abscissas[8]=.956152777134006946598195;
			m_abscissas[9]=.945658237884362956543102;
			m_abscissas[10]=.934072990607095498232952;
			m_abscissas[11]=.921410397371076054635723;
			m_abscissas[12]=.907685062910925043607568;
			m_abscissas[13]=.892912817752776204232152;
			m_abscissas[14]=.877110699939569669620167;
			m_abscissas[15]=.860296935370446882979538;
			m_abscissas[16]=.842490916773780940406622;
			m_abscissas[17]=.823713181336488764363312;
			m_abscissas[18]=.803985387014566437670438;
			m_abscissas[19]=.783330287551689979201918;
			m_abscissas[20]=.76177170623441627570913;
			m_abscissas[21]=.739334508414088037773129;
			m_abscissas[22]=.716044572827033202401025;
			m_abscissas[23]=.691928761746073039103277;
			m_abscissas[24]=.667014889997724077606074;
			m_abscissas[25]=.641331692880801492841388;
			m_abscissas[26]=.614908793023407720149118;
			m_abscissas[27]=.587776666216520372693532;
			m_abscissas[28]=.559966606263577766234105;
			m_abscissas[29]=.53151068888659787843176;
			m_abscissas[30]=.502441734730456513888601;
			m_abscissas[31]=.47279327150799183509643;
			m_abscissas[32]=.442599495329594231680833;
			m_abscissas[33]=.411895231261881708602946;
			m_abscissas[34]=.380715893160950569903095;
			m_abscissas[35]=.349097442826528183794752;
			m_abscissas[36]=.317076348524138107148509;
			m_abscissas[37]=.284689542923116943093885;
			m_abscissas[38]=.251974380498996180444597;
			m_abscissas[39]=.218968594449380152453027;
			m_abscissas[40]=.185710253173012450702391;
			m_abscissas[41]=.152237716362226996438541;
			m_abscissas[42]=.118589590759425929470549;
			m_abscissas[43]=.084804685628614012678252;
			m_abscissas[44]=.050921967993347923430376;
			m_abscissas[45]=.016980517692728234384057;
			m_weights[0]=.000867185178767142135354;
			m_weights[1]=.002017671366262838591880;
			m_weights[2]=.003167535943396097874259;
			m_weights[3]=.004313895331861700472339;
			m_weights[4]=.005455308908000870987157;
			m_weights[5]=.006590439334214895223178;
			m_weights[6]=.007717971837373568504533;
			m_weights[7]=.008836604056467877374548;
			m_weights[8]=.009945045019726082041771;
			m_weights[9]=.011042015922635394223985;
			m_weights[10]=.012126251362637710529297;
			m_weights[11]=.013196500705711138029112;
			m_weights[12]=.014251529488953925265807;
			m_weights[13]=.015290120825796501506906;
			m_weights[14]=.016311076800255958004814;
			m_weights[15]=.017313219843689776361140;
			m_weights[16]=.018295394090267557291187;
			m_weights[17]=.019256466708559474712372;
			m_weights[18]=.020195329207187483749564;
			m_weights[19]=.021110898712762461809976;
			m_weights[20]=.022002119218485857398744;
			m_weights[21]=.022867962801892542404341;
			m_weights[22]=.023707430810281912393537;
			m_weights[23]=.024519555012440974257171;
			m_weights[24]=.025303398715313225697548;
			m_weights[25]=.026058057844314179222458;
			m_weights[26]=.026782661986040323300488;
			m_weights[27]=.027476375392164173395179;
			m_weights[28]=.028138397943354404514451;
			m_weights[29]=.028767966072107175822380;
			m_weights[30]=.029364353643422812612747;
			m_weights[31]=.029926872792311073307868;
			m_weights[32]=.030454874717158320980635;
			m_weights[33]=.030947750428041031668041;
			m_weights[34]=.031404931449122177916140;
			m_weights[35]=.031825890474320085825973;
			m_weights[36]=.032210141975493329535744;
			m_weights[37]=.032557242762440045243162;
			m_weights[38]=.032866792494065660326468;
			m_weights[39]=.033138434140129381822621;
			m_weights[40]=.033371854393036810307801;
			m_weights[41]=.033566784029203676310076;
			m_weights[42]=.033722998219573871693801;
			m_weights[43]=.033840316788933601891418;
			m_weights[44]=.033918604423722549495027;
			m_weights[45]=.033957770828102347967003;
			break;
		case 93:
			m_abscissas[0]=.999669260657168688326821;
			m_abscissas[1]=.998257764171412775752493;
			m_abscissas[2]=.995720046248950179380265;
			m_abscissas[3]=.992058407192062741161222;
			m_abscissas[4]=.987276911268572265931696;
			m_abscissas[5]=.981380939533501606767866;
			m_abscissas[6]=.974377142133410456975732;
			m_abscissas[7]=.966273423040059125621511;
			m_abscissas[8]=.957078929065123352230644;
			m_abscissas[9]=.946804038848475454452155;
			m_abscissas[10]=.935460350875189251367485;
			m_abscissas[11]=.92306067026585348438925;
			m_abscissas[12]=.909618994265676445699939;
			m_abscissas[13]=.89515049641499902748576;
			m_abscissas[14]=.879671509405002962140068;
			m_abscissas[15]=.863199506631601715124534;
			m_abscissas[16]=.84575308246526130294652;
			m_abscissas[17]=.827351931257451916351477;
			m_abscissas[18]=.808016825106586578646116;
			m_abscissas[19]=.787769590408075201199532;
			m_abscissas[20]=.766633083214692959396041;
			m_abscissas[21]=.744631163434914460185216;
			m_abscissas[22]=.721788667898239650908218;
			m_abscissas[23]=.698131382317853409145991;
			m_abscissas[24]=.673686012182227915571455;
			m_abscissas[25]=.648480152608499893383399;
			m_abscissas[26]=.622542257191635632724262;
			m_abscissas[27]=.595901605884536044341298;
			m_abscissas[28]=.568588271945331683331926;
			m_abscissas[29]=.540633087989173217270797;
			m_abscissas[30]=.512067611182835427355167;
			m_abscissas[31]=.482924087621421657446984;
			m_abscissas[32]=.453235415927379745929265;
			m_abscissas[33]=.423035110112918960207129;
			m_abscissas[34]=.392357261747749387298647;
			m_abscissas[35]=.361236501474849727744629;
			m_abscissas[36]=.32970795991770564279237;
			m_abscissas[37]=.297807228023147909018027;
			m_abscissas[38]=.265570316884556881013476;
			m_abscissas[39]=.233033617090786443643251;
			m_abscissas[40]=.20023385764669609657234;
			m_abscissas[41]=.1672080645116634566664;
			m_abscissas[42]=.133993518802880746725403;
			m_abscissas[43]=.100627714710617278030891;
			m_abscissas[44]=.067148317172955104387614;
			m_abscissas[45]=.033593119357776561271188;
			m_abscissas[46]=0.;
			m_weights[0]=.000848737168067911004880;
			m_weights[1]=.001974768768686808388940;
			m_weights[2]=.003100240403099316775465;
			m_weights[3]=.004222357382406607998633;
			m_weights[4]=.005339737098169214613758;
			m_weights[5]=.006451097794311275889058;
			m_weights[6]=.007555179500769820751619;
			m_weights[7]=.008650734035428648314141;
			m_weights[8]=.009736523941887687826947;
			m_weights[9]=.010811323190542489382026;
			m_weights[10]=.011873918327447125098614;
			m_weights[11]=.012923109753185350456027;
			m_weights[12]=.013957713038007970724070;
			m_weights[13]=.014976560240671880953919;
			m_weights[14]=.015978501217782116788317;
			m_weights[15]=.016962404917329010901227;
			m_weights[16]=.017927160652813713178853;
			m_weights[17]=.01887167935550803461443;
			m_weights[18]=.019794894802927928668055;
			m_weights[19]=.020695764821868734488589;
			m_weights[20]=.021573272464499818015058;
			m_weights[21]=.022426427156109571889107;
			m_weights[22]=.023254265813157751689920;
			m_weights[23]=.024055853930344656153065;
			m_weights[24]=.024830286635452581891835;
			m_weights[25]=.025576689710757838922176;
			m_weights[26]=.026294220579853274752298;
			m_weights[27]=.026982069258762733048788;
			m_weights[28]=.027639459270270719713116;
			m_weights[29]=.028265648520433064357429;
			m_weights[30]=.028859930136277706362907;
			m_weights[31]=.029421633263748977485516;
			m_weights[32]=.029950123824993924165878;
			m_weights[33]=.030444805234135309496476;
			m_weights[34]=.030905119070722935908768;
			m_weights[35]=.031330545710102801925915;
			m_weights[36]=.031720604909992308832588;
			m_weights[37]=.032074856352599219581716;
			m_weights[38]=.032392900141672292706303;
			m_weights[39]=.032674377253922415754864;
			m_weights[40]=.032918969944304591132477;
			m_weights[41]=.033126402104703225972940;
			m_weights[42]=.033296439575615789345242;
			m_weights[43]=.033428890410482966294255;
			m_weights[44]=.033523605092366899732467;
			m_weights[45]=.033580476702732908204233;
			m_weights[46]=.033599441042144139694191;
			break;
		case 94:
			m_abscissas[0]=.999676222955049794478464;
			m_abscissas[1]=.998294431021687606243547;
			m_abscissas[2]=.995810084201718879423191;
			m_abscissas[3]=.992225375400876658453786;
			m_abscissas[4]=.98754419832047978867885;
			m_abscissas[5]=.981771709883378229724351;
			m_abscissas[6]=.974914283790465042598315;
			m_abscissas[7]=.966979495881343388944632;
			m_abscissas[8]=.957976113742955023987959;
			m_abscissas[9]=.947914086348761722901954;
			m_abscissas[10]=.93680453280282537821092;
			m_abscissas[11]=.924659729937951474687734;
			m_abscissas[12]=.911493098694057485956076;
			m_abscissas[13]=.897319189258767992575134;
			m_abscissas[14]=.882153664972872260771444;
			m_abscissas[15]=.866013285012198673277729;
			m_abscissas[16]=.84891588586203340015019;
			m_abscissas[17]=.830880361603017377979289;
			m_abscissas[18]=.811926643029482883145495;
			m_abscissas[19]=.79207567562284516570144;
			m_abscissas[20]=.771349396404123699387454;
			m_abscissas[21]=.749770709691013517218792;
			m_abscissas[22]=.727363461786199059394637;
			m_abscissas[23]=.704152414624820241445598;
			m_abscissas[24]=.680163218410172621032254;
			m_abscissas[25]=.655422383267855255435989;
			m_abscissas[26]=.629957249949673190129547;
			m_abscissas[27]=.603795959619657222727321;
			m_abscissas[28]=.576967422755581620592581;
			m_abscissas[29]=.549501287200340399755088;
			m_abscissas[30]=.521427905398483940978552;
			m_abscissas[31]=.492778300854119354040168;
			m_abscissas[32]=.463584133847239278989447;
			m_abscissas[33]=.433877666446363893398923;
			m_abscissas[34]=.403691726856158945048469;
			m_abscissas[35]=.373059673139427839806139;
			m_abscissas[36]=.342015356353567405830147;
			m_abscissas[37]=.310593083142224186454218;
			m_abscissas[38]=.278827577823490286044164;
			m_abscissas[39]=.246753944016534251555295;
			m_abscissas[40]=.214407625849072610544954;
			m_abscissas[41]=.181824368788550945981578;
			m_abscissas[42]=.149040180140319261529605;
			m_abscissas[43]=.116091289256454421500443;
			m_abscissas[44]=.083014107499202232824769;
			m_abscissas[45]=.049845188003282920429641;
			m_abscissas[46]=.016621185281526033776403;
			m_weights[0]=.000830871612682162494642;
			m_weights[1]=.001933219888725418943118;
			m_weights[2]=.003035065891038628027392;
			m_weights[3]=.004133699875407776483297;
			m_weights[4]=.005227794289507767545306;
			m_weights[5]=.006316120091036448223108;
			m_weights[6]=.007397469288142356200863;
			m_weights[7]=.008470645094534635999910;
			m_weights[8]=.009534460832865158250063;
			m_weights[9]=.010587740564954122236724;
			m_weights[10]=.011629320161122414596074;
			m_weights[11]=.012658048497638994444825;
			m_weights[12]=.013672788690606878506440;
			m_weights[13]=.014672419334499464204265;
			m_weights[14]=.015655835732515557860022;
			m_weights[15]=.016621951112665496638329;
			m_weights[16]=.017569697826143251998726;
			m_weights[17]=.018498028525665910953809;
			m_weights[18]=.019405917321982004886052;
			m_weights[19]=.020292360917011132179888;
			m_weights[20]=.021156379712221389815045;
			m_weights[21]=.021997018890940077173397;
			m_weights[22]=.022813349473355236410012;
			m_weights[23]=.023604469343014382280508;
			m_weights[24]=.024369504243668988306347;
			m_weights[25]=.025107608745352405128580;
			m_weights[26]=.025817967178616728164403;
			m_weights[27]=.026499794535891699196694;
			m_weights[28]=.027152337338966564723883;
			m_weights[29]=.027774874471634220620651;
			m_weights[30]=.028366717976576106812729;
			m_weights[31]=.028927213815606255842275;
			m_weights[32]=.029455742592433676397192;
			m_weights[33]=.029951720237143869200088;
			m_weights[34]=.030414598651642712203281;
			m_weights[35]=.030843866315349187411107;
			m_weights[36]=.031239048850467417882191;
			m_weights[37]=.031599709546213200464774;
			m_weights[38]=.031925449841415613925841;
			m_weights[39]=.032215909764960307112818;
			m_weights[40]=.032470768333587679484508;
			m_weights[41]=.032689743906606307152528;
			m_weights[42]=.032872594497129590726148;
			m_weights[43]=.033019118039491655076671;
			m_weights[44]=.033129152612546963216005;
			m_weights[45]=.033202576618606863798766;
			m_weights[46]=.033239308917815320800705;
			break;
		case 95:
			m_abscissas[0]=.999682967707192019483573;
			m_abscissas[1]=.998329952516746331615185;
			m_abscissas[2]=.995897311188423402061516;
			m_abscissas[3]=.992387135003501404489007;
			m_abscissas[4]=.987803155906081287771609;
			m_abscissas[5]=.982150318221625646102971;
			m_abscissas[6]=.975434733411440911686142;
			m_abscissas[7]=.967663666022196524115609;
			m_abscissas[8]=.958845523847909449427667;
			m_abscissas[9]=.948989848174295482042095;
			m_abscissas[10]=.938107303198478037136254;
			m_abscissas[11]=.926209664377719775633338;
			m_abscissas[12]=.913309805634067347397871;
			m_abscissas[13]=.89942168539639320374082;
			m_abscissas[14]=.884560331481440557481542;
			m_abscissas[15]=.868741824824129904742858;
			m_abscissas[16]=.851983282071784280841952;
			m_abscissas[17]=.834302837059601322207667;
			m_abscissas[18]=.81571962118660965386023;
			m_abscissas[19]=.796253742712893188667863;
			m_abscissas[20]=.775926265000223956226838;
			m_abscissas[21]=.754759183719492392253657;
			m_abscissas[22]=.732775403049502118342711;
			m_abscissas[23]=.709998710892823105801743;
			m_abscissas[24]=.686453753135482209243489;
			m_abscissas[25]=.662166006978318026262701;
			m_abscissas[26]=.637161753368840027179725;
			m_abscissas[27]=.611468048563410696074958;
			m_abscissas[28]=.585112694850514089292359;
			m_abscissas[29]=.558124210466784403949682;
			m_abscissas[30]=.530531798738343289256909;
			m_abscissas[31]=.502365316480834043324797;
			m_abscissas[32]=.473655241692343777570581;
			m_abscissas[33]=.444432640574170340460532;
			m_abscissas[34]=.414729133915118515972148;
			m_abscissas[35]=.384576862875699010848323;
			m_abscissas[36]=.354008454209253305823369;
			m_abscissas[37]=.323056984957636889277322;
			m_abscissas[38]=.291755946659662073485054;
			m_abscissas[39]=.260139209111028909464386;
			m_abscissas[40]=.22824098371495810363821;
			m_abscissas[41]=.196095786463182778498271;
			m_abscissas[42]=.163738400587355934951945;
			m_abscissas[43]=.131203838921287136147017;
			m_abscissas[44]=.098527306014734857520207;
			m_abscissas[45]=.065744160039749798572533;
			m_abscissas[46]=.032889874530788938676986;
			m_abscissas[47]=0.;
			m_weights[0]=.000813564249454116501054;
			m_weights[1]=.001892968377922935762779;
			m_weights[2]=.002971924240818190718439;
			m_weights[3]=.004047803316371759906880;
			m_weights[4]=.005119330329927718280032;
			m_weights[5]=.006185326261033323769312;
			m_weights[6]=.007244632443933199672626;
			m_weights[7]=.008296100874530990238146;
			m_weights[8]=.009338593083876397086741;
			m_weights[9]=.010370980703116096840840;
			m_weights[10]=.011392146459085844039242;
			m_weights[11]=.012400985296116061040182;
			m_weights[12]=.013396405534368285441366;
			m_weights[13]=.014377330033659082083575;
			m_weights[14]=.015342697350288356634592;
			m_weights[15]=.016291462880991043265916;
			m_weights[16]=.017222599990716984413340;
			m_weights[17]=.018135101122045144107597;
			m_weights[18]=.019027978884545706393070;
			m_weights[19]=.019900267122657211244872;
			m_weights[20]=.020751021960784901817908;
			m_weights[21]=.021579322824411401206769;
			m_weights[22]=.022384273436069391840419;
			m_weights[23]=.023165002785071391749200;
			m_weights[24]=.023920666069930610077075;
			m_weights[25]=.024650445612442619976130;
			m_weights[26]=.025353551742432012936600;
			m_weights[27]=.026029223652202271532901;
			m_weights[28]=.026676730219761354318968;
			m_weights[29]=.027295370799930222665784;
			m_weights[30]=.027884475982476914243093;
			m_weights[31]=.028443408316454862613119;
			m_weights[32]=.028971562999961011534842;
			m_weights[33]=.029468368534566882375151;
			m_weights[34]=.029933287343714112252400;
			m_weights[35]=.030365816354405066777244;
			m_weights[36]=.030765487541558914752958;
			m_weights[37]=.031131868434443998256823;
			m_weights[38]=.031464562584638402013217;
			m_weights[39]=.031763209995012280290979;
			m_weights[40]=.032027487509267695292957;
			m_weights[41]=.032257109161614414347348;
			m_weights[42]=.032451826486203259266859;
			m_weights[43]=.032611428785982154256707;
			m_weights[44]=.032735743360683932269197;
			m_weights[45]=.032824635693699186693089;
			m_weights[46]=.032878009597631948235576;
			m_weights[47]=.032895807318380677932174;
			break;
		case 96:
			m_abscissas[0]=.99968950388323076682769;
			m_abscissas[1]=.998364375863181677724149;
			m_abscissas[2]=.995981842987209290650399;
			m_abscissas[3]=.992543900323762624571892;
			m_abscissas[4]=.988054126329623799480763;
			m_abscissas[5]=.982517263563014677447046;
			m_abscissas[6]=.975939174585136466452601;
			m_abscissas[7]=.968326828463264212173659;
			m_abscissas[8]=.959688291448742539300068;
			m_abscissas[9]=.950032717784437635756099;
			m_abscissas[10]=.939370339752755216931857;
			m_abscissas[11]=.92771245672230869096469;
			m_abscissas[12]=.915071423120898074205884;
			m_abscissas[13]=.901460635315852341319233;
			m_abscissas[14]=.886894517402420416056877;
			m_abscissas[15]=.871388505909296502873775;
			m_abscissas[16]=.854959033434601455462787;
			m_abscissas[17]=.837623511228187121494303;
			m_abscissas[18]=.819400310737931675539;
			m_abscissas[19]=.800308744139140817228796;
			m_abscissas[20]=.780369043867433217603604;
			m_abscissas[21]=.75960234117664749870297;
			m_abscissas[22]=.738030643744400132851166;
			m_abscissas[23]=.715676812348967626225144;
			m_abscissas[24]=.692564536642171561344246;
			m_abscissas[25]=.668718310043916153952557;
			m_abscissas[26]=.644163403784967106798412;
			m_abscissas[27]=.618925840125468570386369;
			m_abscissas[28]=.593032364777572080683556;
			m_abscissas[29]=.56651041856139716840425;
			m_abscissas[30]=.539388108324357436226803;
			m_abscissas[31]=.51169417715466767358551;
			m_abscissas[32]=.483457973920596359768406;
			m_abscissas[33]=.454709422167743008635676;
			m_abscissas[34]=.425478988407300545364819;
			m_abscissas[35]=.395797649828908603285;
			m_abscissas[36]=.365696861472313635030896;
			m_abscissas[37]=.335208522892625422616326;
			m_abscissas[38]=.30436494435449635302393;
			m_abscissas[39]=.273198812591049141487272;
			m_abscissas[40]=.241743156163840012327932;
			m_abscissas[41]=.210031310460567203602847;
			m_abscissas[42]=.178096882367618602759403;
			m_abscissas[43]=.145973714654896941989107;
			m_abscissas[44]=.113695850110665920911208;
			m_abscissas[45]=.081297495464425558994471;
			m_abscissas[46]=.048812985136049731111958;
			m_abscissas[47]=.016276744849602969579135;
			m_weights[0]=.000796792065552012429438;
			m_weights[1]=.001853960788946921732336;
			m_weights[2]=.002910731817934946408411;
			m_weights[3]=.003964554338444686673733;
			m_weights[4]=.005014202742927517692470;
			m_weights[5]=.006058545504235961683317;
			m_weights[6]=.007096470791153865269144;
			m_weights[7]=.008126876925698759217382;
			m_weights[8]=.009148671230783386632585;
			m_weights[9]=.010160770535008415757588;
			m_weights[10]=.011162102099838498591213;
			m_weights[11]=.012151604671088319635181;
			m_weights[12]=.013128229566961572637064;
			m_weights[13]=.014090941772314860915862;
			m_weights[14]=.015038721026994938005876;
			m_weights[15]=.015970562902562291380616;
			m_weights[16]=.016885479864245172450478;
			m_weights[17]=.017782502316045260837614;
			m_weights[18]=.018660679627411467385157;
			m_weights[19]=.019519081140145022410085;
			m_weights[20]=.020356797154333324595245;
			m_weights[21]=.021172939892191298987674;
			m_weights[22]=.021966644438744349194756;
			m_weights[23]=.022737069658329374001348;
			m_weights[24]=.023483399085926219842236;
			m_weights[25]=.024204841792364691282267;
			m_weights[26]=.024900633222483610288382;
			m_weights[27]=.025570036005349361498797;
			m_weights[28]=.026212340735672413913458;
			m_weights[29]=.026826866725591762198057;
			m_weights[30]=.027412962726029242823421;
			m_weights[31]=.027970007616848334439819;
			m_weights[32]=.028497411065085385645600;
			m_weights[33]=.028994614150555236542679;
			m_weights[34]=.029461089958167905970436;
			m_weights[35]=.029896344136328385984388;
			m_weights[36]=.030299915420827593794089;
			m_weights[37]=.030671376123669149014229;
			m_weights[38]=.031010332586313837423250;
			m_weights[39]=.031316425596861355812784;
			m_weights[40]=.031589330770727168558021;
			m_weights[41]=.031828758894411006534754;
			m_weights[42]=.032034456231992663218139;
			m_weights[43]=.032206204794030250668667;
			m_weights[44]=.032343822568575928428775;
			m_weights[45]=.032447163714064269364013;
			m_weights[46]=.032516118713868835987205;
			m_weights[47]=.032550614492363166241961;
			break;
		case 97:
			m_abscissas[0]=.999695839995246724911968;
			m_abscissas[1]=.998397745860566436733244;
			m_abscissas[2]=.996063789477488413024312;
			m_abscissas[3]=.99269587477855160925815;
			m_abscissas[4]=.988297434501838787317894;
			m_abscissas[5]=.982873019613528351035924;
			m_abscissas[6]=.976428256308619764924245;
			m_abscissas[7]=.96896983303495092210011;
			m_abscissas[8]=.960505491652787246157486;
			m_abscissas[9]=.951044018766639195210907;
			m_abscissas[10]=.940595236358027984371685;
			m_abscissas[11]=.92916999148161914082226;
			m_abscissas[12]=.916780144953188888049796;
			m_abscissas[13]=.903438559010121616414912;
			m_abscissas[14]=.889159083944300362039029;
			m_abscissas[15]=.873956543715414215359203;
			m_abscissas[16]=.857846720556790301260558;
			m_abscissas[17]=.840846338588286767457515;
			m_abscissas[18]=.82297304645248444175042;
			m_abscissas[19]=.804245398991770565478888;
			m_abscissas[20]=.784682837985085728830414;
			m_abscissas[21]=.764305671964182056902006;
			m_abscissas[22]=.743135055130253938043062;
			m_abscissas[23]=.721192965392770524946237;
			m_abscissas[24]=.698502181553271116636779;
			m_abscissas[25]=.675086259657784965824565;
			m_abscissas[26]=.650969508542408255109624;
			m_abscissas[27]=.626176964597413854512989;
			m_abscissas[28]=.600734365776084208376919;
			m_abscissas[29]=.574668124875244104327378;
			m_abscissas[30]=.548005302115227734655986;
			m_abscissas[31]=.52077357704774283654722;
			m_abscissas[32]=.493001219820793197245321;
			m_abscissas[33]=.464717061830488813407276;
			m_abscissas[34]=.435950465790209876438386;
			m_abscissas[35]=.406731295248195902884206;
			m_abscissas[36]=.377089883585204144300973;
			m_abscissas[37]=.347057002524421321481913;
			m_abscissas[38]=.316663830186319188061198;
			m_abscissas[39]=.285941918721616923173769;
			m_abscissas[40]=.254923161555951399025067;
			m_abscissas[41]=.223639760280259517555185;
			m_abscissas[42]=.19212419122124464623572;
			m_abscissas[43]=.16040917172663132738278;
			m_abscissas[44]=.128527626200208545469683;
			m_abscissas[45]=.096512651921921606901657;
			m_abscissas[46]=.064397484688495848021885;
			m_abscissas[47]=.032215464310261708911052;
			m_abscissas[48]=0.;
			m_weights[0]=.000780533221942561245724;
			m_weights[1]=.001816146398210039609508;
			m_weights[2]=.002851409243213055771420;
			m_weights[3]=.003883845329489294421733;
			m_weights[4]=.004912276262166028130835;
			m_weights[5]=.005935615630788222954183;
			m_weights[6]=.006952796096469405526465;
			m_weights[7]=.007962759997865495595599;
			m_weights[8]=.008964458176697999432565;
			m_weights[9]=.009956850427084044948238;
			m_weights[10]=.010938906359195948953967;
			m_weights[11]=.011909606385330756832736;
			m_weights[12]=.012867942742493386675711;
			m_weights[13]=.013812920521853043271149;
			m_weights[14]=.014743558693236950176360;
			m_weights[15]=.015658891119156920520207;
			m_weights[16]=.016557967555342456629028;
			m_weights[17]=.017439854635807804639405;
			m_weights[18]=.018303636840964140822292;
			m_weights[19]=.019148417447528129335257;
			m_weights[20]=.019973319459108046888189;
			m_weights[21]=.020777486516426568497990;
			m_weights[22]=.021560083786191718278435;
			m_weights[23]=.022320298827667132378623;
			m_weights[24]=.023057342436025995796396;
			m_weights[25]=.023770449461603068821042;
			m_weights[26]=.024458879604187847290600;
			m_weights[27]=.025121918181530046735652;
			m_weights[28]=.025758876871256784895352;
			m_weights[29]=.026369094425429349757079;
			m_weights[30]=.026951937356996440673634;
			m_weights[31]=.027506800597430342560096;
			m_weights[32]=.028033108124862677526805;
			m_weights[33]=.028530313562067187518238;
			m_weights[34]=.028997900743668431872052;
			m_weights[35]=.029435384251987320864243;
			m_weights[36]=.029842309920967029034578;
			m_weights[37]=.030218255307656014534521;
			m_weights[38]=.030562830130758583861351;
			m_weights[39]=.030875676675797653824326;
			m_weights[40]=.031156470166469041457751;
			m_weights[41]=.031404919101801723624578;
			m_weights[42]=.031620765558774016042949;
			m_weights[43]=.031803785460071490444954;
			m_weights[44]=.031953788806708641945284;
			m_weights[45]=.032070619875272799349280;
			m_weights[46]=.032154157379585501535780;
			m_weights[47]=.032204314596613505334757;
			m_weights[48]=.032221039456498706326168;
			break;
		case 98:
			m_abscissas[0]=.999701984125490903637295;
			m_abscissas[1]=.998430105047230453778693;
			m_abscissas[2]=.99614325499685929606298;
			m_abscissas[3]=.99284325153690673013188;
			m_abscissas[4]=.988533388989158168372173;
			m_abscissas[5]=.983218036293722991285338;
			m_abscissas[6]=.976902595063886082499069;
			m_abscissas[7]=.969593487100794472319864;
			m_abscissas[8]=.961298146002891057710418;
			m_abscissas[9]=.952025008986401501308772;
			m_abscissas[10]=.941783508062711821098796;
			m_abscissas[11]=.930584060339307042124495;
			m_abscissas[12]=.918438057373568117197214;
			m_abscissas[13]=.905357853559839135953333;
			m_abscissas[14]=.891356753548891066431957;
			m_abscissas[15]=.876448998706844105112915;
			m_abscissas[16]=.860649752624551753159007;
			m_abscissas[17]=.843975085690770467285956;
			m_abscissas[18]=.826441958744047855107785;
			m_abscissas[19]=.808068205819534351294907;
			m_abscissas[20]=.788872516008022266700967;
			m_abscissas[21]=.768874414445517797116596;
			m_abscissas[22]=.748094242452592450640567;
			m_abscissas[23]=.72655313684365848091257;
			m_abscissas[24]=.704273008427177423656977;
			m_abscissas[25]=.681276519718646320796039;
			m_abscissas[26]=.657587061889014905160305;
			m_abscissas[27]=.633228730971969861881956;
			m_abscissas[28]=.608226303354279533514417;
			m_abscissas[29]=.582605210574123959653296;
			m_abscissas[30]=.556391513453040586457398;
			m_abscissas[31]=.529611875587794876248436;
			m_abscissas[32]=.502293536229136859823462;
			m_abscissas[33]=.474464282575028844127023;
			m_abscissas[34]=.446152421506525448597057;
			m_abscissas[35]=.41738675079505433418147;
			m_abscissas[36]=.388196529810383863720304;
			m_abscissas[37]=.358611449759071965248417;
			m_abscissas[38]=.328661603483668158817609;
			m_abscissas[39]=.298377454853387577158828;
			m_abscissas[40]=.267789807777391413732747;
			m_abscissas[41]=.236929774872192150992228;
			m_abscissas[42]=.205828745815053770307665;
			m_abscissas[43]=.174518355415576567874967;
			m_abscissas[44]=.14303045143794287516603;
			m_abscissas[45]=.111397062206553617998208;
			m_abscissas[46]=.079650364028005988212934;
			m_abscissas[47]=.047822648462549322962768;
			m_abscissas[48]=.015946289478309399287564;
			m_weights[0]=.000764766982274313458039;
			m_weights[1]=.001779477041014528741695;
			m_weights[2]=.002793881135722130870630;
			m_weights[3]=.003805574085352359565515;
			m_weights[4]=.004813422398586770918478;
			m_weights[5]=.005816382546439639112763;
			m_weights[6]=.006813429479165215998772;
			m_weights[7]=.007803547379100754890980;
			m_weights[8]=.008785728467392263202699;
			m_weights[9]=.009758973402174096835348;
			m_weights[10]=.010722292083224317120244;
			m_weights[11]=.011674704577138124287429;
			m_weights[12]=.012615242078921952857781;
			m_weights[13]=.013542947881029465143648;
			m_weights[14]=.014456878338304401977569;
			m_weights[15]=.015356103823497755768498;
			m_weights[16]=.016239709670453695652727;
			m_weights[17]=.017106797103089900262354;
			m_weights[18]=.017956484148770628122443;
			m_weights[19]=.018787906534904686561487;
			m_weights[20]=.019600218567726330773237;
			m_weights[21]=.020392593992291914579484;
			m_weights[22]=.021164226832774856911280;
			m_weights[23]=.021914332212178650419019;
			m_weights[24]=.022642147150618433111263;
			m_weights[25]=.023346931341349274712683;
			m_weights[26]=.024027967903745498803241;
			m_weights[27]=.024684564112460996181980;
			m_weights[28]=.025316052102026097343147;
			m_weights[29]=.025921789546162448918468;
			m_weights[30]=.026501160311123639352488;
			m_weights[31]=.027053575082396128277676;
			m_weights[32]=.027578471964122393900100;
			m_weights[33]=.028075317050636138753246;
			m_weights[34]=.028543604969527885703490;
			m_weights[35]=.028982859395688342047449;
			m_weights[36]=.029392633535806492167763;
			m_weights[37]=.029772510582829476266176;
			m_weights[38]=.030122104139921898848531;
			m_weights[39]=.030441058613493258394908;
			m_weights[40]=.030729049574893669920014;
			m_weights[41]=.030985784090409934631043;
			m_weights[42]=.031211001019226264416841;
			m_weights[43]=.031404471279046561517489;
			m_weights[44]=.031565998079108052901451;
			m_weights[45]=.031695417120349251609074;
			m_weights[46]=.031792596762528630198318;
			m_weights[47]=.031857438158124010713099;
			m_weights[48]=.031889875352876467277945;
			break;
		case 99:
			m_abscissas[0]=.999707943952169355863584;
			m_abscissas[1]=.998461493835841338972881;
			m_abscissas[2]=.996220338673088608562895;
			m_abscissas[3]=.992986214133112277329634;
			m_abscissas[4]=.988762282989873612510342;
			m_abscissas[5]=.983552741156280012965562;
			m_abscissas[6]=.977362776750705944569794;
			m_abscissas[7]=.97019855807429021548073;
			m_abscissas[8]=.962067225640901942739071;
			m_abscissas[9]=.952976884453373502575668;
			m_abscissas[10]=.942936595688479410294659;
			m_abscissas[11]=.93195636756149696436165;
			m_abscissas[12]=.92004714530050021871513;
			m_abscissas[13]=.90722080021057356823788;
			m_abscissas[14]=.893490117826414215002245;
			m_abscissas[15]=.878868785159513850270077;
			m_abscissas[16]=.863371377049917044434898;
			m_abscissas[17]=.84701334163477414729914;
			m_abscissas[18]=.82981098494743056502905;
			m_abscissas[19]=.811781454661989134683012;
			m_abscissas[20]=.7929427229993088939085;
			m_abscissas[21]=.773313568811336145198437;
			m_abscissas[22]=.752913558861538118254508;
			m_abscissas[23]=.731763028320043522180818;
			m_abscissas[24]=.709883060492896858206503;
			m_abscissas[25]=.687295465805609097500626;
			m_abscissas[26]=.664022760061938457558342;
			m_abscissas[27]=.64008814199956252510282;
			m_abscissas[28]=.615515470165007177050707;
			m_abscissas[29]=.59032923913087856421665;
			m_abscissas[30]=.564554555079101550743094;
			m_abscissas[31]=.538217110774501032544869;
			m_abscissas[32]=.511343159953671014219137;
			m_abscissas[33]=.483959491154659700177257;
			m_abscissas[34]=.456093401013556637292961;
			m_abscissas[35]=.427772667054599621755168;
			m_abscissas[36]=.399025520000924152379318;
			m_abscissas[37]=.369880615633556193728667;
			m_abscissas[38]=.340367006226699443041391;
			m_abscissas[39]=.310514111587790736323991;
			m_abscissas[40]=.280351689731191267177665;
			m_abscissas[41]=.249909807214746539358834;
			m_abscissas[42]=.219218809168784070389444;
			m_abscissas[43]=.188309289047424476421635;
			m_abscissas[44]=.157212058132358394383799;
			m_abscissas[45]=.125958114819488461718241;
			m_abscissas[46]=.094578613719052031876439;
			m_abscissas[47]=.063104834600026240206508;
			m_abscissas[48]=.031568151209772265203458;
			m_abscissas[49]=0.;
			m_weights[0]=.000749473646737405363375;
			m_weights[1]=.001743906958219244938637;
			m_weights[2]=.002738075873626878091325;
			m_weights[3]=.003729643487243034749199;
			m_weights[4]=.004717519037520830079687;
			m_weights[5]=.005700699773395926875152;
			m_weights[6]=.006678200860575098165184;
			m_weights[7]=.007649046279335257935390;
			m_weights[8]=.008612267615478888991733;
			m_weights[9]=.009566904411326136356899;
			m_weights[10]=.010512004915524745405749;
			m_weights[11]=.011446626951498253761133;
			m_weights[12]=.012369838822175162103433;
			m_weights[13]=.013280720222657283479954;
			m_weights[14]=.014178363149579446066143;
			m_weights[15]=.015061872801990233312953;
			m_weights[16]=.015930368470960849711038;
			m_weights[17]=.016782984416138707089503;
			m_weights[18]=.017618870727924380506757;
			m_weights[19]=.018437194174178499270985;
			m_weights[20]=.019237139030487184798674;
			m_weights[21]=.020017907893086566207948;
			m_weights[22]=.020778722473594211207425;
			m_weights[23]=.021518824374730223818247;
			m_weights[24]=.022237475846239371584356;
			m_weights[25]=.022933960520251055284083;
			m_weights[26]=.023607584125337894046618;
			m_weights[27]=.024257675178557078231640;
			m_weights[28]=.024883585654781946445987;
			m_weights[29]=.025484691632654754650582;
			m_weights[30]=.026060393916515482540929;
			m_weights[31]=.026610118633685851301792;
			m_weights[32]=.027133317806512550926393;
			m_weights[33]=.027629469898599012322076;
			m_weights[34]=.028098080334680911265934;
			m_weights[35]=.028538681993626949726637;
			m_weights[36]=.028950835674073310403739;
			m_weights[37]=.029334130532227503476433;
			m_weights[38]=.029688184491405098448018;
			m_weights[39]=.030012644622891034471903;
			m_weights[40]=.030307187497745803979613;
			m_weights[41]=.030571519509205779992182;
			m_weights[42]=.030805377165356279499179;
			m_weights[43]=.031008527351785595358335;
			m_weights[44]=.031180767563958158370334;
			m_weights[45]=.031321926109075180128175;
			m_weights[46]=.031431862277221546161521;
			m_weights[47]=.031510466481628347713238;
			m_weights[48]=.031557660367911228858092;
			m_weights[49]=.031573396892175653051374;
			break;
		case 100:
			m_abscissas[0]=.999713726773441233678228;
			m_abscissas[1]=.998491950639595818400163;
			m_abscissas[2]=.996295134733125149186132;
			m_abscissas[3]=.99312493703744345965201;
			m_abscissas[4]=.988984395242991748004419;
			m_abscissas[5]=.9838775407060570154961;
			m_abscissas[6]=.977809358486918288553781;
			m_abscissas[7]=.970785775763706331930898;
			m_abscissas[8]=.962813654255815527293659;
			m_abscissas[9]=.953900782925491742849337;
			m_abscissas[10]=.944055870136255977962775;
			m_abscissas[11]=.933288535043079545924334;
			m_abscissas[12]=.921609298145333952666951;
			m_abscissas[13]=.909029570982529690467126;
			m_abscissas[14]=.895561644970726986698521;
			m_abscissas[15]=.881218679385018415573317;
			m_abscissas[16]=.86601468849716462341074;
			m_abscissas[17]=.849964527879591284293363;
			m_abscissas[18]=.833083879888400823542916;
			m_abscissas[19]=.815389238339176254393989;
			m_abscissas[20]=.796897892390314476389573;
			m_abscissas[21]=.777627909649495475627551;
			m_abscissas[22]=.757598118519707176035668;
			m_abscissas[23]=.736828089802020705512428;
			m_abscissas[24]=.715338117573056446459967;
			m_abscissas[25]=.693149199355801965948648;
			m_abscissas[26]=.670283015603141015802587;
			m_abscissas[27]=.64676190851412927983263;
			m_abscissas[28]=.622608860203707771604191;
			m_abscissas[29]=.597847470247178721264807;
			m_abscissas[30]=.57250193262138119131687;
			m_abscissas[31]=.546597012065094167467994;
			m_abscissas[32]=.520158019881763056646816;
			m_abscissas[33]=.493210789208190933569309;
			m_abscissas[34]=.465781649773358042249217;
			m_abscissas[35]=.437897402172031513108978;
			m_abscissas[36]=.409585291678301542528868;
			m_abscissas[37]=.380872981624629956763363;
			m_abscissas[38]=.351788526372421720972344;
			m_abscissas[39]=.322360343900529151722477;
			m_abscissas[40]=.292617188038471964737556;
			m_abscissas[41]=.262588120371503479168929;
			m_abscissas[42]=.23230248184497396964951;
			m_abscissas[43]=.201789864095735997236049;
			m_abscissas[44]=.171080080538603274887532;
			m_abscissas[45]=.140203137236113973207515;
			m_abscissas[46]=.109189203580061115003426;
			m_abscissas[47]=.078068582813436636694817;
			m_abscissas[48]=.046871682421591631614924;
			m_abscissas[49]=.015628984421543082872217;
			m_weights[0]=.000734634490505671730406;
			m_weights[1]=.001709392653518105239529;
			m_weights[2]=.002683925371553482419440;
			m_weights[3]=.003655961201326375182342;
			m_weights[4]=.004624450063422119351096;
			m_weights[5]=.005588428003865515157212;
			m_weights[6]=.006546948450845322764152;
			m_weights[7]=.007499073255464711578829;
			m_weights[8]=.008443871469668971402621;
			m_weights[9]=.009380419653694457951418;
			m_weights[10]=.010307802574868969585782;
			m_weights[11]=.011225114023185977117222;
			m_weights[12]=.012131457662979497407745;
			m_weights[13]=.013025947892971542285559;
			m_weights[14]=.013907710703718772687954;
			m_weights[15]=.014775884527441301768880;
			m_weights[16]=.015629621077546002723937;
			m_weights[17]=.016468086176145212643105;
			m_weights[18]=.017290460568323582439344;
			m_weights[19]=.018095940722128116664391;
			m_weights[20]=.018883739613374904552941;
			m_weights[21]=.019653087494435305865381;
			m_weights[22]=.020403232646209432766839;
			m_weights[23]=.021133442112527641542672;
			m_weights[24]=.021843002416247386313954;
			m_weights[25]=.022531220256336272701797;
			m_weights[26]=.023197423185254121622489;
			m_weights[27]=.023840960265968205962560;
			m_weights[28]=.024461202707957052719975;
			m_weights[29]=.025057544481579589703764;
			m_weights[30]=.025629402910208116075642;
			m_weights[31]=.026176219239545676342309;
			m_weights[32]=.026697459183570962660385;
			m_weights[33]=.027192613446576880136492;
			m_weights[34]=.027661198220792388294204;
			m_weights[35]=.028102755659101173317648;
			m_weights[36]=.028516854322395097990937;
			m_weights[37]=.028903089601125203134876;
			m_weights[38]=.029261084110638276620119;
			m_weights[39]=.029590488059912642511755;
			m_weights[40]=.029890979593332830916837;
			m_weights[41]=.030162265105169144919069;
			m_weights[42]=.030404079526454820016508;
			m_weights[43]=.030616186583980448496459;
			m_weights[44]=.030798379031152590427714;
			m_weights[45]=.030950478850490988234063;
			m_weights[46]=.031072337427566516587810;
			m_weights[47]=.031163835696209906783818;
			m_weights[48]=.031224884254849357732376;
			m_weights[49]=.031255423453863356947642;
			break;
		default:
			m_n=0;
			m_abscissas.resize(0);
			m_weights.resize(0);
			throw ErrN(m_n);
			return;
		}

		for(i=0,j=m_n-1;j>i;i++,j--)
		{
			m_abscissas[j]=-m_abscissas[i];
			m_weights[j]=m_weights[i];
		}
		break;

	// GaussLobatto ****************************************************
	case GaussLobatto:
		switch (m_n)
		{
		case 2:
			m_weights[0]=1.;
			break;
		case 3:
			m_abscissas[1]=0.;
			m_weights[0]=.333333333333333333333333;
			m_weights[1]=1.333333333333333333333333;
			break;
		case 4:
			m_abscissas[1]=.447213595499957939281835;
			m_weights[0]=.166666666666666666666667;
			m_weights[1]=.833333333333333333333333;
			break;
		case 5:
			m_abscissas[1]=.654653670707977143798292;
			m_abscissas[2]=0.;
			m_weights[0]=.1;
			m_weights[1]=.544444444444444444444444;
			m_weights[2]=.711111111111111111111111;
			break;
		case 6:
			m_abscissas[1]=.765055323929464692851003;
			m_abscissas[2]=.285231516480645096314151;
			m_weights[0]=.066666666666666666666667;
			m_weights[1]=.378474956297846980316613;
			m_weights[2]=.554858377035486353016721;
			break;
		case 7:
			m_abscissas[1]=.830223896278566929872032;
			m_abscissas[2]=.468848793470714213803772;
			m_abscissas[3]=0.;
			m_weights[0]=.047619047619047619047619;
			m_weights[1]=.2768260473615659480107;
			m_weights[2]=.431745381209862623417871;
			m_weights[3]=.487619047619047619047619;
			break;
		case 8:
			m_abscissas[1]=.871740148509606615337446;
			m_abscissas[2]=.591700181433142302144511;
			m_abscissas[3]=.209299217902478868768657;
			m_weights[0]=.035714285714285714285714;
			m_weights[1]=.210704227143506039382992;
			m_weights[2]=.341122692483504364764241;
			m_weights[3]=.412458794658703881567053;
			break;
		case 9:
			m_abscissas[1]=.899757995411460157312345;
			m_abscissas[2]=.677186279510737753445885;
			m_abscissas[3]=.363117463826178158710752;
			m_abscissas[4]=0.;
			m_weights[0]=.027777777777777777777778;
			m_weights[1]=.16549536156080552504634;
			m_weights[2]=.274538712500161735280706;
			m_weights[3]=.346428510973046345115132;
			m_weights[4]=.371519274376417233560091;
			break;
		case 10:
			m_abscissas[1]=.919533908166458813828933;
			m_abscissas[2]=.738773865105505075003106;
			m_abscissas[3]=.477924949810444495661175;
			m_abscissas[4]=.16527895766638702462622;
			m_weights[0]=.022222222222222222222222;
			m_weights[1]=.133305990851070111126227;
			m_weights[2]=.224889342063126452119458;
			m_weights[3]=.292042683679683757875582;
			m_weights[4]=.327539761183897456656511;
			break;
		case 11:
			m_abscissas[1]=.934001430408059134332274;
			m_abscissas[2]=.784483473663144418622418;
			m_abscissas[3]=.565235326996205006470964;
			m_abscissas[4]=.295758135586939391431912;
			m_abscissas[5]=0.;
			m_weights[0]=.018181818181818181818182;
			m_weights[1]=.109612273266994864461403;
			m_weights[2]=.187169881780305204108142;
			m_weights[3]=.248048104264028314040085;
			m_weights[4]=.286879124779008088679222;
			m_weights[5]=.300217595455690693785932;
			break;
		case 12:
			m_abscissas[1]=.94489927222288222340758;
			m_abscissas[2]=.819279321644006678348642;
			m_abscissas[3]=.632876153031860677662405;
			m_abscissas[4]=.39953094096534893226435;
			m_abscissas[5]=.136552932854927554864062;
			m_weights[0]=.015151515151515151515152;
			m_weights[1]=.091684517413196130668343;
			m_weights[2]=.157974705564370115164671;
			m_weights[3]=.212508417761021145358302;
			m_weights[4]=.251275603199201280293244;
			m_weights[5]=.271405240910696177000288;
			break;
		case 13:
			m_abscissas[1]=.953309846642163911896905;
			m_abscissas[2]=.846347564651872316865926;
			m_abscissas[3]=.686188469081757426072759;
			m_abscissas[4]=.482909821091336201746937;
			m_abscissas[5]=.249286930106239992568674;
			m_abscissas[6]=0.;
			m_weights[0]=.012820512820512820512821;
			m_weights[1]=.077801686746818927793589;
			m_weights[2]=.134981926689608349119915;
			m_weights[3]=.183646865203550092007494;
			m_weights[4]=.220767793566110086085534;
			m_weights[5]=.244015790306676356458578;
			m_weights[6]=.251930849333446736044139;
			break;
		case 14:
			m_abscissas[1]=.9599350452672609013551;
			m_abscissas[2]=.86780105383034725100022;
			m_abscissas[3]=.728868599091326140584672;
			m_abscissas[4]=.550639402928647055316623;
			m_abscissas[5]=.342724013342712845043903;
			m_abscissas[6]=.116331868883703867658777;
			m_weights[0]=.010989010989010989010989;
			m_weights[1]=.066837284497681284634071;
			m_weights[2]=.116586655898711651540997;
			m_weights[3]=.160021851762952142412821;
			m_weights[4]=.194826149373416118640332;
			m_weights[5]=.219126253009770754871163;
			m_weights[6]=.231612794468457058889628;
			break;
		case 15:
			m_abscissas[1]=.965245926503838572795851;
			m_abscissas[2]=.885082044222976298825402;
			m_abscissas[3]=.763519689951815200704118;
			m_abscissas[4]=.60625320546984571112353;
			m_abscissas[5]=.420638054713672480921897;
			m_abscissas[6]=.215353955363794238225679;
			m_abscissas[7]=0.;
			m_weights[0]=.009523809523809523809524;
			m_weights[1]=.058029893028601249096881;
			m_weights[2]=.101660070325718067603666;
			m_weights[3]=.140511699802428109460447;
			m_weights[4]=.172789647253600949052077;
			m_weights[5]=.1969872359646133560925;
			m_weights[6]=.21197358592682092012743;
			m_weights[7]=.21704811634881564951495;
			break;
		case 16:
			m_abscissas[1]=.969568046270217932952243;
			m_abscissas[2]=.899200533093472092994628;
			m_abscissas[3]=.792008291861815063931088;
			m_abscissas[4]=.652388702882493089467883;
			m_abscissas[5]=.486059421887137611781891;
			m_abscissas[6]=.299830468900763208098353;
			m_abscissas[7]=.101326273521949447843033;
			m_weights[0]=.008333333333333333333333;
			m_weights[1]=.050850361005919905403245;
			m_weights[2]=.089393697325930800991052;
			m_weights[3]=.124255382132514098349536;
			m_weights[4]=.154026980807164280815645;
			m_weights[5]=.177491913391704125301076;
			m_weights[6]=.193690023825203584316914;
			m_weights[7]=.201958308178229871489199;
			break;
		case 17:
			m_abscissas[1]=.97313217663141831415698;
			m_abscissas[2]=.910879995915573595623803;
			m_abscissas[3]=.815696251221770307106751;
			m_abscissas[4]=.691028980627684705394919;
			m_abscissas[5]=.541385399330101539123733;
			m_abscissas[6]=.372174433565477041907235;
			m_abscissas[7]=.189511973518317388304263;
			m_abscissas[8]=0.;
			m_weights[0]=.007352941176470588235294;
			m_weights[1]=.044921940543254209647401;
			m_weights[2]=.079198270503687119190264;
			m_weights[3]=.110592909007028161375773;
			m_weights[4]=.137987746201926559056202;
			m_weights[5]=.160394661997621539516328;
			m_weights[6]=.177004253515657870436946;
			m_weights[7]=.187216339677619235892088;
			m_weights[8]=.190661874753469433299407;
			break;
		case 18:
			m_abscissas[1]=.976105557412198542864519;
			m_abscissas[2]=.920649185347533873837855;
			m_abscissas[3]=.835593535218090213713646;
			m_abscissas[4]=.72367932928324268130621;
			m_abscissas[5]=.588504834318661761173536;
			m_abscissas[6]=.434415036912123975342287;
			m_abscissas[7]=.266362652878280984167665;
			m_abscissas[8]=.089749093484652111022645;
			m_weights[0]=.006535947712418300653595;
			m_weights[1]=.039970628810914066137599;
			m_weights[2]=.070637166885633664999223;
			m_weights[3]=.099016271717502802394424;
			m_weights[4]=.124210533132967100263396;
			m_weights[5]=.145411961573802267983003;
			m_weights[6]=.161939517237602489264327;
			m_weights[7]=.173262109489456226010614;
			m_weights[8]=.179015863439703082293819;
			break;
		case 19:
			m_abscissas[1]=.978611766222080095152634;
			m_abscissas[2]=.92890152815258624371794;
			m_abscissas[3]=.852460577796646093085956;
			m_abscissas[4]=.751494202552613014163637;
			m_abscissas[5]=.628908137265220497766832;
			m_abscissas[6]=.48822928568071350277791;
			m_abscissas[7]=.3335048478244986102985;
			m_abscissas[8]=.169186023409281571375154;
			m_abscissas[9]=0.;
			m_weights[0]=.005847953216374269005848;
			m_weights[1]=.035793365186176477115426;
			m_weights[2]=.063381891762629736851696;
			m_weights[3]=.089131757099207084448009;
			m_weights[4]=.11231534147730504407091;
			m_weights[5]=.132267280448750776926047;
			m_weights[6]=.148413942595938885009681;
			m_weights[7]=.160290924044061241979911;
			m_weights[8]=.167556584527142867270137;
			m_weights[9]=.170001919284827234644673;
			break;
		case 20:
			m_abscissas[1]=.980743704893914171925446;
			m_abscissas[2]=.935934498812665435716182;
			m_abscissas[3]=.866877978089950141309847;
			m_abscissas[4]=.775368260952055870414318;
			m_abscissas[5]=.663776402290311289846403;
			m_abscissas[6]=.534992864031886261648136;
			m_abscissas[7]=.392353183713909299386475;
			m_abscissas[8]=.239551705922986495182401;
			m_abscissas[9]=.080545937238821837975945;
			m_weights[0]=.005263157894736842105263;
			m_weights[1]=.032237123188488941491605;
			m_weights[2]=.057181802127566826004754;
			m_weights[3]=.080631763996119603144777;
			m_weights[4]=.101991499699450815683781;
			m_weights[5]=.12070922762867472509943;
			m_weights[6]=.136300482358724184489781;
			m_weights[7]=.148361554070916825814713;
			m_weights[8]=.15658010264747548715817;
			m_weights[9]=.160743286387845749007727;
			break;
		case 21:
			m_abscissas[1]=.982572296604548028234481;
			m_abscissas[2]=.941976296959745534296103;
			m_abscissas[3]=.879294755323590464451154;
			m_abscissas[4]=.796001926077712404744313;
			m_abscissas[5]=.694051026062223232627316;
			m_abscissas[6]=.575831960261830686927022;
			m_abscissas[7]=.444115783279002101194516;
			m_abscissas[8]=.301989856508764887275352;
			m_abscissas[9]=.152785515802185466006358;
			m_abscissas[10]=0.;
			m_weights[0]=.004761904761904761904762;
			m_weights[1]=.029184840098505458609459;
			m_weights[2]=.051843169000849625072723;
			m_weights[3]=.073273918185074144252548;
			m_weights[4]=.092985467957886065301138;
			m_weights[5]=.110517083219123335267;
			m_weights[6]=.125458121190868948015158;
			m_weights[7]=.1374584628600413435809;
			m_weights[8]=.146236862447977459267271;
			m_weights[9]=.151587575111681384453251;
			m_weights[10]=.153385190332174948551584;
			break;
		case 22:
			m_abscissas[1]=.98415243845764617655229;
			m_abscissas[2]=.947204283999228680524214;
			m_abscissas[3]=.890062290190904470529658;
			m_abscissas[4]=.813948927611921136045442;
			m_abscissas[5]=.720487239961202158119882;
			m_abscissas[6]=.611669438284258971226212;
			m_abscissas[7]=.489814875189902349808751;
			m_abscissas[8]=.357520710138919538060957;
			m_abscissas[9]=.217606585159285041787955;
			m_abscissas[10]=.073054540010898334761089;
			m_weights[0]=.004329004329004329004329;
			m_weights[1]=.026545747682501757911628;
			m_weights[2]=.047214465293740752123776;
			m_weights[3]=.066865605864553076012404;
			m_weights[4]=.085090060391838447815711;
			m_weights[5]=.101500574801647674372437;
			m_weights[6]=.115747644653939066590036;
			m_weights[7]=.127527696653430275530844;
			m_weights[8]=.136589688613741426686177;
			m_weights[9]=.142740492271361400336236;
			m_weights[10]=.14584901944424179361642;
			break;
		case 23:
			m_abscissas[1]=.985527155878732578081463;
			m_abscissas[2]=.95175795571071020413564;
			m_abscissas[3]=.89945855804034501095016;
			m_abscissas[4]=.829651096651285886223201;
			m_abscissas[5]=.743695041172060683945164;
			m_abscissas[6]=.643263644460136208476146;
			m_abscissas[7]=.530311771136844168130115;
			m_abscissas[8]=.40703793791447482919595;
			m_abscissas[9]=.275841548945793067106878;
			m_abscissas[10]=.139276204040668398591863;
			m_abscissas[11]=0.;
			m_weights[0]=.003952569169960474308300;
			m_weights[1]=.024248600771531736517400;
			m_weights[2]=.043175871170241834748876;
			m_weights[3]=.061252477129554206381383;
			m_weights[4]=.078135449475569989741934;
			m_weights[5]=.093497246163512341833501;
			m_weights[6]=.107039101724336511535184;
			m_weights[7]=.118497510662749131302126;
			m_weights[8]=.127649474701758876636149;
			m_weights[9]=.134316872638603819901565;
			m_weights[10]=.138369936385807394523503;
			m_weights[11]=.13972978001274736514016;
			break;
		case 24:
			m_abscissas[1]=.986730553505160883553087;
			m_abscissas[2]=.955748220929886358026977;
			m_abscissas[3]=.907705675113506521995153;
			m_abscissas[4]=.843464070154872040623305;
			m_abscissas[5]=.764170482420493307787375;
			m_abscissas[6]=.671240105264128699835665;
			m_abscissas[7]=.56633135797929531218941;
			m_abscissas[8]=.451316373214322618248218;
			m_abscissas[9]=.328247613375510912033389;
			m_abscissas[10]=.199321253390832667236573;
			m_abscissas[11]=.066837993737228578113642;
			m_weights[0]=.003623188405797101449275;
			m_weights[1]=.022236853464711208992960;
			m_weights[2]=.039631681333467809469262;
			m_weights[3]=.056309848724646199020948;
			m_weights[4]=.071981862055293982215639;
			m_weights[5]=.086369029967929068216563;
			m_weights[6]=.099214827684083587414150;
			m_weights[7]=.110290086892968604110424;
			m_weights[8]=.119397193702491319032294;
			m_weights[9]=.126373642028020800127248;
			m_weights[10]=.13109494187360394235445;
			m_weights[11]=.133476843866986377596786;
			break;
		case 25:
			m_abscissas[1]=.987789944931493709271804;
			m_abscissas[2]=.959264138252534478859845;
			m_abscissas[3]=.91498277073462257832315;
			m_abscissas[4]=.855676465835316577523813;
			m_abscissas[5]=.782319659240716780399188;
			m_abscissas[6]=.696117048815134366760365;
			m_abscissas[7]=.598484147279993268097621;
			m_abscissas[8]=.49102411481887838261896;
			m_abscissas[9]=.375501457859227233228715;
			m_abscissas[10]=.253813064168876580179887;
			m_abscissas[11]=.127957059483106972708985;
			m_abscissas[12]=0.;
			m_weights[0]=.003333333333333333333333;
			m_weights[1]=.020465168932974385308542;
			m_weights[2]=.036504738794271372032383;
			m_weights[3]=.051936228368491474643334;
			m_weights[4]=.066513728675312784693870;
			m_weights[5]=.079998774836292981801626;
			m_weights[6]=.092170139910620421912690;
			m_weights[7]=.102828030347957830827504;
			m_weights[8]=.111797466268320888156244;
			m_weights[9]=.118931179406811825409444;
			m_weights[10]=.124112038937950290695215;
			m_weights[11]=.127254977538331447017114;
			m_weights[12]=.128308389298661928337399;
			break;
		case 26:
			m_abscissas[1]=.988727412311475654429247;
			m_abscissas[2]=.962377874767717329750663;
			m_abscissas[3]=.921435546817557341125131;
			m_abscissas[4]=.866524323959123570974451;
			m_abscissas[5]=.798477183107437439036788;
			m_abscissas[6]=.718325816362665080525405;
			m_abscissas[7]=.627285299492316887800519;
			m_abscissas[8]=.526735742029878545298243;
			m_abscissas[9]=.418201387066246785563684;
			m_abscissas[10]=.303327512859252720774041;
			m_abscissas[11]=.183855495270054901321206;
			m_abscissas[12]=.061596411781919728205487;
			m_weights[0]=.003076923076923076923077;
			m_weights[1]=.018896858024263465581352;
			m_weights[2]=.033732303685955999377523;
			m_weights[3]=.048048399081180627315979;
			m_weights[4]=.06163502514254740278218;
			m_weights[5]=.074287050122291137316084;
			m_weights[6]=.085812863980004362187481;
			m_weights[7]=.096037802353901310803697;
			m_weights[8]=.104806886230737052989906;
			m_weights[9]=.111987194119860335300211;
			m_weights[10]=.117469884093809007076696;
			m_weights[11]=.121171846288443346044287;
			m_weights[12]=.123036963800082876301527;
			break;
		case 27:
			m_abscissas[1]=.989560963728550621148875;
			m_abscissas[2]=.965148402450818919264267;
			m_abscissas[3]=.927183458725115778859747;
			m_abscissas[4]=.876202086214522478666543;
			m_abscissas[5]=.812920486895812288258092;
			m_abscissas[6]=.738227149846459910839427;
			m_abscissas[7]=.653170663696809515178705;
			m_abscissas[8]=.558945060942561180062668;
			m_abscissas[9]=.456873075614082403678613;
			m_abscissas[10]=.348387581989028670437087;
			m_abscissas[11]=.235011483102918133336802;
			m_abscissas[12]=.118336333898521048438708;
			m_abscissas[13]=0.;
			m_weights[0]=.002849002849002849002849;
			m_weights[1]=.017501974876065579019370;
			m_weights[2]=.031262951735202384324785;
			m_weights[3]=.044577657933061698744075;
			m_weights[4]=.057265569680162731739087;
			m_weights[5]=.069149342360043276280635;
			m_weights[6]=.080062321970538458168238;
			m_weights[7]=.089851365259290559972136;
			m_weights[8]=.098379074585952763179309;
			m_weights[9]=.105525747821253011218382;
			m_weights[10]=.111191065257437032928262;
			m_weights[11]=.115295500254651982813753;
			m_weights[12]=.117781436585956159066499;
			m_weights[13]=.11861397766276302708524;
			break;
		case 28:
			m_abscissas[1]=.990305402618454125517093;
			m_abscissas[2]=.967624285857131304343771;
			m_abscissas[3]=.932325167121558524926142;
			m_abscissas[4]=.884871017211302838412306;
			m_abscissas[5]=.825880970056338196044514;
			m_abscissas[6]=.756124194005569765152096;
			m_abscissas[7]=.676510128929573317719381;
			m_abscissas[8]=.588076689837175606478312;
			m_abscissas[9]=.491976753931579380429617;
			m_abscissas[10]=.389463137576362807935884;
			m_abscissas[11]=.281872266621602371954153;
			m_abscissas[12]=.170606755308004360882703;
			m_abscissas[13]=.057117121693512897626544;
			m_weights[0]=.002645502645502645502646;
			m_weights[1]=.016255883957504218198990;
			m_weights[2]=.029054220677979144706853;
			m_weights[3]=.041466915243006721037986;
			m_weights[4]=.053338077047327488400326;
			m_weights[5]=.064513658080354538390246;
			m_weights[6]=.074848123509707702780568;
			m_weights[7]=.084206795121510325739491;
			m_weights[8]=.092467685997712175722035;
			m_weights[9]=.099523110412495672875945;
			m_weights[10]=.105281093761055894482378;
			m_weights[11]=.109666573795976625088586;
			m_weights[12]=.112622380077239012540636;
			m_weights[13]=.114109979672627834533315;
			break;
		case 29:
			m_abscissas[1]=.990972988268569816278745;
			m_abscissas[2]=.969845807287936293712401;
			m_abscissas[3]=.936942718520982511542037;
			m_abscissas[4]=.89266571997608827876423;
			m_abscissas[5]=.837552736281786565242073;
			m_abscissas[6]=.772272897206468741354171;
			m_abscissas[7]=.697618661356367988393974;
			m_abscissas[8]=.614496252203432949970904;
			m_abscissas[9]=.523914674371969083578817;
			m_abscissas[10]=.426973471713494387637576;
			m_abscissas[11]=.324849382841911006260014;
			m_abscissas[12]=.218782058284261019809052;
			m_abscissas[13]=.110059013395592109707788;
			m_abscissas[14]=0.;
			m_weights[0]=.002463054187192118226601;
			m_weights[1]=.015138169859967596790858;
			m_weights[2]=.027070806296824827540851;
			m_weights[3]=.038668439979712979747833;
			m_weights[4]=.049795809093237560922668;
			m_weights[5]=.060318503828522735028064;
			m_weights[6]=.070108938000597999185267;
			m_weights[7]=.079048313027885601784170;
			m_weights[8]=.087028133441135605637626;
			m_weights[9]=.093951542114796312497837;
			m_weights[10]=.099734501637149860653569;
			m_weights[11]=.104306816463676537422707;
			m_weights[12]=.107612985833568539373587;
			m_weights[13]=.109612877845890420132132;
			m_weights[14]=.110282216779682610112458;
			break;
		case 30:
			m_abscissas[1]=.991573942840500293338818;
			m_abscissas[2]=.971846603166269241676593;
			m_abscissas[3]=.94110478095105708230719;
			m_abscissas[4]=.899699218199276859553343;
			m_abscissas[5]=.848099487180198109551424;
			m_abscissas[6]=.78689035723754708044996;
			m_abscissas[7]=.716765398637085131633799;
			m_abscissas[8]=.638519175807558407371093;
			m_abscissas[9]=.55303826009505285238455;
			m_abscissas[10]=.461291190168240685226557;
			m_abscissas[11]=.364317500422448997755985;
			m_abscissas[12]=.263215943719573791267099;
			m_abscissas[13]=.159132042625850467824949;
			m_abscissas[14]=.053245110485486669363008;
			m_weights[0]=.002298850574712643678161;
			m_weights[1]=.014131799327905387640732;
			m_weights[2]=.025283166740551402204268;
			m_weights[3]=.036142094199408535314733;
			m_weights[4]=.046590694533142927401880;
			m_weights[5]=.056511197923080383302194;
			m_weights[6]=.065791336397790054944101;
			m_weights[7]=.074326003324718253834068;
			m_weights[8]=.082018512833406914799722;
			m_weights[9]=.088781712319765210167256;
			m_weights[10]=.094538975193860891781133;
			m_weights[11]=.099225071004299830657676;
			m_weights[12]=.102786905307234989470281;
			m_weights[13]=.105184121596454649856213;
			m_weights[14]=.106389558723667924947582;
			break;
		case 31:
			m_abscissas[1]=.992116844346481088589325;
			m_abscissas[2]=.97365493581573647482607;
			m_abscissas[3]=.944869170208039229806406;
			m_abscissas[4]=.906066951441269813144512;
			m_abscissas[5]=.857659995297455610367005;
			m_abscissas[6]=.800161543192462989272331;
			m_abscissas[7]=.734181136309075242134141;
			m_abscissas[8]=.660418202611525523337041;
			m_abscissas[9]=.579654657208001913003297;
			m_abscissas[10]=.4927466190988324057038;
			m_abscissas[11]=.400615338280561806302664;
			m_abscissas[12]=.304237431272872620959455;
			m_abscissas[13]=.204634529247524501141094;
			m_abscissas[14]=.102862448760682277089115;
			m_abscissas[15]=0.;
			m_weights[0]=.002150537634408602150538;
			m_weights[1]=.013222471025464670302636;
			m_weights[2]=.023666433230270316730046;
			m_weights[3]=.033853940405224057629811;
			m_weights[4]=.043681818160066912829747;
			m_weights[5]=.053046465493448782774281;
			m_weights[6]=.061848741290454623390749;
			m_weights[7]=.069995377594100570450215;
			m_weights[8]=.077400032341475618634284;
			m_weights[9]=.083984220517529730665258;
			m_weights[10]=.089678151045260822248143;
			m_weights[11]=.094421468377857957432190;
			m_weights[12]=.098163893013712757119742;
			m_weights[13]=.100865754798650513395563;
			m_weights[14]=.102498413595470397675906;
			m_weights[15]=.103044562953207333141785;
			break;
		case 32:
			m_abscissas[1]=.992608933972761359371541;
			m_abscissas[2]=.975294690482709228062456;
			m_abscissas[3]=.948284838417232378083328;
			m_abscissas[4]=.911849939063731904074531;
			m_abscissas[5]=.866352476012675519830815;
			m_abscissas[6]=.812244731777442344546822;
			m_abscissas[7]=.750064493936674797717276;
			m_abscissas[8]=.680429755615550815942396;
			m_abscissas[9]=.604032587148421126137173;
			m_abscissas[10]=.521632262881565290606593;
			m_abscissas[11]=.434047717201846939603319;
			m_abscissas[12]=.342149406538881486253812;
			m_abscissas[13]=.246850658850205304416241;
			m_abscissas[14]=.149098596813647494914382;
			m_abscissas[15]=.049864725046593252306298;
			m_weights[0]=.002016129032258064516129;
			m_weights[1]=.012398106501373843788620;
			m_weights[2]=.022199552889291964623832;
			m_weights[3]=.031775135410915465781562;
			m_weights[4]=.041034201586062723330404;
			m_weights[5]=.049885271336221207011960;
			m_weights[6]=.058240497248055869550799;
			m_weights[7]=.066016877257154543932436;
			m_weights[8]=.073137139602679032640371;
			m_weights[9]=.079530525692106252292357;
			m_weights[10]=.085133497949668230527528;
			m_weights[11]=.089890372957357833072125;
			m_weights[12]=.093753875546813813565908;
			m_weights[13]=.096685608948002600560378;
			m_weights[14]=.098656436540761777170651;
			m_weights[15]=.099646771501276777634939;
			break;
		case 33:
			m_abscissas[1]=.993056358433658343667336;
			m_abscissas[2]=.976786163316906301485628;
			m_abscissas[3]=.951393451396995743375524;
			m_abscissas[4]=.917117303450941240825679;
			m_abscissas[5]=.874278100750562220646561;
			m_abscissas[6]=.823275923004067469559071;
			m_abscissas[7]=.764587001793528628006658;
			m_abscissas[8]=.698759316618162595699677;
			m_abscissas[9]=.626407491281268257263831;
			m_abscissas[10]=.548207059919111623110405;
			m_abscissas[11]=.464888161632106755973004;
			m_abscissas[12]=.377228724253393635041935;
			m_abscissas[13]=.286047201487674041605345;
			m_abscissas[14]=.192194931467477225741307;
			m_abscissas[15]=.096548188176107006316958;
			m_abscissas[16]=0.;
			m_weights[0]=.001893939393939393939394;
			m_weights[1]=.011648448392267734651222;
			m_weights[2]=.020864609017603360095812;
			m_weights[3]=.029881045916746477519971;
			m_weights[4]=.038617814771813967563859;
			m_weights[5]=.046993850461024170547973;
			m_weights[6]=.054931059442626967951699;
			m_weights[7]=.062355367852465305441061;
			m_weights[8]=.069197469494016147559970;
			m_weights[9]=.075393486923973828507130;
			m_weights[10]=.080885572193455092179708;
			m_weights[11]=.085622448531813132550192;
			m_weights[12]=.089559889747077400661148;
			m_weights[13]=.092661133442241463529935;
			m_weights[14]=.094897224394591815824355;
			m_weights[15]=.096247284972985461996554;
			m_weights[16]=.096698710102716558960033;
			break;
		case 34:
			m_abscissas[1]=.993464362587516202254561;
			m_abscissas[2]=.978146688879937948945225;
			m_abscissas[3]=.954230648912807528407243;
			m_abscissas[4]=.921928295677448922742807;
			m_abscissas[5]=.881523836512273264398476;
			m_abscissas[6]=.833372442803142614912342;
			m_abscissas[7]=.777897306431867410882037;
			m_abscissas[8]=.715585960720219005997251;
			m_abscissas[9]=.646986007936014624377186;
			m_abscissas[10]=.572700311274922844790932;
			m_abscissas[11]=.493381698673270957436489;
			m_abscissas[12]=.409727226559592744857378;
			m_abscissas[13]=.322472054511488993577467;
			m_abscissas[14]=.232382984869116843063348;
			m_abscissas[15]=.140251724176193858371309;
			m_abscissas[16]=.046887925712970680086839;
			m_weights[0]=.001782531194295900178253;
			m_weights[1]=.010964740818306551401475;
			m_weights[2]=.019646278160713431287947;
			m_weights[3]=.028150536862243503400752;
			m_weights[4]=.036406770950782552350792;
			m_weights[5]=.044342902060244342300273;
			m_weights[6]=.051889291830927120775840;
			m_weights[7]=.058979654026745226695439;
			m_weights[8]=.065551690108265120430438;
			m_weights[9]=.071547649877072822035589;
			m_weights[10]=.076914843063590982105076;
			m_weights[11]=.081606103850422383162089;
			m_weights[12]=.085580205971977586590127;
			m_weights[13]=.088802225283584032118158;
			m_weights[14]=.091243846806914487022440;
			m_weights[15]=.092883613629610646143686;
			m_weights[16]=.093707115504303312001626;
			break;
		case 35:
			m_abscissas[1]=.993837443644386550286625;
			m_abscissas[2]=.979391143429920623188911;
			m_abscissas[3]=.956827058436987882437572;
			m_abscissas[4]=.926333932016081938134665;
			m_abscissas[5]=.888164753717950455092073;
			m_abscissas[6]=.84263589538219978245942;
			m_abscissas[7]=.790124657084795619544918;
			m_abscissas[8]=.731066178378765238953892;
			m_abscissas[9]=.665949843865775957339233;
			m_abscissas[10]=.59531523208008327405403;
			m_abscissas[11]=.519747646062201663551533;
			m_abscissas[12]=.439873264131717002760159;
			m_abscissas[13]=.356353951565454075723553;
			m_abscissas[14]=.269881776381720791894267;
			m_abscissas[15]=.181173274765601666451748;
			m_abscissas[16]=.090963513696700411148869;
			m_abscissas[17]=0.;
			m_weights[0]=.001680672268907563025210;
			m_weights[1]=.010339473314540376211365;
			m_weights[2]=.018531393476282057481564;
			m_weights[3]=.026565395330778215947602;
			m_weights[4]=.034378667890410702169967;
			m_weights[5]=.041906918204587880736195;
			m_weights[6]=.049087863437218273244470;
			m_weights[7]=.055862029889036716723305;
			m_weights[8]=.062173295815379782245191;
			m_weights[9]=.067969368873298345403166;
			m_weights[10]=.073202223274018847533132;
			m_weights[11]=.077828499133511153103592;
			m_weights[12]=.081809862335847833270083;
			m_weights[13]=.085113322437326040267099;
			m_weights[14]=.087711506159071955511312;
			m_weights[15]=.089582884273716908711804;
			m_weights[16]=.090711950036822492939397;
			m_weights[17]=.091089347698489710951096;
			break;
		case 36:
			m_abscissas[1]=.994179475473123927933164;
			m_abscissas[2]=.980532351282529728026215;
			m_abscissas[3]=.959209115868735286121373;
			m_abscissas[4]=.930378323982939160800249;
			m_abscissas[5]=.894265911985616359370014;
			m_abscissas[6]=.85115459340254209235516;
			m_abscissas[7]=.801381804464571473221217;
			m_abscissas[8]=.74533709877172902379729;
			m_abscissas[9]=.683459109150045999754471;
			m_abscissas[10]=.616232118603912320494491;
			m_abscissas[11]=.544182271740415312747137;
			m_abscissas[12]=.46787345770807005501504;
			m_abscissas[13]=.387902897366036242953403;
			m_abscissas[14]=.304896469417054388841989;
			m_abscissas[15]=.219503812169268561244858;
			m_abscissas[16]=.132393239302661031261253;
			m_abscissas[17]=.044246509451450338451956;
			m_weights[0]=.001587301587301587301587;
			m_weights[1]=.009766174271023463721364;
			m_weights[2]=.017508591322167709600742;
			m_weights[3]=.025109861490694849103947;
			m_weights[4]=.032514046428104876118530;
			m_weights[5]=.039663623505124960590380;
			m_weights[6]=.046502730682987540520255;
			m_weights[7]=.052977870870747266275504;
			m_weights[8]=.059038377820917483040875;
			m_weights[9]=.064636824372075271685496;
			m_weights[10]=.069729397256839424743217;
			m_weights[11]=.074276241315475396764033;
			m_weights[12]=.078241771954180374387712;
			m_weights[13]=.081594953883979392564245;
			m_weights[14]=.084309544130491856057057;
			m_weights[15]=.086364297480506395718737;
			m_weights[16]=.087743132786509785840055;
			m_weights[17]=.088435258840872365966263;
			break;
		case 37:
			m_abscissas[1]=.994493809453261847978516;
			m_abscissas[2]=.98158141498697708404654;
			m_abscissas[3]=.961399734006848086105398;
			m_abscissas[4]=.934099765469704994129261;
			m_abscissas[5]=.899883908976607050629021;
			m_abscissas[6]=.859005567046514524498474;
			m_abscissas[7]=.811767418583655083661577;
			m_abscissas[8]=.75851921157356837558914;
			m_abscissas[9]=.699655183732329432150335;
			m_abscissas[10]=.635611147346434047973207;
			m_abscissas[11]=.5668612641786996769424;
			m_abscissas[12]=.493914535633420832714288;
			m_abscissas[13]=.417311034633830771351283;
			m_abscissas[14]=.337617907295764718093032;
			m_abscissas[15]=.255425174080796752140335;
			m_abscissas[16]=.171341361553725398400625;
			m_abscissas[17]=.085988997104455259538865;
			m_abscissas[18]=0.;
			m_weights[0]=.001501501501501501501502;
			m_weights[1]=.009239243242841929655483;
			m_weights[2]=.016568023462080861143385;
			m_weights[3]=.023770244581551458694408;
			m_weights[4]=.030795943408698084411215;
			m_weights[5]=.037593506203916828019662;
			m_weights[6]=.044112694216875663981525;
			m_weights[7]=.050305268315041556949868;
			m_weights[8]=.056125390773875629887112;
			m_weights[9]=.061529975716485608633587;
			m_weights[10]=.066479011534697336399423;
			m_weights[11]=.070935858364453643757075;
			m_weights[12]=.074867519862718237396340;
			m_weights[13]=.078244887728851382163864;
			m_weights[14]=.081042957322503717113241;
			m_weights[15]=.083241012845791743941462;
			m_weights[16]=.084822780746207937207977;
			m_weights[17]=.085776550217099386399959;
			m_weights[18]=.086095259909614985485823;
			break;
		case 38:
			m_abscissas[1]=.994783356795452313328635;
			m_abscissas[2]=.982547985198530559900084;
			m_abscissas[3]=.963418850143560879082852;
			m_abscissas[4]=.937531626139819667669758;
			m_abscissas[5]=.90506815879723055435989;
			m_abscissas[6]=.86625623202567185762158;
			m_abscissas[7]=.821368111961487261722685;
			m_abscissas[8]=.770718669059692832716614;
			m_abscissas[9]=.714663179030380719539022;
			m_abscissas[10]=.653594834278909216916995;
			m_abscissas[11]=.587941987370534199612231;
			m_abscissas[12]=.518165147100536557740726;
			m_abscissas[13]=.444753748682843433303313;
			m_abscissas[14]=.368222720888047142657807;
			m_abscissas[15]=.28910887428658497090109;
			m_abscissas[16]=.207967135966959799821922;
			m_abscissas[17]=.125366657157768623782895;
			m_abscissas[18]=.041886821065351953400000;
			m_weights[0]=.001422475106685633001422;
			m_weights[1]=.008753813994360225851613;
			m_weights[2]=.015701121356300385369062;
			m_weights[3]=.022534606874418874728817;
			m_weights[4]=.029209520950421723821436;
			m_weights[5]=.035679424686142199383564;
			m_weights[6]=.041899019716493155527199;
			m_weights[7]=.047824705288195613033525;
			m_weights[8]=.053414926372061171192042;
			m_weights[9]=.058630475628248269850398;
			m_weights[10]=.063434771663867566047072;
			m_weights[11]=.067794116802442773102133;
			m_weights[12]=.071677933929969965507874;
			m_weights[13]=.075058981187236438837269;
			m_weights[14]=.077913543156257632011343;
			m_weights[15]=.080221597260513986032533;
			m_weights[16]=.081966954237835647749394;
			m_weights[17]=.083137371712546614227294;
			m_weights[18]=.083724640076002124726007;
			break;
		case 39:
			m_abscissas[1]=.995050656121478376915725;
			m_abscissas[2]=.983440482502854223062303;
			m_abscissas[3]=.96528387707647646365781;
			m_abscissas[4]=.940703089395317434290379;
			m_abscissas[5]=.909861952180285172821212;
			m_abscissas[6]=.872965780179587851504941;
			m_abscissas[7]=.830260138207092607143013;
			m_abscissas[8]=.782029237149468930030148;
			m_abscissas[9]=.728594051623129742968536;
			m_abscissas[10]=.670310187206353436901519;
			m_abscissas[11]=.60756551529529640518291;
			m_abscissas[12]=.540777592505078197933826;
			m_abscissas[13]=.47039088220905566989699;
			m_abscissas[14]=.396873796873427547368441;
			m_abscissas[15]=.320715580942638355587055;
			m_abscissas[16]=.242423055053562184848275;
			m_abscissas[17]=.162517243262224063032531;
			m_abscissas[18]=.081529905737327456128966;
			m_abscissas[19]=0.;
			m_weights[0]=.001349527665317139001350;
			m_weights[1]=.008305642007788296850663;
			m_weights[2]=.014900402058706222951911;
			m_weights[3]=.021392502271216358625359;
			m_weights[4]=.027741757700206936732476;
			m_weights[5]=.033906276321207934698432;
			m_weights[6]=.039845113337566719409541;
			m_weights[7]=.045518770698763460405058;
			m_weights[8]=.050889500060123478750620;
			m_weights[9]=.055921563937307437061780;
			m_weights[10]=.060581476611788699818740;
			m_weights[11]=.064838228089039192113622;
			m_weights[12]=.068663490921010460282781;
			m_weights[13]=.072031808925239211720637;
			m_weights[14]=.074920766690497159267913;
			m_weights[15]=.077311138798527091998650;
			m_weights[16]=.079187017793501901285125;
			m_weights[17]=.080535920058912584106831;
			m_weights[18]=.081348868902776145914834;
			m_weights[19]=.081620454301007138007350;
			break;
		case 40:
			m_abscissas[1]=.995297929244348896904262;
			m_abscissas[2]=.984266280717503354734035;
			m_abscissas[3]=.967010076487988520647637;
			m_abscissas[4]=.943639764943601642574207;
			m_abscissas[5]=.914303339690209451072667;
			m_abscissas[6]=.879186343479339837890063;
			m_abscissas[7]=.838510822778106440735651;
			m_abscissas[8]=.792533952601551886811079;
			m_abscissas[9]=.741546419147384417491602;
			m_abscissas[10]=.685870585084313713826865;
			m_abscissas[11]=.625858452755257513399844;
			m_abscissas[12]=.561889439294722648770394;
			m_abscissas[13]=.494367978125253606653351;
			m_abscissas[14]=.423720962155550984750338;
			m_abscissas[15]=.350395044914180877980388;
			m_abscissas[16]=.274853816714324366517615;
			m_abscissas[17]=.197574873718910771843626;
			m_abscissas[18]=.119046798444971093517345;
			m_abscissas[19]=.039766070802181900147366;
			m_weights[0]=.001282051282051282051282;
			m_weights[1]=.007891011588600613761295;
			m_weights[2]=.014159307549919773151301;
			m_weights[3]=.020334759063387161851410;
			m_weights[4]=.026381190653141486262548;
			m_weights[5]=.032260717927117395492625;
			m_weights[6]=.037936243700708446992685;
			m_weights[7]=.043371908194757981008874;
			m_weights[8]=.048533353845914325138542;
			m_weights[9]=.053387951971494182954950;
			m_weights[10]=.057905011981786083403241;
			m_weights[11]=.062055976475709524970227;
			m_weights[12]=.065814602222895902663174;
			m_weights[13]=.069157126276081134455988;
			m_weights[14]=.072062416302054295220585;
			m_weights[15]=.074512104235389338578625;
			m_weights[16]=.076490702433396506255176;
			m_weights[17]=.077985701608680580727100;
			m_weights[18]=.078987649925364344189224;
			m_weights[19]=.079490212761549640871149;
			break;
		case 41:
			m_abscissas[1]=.995527127445284361488009;
			m_abscissas[2]=.985031859124724776253906;
			m_abscissas[3]=.968610869567681830065937;
			m_abscissas[4]=.946364199647492121200534;
			m_abscissas[5]=.918425870697903813199858;
			m_abscissas[6]=.884963972169615122780336;
			m_abscissas[7]=.846179772103991842218771;
			m_abscissas[8]=.802306533958091781416208;
			m_abscissas[9]=.7536081218869513649149;
			m_abscissas[10]=.700377416780682957795393;
			m_abscissas[11]=.642934556064585422876989;
			m_abscissas[12]=.581625008915212688025598;
			m_abscissas[13]=.516817498847129316977525;
			m_abscissas[14]=.448901786314828498959628;
			m_abscissas[15]=.378286324731083204659618;
			m_abscissas[16]=.305395804029445742124811;
			m_abscissas[17]=.230668596559530258849076;
			m_abscissas[18]=.154554120683840000537642;
			m_abscissas[19]=.07751013793625427559391;
			m_abscissas[20]=0.;
			m_weights[0]=.001219512195121951219512;
			m_weights[1]=.007506658762851588529150;
			m_weights[2]=.013472071094470606617569;
			m_weights[3]=.019353298566251679905904;
			m_weights[4]=.025117698393601086512259;
			m_weights[5]=.030730929097922458567985;
			m_weights[6]=.036159303444689512510506;
			m_weights[7]=.041370196155572640757874;
			m_weights[8]=.046332276348651614479611;
			m_weights[9]=.051015704947986513457527;
			m_weights[10]=.055392316918914310539246;
			m_weights[11]=.059435791663543655744791;
			m_weights[12]=.063121811727549795415230;
			m_weights[13]=.066428209232842706043724;
			m_weights[14]=.069335099287094895125013;
			m_weights[15]=.071824999619763970466299;
			m_weights[16]=.073882935747389606369092;
			m_weights[17]=.075496531045917972790580;
			m_weights[18]=.076656081193317470597834;
			m_weights[19]=.077354612537305548536921;
			m_weights[20]=.077587924038480831626749;
			break;
		case 42:
			m_abscissas[1]=.995739970053516275193522;
			m_abscissas[2]=.985742929506699610888432;
			m_abscissas[3]=.970098096627757617117713;
			m_abscissas[4]=.948896305445434491595154;
			m_abscissas[5]=.922259214258616176830188;
			m_abscissas[6]=.890339459837991516007911;
			m_abscissas[7]=.853319898720650343162467;
			m_abscissas[8]=.811412585793422141220008;
			m_abscissas[9]=.764857567426475150624244;
			m_abscissas[10]=.713921509314801542480233;
			m_abscissas[11]=.658896170202595526257923;
			m_abscissas[12]=.60009673124863918596173;
			m_abscissas[13]=.537859990962593292359701;
			m_abscissas[14]=.472542436194847582795195;
			m_abscissas[15]=.404518200291535767906787;
			m_abscissas[16]=.334176920139459352818578;
			m_abscissas[17]=.261921504390150265339525;
			m_abscissas[18]=.188165825654151412552726;
			m_abscissas[19]=.113332349888654096636446;
			m_abscissas[20]=.037849716559603622513665;
			m_weights[0]=.001161440185830429732869;
			m_weights[1]=.007149706972466647355627;
			m_weights[2]=.012833605558313824672050;
			m_weights[3]=.018440983044666174834525;
			m_weights[4]=.023942318410790523580436;
			m_weights[5]=.029306411216165701870416;
			m_weights[6]=.034502604399116890850701;
			m_weights[7]=.039501155201751501059319;
			m_weights[8]=.044273440030289172481192;
			m_weights[9]=.048792126960380732316335;
			m_weights[10]=.053031334938156928621913;
			m_weights[11]=.056966782983790267044759;
			m_weights[12]=.060575929662544087860717;
			m_weights[13]=.063838102375796548532158;
			m_weights[14]=.066734615856802646382143;
			m_weights[15]=.069248879241938953640741;
			m_weights[16]=.071366491125350074362376;
			m_weights[17]=.073075322062271661036220;
			m_weights[18]=.074365584053192820801492;
			m_weights[19]=.075229886613271169194800;
			m_weights[20]=.075663279107113243769212;
			break;
		case 43:
			m_abscissas[1]=.995937976757778639619215;
			m_abscissas[2]=.986404542635975161654278;
			m_abscissas[3]=.971482235047683994837946;
			m_abscissas[4]=.951253719327373436734783;
			m_abscissas[5]=.925829682853097678255633;
			m_abscissas[6]=.89534904191295572437263;
			m_abscissas[7]=.859978293183755396916331;
			m_abscissas[8]=.819910629138955177820385;
			m_abscissas[9]=.775364890299354078139868;
			m_abscissas[10]=.726584372675621422699255;
			m_abscissas[11]=.673835500076456115796365;
			m_abscissas[12]=.617406369501964623067828;
			m_abscissas[13]=.557605177910553930577022;
			m_abscissas[14]=.494758539088465517200573;
			m_abscissas[15]=.429209699874093222511203;
			m_abscissas[16]=.361316665507305544757603;
			m_abscissas[17]=.291450244356345478921723;
			m_abscissas[18]=.219992022708627064264408;
			m_abscissas[19]=.147332280690440653416068;
			m_abscissas[20]=.073867860700411845220011;
			m_abscissas[21]=0.;
			m_weights[0]=.001107419712070874861573;
			m_weights[1]=.006817613198072901863641;
			m_weights[2]=.012239409664139598703235;
			m_weights[3]=.017591487668267928581135;
			m_weights[4]=.022847092559028573947016;
			m_weights[5]=.027977816257611357622939;
			m_weights[6]=.032955701327542518084296;
			m_weights[7]=.037753579888412150008127;
			m_weights[8]=.042345254883842158530254;
			m_weights[9]=.046705651312898933559924;
			m_weights[10]=.050810955703379603478861;
			m_weights[11]=.054638747086953435804089;
			m_weights[12]=.058168119828086677424029;
			m_weights[13]=.061379797970822127502072;
			m_weights[14]=.064256240598347203487658;
			m_weights[15]=.066781737677068672515969;
			m_weights[16]=.068942495881944086617019;
			m_weights[17]=.070726713943604273670516;
			m_weights[18]=.072124647110261263100154;
			m_weights[19]=.073128660374636309801417;
			m_weights[20]=.073733270176421773657136;
			m_weights[21]=.073935174353175154357878;
			break;
		case 44:
			m_abscissas[1]=.996122494786791329907665;
			m_abscissas[2]=.987021177929833198150203;
			m_abscissas[3]=.97277258300811034082272;
			m_abscissas[4]=.953452107383077539727413;
			m_abscissas[5]=.929160675884153499380638;
			m_abscissas[6]=.900024989097820046223142;
			m_abscissas[7]=.866196967976935704740983;
			m_abscissas[8]=.827852985419625160640995;
			m_abscissas[9]=.785192953804879288459515;
			m_abscissas[10]=.738439285292858486684466;
			m_abscissas[11]=.687835733326544180377981;
			m_abscissas[12]=.633646122298528871523774;
			m_abscissas[13]=.576152972333864850757971;
			m_abscissas[14]=.515656026489114936578085;
			m_abscissas[15]=.452470688102838179389732;
			m_abscissas[16]=.386926376471010785295988;
			m_abscissas[17]=.319364809433364706093233;
			m_abscissas[18]=.250138221831394900453932;
			m_abscissas[19]=.179607529129996240337672;
			m_abscissas[20]=.108140445778956564912556;
			m_abscissas[21]=.036109568125621549814644;
			m_weights[0]=.001057082452431289640592;
			m_weights[1]=.006508122621476111621059;
			m_weights[2]=.011685488970917214518240;
			m_weights[3]=.016799192269921774573942;
			m_weights[4]=.021824935857153263992949;
			m_weights[5]=.026736800540116416609428;
			m_weights[6]=.031509239967195059317805;
			m_weights[7]=.036117391552679905856959;
			m_weights[8]=.040537237485392991291375;
			m_weights[9]=.044745737726254162123736;
			m_weights[10]=.048720952535292529584238;
			m_weights[11]=.052442157727594372667410;
			m_weights[12]=.055889953075826156227393;
			m_weights[13]=.059046363613777589270938;
			m_weights[14]=.061894933426761269231455;
			m_weights[15]=.064420811484882628504942;
			m_weights[16]=.066610829090974535804732;
			m_weights[17]=.068453568548275647380910;
			m_weights[18]=.069939422694159254680508;
			m_weights[19]=.071060644991731395819342;
			m_weights[20]=.071811389919312993244424;
			m_weights[21]=.072187743447873438037624;
			break;
		case 45:
			m_abscissas[1]=.996294721869088471619553;
			m_abscissas[2]=.987596819241306571023692;
			m_abscissas[3]=.973977415006971872449606;
			m_abscissas[4]=.955505422608103099566301;
			m_abscissas[5]=.932273056605535806146573;
			m_abscissas[6]=.904396113264306571387885;
			m_abscissas[7]=.872013494142448260348549;
			m_abscissas[8]=.835286536659991177273869;
			m_abscissas[9]=.794398217070301052141619;
			m_abscissas[10]=.749552241318063912336641;
			m_abscissas[11]=.700972031194706270251557;
			m_abscissas[12]=.648899611724784859191826;
			m_abscissas[13]=.593594405639634335240886;
			m_abscissas[14]=.535331941068692006438943;
			m_abscissas[15]=.474402478940895817551791;
			m_abscissas[16]=.411109566959743199685829;
			m_abscissas[17]=.345768527368669308481774;
			m_abscissas[18]=.278704886047254116788434;
			m_abscissas[19]=.210252750768068475736425;
			m_abscissas[20]=.140753146695693408788093;
			m_abscissas[21]=.070552317421623826527036;
			m_abscissas[22]=0.;
			m_weights[0]=.001010101010101010101010;
			m_weights[1]=.006219230315748659498528;
			m_weights[2]=.011168288996584963868274;
			m_weights[3]=.016059089495837321372108;
			m_weights[4]=.020869524716397265463232;
			m_weights[5]=.025575899418151970997066;
			m_weights[6]=.030154825753535939195392;
			m_weights[7]=.034583509607989822671737;
			m_weights[8]=.038839894146310447526490;
			m_weights[9]=.042902777131137507705848;
			m_weights[10]=.046751918850698062381251;
			m_weights[11]=.050368143785247741853707;
			m_weights[12]=.053733436478770827087426;
			m_weights[13]=.056831031443630642039527;
			m_weights[14]=.059645496759154213583389;
			m_weights[15]=.062162810990644116445268;
			m_weights[16]=.064370433064034661009122;
			m_weights[17]=.066257364756539747284136;
			m_weights[18]=.067814205496082610677183;
			m_weights[19]=.069033199198602262027678;
			m_weights[20]=.069908272911014880663010;
			m_weights[21]=.070435067067928043537414;
			m_weights[22]=.070610957211714566022408;
			break;
		case 46:
			m_abscissas[1]=.996455725704815902552829;
			m_abscissas[2]=.988135019182210394220159;
			m_abscissas[3]=.975104113996806795002877;
			m_abscissas[4]=.957426124311394757140317;
			m_abscissas[5]=.935185473606613864753713;
			m_abscissas[6]=.908488200139875120810622;
			m_abscissas[7]=.877461547724035379083377;
			m_abscissas[8]=.8422533809366427793469;
			m_abscissas[9]=.803031486928946230108558;
			m_abscissas[10]=.759982778173328678085711;
			m_abscissas[11]=.713312402704111991782957;
			m_abscissas[12]=.663242766935862123946141;
			m_abscissas[13]=.610012476015493433850575;
			m_abscissas[14]=.5538751968769602665665;
			m_abscissas[15]=.49509844946839523972304;
			m_abscissas[16]=.433962331936299107755247;
			m_abscissas[17]=.370758185853935427192059;
			m_abscissas[18]=.305787207861060975210511;
			m_abscissas[19]=.239359014334641604037935;
			m_abscissas[20]=.171790165932544020789891;
			m_abscissas[21]=.103402659042611870517287;
			m_abscissas[22]=.034522391326796098360574;
			m_weights[0]=.000966183574879227053140;
			m_weights[1]=.005949148745440358215761;
			m_weights[2]=.010684638400132985608567;
			m_weights[3]=.015366706580523116580788;
			m_weights[4]=.019975201401608782729693;
			m_weights[5]=.024488419616659377443578;
			m_weights[6]=.028884910179304030133472;
			m_weights[7]=.033143738893585783362692;
			m_weights[8]=.037244616854397545030950;
			m_weights[9]=.041168004241210667040078;
			m_weights[10]=.044895205632908466506771;
			m_weights[11]=.048408459896685135399934;
			m_weights[12]=.051691025153219682167053;
			m_weights[13]=.054727258705199017822139;
			m_weights[14]=.057502691652412540476584;
			m_weights[15]=.060004097878936433070833;
			m_weights[16]=.062219557101267637017682;
			m_weights[17]=.064138511685042370350495;
			m_weights[18]=.065751816963516745857487;
			m_weights[19]=.067051784820030349204827;
			m_weights[20]=.068032220327834335081486;
			m_weights[21]=.068688451273269446847098;
			m_weights[22]=.069017350421935966998893;
			break;
		case 47:
			m_abscissas[1]=.996606460542120689100204;
			m_abscissas[2]=.988638953918082671321448;
			m_abscissas[3]=.97615928406971716999016;
			m_abscissas[4]=.959225365506996570193695;
			m_abscissas[5]=.937914635938644823236032;
			m_abscissas[6]=.912324380558637882417765;
			m_abscissas[7]=.882571380155343924329266;
			m_abscissas[8]=.848791398929183857825307;
			m_abscissas[9]=.811138571206589038773796;
			m_abscissas[10]=.769784700390082013731467;
			m_abscissas[11]=.724918475981148510130373;
			m_abscissas[12]=.676744613056357123175095;
			m_abscissas[13]=.625482918409846591751866;
			m_abscissas[14]=.571367287737130474715516;
			m_abscissas[15]=.514644638485409135738924;
			m_abscissas[16]=.455573783262754294328386;
			m_abscissas[17]=.394424248958094301218858;
			m_abscissas[18]=.331475046966145101677078;
			m_abscissas[19]=.267013400131795535147989;
			m_abscissas[20]=.201333432224510103691965;
			m_abscissas[21]=.134734825923506926548638;
			m_abscissas[22]=.067521455437714549841933;
			m_abscissas[23]=0.;
			m_weights[0]=.000925069380203515263645;
			m_weights[1]=.005696280091592265538099;
			m_weights[2]=.010231700532574725125926;
			m_weights[3]=.014718038492394861880805;
			m_weights[4]=.019136892105083802168289;
			m_weights[5]=.023468346462881329068041;
			m_weights[6]=.027692692211754887358068;
			m_weights[7]=.031790670982680377619637;
			m_weights[8]=.035743590718663316619882;
			m_weights[9]=.039533417766413467251517;
			m_weights[10]=.043142861264354087302849;
			m_weights[11]=.046555452799402833495025;
			m_weights[12]=.049755621859878704642921;
			m_weights[13]=.052728767019742942734134;
			m_weights[14]=.055461322629035284095042;
			m_weights[15]=.057940820745543195139935;
			m_weights[16]=.060155948041983030272346;
			m_weights[17]=.062096597436779087093703;
			m_weights[18]=.063753914216627272671597;
			m_weights[19]=.065120336442322086000565;
			m_weights[20]=.066189629454511168970625;
			m_weights[21]=.066956914322502746641285;
			m_weights[22]=.067418690106642322353544;
			m_weights[23]=.067572849832865381385037;
			break;
		case 48:
			m_abscissas[1]=.996747781339857464404616;
			m_abscissas[2]=.98911147001363572789155;
			m_abscissas[3]=.97714884689083677287986;
			m_abscissas[4]=.960913153506381590105512;
			m_abscissas[5]=.940475549335081295199323;
			m_abscissas[6]=.915925449976246420077534;
			m_abscissas[7]=.887370224318220097195663;
			m_abscissas[8]=.854934744838774070269484;
			m_abscissas[9]=.818760847485605868697692;
			m_abscissas[10]=.779006713599843821970579;
			m_abscissas[11]=.735846179112848368762404;
			m_abscissas[12]=.689467974811090710372902;
			m_abscissas[13]=.6400749012659571378282;
			m_abscissas[14]=.587882942145485181026237;
			m_abscissas[15]=.533120319832776961104784;
			m_abscissas[16]=.476026497502857078021684;
			m_abscissas[17]=.416851132032702669913078;
			m_abscissas[18]=.355852982328906043821908;
			m_abscissas[19]=.293298777849759265359083;
			m_abscissas[20]=.229462052271244570326006;
			m_abscissas[21]=.16462194739809062747786;
			m_abscissas[22]=.099061992550750112347025;
			m_abscissas[23]=.033068864766152911091070;
			m_weights[0]=.000886524822695035460993;
			m_weights[1]=.005459192600248119220190;
			m_weights[2]=.009806931978900328053140;
			m_weights[3]=.014109490605488713237007;
			m_weights[4]=.018350036475219081468504;
			m_weights[5]=.022510263736936084498767;
			m_weights[6]=.026572032590911317122306;
			m_weights[7]=.030517597621303642529485;
			m_weights[8]=.034329711700673836980184;
			m_weights[9]=.037991707926728156525686;
			m_weights[10]=.041487574518847291124541;
			m_weights[11]=.044802025564071610480353;
			m_weights[12]=.047920568156735500660645;
			m_weights[13]=.050829565905944697254438;
			m_weights[14]=.053516298628674931225426;
			m_weights[15]=.055969018005148835582721;
			m_weights[16]=.058176998969274631657866;
			m_weights[17]=.060130586616844023050871;
			m_weights[18]=.061821238429979366807228;
			m_weights[19]=.063241561635037985662260;
			m_weights[20]=.064385345531601176532962;
			m_weights[21]=.065247588651725823576077;
			m_weights[22]=.065824520631008017095122;
			m_weights[23]=.066113618696001794193227;
			break;
		case 49:
			m_abscissas[1]=.99688045591002292771758;
			m_abscissas[2]=.989555124619230735708329;
			m_abscissas[3]=.978078124501372015878816;
			m_abscissas[4]=.96249848799325272995159;
			m_abscissas[5]=.942881719663199757959287;
			m_abscissas[6]=.919310144272584490655615;
			m_abscissas[7]=.89188264602644242708591;
			m_abscissas[8]=.860714272804789485146972;
			m_abscissas[9]=.825935759285070737569338;
			m_abscissas[10]=.787692980643654009136262;
			m_abscissas[11]=.746146341551779682652766;
			m_abscissas[12]=.701470103771052954136129;
			m_abscissas[13]=.653851655432212468641442;
			m_abscissas[14]=.603490725166766500593951;
			m_abscissas[15]=.55059854443322801224072;
			m_abscissas[16]=.495396961572752400401354;
			m_abscissas[17]=.438117511320741840625797;
			m_abscissas[18]=.379000443682768873889685;
			m_abscissas[19]=.31829371625110155960571;
			m_abscissas[20]=.256251954190146188459633;
			m_abscissas[21]=.193135382254020709956876;
			m_abscissas[22]=.129208733316379932985719;
			m_abscissas[23]=.064740137990899129063959;
			m_abscissas[24]=0.;
			m_weights[0]=.000850340136054421768707;
			m_weights[1]=.005236600299777745191279;
			m_weights[2]=.009408046962855594478388;
			m_weights[3]=.013537829381581817197829;
			m_weights[4]=.017610526816783106317169;
			m_weights[5]=.021609284240822007705373;
			m_weights[6]=.025517379167136704734213;
			m_weights[7]=.029318434709843002710351;
			m_weights[8]=.032996513505725078216615;
			m_weights[9]=.036536190813376634842952;
			m_weights[10]=.039922621143590443425752;
			m_weights[11]=.043141601227016470213012;
			m_weights[12]=.046179629873557273465523;
			m_weights[13]=.049023964729225407155620;
			m_weights[14]=.051662675783964653525349;
			m_weights[15]=.054084695442135281724517;
			m_weights[16]=.056279864961135846888675;
			m_weights[17]=.058238977070491023949602;
			m_weights[18]=.059953814596105188766986;
			m_weights[19]=.061417184929446926959617;
			m_weights[20]=.062622950198032684894111;
			m_weights[21]=.063566053011186641356552;
			m_weights[22]=.064242537673388332370527;
			m_weights[23]=.064649566776405844555923;
			m_weights[24]=.064785433100723735170717;
			break;
		case 50:
			m_abscissas[1]=.997005175362643586236157;
			m_abscissas[2]=.989972220058020802022452;
			m_abscissas[3]=.97895191064671980926731;
			m_abscissas[4]=.963989480109968612133874;
			m_abscissas[5]=.945145328680881413768193;
			m_abscissas[6]=.922495378502752746656972;
			m_abscissas[7]=.896130849076832632761554;
			m_abscissas[8]=.866157908177353310967717;
			m_abscissas[9]=.832697249796457952787037;
			m_abscissas[10]=.795883610117728469387034;
			m_abscissas[11]=.755865225794053039060412;
			m_abscissas[12]=.71280323742364585985443;
			m_abscissas[13]=.666871040879044815895628;
			m_abscissas[14]=.618253589201529170833818;
			m_abscissas[15]=.567146647915670425728938;
			m_abscissas[16]=.513756006783048744067551;
			m_abscissas[17]=.458296651179299759274824;
			m_abscissas[18]=.400991896436398480899217;
			m_abscissas[19]=.342072488638803843629072;
			m_abscissas[20]=.281775675495909512022964;
			m_abscissas[21]=.220344251033056476065653;
			m_abscissas[22]=.158025577948402902986515;
			m_abscissas[23]=.095070591572656647761643;
			m_abscissas[24]=.031732789442623712911932;
			m_weights[0]=.000816326530612244897959;
			m_weights[1]=.005027345549741657726368;
			m_weights[2]=.009032986686899602099342;
			m_weights[3]=.013000139812211760656502;
			m_weights[4]=.016914655483863766023195;
			m_weights[5]=.020760989497015993439296;
			m_weights[6]=.024523701718169494314950;
			m_weights[7]=.028187655451060249166250;
			m_weights[8]=.031738102606570638750900;
			m_weights[9]=.035160749089347249629539;
			m_weights[10]=.038441814210997734562458;
			m_weights[11]=.041568086852368492732712;
			m_weights[12]=.044526978934472742230458;
			m_weights[13]=.047306576229236722853514;
			m_weights[14]=.049895686393343376046454;
			m_weights[15]=.052283884066433463415596;
			m_weights[16]=.054461552866930615292671;
			m_weights[17]=.056419924123210141027618;
			m_weights[18]=.058151112187474100160947;
			m_weights[19]=.059648146191833454146703;
			m_weights[20]=.060904998119632791002923;
			m_weights[21]=.061916607079486179018278;
			m_weights[22]=.062678899684568010746858;
			m_weights[23]=.063188806455266633214218;
			m_weights[24]=.063444274179252886844291;
			break;
		case 51:
			m_abscissas[1]=.997122563118988893998443;
			m_abscissas[2]=.990364833687511589578433;
			m_abscissas[3]=.979774532409592090331641;
			m_abscissas[4]=.965393455475887309534171;
			m_abscissas[5]=.947277386308895042322794;
			m_abscissas[6]=.925496454144103561198784;
			m_abscissas[7]=.900134940683224432581319;
			m_abscissas[8]=.871290971531206066882681;
			m_abscissas[9]=.839076141797774265639416;
			m_abscissas[10]=.803615086245640300100849;
			m_abscissas[11]=.765044997883568215875303;
			m_abscissas[12]=.723515097551191431768244;
			m_abscissas[13]=.679186056790513039738486;
			m_abscissas[14]=.632229376333391998070375;
			m_abscissas[15]=.582826722651486238585976;
			m_abscissas[16]=.531169225155048514064366;
			m_abscissas[17]=.477456736769347268831036;
			m_abscissas[18]=.421897060754520599134891;
			m_abscissas[19]=.364705146762955286715757;
			m_abscissas[20]=.306102259246090230158172;
			m_abscissas[21]=.246315121428862910993176;
			m_abscissas[22]=.185575038164172938164973;
			m_abscissas[23]=.124117001061232276799853;
			m_abscissas[24]=.062178779350124090136338;
			m_abscissas[25]=0.;
			m_weights[0]=.000784313725490196078431;
			m_weights[1]=.004830383978507999852691;
			m_weights[2]=.008679892841421933151252;
			m_weights[3]=.012493788584867125902766;
			m_weights[4]=.016259069234618171293524;
			m_weights[5]=.019961377244536783766977;
			m_weights[6]=.023586434922498588683672;
			m_weights[7]=.027120231473672400038506;
			m_weights[8]=.030549100473198493004480;
			m_weights[9]=.033859778504004803810898;
			m_weights[10]=.037039458253170415119462;
			m_weights[11]=.040075838706706095741385;
			m_weights[12]=.042957173003353664119186;
			m_weights[13]=.045672313999034519204403;
			m_weights[14]=.048210757450078384095660;
			m_weights[15]=.050562682681471095465570;
			m_weights[16]=.052718990597057519621050;
			m_weights[17]=.054671338891197740866053;
			m_weights[18]=.056412174328837261210352;
			m_weights[19]=.057934761970736018676324;
			m_weights[20]=.059233211231659749718833;
			m_weights[21]=.060302498671201867054873;
			m_weights[22]=.061138487429338506605224;
			m_weights[23]=.061737943231686401764165;
			m_weights[24]=.062098546902639266272766;
			m_weights[25]=.062218903338029997762994;
			break;
		case 52:
			m_abscissas[1]=.997233182713042046092088;
			m_abscissas[2]=.990734843758915833096626;
			m_abscissas[3]=.980549903623744471141292;
			m_abscissas[4]=.966717043563305425728437;
			m_abscissas[5]=.949287862925113671143307;
			m_abscissas[6]=.928327239475647088222969;
			m_abscissas[7]=.903913163015145632477465;
			m_abscissas[8]=.876136462081327188519695;
			m_abscissas[9]=.845100471089530405736214;
			m_abscissas[10]=.810920647736802684339218;
			m_abscissas[11]=.773724144242141254517305;
			m_abscissas[12]=.733649334675882368578759;
			m_abscissas[13]=.690845300370072093663387;
			m_abscissas[14]=.645471275416892417965325;
			m_abscissas[15]=.597696054358257180653434;
			m_abscissas[16]=.547697364288884839950607;
			m_abscissas[17]=.495661203718056344576499;
			m_abscissas[18]=.441781150654457638927306;
			m_abscissas[19]=.386257642490774664172299;
			m_abscissas[20]=.329297230368455637786073;
			m_abscissas[21]=.271111810797360807370911;
			m_abscissas[22]=.211917837389278654143957;
			m_abscissas[23]=.151935515638059628368919;
			m_abscissas[24]=.091387983742063984477523;
			m_abscissas[25]=.030500482516469938402141;
			m_weights[0]=.000754147812971342383107;
			m_weights[1]=.004644771443213116889016;
			m_weights[2]=.008347084648331167365107;
			m_weights[3]=.012016392112302461464313;
			m_weights[4]=.015640729519941051088035;
			m_weights[5]=.019206815612478106754648;
			m_weights[6]=.022701428364924928948378;
			m_weights[7]=.026111580901739542893132;
			m_weights[8]=.029424592179455154096601;
			m_weights[9]=.032628139703563413354319;
			m_weights[10]=.035710307082888823748700;
			m_weights[11]=.038659628980130446659188;
			m_weights[12]=.041465134017019775698673;
			m_weights[13]=.044116385702106620348394;
			m_weights[14]=.046603521310066087026075;
			m_weights[15]=.048917288599910145483656;
			m_weights[16]=.051049080249230072052854;
			m_weights[17]=.052990965882670137798461;
			m_weights[18]=.054735721578548513637239;
			m_weights[19]=.056276856745425634553722;
			m_weights[20]=.057608638269473938479785;
			m_weights[21]=.058726111843282853058809;
			m_weights[22]=.059625120397019955606933;
			m_weights[23]=.060302319563539377757957;
			m_weights[24]=.060755190120000070875005;
			m_weights[25]=.060982047359767261977893;
			break;
		case 53:
			m_abscissas[1]=.997337544563873296139261;
			m_abscissas[2]=.991083951875506014274512;
			m_abscissas[3]=.981281571296245541691989;
			m_abscissas[4]=.967966255453209737078924;
			m_abscissas[5]=.951185804611039587172442;
			m_abscissas[6]=.93100032697697855123848;
			m_abscissas[7]=.907482095658754747321456;
			m_abscissas[8]=.880715306139273234711328;
			m_abscissas[9]=.850795779716250583956605;
			m_abscissas[10]=.817830622232804379910573;
			m_abscissas[11]=.781937841388230068759524;
			m_abscissas[12]=.743245924631939271801495;
			m_abscissas[13]=.701893379376199715724517;
			m_abscissas[14]=.658028237262838171367119;
			m_abscissas[15]=.611807524297300735680325;
			m_abscissas[16]=.563396698765007172941586;
			m_abscissas[17]=.51296905895114477113094;
			m_abscissas[18]=.460705122788844551221566;
			m_abscissas[19]=.40679198165903269776577;
			m_abscissas[20]=.351422630656664715515794;
			m_abscissas[21]=.294795277721687971537481;
			m_abscissas[22]=.237112634108395913245852;
			m_abscissas[23]=.178581188733442952199615;
			m_abscissas[24]=.11941046900038129180444;
			m_abscissas[25]=.059812290746919925058736;
			m_abscissas[26]=0.;
			m_weights[0]=.000725689404934687953556;
			m_weights[1]=.004469652707677491545451;
			m_weights[2]=.008033038910742344715400;
			m_weights[3]=.011565788706571830929712;
			m_weights[4]=.015056877844866209346080;
			m_weights[5]=.018494003026789782362491;
			m_weights[6]=.021864902617516716900834;
			m_weights[7]=.025157522479674316806872;
			m_weights[8]=.028360080648519956137105;
			m_weights[9]=.031461114841032122859520;
			m_weights[10]=.034449525136823758274533;
			m_weights[11]=.037314614310195046461108;
			m_weights[12]=.040046126367143286541408;
			m_weights[13]=.042634283368374187551790;
			m_weights[14]=.045069820484567010286621;
			m_weights[15]=.047344019189206154268418;
			m_weights[16]=.049448738483356346294638;
			m_weights[17]=.051376444046663164363728;
			m_weights[18]=.053120235213173155320642;
			m_weights[19]=.054673869676914205474341;
			m_weights[20]=.056031785839601260975093;
			m_weights[21]=.057189122720914507787247;
			m_weights[22]=.058141737360331926275466;
			m_weights[23]=.058886219648372720267941;
			m_weights[24]=.059419904534250328685098;
			m_weights[25]=.059740881566290257385570;
			m_weights[26]=.059848001730994448458669;
			break;
		case 54:
			m_abscissas[1]=.997436111871166556317477;
			m_abscissas[2]=.991413702551490313213614;
			m_abscissas[3]=.981972756063850978303175;
			m_abscissas[4]=.969146551662650363117137;
			m_abscissas[5]=.952979433807829466913837;
			m_abscissas[6]=.933527171015984691480693;
			m_abscissas[7]=.910856833067617733651716;
			m_abscissas[8]=.885046575395687500876359;
			m_abscissas[9]=.856185374325254212744589;
			m_abscissas[10]=.824372722034548961626521;
			m_abscissas[11]=.789718284279766083018448;
			m_abscissas[12]=.75234152267297620360852;
			m_abscissas[13]=.712371283031494022495908;
			m_abscissas[14]=.669945351303661499606944;
			m_abscissas[15]=.625209978639211783866186;
			m_abscissas[16]=.578319377258870147066267;
			m_abscissas[17]=.529435188869727418337599;
			m_abscissas[18]=.478725927463423948310879;
			m_abscissas[19]=.426366398420430805037874;
			m_abscissas[20]=.372537095924318768457758;
			m_abscissas[21]=.317423580764092209474948;
			m_abscissas[22]=.2612158406699428655231;
			m_abscissas[23]=.204107635387798683892219;
			m_abscissas[24]=.146295828750537683422381;
			m_abscissas[25]=.087979710048492690585791;
			m_abscissas[26]=.029360307038717604364620;
			m_weights[0]=.000698812019566736547869;
			m_weights[1]=.004304251584463480417659;
			m_weights[2]=.007736372627599463903550;
			m_weights[3]=.011140014296305147055445;
			m_weights[4]=.014505005478884170177732;
			m_weights[5]=.017819933052124230212273;
			m_weights[6]=.021073410580106739266479;
			m_weights[7]=.024254234983220790870179;
			m_weights[8]=.027351445883033061516578;
			m_weights[9]=.030354368519553280095034;
			m_weights[10]=.033252652136918527234361;
			m_weights[11]=.036036306239565543939993;
			m_weights[12]=.038695735266165073119030;
			m_weights[13]=.041221771772643540192121;
			m_weights[14]=.043605708085066802238094;
			m_weights[15]=.045839326342925629199573;
			m_weights[16]=.047914926841964039572332;
			m_weights[17]=.049825354584678980118855;
			m_weights[18]=.051564023949803378419653;
			m_weights[19]=.053124941397182469110682;
			m_weights[20]=.054502726130549555237472;
			m_weights[21]=.055692628647401799167016;
			m_weights[22]=.056690547112276877548226;
			m_weights[23]=.057493041497133530571343;
			m_weights[24]=.058097345440178801295244;
			m_weights[25]=.058501375782315968692535;
			m_weights[26]=.058703739748372384280673;
			break;
		case 55:
			m_abscissas[1]=.997529305761273055310023;
			m_abscissas[2]=.991725500291280025163236;
			m_abscissas[3]=.982626387543026509356851;
			m_abscissas[4]=.970262901463928416817998;
			m_abscissas[5]=.954676237448927701620853;
			m_abscissas[6]=.93591820858327983182462;
			m_abscissas[7]=.914051140447173084545352;
			m_abscissas[8]=.88914767909967216179311;
			m_abscissas[9]=.861290554243305483911712;
			m_abscissas[10]=.830572306023714114034022;
			m_abscissas[11]=.797094978290066162178241;
			m_abscissas[12]=.760969779922448199853132;
			m_abscissas[13]=.722316715559614771487785;
			m_abscissas[14]=.681264187036553262524134;
			m_abscissas[15]=.637948566891829496951514;
			m_abscissas[16]=.592513745378313939360299;
			m_abscissas[17]=.54511065249044732148769;
			m_abscissas[18]=.495896756600210923838521;
			m_abscissas[19]=.445035541369686770838633;
			m_abscissas[20]=.392695962679221496534149;
			m_abscissas[21]=.339051887376031760789106;
			m_abscissas[22]=.284281515708163298013364;
			m_abscissas[23]=.228566789362727442007433;
			m_abscissas[24]=.172092787075050373308619;
			m_abscissas[25]=.115047109816591527416801;
			m_abscissas[26]=.057619257604064327757598;
			m_abscissas[27]=0.;
			m_weights[0]=.000673400673400673400673;
			m_weights[1]=.004147862328671043123185;
			m_weights[2]=.007455827803521358311855;
			m_weights[3]=.010737281181964296770952;
			m_weights[4]=.013982826904940361822863;
			m_weights[5]=.017181863492228658796684;
			m_weights[6]=.020323803383333689428618;
			m_weights[7]=.023398221250648127813041;
			m_weights[8]=.026394908601620216102565;
			m_weights[9]=.029303912637618080632948;
			m_weights[10]=.032115570845726109235315;
			m_weights[11]=.034820543650590016717662;
			m_weights[12]=.037409845665582207589004;
			m_weights[13]=.039874875642630917941621;
			m_weights[14]=.042207445093640032995372;
			m_weights[15]=.044399805516991253776847;
			m_weights[16]=.046444674150926229744296;
			m_weights[17]=.048335258173876831561601;
			m_weights[18]=.050065277274085500628483;
			m_weights[19]=.051628984514939112374775;
			m_weights[20]=.053021185427449940669032;
			m_weights[21]=.054237255266874140992069;
			m_weights[22]=.055273154376380133691153;
			m_weights[23]=.056125441606870230862404;
			m_weights[24]=.056791285748460807975211;
			m_weights[25]=.057268474935698951565608;
			m_weights[26]=.057555423995304932890923;
			m_weights[27]=.057651179712052285170478;
			break;
		case 56:
			m_abscissas[1]=.997617509790724140491764;
			m_abscissas[2]=.992020624541862913171079;
			m_abscissas[3]=.98324513529644732298199;
			m_abscissas[4]=.971319834891684551315004;
			m_abscissas[5]=.956283044313714640292072;
			m_abscissas[6]=.93818296540684856078622;
			m_abscissas[7]=.917077590996499164267218;
			m_abscissas[8]=.893034533605793606640643;
			m_abscissas[9]=.866130813198811790342735;
			m_abscissas[10]=.836452612023649289538438;
			m_abscissas[11]=.80409499919007531074864;
			m_abscissas[12]=.769161626430247177447096;
			m_abscissas[13]=.731764396217919324373797;
			m_abscissas[14]=.692023103389059096091198;
			m_abscissas[15]=.650065051446553175293882;
			m_abscissas[16]=.606024644794324021997723;
			m_abscissas[17]=.560042958215125198936865;
			m_abscissas[18]=.512267284975337213289178;
			m_abscissas[19]=.462850665006637655554165;
			m_abscissas[20]=.411951394677242021576961;
			m_abscissas[21]=.359732519723843995453982;
			m_abscissas[22]=.306361312969020251435207;
			m_abscissas[23]=.252008738497424292964662;
			m_abscissas[24]=.196848904007364539338577;
			m_abscissas[25]=.141058503092174943956693;
			m_abscissas[26]=.084816249238004037359122;
			m_abscissas[27]=.028302303351157746960854;
			m_weights[0]=.000649350649350649350649;
			m_weights[1]=.003999842105106366403202;
			m_weights[2]=.007190258142992979683703;
			m_weights[3]=.010355959403268498135341;
			m_weights[4]=.013488256491098847160156;
			m_weights[5]=.016577289173990204195047;
			m_weights[6]=.019613200257610404116916;
			m_weights[7]=.022586276257862587276697;
			m_weights[8]=.025486997771695898439006;
			m_weights[9]=.028306074740684041270074;
			m_weights[10]=.031034477684631203654553;
			m_weights[11]=.033663467159778755532859;
			m_weights[12]=.036184621971786739088954;
			m_weights[13]=.038589866248989719990826;
			m_weights[14]=.040871495359022179570114;
			m_weights[15]=.043022200613337808518667;
			m_weights[16]=.045035092692287905780105;
			m_weights[17]=.046903723721137408324006;
			m_weights[18]=.048622107928936950729087;
			m_weights[19]=.050184740825421412711713;
			m_weights[20]=.051586616835224780268052;
			m_weights[21]=.052823245333322424251565;
			m_weights[22]=.053890665030563063208178;
			m_weights[23]=.05478545666334415625138;
			m_weights[24]=.055504753946863551376300;
			m_weights[25]=.056046252756911300618943;
			m_weights[26]=.056408218510821248646302;
			m_weights[27]=.056589491723958915447606;
			break;
		case 57:
			m_abscissas[1]=.997701073897274420555761;
			m_abscissas[2]=.992300242815563629040718;
			m_abscissas[3]=.983831436025803592276034;
			m_abscissas[4]=.972321488448411695595862;
			m_abscissas[5]=.957806093078716197277348;
			m_abscissas[6]=.940330149426880326911658;
			m_abscissas[7]=.919947686995880148205485;
			m_abscissas[8]=.896721712254817120461437;
			m_abscissas[9]=.870724018063628598483215;
			m_abscissas[10]=.842034963271610976584837;
			m_abscissas[11]=.810743224955391163030433;
			m_abscissas[12]=.776945524608426001472801;
			m_abscissas[13]=.740746329322149120102489;
			m_abscissas[14]=.702257528959379055210872;
			m_abscissas[15]=.66159809035128810747069;
			m_abscissas[16]=.61889368960244276810071;
			m_abscissas[17]=.574276323648225524554546;
			m_abscissas[18]=.527883902269370854163936;
			m_abscissas[19]=.479859821826909835242525;
			m_abscissas[20]=.430352522036357132950652;
			m_abscissas[21]=.379515027151876282085187;
			m_abscissas[22]=.327504472979052773134001;
			m_abscissas[23]=.274481621178536368532403;
			m_abscissas[24]=.220610362361995063682565;
			m_abscissas[25]=.166057209516402876270381;
			m_abscissas[26]=.1109907833225391740172;
			m_abscissas[27]=.055581290958607085886509;
			m_abscissas[28]=0.;
			m_weights[0]=.000626566416040100250627;
			m_weights[1]=.003859604378543614656658;
			m_weights[2]=.006938617366645615997914;
			m_weights[3]=.009994560358819114607455;
			m_weights[4]=.013019387947265630030192;
			m_weights[5]=.016003917925357171788405;
			m_weights[6]=.018938961856321941528612;
			m_weights[7]=.021815458738125021218673;
			m_weights[8]=.024624521589129164100614;
			m_weights[9]=.027357469519451914170747;
			m_weights[10]=.030005855985896860674694;
			m_weights[11]=.032561495414958661977186;
			m_weights[12]=.035016488714174596641030;
			m_weights[13]=.037363247781945434451797;
			m_weights[14]=.039594519007425480508837;
			m_weights[15]=.041703405714416465839682;
			m_weights[16]=.043683389491276987634509;
			m_weights[17]=.045528350346142624210832;
			m_weights[18]=.047232585627699397912912;
			m_weights[19]=.048790827654326899098111;
			m_weights[20]=.050198259997814124064789;
			m_weights[21]=.051450532371700677191014;
			m_weights[22]=.052543774078444874427290;
			m_weights[23]=.053474605973984535475434;
			m_weights[24]=.054240150912787569286051;
			m_weights[25]=.054838042641156126209898;
			m_weights[26]=.055266433111326354822985;
			m_weights[27]=.055523998193775814691071;
			m_weights[28]=.055609941770094453063964;
			break;
		case 58:
			m_abscissas[1]=.997780317874588336423718;
			m_abscissas[2]=.992565422234548810956107;
			m_abscissas[3]=.984387517507017675476957;
			m_abscissas[4]=.973271645364740615715629;
			m_abscissas[5]=.95925109231985553689167;
			m_abscissas[6]=.942367733315503702103781;
			m_abscissas[7]=.922671966863743560882166;
			m_abscissas[8]=.900222578128748632269899;
			m_abscissas[9]=.875086567518358027214383;
			m_abscissas[10]=.847338952182446563142523;
			m_abscissas[11]=.817062542733214836330668;
			m_abscissas[12]=.784347696381111963476667;
			m_abscissas[13]=.749292047410204890808774;
			m_abscissas[14]=.712000215871674185938465;
			m_abscissas[15]=.672583495397108703541042;
			m_abscissas[16]=.631159521078396272810062;
			m_abscissas[17]=.587851918412902047763602;
			m_abscissas[18]=.542789934365556886182933;
			m_abscissas[19]=.496108051651068126453461;
			m_abscissas[20]=.447945587388614525963714;
			m_abscissas[21]=.398446277327521409508405;
			m_abscissas[22]=.347757846885189904784426;
			m_abscissas[23]=.296031570277746451523832;
			m_abscissas[24]=.243421819059316079145457;
			m_abscissas[25]=.190085601417371326650184;
			m_abscissas[26]=.136182093599158388742361;
			m_abscissas[27]=.081872164867662033183911;
			m_abscissas[28]=.027317897404865878759635;
			m_weights[0]=.000604960677555958862674;
			m_weights[1]=.003726613100057387030162;
			m_weights[2]=.006699948927693657458810;
			m_weights[3]=.009651722372699157390294;
			m_weights[4]=.012574476194818145560036;
			m_weights[5]=.015459649328833407858377;
			m_weights[6]=.01829866659352221684275;
			m_weights[7]=.021083065913302917452816;
			m_weights[8]=.023804541509152484820335;
			m_weights[9]=.026454973129381841511269;
			m_weights[10]=.029026451657207376072971;
			m_weights[11]=.031511303212033336666484;
			m_weights[12]=.033902112254246458695322;
			m_weights[13]=.036191743807007028336091;
			m_weights[14]=.038373364793743459312340;
			m_weights[15]=.040440464453321367286866;
			m_weights[16]=.042386873782949744132492;
			m_weights[17]=.044206783955842146142788;
			m_weights[18]=.045894763661120301683017;
			m_weights[19]=.047445775315465169911832;
			m_weights[20]=.048855190098803076020512;
			m_weights[21]=.050118801769525812300978;
			m_weights[22]=.051232839218227538320796;
			m_weights[23]=.052193977722618907953020;
			m_weights[24]=.052999348870105525275450;
			m_weights[25]=.053646549118465073827548;
			m_weights[26]=.054133646969104491636201;
			m_weights[27]=.054459188731508404014796;
			m_weights[28]=.054622202861687607622977;
			break;
		case 59:
			m_abscissas[1]=.997855534435087570305881;
			m_abscissas[2]=.992817139710257168738895;
			m_abscissas[3]=.98491541970597601147187;
			m_abscissas[4]=.97417377114232796215602;
			m_abscissas[5]=.960623273532877767447679;
			m_abscissas[6]=.94430302747874669772669;
			m_abscissas[7]=.925260099999668016500479;
			m_abscissas[8]=.90354940186116704666996;
			m_abscissas[9]=.879233533145069705203498;
			m_abscissas[10]=.852382604161637295542541;
			m_abscissas[11]=.823074033883831614119934;
			m_abscissas[12]=.791392326994297348250286;
			m_abscissas[13]=.757428830368598352965463;
			m_abscissas[14]=.721281469768609467791518;
			m_abscissas[15]=.683054467536442193519063;
			m_abscissas[16]=.642858042117458840799831;
			m_abscissas[17]=.600808090285985686106285;
			m_abscissas[18]=.557025852993752558546697;
			m_abscissas[19]=.511637565806586892208532;
			m_abscissas[20]=.464774094938429741617048;
			m_abscissas[21]=.416570559932780810763495;
			m_abscissas[22]=.367165944079904149559995;
			m_abscissas[23]=.316702693693327679691828;
			m_abscissas[24]=.265326307401195906079733;
			m_abscissas[25]=.21318491663676650117151;
			m_abscissas[26]=.160428858537680018604192;
			m_abscissas[27]=.107210242485496176028571;
			m_abscissas[28]=.053682511535311913832803;
			m_abscissas[29]=0.;
			m_weights[0]=.000584453535943892460549;
			m_weights[1]=.003600377581730126576846;
			m_weights[2]=.006473376940152252552053;
			m_weights[3]=.009326197948490124466567;
			m_weights[4]=.012151921331788712610432;
			m_weights[5]=.014942555892415815126659;
			m_weights[6]=.017690089617496036602005;
			m_weights[7]=.020386610960599044531155;
			m_weights[8]=.023024348980387446839479;
			m_weights[9]=.025595700038299813786809;
			m_weights[10]=.028093251048625339524784;
			m_weights[11]=.030509801332777334213674;
			m_weights[12]=.032838383576712233563555;
			m_weights[13]=.035072284007301495726029;
			m_weights[14]=.037205061792294917036596;
			m_weights[15]=.039230567632716818953754;
			m_weights[16]=.041142961504703147486697;
			m_weights[17]=.042936729504496210001718;
			m_weights[18]=.044606699750397694165963;
			m_weights[19]=.046148057297041744053910;
			m_weights[20]=.047556358019633026861215;
			m_weights[21]=.048827541428476511179645;
			m_weights[22]=.049957942377056948931484;
			m_weights[23]=.050944301630032067042411;
			m_weights[24]=.051783775260742829501629;
			m_weights[25]=.052473942851190980241801;
			m_weights[26]=.053012814470870080118249;
			m_weights[27]=.053398836414346775497988;
			m_weights[28]=.053630895681061314281556;
			m_weights[29]=.053708323184438532129573;
			break;
		case 60:
			m_abscissas[1]=.997926991915810995726646;
			m_abscissas[2]=.993056290939090939166247;
			m_abscissas[3]=.985417013447747333371259;
			m_abscissas[4]=.975031044999701881705506;
			m_abscissas[5]=.961927438082924720232712;
			m_abscissas[6]=.946142744770198131787999;
			m_abscissas[7]=.92772097097142251467716;
			m_abscissas[8]=.906713466379259730345336;
			m_abscissas[9]=.883178785110556111285507;
			m_abscissas[10]=.85718252386284242454239;
			m_abscissas[11]=.828797139645535966033604;
			m_abscissas[12]=.798101748086244723596972;
			m_abscissas[13]=.765181903048986547143759;
			m_abscissas[14]=.730129358247941497376554;
			m_abscissas[15]=.693041811551313299479703;
			m_abscissas[16]=.654022632702073894357256;
			m_abscissas[17]=.613180575221495722197461;
			m_abscissas[18]=.570629473302122095639068;
			m_abscissas[19]=.526487924537001623427985;
			m_abscissas[20]=.480878959370623090353581;
			m_abscissas[21]=.433929698193525502814792;
			m_abscissas[22]=.38577099703673222260875;
			m_abscissas[23]=.336537082853778338900848;
			m_abscissas[24]=.286365179407029518449517;
			m_abscissas[25]=.235395124801121599605699;
			m_abscissas[26]=.183768981729595768895294;
			m_abscissas[27]=.131630641521091206475855;
			m_abscissas[28]=.079125423088724234157123;
			m_abscissas[29]=.026399667900479246653205;
			m_weights[0]=.000564971751412429378531;
			m_weights[1]=.003480447968153144994951;
			m_weights[2]=.006258098158477169842527;
			m_weights[3]=.009016842489363824458183;
			m_weights[4]=.011750254422266674931183;
			m_weights[5]=.014450866330632214106734;
			m_weights[6]=.017111184093521227867795;
			m_weights[7]=.019723802887793956220892;
			m_weights[8]=.022281444575675224009847;
			m_weights[9]=.024776982138835830547408;
			m_weights[10]=.027203460825102394952693;
			m_weights[11]=.029554117996938696383815;
			m_weights[12]=.031822402169556473307371;
			m_weights[13]=.034001991355932717972062;
			m_weights[14]=.036086810728312709879901;
			m_weights[15]=.038071049570919672450833;
			m_weights[16]=.039949177486879666876056;
			m_weights[17]=.041715959818897681200113;
			m_weights[18]=.043366472242994262501775;
			m_weights[19]=.044896114495797823190767;
			m_weights[20]=.046300623197757076197224;
			m_weights[21]=.047576083736879506332668;
			m_weights[22]=.048718941180072447212051;
			m_weights[23]=.049726010181792254038865;
			m_weights[24]=.050594483862455558381997;
			m_weights[25]=.051321941631911051493248;
			m_weights[26]=.051906355936194341264761;
			m_weights[27]=.052346097908779451106287;
			m_weights[28]=.052639941910587587191884;
			m_weights[29]=.052787068946106931707579;
			break;
		case 61:
			m_abscissas[1]=.997994936674055614183764;
			m_abscissas[2]=.993283698369058104391363;
			m_abscissas[3]=.985894016957702268078477;
			m_abscissas[4]=.975846387751266865222439;
			m_abscissas[5]=.9631679988629980090368;
			m_abscissas[6]=.947893057970718066383364;
			m_abscissas[7]=.930062754385287097518778;
			m_abscissas[8]=.909725160195428224881826;
			m_abscissas[9]=.886935104312303851491469;
			m_abscissas[10]=.861754025973069297609092;
			m_abscissas[11]=.8342498096540052862153;
			m_abscissas[12]=.804496602314463166167747;
			m_abscissas[13]=.772574613633207541067145;
			m_abscissas[14]=.738569899842772346568372;
			m_abscissas[15]=.702574131773730816901415;
			m_abscissas[16]=.664684347747828191864609;
			m_abscissas[17]=.625002691992915166579726;
			m_abscissas[18]=.583636139288421332164174;
			m_abscissas[19]=.540696206585621273528426;
			m_abscissas[20]=.496298652381221881810864;
			m_abscissas[21]=.450563164655359204524532;
			m_abscissas[22]=.403613038215665185091201;
			m_abscissas[23]=.355574842317474997984462;
			m_abscissas[24]=.30657807945637037608324;
			m_abscissas[25]=.256754836252997496813798;
			m_abscissas[26]=.20623942737138155938839;
			m_abscissas[27]=.155168033430718756590587;
			m_abscissas[28]=.103678333886804674984458;
			m_abscissas[29]=.051909135872809846330257;
			m_abscissas[30]=0.;
			m_weights[0]=.000546448087431693989071;
			m_weights[1]=.003366411226619655863291;
			m_weights[2]=.006053374871734220926052;
			m_weights[3]=.008722604294982956734849;
			m_weights[4]=.011368124877434198941626;
			m_weights[5]=.013982950691308667738721;
			m_weights[6]=.016560064513450823970427;
			m_weights[7]=.019092528532347466860203;
			m_weights[8]=.021573519250141790928096;
			m_weights[9]=.023996349889820170602463;
			m_weights[10]=.026354489661820359749976;
			m_weights[11]=.028641581818932473460421;
			m_weights[12]=.030851460976146763107074;
			m_weights[13]=.032978169813494282438666;
			m_weights[14]=.035015975175564419402399;
			m_weights[15]=.036959383547455665570319;
			m_weights[16]=.038803155875359808763026;
			m_weights[17]=.040542321696378078011141;
			m_weights[18]=.042172192541691657429329;
			m_weights[19]=.043688374578085777836851;
			m_weights[20]=.045086780454352270852183;
			m_weights[21]=.046363640320969803907694;
			m_weights[22]=.047515511993546787831116;
			m_weights[23]=.048539290232741587896497;
			m_weights[24]=.049432215115711843078797;
			m_weights[25]=.050191879476567684041118;
			m_weights[26]=.050816235395797786532164;
			m_weights[27]=.051303599721191368615668;
			m_weights[28]=.051652658605384085515993;
			m_weights[29]=.051862471047802855265991;
			m_weights[30]=.051932471431465988277553;
			break;
		case 62:
			m_abscissas[1]=.998059595212785754431992;
			m_abscissas[2]=.99350011826968708621755;
			m_abscissas[3]=.986348010547110783220756;
			m_abscissas[4]=.976622486573820090017543;
			m_abscissas[5]=.964349017330432504548379;
			m_abscissas[6]=.949559650940589091236508;
			m_abscissas[7]=.932292981593562696111643;
			m_abscissas[8]=.912594060646369511562151;
			m_abscissas[9]=.890514282610571241682242;
			m_abscissas[10]=.866111252339923618174545;
			m_abscissas[11]=.839448635270737000511338;
			m_abscissas[12]=.810595991566435397050734;
			m_abscissas[13]=.77962859476244837829001;
			m_abscissas[14]=.746627235449484454063688;
			m_abscissas[15]=.711678010535563816534812;
			m_abscissas[16]=.674872098649795928507261;
			m_abscissas[17]=.636305522280408845754361;
			m_abscissas[18]=.596078897271021447558381;
			m_abscissas[19]=.55429717033057626831712;
			m_abscissas[20]=.511069345242811110360932;
			m_abscissas[21]=.46650819849018893795512;
			m_abscissas[22]=.420729985034574071811907;
			m_abscissas[23]=.373854135022478473831351;
			m_abscissas[24]=.32600294220629245304104;
			m_abscissas[25]=.277301244894473288355553;
			m_abscissas[26]=.227876100263123186538954;
			m_abscissas[27]=.17785645287868615595713;
			m_abscissas[28]=.127372798296582121638776;
			m_abscissas[29]=.076556842613433868348416;
			m_abscissas[30]=.025541158861093044252936;
			m_weights[0]=.000528820729772607086198;
			m_weights[1]=.003257887589219370540066;
			m_weights[2]=.005858528595120213223690;
			m_weights[3]=.008442515672946951724027;
			m_weights[4]=.011004289228371536056141;
			m_weights[5]=.013537307100143246301915;
			m_weights[6]=.016034991787545405180251;
			m_weights[7]=.018490836436195775981664;
			m_weights[8]=.020898437489170430711832;
			m_weights[9]=.023251515275652286588307;
			m_weights[10]=.025543931593973601826630;
			m_weights[11]=.027769706159981799053502;
			m_weights[12]=.029923032386157679108898;
			m_weights[13]=.031998292609727308198441;
			m_weights[14]=.033990072786850009774235;
			m_weights[15]=.035893176636954393029701;
			m_weights[16]=.037702639209920314253775;
			m_weights[17]=.039413739845115966118707;
			m_weights[18]=.041022014490623408935348;
			m_weights[19]=.042523267351608064560108;
			m_weights[20]=.043913581838027434988477;
			m_weights[21]=.045189330783441834067688;
			m_weights[22]=.046347185908453263873437;
			m_weights[23]=.047384126504193187056188;
			m_weights[24]=.048297447313271564377273;
			m_weights[25]=.049084765587668026568779;
			m_weights[26]=.049744027305178180234631;
			m_weights[27]=.050273512528213979892241;
			m_weights[28]=.050671839890988745000945;
			m_weights[29]=.050937970203387570759211;
			m_weights[30]=.051071209162125844927695;
			break;
		case 63:
			m_abscissas[1]=.998121176070091688658659;
			m_abscissas[2]=.993706247018703528180515;
			m_abscissas[3]=.986780449677166211551299;
			m_abscissas[4]=.977361817050932409842166;
			m_abscissas[5]=.965474236497008878662747;
			m_abscissas[6]=.951147764225168628894519;
			m_abscissas[7]=.934418600235728918145453;
			m_abscissas[8]=.915329008289992860604825;
			m_abscissas[9]=.893927212556823498898521;
			m_abscissas[10]=.870267277023763918337459;
			m_abscissas[11]=.844408969434797842504809;
			m_abscissas[12]=.816417610544564597841317;
			m_abscissas[13]=.786363909228053666934832;
			m_abscissas[14]=.75432378392514933230748;
			m_abscissas[15]=.72037817089838125991254;
			m_abscissas[16]=.684612819801011034072112;
			m_abscissas[17]=.64711807707821842034162;
			m_abscissas[18]=.607988657751861086708052;
			m_abscissas[19]=.567323406167120951259999;
			m_abscissas[20]=.525225046306449002757435;
			m_abscissas[21]=.481799922302148165776647;
			m_abscissas[22]=.437157729803452872950613;
			m_abscissas[23]=.39141123887692680925922;
			m_abscissas[24]=.344676009140304407737047;
			m_abscissas[25]=.297070097849469470769483;
			m_abscissas[26]=.248713761676033151641629;
			m_abscissas[27]=.199729152928891063143146;
			m_abscissas[28]=.150240010987160889557246;
			m_abscissas[29]=.100371349723989526678786;
			m_abscissas[30]=.050249141710840333690223;
			m_abscissas[31]=0.;
			m_weights[0]=.000512032770097286226318;
			m_weights[1]=.003154527389564406040687;
			m_weights[2]=.005672934458269399190819;
			m_weights[3]=.008175685025316198925248;
			m_weights[4]=.010657601118471009159281;
			m_weights[5]=.013112549926151271301674;
			m_weights[6]=.015534359906328990349184;
			m_weights[7]=.017916922379588209765372;
			m_weights[8]=.020254222137224677869336;
			m_weights[9]=.022540356395139317746248;
			m_weights[10]=.024769550866660007011788;
			m_weights[11]=.026936174769537629932035;
			m_weights[12]=.029034755221373856668212;
			m_weights[13]=.031059991141448415969987;
			m_weights[14]=.033006766678829398483248;
			m_weights[15]=.034870164154538920837811;
			m_weights[16]=.036645476494373641158674;
			m_weights[17]=.038328219125237480442085;
			m_weights[18]=.039914141307008859899902;
			m_weights[19]=.041399236872377105162877;
			m_weights[20]=.042779754348084062501260;
			m_weights[21]=.044052206432317512723125;
			m_weights[22]=.045213378804495546804027;
			m_weights[23]=.046260338245293887762391;
			m_weights[24]=.047190440046468733745035;
			m_weights[25]=.048001334691797319117802;
			m_weights[26]=.048690973792284970362565;
			m_weights[27]=.049257615260662228602069;
			m_weights[28]=.049699827712111577528696;
			m_weights[29]=.050016494080114290749109;
			m_weights[30]=.050206814438288188640664;
			m_weights[31]=.050270308021091198644938;
			break;
		case 64:
			m_abscissas[1]=.998179871502163215183455;
			m_abscissas[2]=.993902726703057292374784;
			m_abscissas[3]=.987192676602740242647719;
			m_abscissas[4]=.978066662831396073956635;
			m_abscissas[5]=.966547110369099233518165;
			m_abscissas[6]=.952662235788662915462359;
			m_abscissas[7]=.936446027475634162452937;
			m_abscissas[8]=.917938173510281630834327;
			m_abscissas[9]=.897183967845850042836223;
			m_abscissas[10]=.874234200657627491768543;
			m_abscissas[11]=.849145034542990985005142;
			m_abscissas[12]=.821977867307517050075859;
			m_abscissas[13]=.792799181826208137348025;
			m_abscissas[14]=.761680383408119967788661;
			m_abscissas[15]=.728697625088836939572237;
			m_abscissas[16]=.693931621290704840810228;
			m_abscissas[17]=.65746745031297650421033;
			m_abscissas[18]=.619394346138431547543277;
			m_abscissas[19]=.579805480067718423727434;
			m_abscissas[20]=.538797732716800438993907;
			m_abscissas[21]=.496471456936057753503304;
			m_abscissas[22]=.452930232231581181989846;
			m_abscissas[23]=.408280611289854044527904;
			m_abscissas[24]=.36263185922626182368689;
			m_abscissas[25]=.316095686195625806396437;
			m_abscissas[26]=.268785974019170003988887;
			m_abscissas[27]=.220818497496953503207563;
			m_abscissas[28]=.172310641087792971318875;
			m_abscissas[29]=.123381111650027986160475;
			m_abscissas[30]=.074149647946115918734470;
			m_abscissas[31]=.024736727621958728494991;
			m_weights[0]=.000496031746031746031746;
			m_weights[1]=.003056008244912490378576;
			m_weights[2]=.005496016203817156897805;
			m_weights[3]=.007921289790046634035652;
			m_weights[4]=.010327002366815328460713;
			m_weights[5]=.012707399197454735198938;
			m_weights[6]=.015056683987961442732665;
			m_weights[7]=.017369116384542181594405;
			m_weights[8]=.019639040723241718382364;
			m_weights[9]=.021860903511518060110567;
			m_weights[10]=.024029268144023826713018;
			m_weights[11]=.026138828614338437745033;
			m_weights[12]=.028184422665848517475030;
			m_weights[13]=.030161044499089450675899;
			m_weights[14]=.032063857057727024512136;
			m_weights[15]=.033888203884125397608179;
			m_weights[16]=.035629620524489486277117;
			m_weights[17]=.037283845459801172591796;
			m_weights[18]=.038846830537807736696313;
			m_weights[19]=.040314750881560237117376;
			m_weights[20]=.041684014250801952190132;
			m_weights[21]=.042951269833601818607672;
			m_weights[22]=.044113416446892470918715;
			m_weights[23]=.045167610125947702037459;
			m_weights[24]=.046111271084289059449325;
			m_weights[25]=.046942090027028316445095;
			m_weights[26]=.047658033802220636787024;
			m_weights[27]=.048257350376414548899827;
			m_weights[28]=.048738573122233185119818;
			m_weights[29]=.049100524407501307932345;
			m_weights[30]=.049342318477139573936072;
			m_weights[31]=.049463363620776646441190;
			break;
		case 65:
			m_abscissas[1]=.998235858985168158701431;
			m_abscissas[2]=.994090150118423121239637;
			m_abscissas[3]=.987585930769509135514003;
			m_abscissas[4]=.978739133191998889494039;
			m_abscissas[5]=.967570830266913614343894;
			m_abscissas[6]=.954107537460022212778817;
			m_abscissas[7]=.938381197682665556981809;
			m_abscissas[8]=.920429116242990778477444;
			m_abscissas[9]=.900293875561607010882469;
			m_abscissas[10]=.878023235324945710036145;
			m_abscissas[11]=.853670019681449308699123;
			m_abscissas[12]=.827291992166940420597885;
			m_abscissas[13]=.798951718804367380623093;
			m_abscissas[14]=.768716419761627539364945;
			m_abscissas[15]=.736657809944952395690901;
			m_abscissas[16]=.702851928917937019446909;
			m_abscissas[17]=.667378960555587341343276;
			m_abscissas[18]=.630323042864267422910976;
			m_abscissas[19]=.591772068420342079233737;
			m_abscissas[20]=.55181747590182537996603;
			m_abscissas[21]=.510554033208072030401671;
			m_abscissas[22]=.468079612682275662640837;
			m_abscissas[23]=.424494958970138408210523;
			m_abscissas[24]=.379903450065463996209782;
			m_abscissas[25]=.334410852109538432458408;
			m_abscissas[26]=.288125068525947475729759;
			m_abscissas[27]=.241155884085896520507915;
			m_abscissas[28]=.193614704511110181853616;
			m_abscissas[29]=.145614292231964904959301;
			m_abscissas[30]=.097268498927621703334126;
			m_abscissas[31]=.048691995482555117357182;
			m_abscissas[32]=0.;
			m_weights[0]=.000480769230769230769231;
			m_weights[1]=.002962032541256216029315;
			m_weights[2]=.005327241721587964074182;
			m_weights[3]=.007678570133548121116917;
			m_weights[4]=.010011514973888304411754;
			m_weights[5]=.012320671119444298136523;
			m_weights[6]=.014600589550422213461398;
			m_weights[7]=.016845871022064948300681;
			m_weights[8]=.019051193120028274070329;
			m_weights[9]=.021211326413491117959350;
			m_weights[10]=.023321147949507752520072;
			m_weights[11]=.025375653798947578333281;
			m_weights[12]=.027369971086316568133448;
			m_weights[13]=.029299369619809824458194;
			m_weights[14]=.031159273145641122234465;
			m_weights[15]=.032945270220318297735651;
			m_weights[16]=.034653124683798986293211;
			m_weights[17]=.036278785712687434895212;
			m_weights[18]=.037818397431573346283540;
			m_weights[19]=.039268308060710440688000;
			m_weights[20]=.040625078578868414031619;
			m_weights[21]=.041885490881105350502982;
			m_weights[22]=.043046555412280044711842;
			m_weights[23]=.044105518258298097938899;
			m_weights[24]=.045059867678332635707339;
			m_weights[25]=.045907340062563537896255;
			m_weights[26]=.046645925301327952533556;
			m_weights[27]=.047273871552962506782316;
			m_weights[28]=.047789689399038578190335;
			m_weights[29]=.048192155377141635080858;
			m_weights[30]=.048480314882819829805823;
			m_weights[31]=.048653484433821797399000;
			m_weights[32]=.048711253291255159028844;
			break;
		case 66:
			m_abscissas[1]=.998289302557966964184481;
			m_abscissas[2]=.994269065239866109987532;
			m_abscissas[3]=.987961358114584882645469;
			m_abscissas[4]=.979381178755804385662737;
			m_abscissas[5]=.968548348394562378112213;
			m_abscissas[6]=.955487807597615792278707;
			m_abscissas[7]=.940229605206947566083484;
			m_abscissas[8]=.922808839617329283457458;
			m_abscissas[9]=.903265581151352774067947;
			m_abscissas[10]=.881644781015440349649282;
			m_abscissas[11]=.857996168372987894937567;
			m_abscissas[12]=.832374136177444646285824;
			m_abscissas[13]=.804837616171933170232307;
			m_abscissas[14]=.775449943400108296921004;
			m_abscissas[15]=.744278710564736385345021;
			m_abscissas[16]=.711395612580596901872944;
			m_abscissas[17]=.676876281685017985058408;
			m_abscissas[18]=.640800113488323090022546;
			m_abscissas[19]=.603250084365935249666287;
			m_abscissas[20]=.564312560613086175665856;
			m_abscissas[21]=.524077099801635990522631;
			m_abscissas[22]=.482636244796226462401802;
			m_abscissas[23]=.440085310903747462025823;
			m_abscissas[24]=.396522166645814041955728;
			m_abscissas[25]=.352047008658572233789097;
			m_abscissas[26]=.306762131237629743790232;
			m_abscissas[27]=.260771691058205369494934;
			m_abscissas[28]=.214181467611675881304199;
			m_abscissas[29]=.167098619909543430910151;
			m_abscissas[30]=.119631440014426163239548;
			m_abscissas[31]=.071889103964968982041901;
			m_abscissas[32]=.023981420667563004504300;
			m_weights[0]=.000466200466200466200466;
			m_weights[1]=.002872325184721454831482;
			m_weights[2]=.005166119053881634443437;
			m_weights[3]=.007446823304521234294072;
			m_weights[4]=.009710233958058453947421;
			m_weights[5]=.011951269566650912343266;
			m_weights[6]=.014164802868276553784430;
			m_weights[7]=.016345750877715204891015;
			m_weights[8]=.018489100394119414641849;
			m_weights[9]=.020589922953300905621242;
			m_weights[10]=.022643387220926765262463;
			m_weights[11]=.024644770487593177943226;
			m_weights[12]=.026589469687358721056531;
			m_weights[13]=.028473012054875029668130;
			m_weights[14]=.030291065446664398236242;
			m_weights[15]=.032039448322548396891684;
			m_weights[16]=.033714139372726653503666;
			m_weights[17]=.035311286772243154161753;
			m_weights[18]=.036827217043441167295032;
			m_weights[19]=.03825844350698721413549;
			m_weights[20]=.039601674302541197116149;
			m_weights[21]=.040853819960910270430397;
			m_weights[22]=.042012000510433880690026;
			m_weights[23]=.043073552101351724389209;
			m_weights[24]=.044036033132977012381242;
			m_weights[25]=.044897229869618796459949;
			m_weights[26]=.045655161532359703523320;
			m_weights[27]=.046308084854992809189562;
			m_weights[28]=.046854498093648713628697;
			m_weights[29]=.047293144480897132662920;
			m_weights[30]=.047623015116382898469927;
			m_weights[31]=.047843351287350818443405;
			m_weights[32]=.047953646213724129462301;
			break;
		case 67:
			m_abscissas[1]=.998340354024651460740539;
			m_abscissas[2]=.994439979226630881535672;
			m_abscissas[3]=.988320019400759477223783;
			m_abscissas[4]=.979994605583790181291722;
			m_abscissas[5]=.969482398983672125528172;
			m_abscissas[6]=.956806880413182160225127;
			m_abscissas[7]=.941996342814681027922407;
			m_abscissas[8]=.925083838207541009357941;
			m_abscissas[9]=.906107106945742571858732;
			m_abscissas[10]=.885108494589883565609832;
			m_abscissas[11]=.8621348578661590383107;
			m_abscissas[12]=.837237460316069251744751;
			m_abscissas[13]=.810471858009404650326168;
			m_abscissas[14]=.781897775631006247400227;
			m_abscissas[15]=.751578973241906711492583;
			m_abscissas[16]=.719583104023433166176902;
			m_abscissas[17]=.685981563327304885178555;
			m_abscissas[18]=.650849329371489103672647;
			m_abscissas[19]=.614264795938891837941752;
			m_abscissas[20]=.576309597453113097898224;
			m_abscissas[21]=.537068426822129594624899;
			m_abscissas[22]=.496628846456693889737545;
			m_abscissas[23]=.455081092885345715735457;
			m_abscissas[24]=.412517875402146029223383;
			m_abscissas[25]=.369034169196514506468838;
			m_abscissas[26]=.324727003426835406863743;
			m_abscissas[27]=.279695244710760976266498;
			m_abscissas[28]=.234039376515356831078828;
			m_abscissas[29]=.187861274939374944718126;
			m_abscissas[30]=.141263981387985138969071;
			m_abscissas[31]=.094351472647226531928348;
			m_abscissas[32]=.047228428871240192154526;
			m_abscissas[33]=0.;
			m_weights[0]=.000452284034373586612393;
			m_weights[1]=.002786631587532444205438;
			m_weights[2]=.005012192815941463439094;
			m_weights[3]=.007225398571121279637334;
			m_weights[4]=.009422320925875045499764;
			m_weights[5]=.011598178436229243727118;
			m_weights[6]=.013748142291443545962720;
			m_weights[7]=.015867423047799728832374;
			m_weights[8]=.017951294719242502195909;
			m_weights[9]=.019995108642310574091664;
			m_weights[10]=.021994304875033753380174;
			m_weights[11]=.023944422743325532570792;
			m_weights[12]=.025841110945958424965369;
			m_weights[13]=.027680137331850403686594;
			m_weights[14]=.029457398376407633409476;
			m_weights[15]=.031168928354929759910933;
			m_weights[16]=.032810908200811444174999;
			m_weights[17]=.034379674032537520068026;
			m_weights[18]=.035871725332274714749759;
			m_weights[19]=.037283732758745654210275;
			m_weights[20]=.038612545577450824258203;
			m_weights[21]=.039855198691935543227742;
			m_weights[22]=.041008919260571132877492;
			m_weights[23]=.042071132884179406648445;
			m_weights[24]=.043039469350750346348776;
			m_weights[25]=.043911767924469608950724;
			m_weights[26]=.044686082167276428633962;
			m_weights[27]=.045360684282207672992275;
			m_weights[28]=.045934068968845932401282;
			m_weights[29]=.046404956782275032238056;
			m_weights[30]=.046772296988052152563131;
			m_weights[31]=.047035269906828940778318;
			m_weights[32]=.047193288743391851139696;
			m_weights[33]=.047246000896041743223384;
			break;
		case 68:
			m_abscissas[1]=.998389154033385125542563;
			m_abscissas[2]=.994603362015707223881472;
			m_abscissas[3]=.988662897697367439011241;
			m_abscissas[4]=.980581087829205359661656;
			m_abscissas[5]=.970375517291367991160472;
			m_abscissas[6]=.958068312338819215811808;
			m_abscissas[7]=.943686136277611312712065;
			m_abscissas[8]=.927260141501079140186609;
			m_abscissas[9]=.908825904941494684085892;
			m_abscissas[10]=.888423352071797719887126;
			m_abscissas[11]=.866096670871137076985063;
			m_abscissas[12]=.84189421632285383570275;
			m_abscissas[13]=.815868405787322648633165;
			m_abscissas[14]=.788075605530091379740827;
			m_abscissas[15]=.758576008674467728048384;
			m_abscissas[16]=.727433504853819869078967;
			m_abscissas[17]=.694715541851333874696916;
			m_abscissas[18]=.660492979529733336086914;
			m_abscissas[19]=.624839936368876241591158;
			m_abscissas[20]=.58783362894448205625975;
			m_abscissas[21]=.549554204696165009480689;
			m_abscissas[22]=.510084568347278013163947;
			m_abscissas[23]=.469510202352700959565286;
			m_abscissas[24]=.427918981763569848672313;
			m_abscissas[25]=.385400983909991018384102;
			m_abscissas[26]=.342048293313980312962337;
			m_abscissas[27]=.297954802255179648343216;
			m_abscissas[28]=.253216007421307023700672;
			m_abscissas[29]=.207928803083767972979608;
			m_abscissas[30]=.162191271246377026445092;
			m_abscissas[31]=.116102469221689818226724;
			m_abscissas[32]=.069762215095015300898560;
			m_abscissas[33]=.023270871540750706034374;
			m_weights[0]=.000438981562774363476734;
			m_weights[1]=.002704715860994273205088;
			m_weights[2]=.004865040983040638535546;
			m_weights[3]=.007013692673671004198487;
			m_weights[4]=.009146998291729080677896;
			m_weights[5]=.011260454765196790396957;
			m_weights[6]=.013349510418624054393597;
			m_weights[7]=.015409648553898581984877;
			m_weights[8]=.017436410241159951025471;
			m_weights[9]=.019425407197429312596511;
			m_weights[10]=.021372332286306828433678;
			m_weights[11]=.023272969206686598683936;
			m_weights[12]=.025123201771317972329840;
			m_weights[13]=.026919022887376261969714;
			m_weights[14]=.028656543266722332112474;
			m_weights[15]=.030331999865572848748201;
			m_weights[16]=.031941764043263525180596;
			m_weights[17]=.033482349426087511988814;
			m_weights[18]=.034950419460953453851870;
			m_weights[19]=.036342794643411326827845;
			m_weights[20]=.037656459404876208865539;
			m_weights[21]=.038888568644402364885289;
			m_weights[22]=.040036453891015258520968;
			m_weights[23]=.041097629083346211739582;
			m_weights[24]=.042069795954107501616825;
			m_weights[25]=.042950849007780624314052;
			m_weights[26]=.043738880080758951608026;
			m_weights[27]=.044432182474082639909870;
			m_weights[28]=.045029254649824510030096;
			m_weights[29]=.045528803483127572752657;
			m_weights[30]=.045929747062855226927629;
			m_weights[31]=.046231217034791425204647;
			m_weights[32]=.046432560482317929183282;
			m_weights[33]=.046533341340496863823404;
			break;
		case 69:
			m_abscissas[1]=.998435833045874936825594;
			m_abscissas[2]=.994759649551718452185918;
			m_abscissas[3]=.988990905106129216252274;
			m_abscissas[4]=.981142179119592302282197;
			m_abscissas[5]=.971230056697463640242752;
			m_abscissas[6]=.959275405772027062229107;
			m_abscissas[7]=.945303375548510971972638;
			m_abscissas[8]=.929343353114779598070261;
			m_abscissas[9]=.911428904475732797768471;
			m_abscissas[10]=.891597704986268988540885;
			m_abscissas[11]=.869891460541976241732543;
			m_abscissas[12]=.846355820064479847749267;
			m_abscissas[13]=.821040279597128754529348;
			m_abscissas[14]=.793998078264983434337059;
			m_abscissas[15]=.765286086340615493892576;
			m_abscissas[16]=.734964685661739005220156;
			m_abscissas[17]=.703097642657437104547692;
			m_abscissas[18]=.669751974252776432030359;
			m_abscissas[19]=.634997806935327835532118;
			m_abscissas[20]=.598908229280840195177251;
			m_abscissas[21]=.5615591382487155092347;
			m_abscissas[22]=.523029079570836142729213;
			m_abscissas[23]=.483399082569600969837533;
			m_abscissas[24]=.442752489752673251269427;
			m_abscissas[25]=.40117478154288518784051;
			m_abscissas[26]=.358753396511948209014194;
			m_abscissas[27]=.315577547496056487738298;
			m_abscissas[28]=.271738033980120422576878;
			m_abscissas[29]=.227327051145206522730564;
			m_abscissas[30]=.182437995980772450478818;
			m_abscissas[31]=.137165270869455341540126;
			m_abscissas[32]=.091604085057484437044920;
			m_abscissas[33]=.045850254428234011129140;
			m_abscissas[34]=0.;
			m_weights[0]=.000426257459505541346974;
			m_weights[1]=.002626359191530248361978;
			m_weights[2]=.004724272001917838640348;
			m_weights[3]=.006811145733864212242711;
			m_weights[4]=.008883544073180843941164;
			m_weights[5]=.010937222525842333692138;
			m_weights[6]=.012967887031228867847195;
			m_weights[7]=.014971274576810040389253;
			m_weights[8]=.016943174794505480945283;
			m_weights[9]=.018879441943171713694121;
			m_weights[10]=.020776004593876560042433;
			m_weights[11]=.022628874542110390402518;
			m_weights[12]=.024434155337754570123533;
			m_weights[13]=.026188050543283856606808;
			m_weights[14]=.027886871748594797234828;
			m_weights[15]=.029527046343650373863015;
			m_weights[16]=.031105125040323922605690;
			m_weights[17]=.032617789131169579978544;
			m_weights[18]=.034061857471578013200304;
			m_weights[19]=.035434293171515019379876;
			m_weights[20]=.036732209983240460525993;
			m_weights[21]=.037952878371834785115201;
			m_weights[22]=.039093731255916199875344;
			m_weights[23]=.040152369406563971694153;
			m_weights[24]=.041126566493147415082233;
			m_weights[25]=.042014273765482408915298;
			m_weights[26]=.042813624362489948427413;
			m_weights[27]=.043522937238309260758732;
			m_weights[28]=.044140720697617842731650;
			m_weights[29]=.044665675532729638222848;
			m_weights[30]=.045096697755878093486753;
			m_weights[31]=.045432880920940868727953;
			m_weights[32]=.045673518029725518872503;
			m_weights[33]=.045818103018808523887580;
			m_weights[34]=.045866331823801718275264;
			break;
		case 70:
			m_abscissas[1]=.998480512209964108910656;
			m_abscissas[2]=.99490924669458407620112;
			m_abscissas[3]=.98930488881777240921131;
			m_abscissas[4]=.981679322810447485833615;
			m_abscissas[5]=.972048204114767247880846;
			m_abscissas[6]=.960431230491829662991445;
			m_abscissas[7]=.946852142900975991488504;
			m_abscissas[8]=.931338686225320990812237;
			m_abscissas[9]=.913922555345199997488204;
			m_abscissas[10]=.894639331380805061904575;
			m_abscissas[11]=.873528409412410986268571;
			m_abscissas[12]=.850632918186414702912456;
			m_abscissas[13]=.82599963209908898126041;
			m_abscissas[14]=.799678875688637628442149;
			m_abscissas[15]=.771724420852730049222856;
			m_abscissas[16]=.742193377011809342414564;
			m_abscissas[17]=.711146074447692134274747;
			m_abscissas[18]=.678645941058477871798882;
			m_abscissas[19]=.644759372783020354535022;
			m_abscissas[20]=.609555597960513586077901;
			m_abscissas[21]=.573106535902790222159513;
			m_abscissas[22]=.535486649968560293321586;
			m_abscissas[23]=.496772795439937502964164;
			m_abscissas[24]=.457044062512151681625047;
			m_abscissas[25]=.416381614717288130959927;
			m_abscissas[26]=.374868523112196659424025;
			m_abscissas[27]=.332589596569350029078148;
			m_abscissas[28]=.289631208517382110829528;
			m_abscissas[29]=.246081120485281921155747;
			m_abscissas[30]=.202028302810744788532099;
			m_abscissas[31]=.157562752878972035103582;
			m_abscissas[32]=.112775311263253418528749;
			m_abscissas[33]=.067757476142951529697497;
			m_abscissas[34]=.022601216378025417174315;
			m_weights[0]=.000414078674948240165631;
			m_weights[1]=.002551358378886996726707;
			m_weights[2]=.004589522189694405225985;
			m_weights[3]=.006617237568895839146327;
			m_weights[4]=.008631287197509854505010;
			m_weights[5]=.010627667024309675439095;
			m_weights[6]=.012602322705070883023301;
			m_weights[7]=.014551227422679706318364;
			m_weights[8]=.016470402383506221084662;
			m_weights[9]=.018355927984326261017396;
			m_weights[10]=.02020395275886994571295;
			m_weights[11]=.022010701586047405059249;
			m_weights[12]=.023772483540922281010343;
			m_weights[13]=.025485699497131642149180;
			m_weights[14]=.027146849509668926612901;
			m_weights[15]=.028752539980475433283394;
			m_weights[16]=.030299490599708718199064;
			m_weights[17]=.031784541051951272079279;
			m_weights[18]=.033204657475333598915347;
			m_weights[19]=.034556938661231900612449;
			m_weights[20]=.035838621982331499726959;
			m_weights[21]=.037047089037198517560555;
			m_weights[22]=.038179870999973436209184;
			m_weights[23]=.039234653664343176641927;
			m_weights[24]=.040209282171539175420870;
			m_weights[25]=.041101765412734770482905;
			m_weights[26]=.041910280096868594915748;
			m_weights[27]=.042633174475596959827556;
			m_weights[28]=.043268971717774141022364;
			m_weights[29]=.043816372926572581575475;
			m_weights[30]=.044274259793083201492747;
			m_weights[31]=.044641696880977403921649;
			m_weights[32]=.044917933537565240733533;
			m_weights[33]=.045102405427346867258218;
			m_weights[34]=.045194735684925226923675;
			break;
		case 71:
			m_abscissas[1]=.998523304146253224048308;
			m_abscissas[2]=.99505252984116928930128;
			m_abscissas[3]=.989605636574419651296521;
			m_abscissas[4]=.982193861236442734273098;
			m_abscissas[5]=.972831993899752964103752;
			m_abscissas[6]=.961538643002764975025767;
			m_abscissas[7]=.948336238365504122864166;
			m_abscissas[8]=.933250995623832036731882;
			m_abscissas[9]=.916312866857578960703673;
			m_abscissas[10]=.897555482088836129656598;
			m_abscissas[11]=.877016082911460205635454;
			m_abscissas[12]=.854735448733941757630924;
			m_abscissas[13]=.830757815906328118694453;
			m_abscissas[14]=.805130789941106033661902;
			m_abscissas[15]=.777905251023757782489584;
			m_abscissas[16]=.749135253010614171672124;
			m_abscissas[17]=.718877916119517572282758;
			m_abscissas[18]=.687193313528956733504498;
			m_abscissas[19]=.654144352112248651107091;
			m_abscissas[20]=.619796647544371637134603;
			m_abscissas[21]=.584218394029892064254763;
			m_abscissas[22]=.547480228910918386213612;
			m_abscissas[23]=.509655092424071871007687;
			m_abscissas[24]=.470818082885029229135695;
			m_abscissas[25]=.431046307588230300043238;
			m_abscissas[26]=.390418729717825888236676;
			m_abscissas[27]=.349016011573843985105508;
			m_abscissas[28]=.306920354424857839154571;
			m_abscissas[29]=.264215335305130082286372;
			m_abscissas[30]=.22098574108026884018286;
			m_abscissas[31]=.177317400110851631344315;
			m_abscissas[32]=.133297011848239634090487;
			m_abscissas[33]=.089011974700908789961399;
			m_abscissas[34]=.044550212513056859416102;
			m_abscissas[35]=0.;
			m_weights[0]=.000402414486921529175050;
			m_weights[1]=.002479524518263225452308;
			m_weights[2]=.004460453388051174820460;
			m_weights[3]=.006431484365407057185491;
			m_weights[4]=.008389603263027245488168;
			m_weights[5]=.010331029836525023959186;
			m_weights[6]=.012251933027002253431654;
			m_weights[7]=.014148506144263445635819;
			m_weights[8]=.016016986348395649582988;
			m_weights[9]=.017853665073263293901738;
			m_weights[10]=.019654896302028084768388;
			m_weights[11]=.021417104136272339988836;
			m_weights[12]=.023136790030310674551451;
			m_weights[13]=.024810539797542890943045;
			m_weights[14]=.026435030418129896629728;
			m_weights[15]=.028007036651509293906769;
			m_weights[16]=.029523437447917005545663;
			m_weights[17]=.030981222149531956959585;
			m_weights[18]=.032377496470559100575166;
			m_weights[19]=.033709488245209441134849;
			m_weights[20]=.034974552932608778387888;
			m_weights[21]=.036170178867951700615887;
			m_weights[22]=.037293992249616187944866;
			m_weights[23]=.038343761852420681540033;
			m_weights[24]=.039317403457716431078339;
			m_weights[25]=.040212983991551001633000;
			m_weights[26]=.041028725362706960070139;
			m_weights[27]=.041763007993008604620031;
			m_weights[28]=.042414374032896172615661;
			m_weights[29]=.042981530255889020625863;
			m_weights[30]=.043463350626194975936985;
			m_weights[31]=.043858878534370782338020;
			m_weights[32]=.044167328696596797672843;
			m_weights[33]=.044388088713796399743046;
			m_weights[34]=.044520720287505507477959;
			m_weights[35]=.044564960090078828126201;
			break;
		case 72:
			m_abscissas[1]=.998564313658294984045433;
			m_abscissas[2]=.995189849292620083462805;
			m_abscissas[3]=.989893881603414914423454;
			m_abscissas[4]=.982687044070553353694944;
			m_abscissas[5]=.973583320427823265240812;
			m_abscissas[6]=.962600304032194633248356;
			m_abscissas[7]=.949759202753687388008716;
			m_abscissas[8]=.935084806755521599355699;
			m_abscissas[9]=.91860544324473986134087;
			m_abscissas[10]=.900352922731547384481944;
			m_abscissas[11]=.880362478014148542975205;
			m_abscissas[12]=.858672696347480043356991;
			m_abscissas[13]=.835325445047564133827571;
			m_abscissas[14]=.810365790723049349265524;
			m_abscissas[15]=.783841912310681929858001;
			m_abscissas[16]=.75580500809231629365747;
			m_abscissas[17]=.726309196877790230613118;
			m_abscissas[18]=.695411413546944415771559;
			m_abscissas[19]=.663171299153806578422626;
			m_abscissas[20]=.629651085805858903128136;
			m_abscissas[21]=.594915476541067833810666;
			m_abscissas[22]=.559031520434828193807056;
			m_abscissas[23]=.522068483178074752217824;
			m_abscissas[24]=.484097713376492633952928;
			m_abscissas[25]=.445192504828978374973165;
			m_abscissas[26]=.405427955051240043836844;
			m_abscissas[27]=.364880820317657025592495;
			m_abscissas[28]=.32362936750123048749399;
			m_abscissas[29]=.281753222997629262662051;
			m_abscissas[30]=.239333219024959756888121;
			m_abscissas[31]=.196451237595950887019111;
			m_abscissas[32]=.153190052463735741318592;
			m_abscissas[33]=.109633169346321633189995;
			m_abscissas[34]=.065864664738161749242096;
			m_abscissas[35]=.021969023619968125845270;
			m_weights[0]=.000391236306729264475743;
			m_weights[1]=.002410681810394715533305;
			m_weights[2]=.004336750844447662699333;
			m_weights[3]=.006253435673703185191143;
			m_weights[4]=.008157910705593381995234;
			m_weights[5]=.010046604223577316433636;
			m_weights[6]=.011915893352301624011443;
			m_weights[7]=.013762176749173156355815;
			m_weights[8]=.015581893148042753318552;
			m_weights[9]=.017371531103962238694813;
			m_weights[10]=.019127636658312140305256;
			m_weights[11]=.020846820326960201153004;
			m_weights[12]=.022525763773816434371609;
			m_weights[13]=.024161226274744243014481;
			m_weights[14]=.025750051001339682937346;
			m_weights[15]=.027289171129007412285051;
			m_weights[16]=.028775615764634546986949;
			m_weights[17]=.030206515685673523290901;
			m_weights[18]=.031579108881137634516620;
			m_weights[19]=.032890745884621947305434;
			m_weights[20]=.034138894889486816820091;
			m_weights[21]=.035321146636569453648603;
			m_weights[22]=.036435219065126032194691;
			m_weights[23]=.037478961718107746343169;
			m_weights[24]=.038450359893316575210834;
			m_weights[25]=.039347538532458450281765;
			m_weights[26]=.040168765840606306952304;
			m_weights[27]=.040912456629099056331926;
			m_weights[28]=.041577175375432078663698;
			m_weights[29]=.042161638994238389219884;
			m_weights[30]=.042664719314015521518853;
			m_weights[31]=.043085445254819984396340;
			m_weights[32]=.043423004702727568362195;
			m_weights[33]=.043676746077442555194653;
			m_weights[34]=.043846179590030808294346;
			m_weights[35]=.043930978188349591690980;
			break;
		case 73:
			m_abscissas[1]=.998603638374731958860797;
			m_abscissas[2]=.995321531395185504519922;
			m_abscissas[3]=.990170307080210919406522;
			m_abscissas[4]=.983160035887973790978607;
			m_abscissas[5]=.974303949477461350378959;
			m_abscissas[6]=.963618694379227207947492;
			m_abscissas[7]=.951124338527513701301808;
			m_abscissas[8]=.936844342062603217356573;
			m_abscissas[9]=.920805515817681358847594;
			m_abscissas[10]=.90303797189696363480214;
			m_abscissas[11]=.883575067520369576727246;
			m_abscissas[12]=.862453342571441614719166;
			m_abscissas[13]=.839712451083228072229183;
			m_abscissas[14]=.815395086837403563497811;
			m_abscissas[15]=.78954690323657161949454;
			m_abscissas[16]=.762216427609656082683357;
			m_abscissas[17]=.733454970115973956413339;
			m_abscissas[18]=.703316527421478817757034;
			m_abscissas[19]=.671857681329362597907502;
			m_abscissas[20]=.639137492556094450838386;
			m_abscissas[21]=.605217389852773433303113;
			m_abscissas[22]=.570161054680232537126952;
			m_abscissas[23]=.534034301654575591109041;
			m_abscissas[24]=.496904954987707282333308;
			m_abscissas[25]=.45884272115489758337163;
			m_abscissas[26]=.419919058028481391356301;
			m_abscissas[27]=.380207040723413365324449;
			m_abscissas[28]=.33978122440656110839399;
			m_abscissas[29]=.29871750432731346340597;
			m_abscissas[30]=.257092973332292895324133;
			m_abscissas[31]=.214985777131681193764683;
			m_abscissas[32]=.172474967588886712176664;
			m_abscissas[33]=.129640354308990842461774;
			m_abscissas[34]=.086562354804604242993300;
			m_abscissas[35]=.043321843520433369219223;
			m_abscissas[36]=0.;
			m_weights[0]=.000380517503805175038052;
			m_weights[1]=.002344666485593482692246;
			m_weights[2]=.004218121295625722336675;
			m_weights[3]=.006082671687518730385022;
			m_weights[4]=.007935667326763456814884;
			m_weights[5]=.009773730975556105744840;
			m_weights[6]=.011593434046129251533788;
			m_weights[7]=.013391366934741235093995;
			m_weights[8]=.015164156697001472444234;
			m_weights[9]=.016908476171993488499188;
			m_weights[10]=.018621051091404579152189;
			m_weights[11]=.020298666538887818714803;
			m_weights[12]=.021938173113069590251682;
			m_weights[13]=.023536492897237966270973;
			m_weights[14]=.025090625265340136979122;
			m_weights[15]=.026597652529491525342520;
			m_weights[16]=.028054745425290380200442;
			m_weights[17]=.029459168427803310709017;
			m_weights[18]=.030808284889779630645845;
			m_weights[19]=.032099561993234087095820;
			m_weights[20]=.033330575505521377088913;
			m_weights[21]=.034499014331205948721414;
			m_weights[22]=.035602684851314725658711;
			m_weights[23]=.036639515041904996396676;
			m_weights[24]=.037607558364263057679911;
			m_weights[25]=.038504997419459863126481;
			m_weights[26]=.039330147360421494945802;
			m_weights[27]=.040081459055120878192589;
			m_weights[28]=.040757521994960242489813;
			m_weights[29]=.041357066942889557497801;
			m_weights[30]=.041878968316293058571720;
			m_weights[31]=.042322246300172780024757;
			m_weights[32]=.042686068686663582202976;
			m_weights[33]=.042969752437427414948385;
			m_weights[34]=.043172764965994457155421;
			m_weights[35]=.043294725137644280075264;
			m_weights[36]=.043335403984950278556058;
			break;
		case 74:
			m_abscissas[1]=.998641369330729621550898;
			m_abscissas[2]=.995447880478956364296216;
			m_abscissas[3]=.990435550170972065928858;
			m_abscissas[4]=.983613923020035253647103;
			m_abscissas[5]=.974995528550292740361661;
			m_abscissas[6]=.964596129289969513470913;
			m_abscissas[7]=.952434728740478215276418;
			m_abscissas[8]=.938533544911668022172602;
			m_abscissas[9]=.922917972199216716726468;
			m_abscissas[10]=.905616535885488462429625;
			m_abscissas[11]=.886660840400164170439327;
			m_abscissas[12]=.866085511757453888224386;
			m_abscissas[13]=.843928134389327553038474;
			m_abscissas[14]=.820229182535528831048074;
			m_abscissas[15]=.795031946335414482879951;
			m_abscissas[16]=.768382452765842524917666;
			m_abscissas[17]=.740329381574111166317021;
			m_abscissas[18]=.710923976361909818893585;
			m_abscissas[19]=.680219950984015275226769;
			m_abscissas[20]=.64827339143345861579696;
			m_abscissas[21]=.61514265339282577986884;
			m_abscissas[22]=.58088825563909823387306;
			m_abscissas[23]=.545572769496913792828293;
			m_abscissas[24]=.509260704542285712903812;
			m_abscissas[25]=.472018390765630421087523;
			m_abscissas[26]=.433913857409398726932813;
			m_abscissas[27]=.395016708701664882849792;
			m_abscissas[28]=.355397996712688188915325;
			m_abscissas[29]=.315130091566710742064995;
			m_abscissas[30]=.274286549246081662010092;
			m_abscissas[31]=.23294197722919314877673;
			m_abscissas[32]=.191171898207668553226143;
			m_abscissas[33]=.14905261213174975288543;
			m_abscissas[34]=.106661056835883921533031;
			m_abscissas[35]=.064074667499102558160307;
			m_abscissas[36]=.021371235196913578031537;
			m_weights[0]=.000370233246945575712699;
			m_weights[1]=.002281325829439275453877;
			m_weights[2]=.004104291231634571622919;
			m_weights[3]=.005918800778778059240083;
			m_weights[4]=.007722367145174458584191;
			m_weights[5]=.009511794638850163866101;
			m_weights[6]=.011283836158923213826525;
			m_weights[7]=.013035261295962645527959;
			m_weights[8]=.014762873201952474024753;
			m_weights[9]=.016463517146052541503745;
			m_weights[10]=.018134087116596490013307;
			m_weights[11]=.019771531798535078733084;
			m_weights[12]=.021372860271050282301566;
			m_weights[13]=.022935147526434263516882;
			m_weights[14]=.024455539839881167358453;
			m_weights[15]=.025931259996049114018637;
			m_weights[16]=.027359612369554511098577;
			m_weights[17]=.028737987853196006184867;
			m_weights[18]=.030063868626402155153581;
			m_weights[19]=.031334832755957113415858;
			m_weights[20]=.032548558621008433666838;
			m_weights[21]=.033702829154500159886977;
			m_weights[22]=.034795535893413205439075;
			m_weights[23]=.035824682830491169713046;
			m_weights[24]=.036788390060462246412891;
			m_weights[25]=.037684897214125678749574;
			m_weights[26]=.038512566674048184720066;
			m_weights[27]=.039269886566008243739911;
			m_weights[28]=.039955473520731725190765;
			m_weights[29]=.040568075200879357068505;
			m_weights[30]=.041106572588673658882444;
			m_weights[31]=.041569982029989082549860;
			m_weights[32]=.041957457031173217110036;
			m_weights[33]=.042268289805318078996087;
			m_weights[34]=.042501912565157824383564;
			m_weights[35]=.042657898560231796182558;
			m_weights[36]=.042735962856418776150141;
			break;
		case 75:
			m_abscissas[1]=.998677591495175366570377;
			m_abscissas[2]=.995569180616025852904221;
			m_abscissas[3]=.990690205699502408681585;
			m_abscissas[4]=.984049719773212312942148;
			m_abscissas[5]=.975659596239060615434059;
			m_abscissas[6]=.965534771513308096236744;
			m_abscissas[7]=.953693254250812810121647;
			m_abscissas[8]=.940156101354264828000398;
			m_abscissas[9]=.924947382932221387808996;
			m_abscissas[10]=.90809414036749346639309;
			m_abscissas[11]=.889626338595502192654189;
			m_abscissas[12]=.869576812991138634966082;
			m_abscissas[13]=.847981211069792700530555;
			m_abscissas[14]=.824877929150370595388683;
			m_abscissas[15]=.80030804411209879667461;
			m_abscissas[16]=.774315240375419262613279;
			m_abscissas[17]=.746945732241260280719745;
			m_abscissas[18]=.718248181729095430400604;
			m_abscissas[19]=.68827361206114955473468;
			m_abscissas[20]=.657075316947300885646364;
			m_abscissas[21]=.624708765832395595240814;
			m_abscissas[22]=.591231505274701355512793;
			m_abscissas[23]=.556703056631007315098089;
			m_abscissas[24]=.521184810230386017891518;
			m_abscissas[25]=.484739916224839903912303;
			m_abscissas[26]=.447433172310940856595827;
			m_abscissas[27]=.409330908522120056696609;
			m_abscissas[28]=.370500869296464224958272;
			m_abscissas[29]=.331012093029712097108724;
			m_abscissas[30]=.290934789327611947490661;
			m_abscissas[31]=.250340214175888517251306;
			m_abscissas[32]=.209300543249768096244005;
			m_abscissas[33]=.16788874358831688796112;
			m_abscissas[34]=.126178443861754087166433;
			m_abscissas[35]=.084243803462402028528681;
			m_abscissas[36]=.042159380652026791664366;
			m_abscissas[37]=0.;
			m_weights[0]=.000360360360360360360360;
			m_weights[1]=.002220517299295509506406;
			m_weights[2]=.003995005321211884447826;
			m_weights[3]=.005761457260424863191463;
			m_weights[4]=.007517537537301114406619;
			m_weights[5]=.009260220087137748211213;
			m_weights[6]=.010986427491339444109243;
			m_weights[7]=.012693096958960389772517;
			m_weights[8]=.014377196448480599621057;
			m_weights[9]=.016035732702302637934479;
			m_weights[10]=.017665757385581500072837;
			m_weights[11]=.019264372623930965803714;
			m_weights[12]=.020828736276305757807549;
			m_weights[13]=.022356067042206738724584;
			m_weights[14]=.023843649432795655489456;
			m_weights[15]=.025288838612328093116570;
			m_weights[16]=.026689065107826466263080;
			m_weights[17]=.028041839381614675388640;
			m_weights[18]=.029344756260040815256954;
			m_weights[19]=.030595499211257983668850;
			m_weights[20]=.031791844464854567832269;
			m_weights[21]=.032931664966229576983004;
			m_weights[22]=.034012934158808466062511;
			m_weights[23]=.035033729587449335523619;
			m_weights[24]=.035992236316677997904852;
			m_weights[25]=.036886750157702517763049;
			m_weights[26]=.037715680698487565850934;
			m_weights[27]=.038477554131512797732965;
			m_weights[28]=.039171015874195242550119;
			m_weights[29]=.039794832977321810413789;
			m_weights[30]=.040347896317213302775440;
			m_weights[31]=.040829222567724746394031;
			m_weights[32]=.041237955948577587527340;
			m_weights[33]=.041573369746916460116688;
			m_weights[34]=.041834867609386095280177;
			m_weights[35]=.042021984602431702410728;
			m_weights[36]=.042134388038938066268748;
			m_weights[37]=.042171878069737914912636;
			break;
		case 76:
			m_abscissas[1]=.998712384249348518288221;
			m_abscissas[2]=.995685697217037257841348;
			m_abscissas[3]=.990934829477853933887195;
			m_abscissas[4]=.98446837407949674292728;
			m_abscissas[5]=.976297590742448251437118;
			m_abscissas[6]=.966436643173545736003999;
			m_abscissas[7]=.95490260938410835990496;
			m_abscissas[8]=.941715459941124985429572;
			m_abscissas[9]=.926898025727761018527978;
			m_abscissas[10]=.910475959260251683231539;
			m_abscissas[11]=.892477690626197859587115;
			m_abscissas[12]=.872934378425957762006982;
			m_abscissas[13]=.851879855910361597549083;
			m_abscissas[14]=.829350572450994683724392;
			m_abscissas[15]=.805385530463054916778228;
			m_abscissas[16]=.780026217898716348412259;
			m_abscissas[17]=.753316536432206371619183;
			m_abscissas[18]=.725302725463194052489699;
			m_abscissas[19]=.696033282071297795217579;
			m_abscissas[20]=.665558877060993721750249;
			m_abscissas[21]=.633932267242683225750415;
			m_abscissas[22]=.601208204102030151246078;
			m_abscissas[23]=.56744333901583534524021;
			m_abscissas[24]=.532696125178637934731813;
			m_abscissas[25]=.497026716409892054220771;
			m_abscissas[26]=.460496863016946092445474;
			m_abscissas[27]=.423169804894134394392031;
			m_abscissas[28]=.38511016204306694813807;
			m_abscissas[29]=.346383822703660844082226;
			m_abscissas[30]=.307057829289589401640909;
			m_abscissas[31]=.267200262325622978335049;
			m_abscissas[32]=.226880122587792599082129;
			m_abscissas[33]=.186167211650417404515481;
			m_abscissas[34]=.145132011046793949193284;
			m_abscissas[35]=.103845560252744689132617;
			m_abscissas[36]=.062379333704260348140585;
			m_abscissas[37]=.020805117062142662036204;
			m_weights[0]=.000350877192982456140351;
			m_weights[1]=.002162107722102053256731;
			m_weights[2]=.003890024981613587402622;
			m_weights[3]=.005610299353547198243703;
			m_weights[4]=.007320736637641664942013;
			m_weights[5]=.009018469400863626741796;
			m_weights[6]=.010700579009354140294284;
			m_weights[7]=.012364159597680721726433;
			m_weights[8]=.014006333494405186273199;
			m_weights[9]=.015624258777817107925332;
			m_weights[10]=.017215134991203987188546;
			m_weights[11]=.018776208278827699445631;
			m_weights[12]=.020304776270737773410859;
			m_weights[13]=.021798192813625642081716;
			m_weights[14]=.023253872577176730803689;
			m_weights[15]=.024669295542790605807021;
			m_weights[16]=.026042011373255646563682;
			m_weights[17]=.027369643658729628085395;
			m_weights[18]=.028649894033093318395706;
			m_weights[19]=.029880546154275388195502;
			m_weights[20]=.031059469542044556979390;
			m_weights[21]=.032184623266839338628091;
			m_weights[22]=.033254059483372176952821;
			m_weights[23]=.034265926802963201278291;
			m_weights[24]=.035218473498809486675270;
			m_weights[25]=.036110050538668265411004;
			m_weights[26]=.03693911443972133786325;
			m_weights[27]=.037704229940689652207313;
			m_weights[28]=.038404072486579565750307;
			m_weights[29]=.039037430521764223953361;
			m_weights[30]=.039603207587433709868833;
			m_weights[31]=.040100424219785238463424;
			m_weights[32]=.040528219645668918988842;
			m_weights[33]=.040885853272754767670840;
			m_weights[34]=.041172705971642039907855;
			m_weights[35]=.041388281147691901389889;
			m_weights[36]=.041532205600728315597432;
			m_weights[37]=.041604230171119139489575;
			break;
		case 77:
			m_abscissas[1]=.998745821822100005988076;
			m_abscissas[2]=.995797678482872756647468;
			m_abscissas[3]=.991169941335394515168962;
			m_abscissas[4]=.984870772636740466111087;
			m_abscissas[5]=.976910857614281440410767;
			m_abscissas[6]=.967303636580608826984692;
			m_abscissas[7]=.956065316202466556481515;
			m_abscissas[8]=.943214849785668819936184;
			m_abscissas[9]=.928773907588026951895724;
			m_abscissas[10]=.91276684109782025259227;
			m_abscissas[11]=.895220642310131645504027;
			m_abscissas[12]=.876164898367183166886158;
			m_abscissas[13]=.855631741744635508979826;
			m_abscissas[14]=.833655796109728329936274;
			m_abscissas[15]=.810274117960749884384578;
			m_abscissas[16]=.785526134154750777206074;
			m_abscissas[17]=.759453575433071992189657;
			m_abscissas[18]=.732100406058991500834014;
			m_abscissas[19]=.703512749687347836450469;
			m_abscissas[20]=.673738811591830603069073;
			m_abscissas[21]=.642828797381486141798903;
			m_abscissas[22]=.610834828343746934006495;
			m_abscissas[23]=.577810853556888989374427;
			m_abscissas[24]=.543812558920213457958475;
			m_abscissas[25]=.508897273255412075452563;
			m_abscissas[26]=.4731238716374933189619;
			m_abscissas[27]=.436552676118304027729835;
			m_abscissas[28]=.399245354010068954705524;
			m_abscissas[29]=.361264813900479114527966;
			m_abscissas[30]=.322675099574680885589154;
			m_abscissas[31]=.28354128202304449917377;
			m_abscissas[32]=.243929349716816437769182;
			m_abscissas[33]=.203906097336679655131689;
			m_abscissas[34]=.163539013141853321150486;
			m_abscissas[35]=.122896165169655469057803;
			m_abscissas[36]=.082046086457423537243932;
			m_abscissas[37]=.041057659480335974470632;
			m_abscissas[38]=0.;
			m_weights[0]=.000341763499658236500342;
			m_weights[1]=.002105972565011591481666;
			m_weights[2]=.003789127077951330835066;
			m_weights[3]=.005465007337773833492171;
			m_weights[4]=.007131550971827646884155;
			m_weights[5]=.008786039023989592376154;
			m_weights[6]=.010425701574545074529615;
			m_weights[7]=.012047779796153127892939;
			m_weights[8]=.013649540730543248477868;
			m_weights[9]=.015228284403894977759359;
			m_weights[10]=.016781349154282032580844;
			m_weights[11]=.018306116400209357578642;
			m_weights[12]=.019800015170457010094070;
			m_weights[13]=.021260526490503396077344;
			m_weights[14]=.022685187654797980673235;
			m_weights[15]=.024071596392137405489221;
			m_weights[16]=.025417414923312803275978;
			m_weights[17]=.026720373907026238499771;
			m_weights[18]=.027978276268803244525603;
			m_weights[19]=.029189000907150585656573;
			m_weights[20]=.030350506271086317680846;
			m_weights[21]=.031460833803218303595663;
			m_weights[22]=.032518111242684992612305;
			m_weights[23]=.033520555782459604822856;
			m_weights[24]=.034466477075736652681883;
			m_weights[25]=.035354280086357990654175;
			m_weights[26]=.036182467778488854083159;
			m_weights[27]=.036949643641019453810166;
			m_weights[28]=.037654514042442653919316;
			m_weights[29]=.038295890412241734323763;
			m_weights[30]=.038872691245113234826811;
			m_weights[31]=.039383943924647593623046;
			m_weights[32]=.039828786363394040407903;
			m_weights[33]=.040206468456545352777951;
			m_weights[34]=.040516353346792038767770;
			m_weights[35]=.040757918498213690868522;
			m_weights[36]=.040930756577397099784339;
			m_weights[37]=.041034576140295655865173;
			m_weights[38]=.041069202123672040427474;
			break;
		case 78:
			m_abscissas[1]=.998777973685999410508424;
			m_abscissas[2]=.995905356726308746109961;
			m_abscissas[3]=.991396027877122376988439;
			m_abscissas[4]=.985257745590867531402087;
			m_abscissas[5]=.977500656824682431312287;
			m_abscissas[6]=.968137524084888979205224;
			m_abscissas[7]=.957183737519682037768556;
			m_abscissas[8]=.944657297050677188659591;
			m_abscissas[9]=.930578785013177395636122;
			m_abscissas[10]=.914971333137803996424244;
			m_abscissas[11]=.897860584874865689790819;
			m_abscissas[12]=.879274653413189999028216;
			m_abscissas[13]=.859244075565158448326667;
			m_abscissas[14]=.837801761634514615171292;
			m_abscissas[15]=.814982941367037012615127;
			m_abscissas[16]=.790825106081166381760585;
			m_abscissas[17]=.765367947077785395560552;
			m_abscissas[18]=.738653290432499555986381;
			m_abscissas[19]=.710725028278735654148998;
			m_abscissas[20]=.681629046695231050254761;
			m_abscissas[21]=.651413150316789779936559;
			m_abscissas[22]=.620126983792409659032575;
			m_abscissas[23]=.587821950219973942454891;
			m_abscissas[24]=.554551126691614540219848;
			m_abscissas[25]=.520369177088567789793779;
			m_abscissas[26]=.485332262268842035543049;
			m_abscissas[27]=.44949794779528651434874;
			m_abscissas[28]=.412925109355683430260907;
			m_abscissas[29]=.375673836030271293001653;
			m_abscissas[30]=.337805331565640333916618;
			m_abscissas[31]=.299381813817213645604112;
			m_abscissas[32]=.260466412525534828834876;
			m_abscissas[33]=.221123065594319165990688;
			m_abscissas[34]=.181416414040686004306588;
			m_abscissas[35]=.141411695790170966170014;
			m_abscissas[36]=.101174638491014146634346;
			m_abscissas[37]=.060771351523831439816062;
			m_abscissas[38]=.020268217384097873262030;
			m_weights[0]=.000333000333000333000333;
			m_weights[1]=.002051995271407972094747;
			m_weights[2]=.003692102738812200812235;
			m_weights[3]=.005325281866318271555987;
			m_weights[4]=.006949593299154106357857;
			m_weights[5]=.008562457170302024150652;
			m_weights[6]=.010161242958430579438969;
			m_weights[7]=.011743329722728395695409;
			m_weights[8]=.013306120273912448113696;
			m_weights[9]=.014847047884016059863631;
			m_weights[10]=.016363581258288347390506;
			m_weights[11]=.017853228967268965294782;
			m_weights[12]=.019313543651581745953415;
			m_weights[13]=.020742126092805867833326;
			m_weights[14]=.0221366291794603503844;
			m_weights[15]=.023494761775671939462682;
			m_weights[16]=.024814292492202426070994;
			m_weights[17]=.026093053356406534900405;
			m_weights[18]=.027328943376435762405337;
			m_weights[19]=.028519931994519540049522;
			m_weights[20]=.029664062424016772461059;
			m_weights[21]=.030759454864958277833283;
			m_weights[22]=.031804309592913529683472;
			m_weights[23]=.032796909916175521056066;
			m_weights[24]=.033735624996446855428334;
			m_weights[25]=.034618912528418633870519;
			m_weights[26]=.035445321273856083202432;
			m_weights[27]=.036213493446038113793204;
			m_weights[28]=.036922166940640231251232;
			m_weights[29]=.037570177409400200012867;
			m_weights[30]=.038156460173162656450791;
			m_weights[31]=.038680051971161772174489;
			m_weights[32]=.039140092543669459106906;
			m_weights[33]=.039535826045409931415963;
			m_weights[34]=.039866602287419173980122;
			m_weights[35]=.040131877805309509826029;
			m_weights[36]=.040331216752184516633434;
			m_weights[37]=.040464291614737526293476;
			m_weights[38]=.040530883751357364697437;
			break;
		case 79:
			m_abscissas[1]=.99880890491839960060491;
			m_abscissas[2]=.996008949576778147559856;
			m_abscissas[3]=.99161354499852393997726;
			m_abscissas[4]=.985630070805993305208968;
			m_abscissas[5]=.978068169202027650428462;
			m_abscissas[6]=.968939967071820301092868;
			m_abscissas[7]=.958260088786484813408687;
			m_abscissas[8]=.946045640012901027079492;
			m_abscissas[9]=.932316182478450533123024;
			m_abscissas[10]=.917093703422759140436424;
			m_abscissas[11]=.900402580708479171976415;
			m_abscissas[12]=.882269543929455820726745;
			m_abscissas[13]=.86272363167871762854223;
			m_abscissas[14]=.84179614508448886224129;
			m_abscissas[15]=.819520597705906195221623;
			m_abscissas[16]=.795932661876753159223911;
			m_abscissas[17]=.771070111587152002083937;
			m_abscissas[18]=.744972761996784068228178;
			m_abscissas[19]=.717682405677652995803481;
			m_abscissas[20]=.689242745689146625687869;
			m_abscissas[21]=.659699325592957114941095;
			m_abscissas[22]=.629099456520167753980946;
			m_abscissas[23]=.597492141407447604006086;
			m_abscissas[24]=.564927996523775820485244;
			m_abscissas[25]=.531459170413424185096407;
			m_abscissas[26]=.497139260385043705334456;
			m_abscissas[27]=.462023226680617820344446;
			m_abscissas[28]=.426167304461751728634799;
			m_abscissas[29]=.389628913754257032985803;
			m_abscissas[30]=.352466567495256677672388;
			m_abscissas[31]=.314739777830071171019185;
			m_abscissas[32]=.276508960808948079043472;
			m_abscissas[33]=.237835339636257998912637;
			m_abscissas[34]=.198780846627097402790271;
			m_abscissas[35]=.159408024028308059795226;
			m_abscissas[36]=.119779923862740806558954;
			m_abscissas[37]=.079960006957155282252918;
			m_abscissas[38]=.040012041315454332398912;
			m_abscissas[39]=0.;
			m_weights[0]=.000324569944823109380071;
			m_weights[1]=.002000066655693811950907;
			m_weights[2]=.003598756276434813202070;
			m_weights[3]=.005190842429146995580878;
			m_weights[4]=.006774500643656519002294;
			m_weights[5]=.008347281454628460964517;
			m_weights[6]=.009906685113141957190664;
			m_weights[7]=.011450220086306295908266;
			m_weights[8]=.012975416662037399774332;
			m_weights[9]=.014479833284760545885356;
			m_weights[10]=.015961061200966027767175;
			m_weights[11]=.017416728582840545158027;
			m_weights[12]=.018844504435087638466321;
			m_weights[13]=.020242102376404730639532;
			m_weights[14]=.021607284324378811582750;
			m_weights[15]=.022937864091628290521112;
			m_weights[16]=.024231710893311147029049;
			m_weights[17]=.025486752763079023254662;
			m_weights[18]=.026700979873317817368245;
			m_weights[19]=.027872447755027573469772;
			m_weights[20]=.028999280412542761836327;
			m_weights[21]=.030079673328303119313655;
			m_weights[22]=.031111896352976735451687;
			m_weights[23]=.032094296476374177529287;
			m_weights[24]=.033025300474756970343351;
			m_weights[25]=.033903417430326291280303;
			m_weights[26]=.034727241118873211148466;
			m_weights[27]=.035495452261777285812957;
			m_weights[28]=.036206820638753998050466;
			m_weights[29]=.036860207057972287288063;
			m_weights[30]=.037454565180390344712632;
			m_weights[31]=.037988943195390336399786;
			m_weights[32]=.038462485345030196382214;
			m_weights[33]=.038874433294472613282779;
			m_weights[34]=.039224127346397360233608;
			m_weights[35]=.039511007497452748895004;
			m_weights[36]=.039734614335054795018316;
			m_weights[37]=.039894589773098240581244;
			m_weights[38]=.039990677625401463465501;
			m_weights[39]=.040022724015967097756710;
			break;
		case 80:
			m_abscissas[1]=.998838676530925078355099;
			m_abscissas[2]=.996108661079905199383947;
			m_abscissas[3]=.991822920181215868187135;
			m_abscissas[4]=.985988477763353179517021;
			m_abscissas[5]=.978614502316916237354856;
			m_abscissas[6]=.969712524180795617358871;
			m_abscissas[7]=.95929644895627958345874;
			m_abscissas[8]=.947382542843566522957115;
			m_abscissas[9]=.933989409347880203104772;
			m_abscissas[10]=.919137960990880499721068;
			m_abscissas[11]=.902851386971541533959915;
			m_abscissas[12]=.885155117102388346349455;
			m_abscissas[13]=.866076782175049819116344;
			m_abscissas[14]=.845646170855781575384207;
			m_abscissas[15]=.82389518319510250605021;
			m_abscissas[16]=.800857780832007330696272;
			m_abscissas[17]=.7765699339744185249253;
			m_abscissas[18]=.751069565240709811021182;
			m_abscissas[19]=.72439649045110723775404;
			m_abscissas[20]=.696592356462052999082075;
			m_abscissas[21]=.667700576140971784667015;
			m_abscissas[22]=.637766260583196657267152;
			m_abscissas[23]=.606836148677031639425659;
			m_abscissas[24]=.574958534127017420342741;
			m_abscissas[25]=.542183190049403807573462;
			m_abscissas[26]=.508561291257603478978813;
			m_abscissas[27]=.474145334358995725478628;
			m_abscissas[28]=.438989055787857900064847;
			m_abscissas[29]=.403147347902419263421981;
			m_abscissas[30]=.366676173277050829975379;
			m_abscissas[31]=.329632477323420340298729;
			m_abscissas[32]=.292074099377048829203835;
			m_abscissas[33]=.254059682388100099878332;
			m_abscissas[34]=.215648581357412848308747;
			m_abscissas[35]=.176900770660743697513687;
			m_abscissas[36]=.137876750405924865766332;
			m_abscissas[37]=.098637451969149846919260;
			m_abscissas[38]=.059244142857881932493213;
			m_abscissas[39]=.019758331048931629780230;
			m_weights[0]=.000316455696202531645570;
			m_weights[1]=.001950084350977559340474;
			m_weights[2]=.003508904201029018399793;
			m_weights[3]=.005061425949589057683013;
			m_weights[4]=.006605932495168233793061;
			m_weights[5]=.008140096727011971659965;
			m_weights[6]=.009661541673697277599664;
			m_weights[7]=.011167897347743279416050;
			m_weights[8]=.012656813820266284299011;
			m_weights[9]=.014125967211192374290775;
			m_weights[10]=.015573064034917646948155;
			m_weights[11]=.016995845040870654249112;
			m_weights[12]=.018392088846893499823931;
			m_weights[13]=.019759615455044966282947;
			m_weights[14]=.021096289678281489680299;
			m_weights[15]=.022400024486050025554541;
			m_weights[16]=.023668784269300632189419;
			m_weights[17]=.024900588022449768234019;
			m_weights[18]=.026093512438604030628558;
			m_weights[19]=.027245694913864537361425;
			m_weights[20]=.028355336456369543553869;
			m_weights[21]=.029420704495726294269795;
			m_weights[22]=.030440135588556215114546;
			m_weights[23]=.031412038015994427315518;
			m_weights[24]=.032334894269127558833428;
			m_weights[25]=.033207263418513783367375;
			m_weights[26]=.034027783364101021579201;
			m_weights[27]=.034795172962040472781670;
			m_weights[28]=.035508234025081408612965;
			m_weights[29]=.036165853193428345382558;
			m_weights[30]=.036767003673142537573464;
			m_weights[31]=.037310746839375602430654;
			m_weights[32]=.037796233701933498811388;
			m_weights[33]=.038222706230883605435152;
			m_weights[34]=.038589498540135870664241;
			m_weights[35]=.038896037927150553000195;
			m_weights[36]=.039141845767149563761486;
			m_weights[37]=.039326538260435491200255;
			m_weights[38]=.039449827031651661131343;
			m_weights[39]=.039511519580047706101119;
			break;
		case 81:
			m_abscissas[1]=.998867345770500841100192;
			m_abscissas[2]=.996204682702184960470576;
			m_abscissas[3]=.99202455459093139926184;
			m_abscissas[4]=.986333651125435510374692;
			m_abscissas[5]=.979140695862641698268511;
			m_abscissas[6]=.970456658823796134193629;
			m_abscissas[7]=.960294770429855704815294;
			m_abscissas[8]=.948670508227915468560794;
			m_abscissas[9]=.935601575373260056421224;
			m_abscissas[10]=.921107874410171768350035;
			m_abscissas[11]=.905211477269173116591232;
			m_abscissas[12]=.887936591794978293100687;
			m_abscissas[13]=.869309524951357764395725;
			m_abscissas[14]=.849358642796758156385449;
			m_abscissas[15]=.828114327308051811762871;
			m_abscissas[16]=.805608930125844346226574;
			m_abscissas[17]=.781876723295594503099018;
			m_abscissas[18]=.756953847081557058237751;
			m_abscissas[19]=.730878254934112899975292;
			m_abscissas[20]=.703689655694914046223517;
			m_abscissas[21]=.675429453128221811751892;
			m_abscissas[22]=.646140682870744057390973;
			m_abscissas[23]=.615867946896124935416579;
			m_abscissas[24]=.58465734559397525646683;
			m_abscissas[25]=.55255640756693354887951;
			m_abscissas[26]=.519614017252703657078819;
			m_abscissas[27]=.485880340481314559105281;
			m_abscissas[28]=.451406748080984313936634;
			m_abscissas[29]=.416245737648936324129461;
			m_abscissas[30]=.380450853606306937499523;
			m_abscissas[31]=.344076605658893962126213;
			m_abscissas[32]=.30717838578792157494801;
			m_abscissas[33]=.269812383897234425048037;
			m_abscissas[34]=.232035502245378899863332;
			m_abscissas[35]=.19390526879287931417384;
			m_abscissas[36]=.155479749596668308617944;
			m_abscissas[37]=.116817460385081442421644;
			m_abscissas[38]=.077977277448073591043864;
			m_abscissas[39]=.039018347978357390263399;
			m_abscissas[40]=0.;
			m_weights[0]=.000308641975308641975309;
			m_weights[1]=.001901952304441548333185;
			m_weights[2]=.003422374319978231311735;
			m_weights[3]=.004936785501319891524640;
			m_weights[4]=.006443569163820328852990;
			m_weights[5]=.007940513090266067115098;
			m_weights[6]=.009425355669784103564330;
			m_weights[7]=.010895841162442700805907;
			m_weights[8]=.012349732276884802987120;
			m_weights[9]=.013784815841038976893845;
			m_weights[10]=.015198906871869646907807;
			m_weights[11]=.016589852155877629562367;
			m_weights[12]=.017955533631297165600375;
			m_weights[13]=.019293871659758731348521;
			m_weights[14]=.020602828215540756039235;
			m_weights[15]=.021880410000606579500917;
			m_weights[16]=.023124671486272563520634;
			m_weights[17]=.024333717879439170183180;
			m_weights[18]=.025505708010114218420441;
			m_weights[19]=.026638857136467921747019;
			m_weights[20]=.027731439663487871837948;
			m_weights[21]=.028781791771282234063462;
			m_weights[22]=.029788313949136625549039;
			m_weights[23]=.030749473431529472622314;
			m_weights[24]=.031663806532434869134677;
			m_weights[25]=.032529920874382211161514;
			m_weights[26]=.033346497508893386070601;
			m_weights[27]=.034112292925078335171105;
			m_weights[28]=.034826140943336761528049;
			m_weights[29]=.03548695449128657126638;
			m_weights[30]=.036093727259217589987423;
			m_weights[31]=.036645535232551636465016;
			m_weights[32]=.037141538098976715543393;
			m_weights[33]=.037580980528113521005051;
			m_weights[34]=.037963193321766260115129;
			m_weights[35]=.038287594433006686307127;
			m_weights[36]=.038553689852539827074446;
			m_weights[37]=.038761074361001897928360;
			m_weights[38]=.038909432146044980071968;
			m_weights[39]=.038998537283268889811402;
			m_weights[40]=.039028254080267962181883;
			break;
		case 82:
			m_abscissas[1]=.998894966394696698815605;
			m_abscissas[2]=.996297194250043799865703;
			m_abscissas[3]=.99221882499705555093116;
			m_abscissas[4]=.986666233997750741070277;
			m_abscissas[5]=.979647726580751824185994;
			m_abscissas[6]=.971173746070983293801368;
			m_abscissas[7]=.961256888166997334925859;
			m_abscissas[8]=.949911888933842704619555;
			m_abscissas[9]=.937155604911370341385666;
			m_abscissas[10]=.923006988792194317328992;
			m_abscissas[11]=.907487061561756742817092;
			m_abscissas[12]=.890618881403910931940402;
			m_abscissas[13]=.872427509511123607565274;
			m_abscissas[14]=.852939972886976375387955;
			m_abscissas[15]=.832185224212252659257169;
			m_abscissas[16]=.81019409884172699653514;
			m_abscissas[17]=.786999268999269027481455;
			m_abscissas[18]=.762635195241263068011717;
			m_abscissas[19]=.737138075261520534304048;
			m_abscissas[20]=.710545790114352497924644;
			m_abscissas[21]=.682897847936055862838255;
			m_abscissas[22]=.654235325248642747640185;
			m_abscissas[23]=.624600805933153053695664;
			m_abscissas[24]=.594038317963303664814702;
			m_abscissas[25]=.562593267993525229208906;
			m_abscissas[26]=.530312373898606395868827;
			m_abscissas[27]=.497243595365196607888382;
			m_abscissas[28]=.463436062638305020340167;
			m_abscissas[29]=.428940003528668983594753;
			m_abscissas[30]=.393806668789445809609376;
			m_abscissas[31]=.358088255973101799002335;
			m_abscissas[32]=.321837831881628782746056;
			m_abscissas[33]=.285109253725307131639182;
			m_abscissas[34]=.247957089107152048644977;
			m_abscissas[35]=.210436534951924029583053;
			m_abscissas[36]=.172603335500151994800979;
			m_abscissas[37]=.134513699489006381242948;
			m_abscissas[38]=.096224216643067344382673;
			m_abscissas[39]=.057791773599058335377344;
			m_abscissas[40]=.019273469389456151616600;
			m_weights[0]=.000301114122252333634447;
			m_weights[1]=.001855580315744599796964;
			m_weights[2]=.003339004913675408060968;
			m_weights[3]=.004816689134073012918872;
			m_weights[4]=.006287110273231084153392;
			m_weights[5]=.007748164083425875799598;
			m_weights[6]=.009197697427290524946622;
			m_weights[7]=.010633562032620982961726;
			m_weights[8]=.012053626603374698460348;
			m_weights[9]=.013455782194530971975302;
			m_weights[10]=.014837946027709788072788;
			m_weights[11]=.016198064832495504266826;
			m_weights[12]=.017534117997660981332801;
			m_weights[13]=.018844120618259958077543;
			m_weights[14]=.020126126466356761611963;
			m_weights[15]=.021378230893713586412633;
			m_weights[16]=.022598573667575668777677;
			m_weights[17]=.023785341737842247091297;
			m_weights[18]=.024936771932727891385110;
			m_weights[19]=.026051153579530478486382;
			m_weights[20]=.027126831046943608377995;
			m_weights[21]=.028162206205320058757002;
			m_weights[22]=.029155740801336359309569;
			m_weights[23]=.030105958743592630313875;
			m_weights[24]=.031011448295789692719126;
			m_weights[25]=.031870864174248491430133;
			m_weights[26]=.032682929546670441958853;
			m_weights[27]=.033446437929178796211226;
			m_weights[28]=.034160254978828965833656;
			m_weights[29]=.034823320178928948512609;
			m_weights[30]=.035434648414668893061192;
			m_weights[31]=.035993331436720906984004;
			m_weights[32]=.036498539210636048832114;
			m_weights[33]=.036949521150034707432194;
			m_weights[34]=.037345607231758934180044;
			m_weights[35]=.037686208991330464907822;
			m_weights[36]=.037970820397235856655826;
			m_weights[37]=.038199018602740089808816;
			m_weights[38]=.038370464574111867933797;
			m_weights[39]=.038484903594327407692613;
			m_weights[40]=.038542165641504470864272;
			break;
		case 83:
			m_abscissas[1]=.998921588923861277567092;
			m_abscissas[2]=.996386364711517740044922;
			m_abscissas[3]=.992406085530840691830513;
			m_abscissas[4]=.986986830917179220037428;
			m_abscissas[5]=.980136512775077015730332;
			m_abscissas[6]=.97186507896333575973954;
			m_abscissas[7]=.962184528043615338013599;
			m_abscissas[8]=.951108898428134088385066;
			m_abscissas[9]=.938654249978645141576373;
			m_abscissas[10]=.924838641425436563989233;
			m_abscissas[11]=.909682104474748287078922;
			m_abscissas[12]=.893206614897845783530992;
			m_abscissas[13]=.875436060734328851426454;
			m_abscissas[14]=.856396207691770955313956;
			m_abscissas[15]=.836114661807490909714866;
			m_abscissas[16]=.814620829433902569912719;
			m_abscissas[17]=.791945874609092352811323;
			m_abscissas[18]=.768122673876334582934876;
			m_abscissas[19]=.743185768619092722407999;
			m_abscissas[20]=.71717131498120825470085;
			m_abscissas[21]=.690117031445236664671839;
			m_abscissas[22]=.662062144145148315961363;
			m_abscissas[23]=.63304732999281730194859;
			m_abscissas[24]=.603114657700843498468359;
			m_abscissas[25]=.572307526787273930443091;
			m_abscissas[26]=.540670604650697262906419;
			m_abscissas[27]=.508249761806971231139434;
			m_abscissas[28]=.475092005381500586917953;
			m_abscissas[29]=.441245410953507303946699;
			m_abscissas[30]=.406759052851120778275678;
			m_abscissas[31]=.37168293299835954745688;
			m_abscissas[32]=.33606790841717401900682;
			m_abscissas[33]=.299965617489668568098153;
			m_abscissas[34]=.263428405087418156912305;
			m_abscissas[35]=.226509246676436636322379;
			m_abscissas[36]=.189261671507838664890527;
			m_abscissas[37]=.151739685005562520805637;
			m_abscissas[38]=.113997690463685035392832;
			m_abscissas[39]=.076090410166860732126301;
			m_abscissas[40]=.038072806048253545877765;
			m_abscissas[41]=0.;
			m_weights[0]=.000293858360270349691449;
			m_weights[1]=.001810883614315921626372;
			m_weights[2]=.003258643980634346688659;
			m_weights[3]=.004700918797684617046260;
			m_weights[4]=.006136273379206363971721;
			m_weights[5]=.007562705015459330308044;
			m_weights[6]=.008978162641886783867461;
			m_weights[7]=.010380599149952347285984;
			m_weights[8]=.011767983059442399568825;
			m_weights[9]=.013138303619028023729168;
			m_weights[10]=.014489574387551097542067;
			m_weights[11]=.015819836355151715412622;
			m_weights[12]=.017127160881894571564311;
			m_weights[13]=.018409652538095154033259;
			m_weights[14]=.019665451873743852442305;
			m_weights[15]=.020892738125438802360140;
			m_weights[16]=.022089731862223725455325;
			m_weights[17]=.023254697568935086292048;
			m_weights[18]=.024385946164499308634913;
			m_weights[19]=.025481837452134937943469;
			m_weights[20]=.026540782498230628286797;
			m_weights[21]=.027561245936629061822983;
			m_weights[22]=.02854174819507854693111;
			m_weights[23]=.029480867640684824211956;
			m_weights[24]=.030377242641289183867343;
			m_weights[25]=.031229573539806965499896;
			m_weights[26]=.032036624538678368064142;
			m_weights[27]=.032797225491708713005238;
			m_weights[28]=.033510273600706378785740;
			m_weights[29]=.034174735014462615636671;
			m_weights[30]=.034789646327757716077635;
			m_weights[31]=.035354115978222091274481;
			m_weights[32]=.035867325539028314910181;
			m_weights[33]=.036328530905540830346156;
			m_weights[34]=.036737063374203491289624;
			m_weights[35]=.037092330612101157426568;
			m_weights[36]=.037393817515789941491001;
			m_weights[37]=.037641086958151155647565;
			m_weights[38]=.037833780422186288301724;
			m_weights[39]=.037971618520834214091128;
			m_weights[40]=.038054401402057056616211;
			m_weights[41]=.038082009038607441902893;
			break;
		case 84:
			m_abscissas[1]=.998947260872255005807015;
			m_abscissas[2]=.996472353027904818547034;
			m_abscissas[3]=.992586669297606460556407;
			m_abscissas[4]=.987296010592762689873231;
			m_abscissas[5]=.980607918453013685492285;
			m_abscissas[6]=.972531874306085590026333;
			m_abscissas[7]=.963079314524793732751603;
			m_abscissas[8]=.952263620628578004844581;
			m_abscissas[9]=.940100102250194065456954;
			m_abscissas[10]=.926605976154092982489685;
			m_abscissas[11]=.911800342151917573259302;
			m_abscissas[12]=.895704156198804296271344;
			m_abscissas[13]=.878340200797066947903203;
			m_abscissas[14]=.859733052784286464646288;
			m_abscissas[15]=.839909048566660305905079;
			m_abscissas[16]=.818896246853953839431739;
			m_abscissas[17]=.796724388952343768277165;
			m_abscissas[18]=.773424856673210539005279;
			m_abscissas[19]=.749030627918470877659969;
			m_abscissas[20]=.72357623000589233387422;
			m_abscissas[21]=.697097690800792773539954;
			m_abscissas[22]=.669632487723498757684324;
			m_abscissas[23]=.641219494704865602093301;
			m_abscissas[24]=.61189892716501969814064;
			m_abscissas[25]=.581712285093253053165171;
			m_abscissas[26]=.550702294309669701684771;
			m_abscissas[27]=.518912845991745678352359;
			m_abscissas[28]=.48638893455141255523422;
			m_abscissas[29]=.453176593950604199664442;
			m_abscissas[30]=.419322832545413211883429;
			m_abscissas[31]=.384875566551083772206299;
			m_abscissas[32]=.349883552222018066788752;
			m_abscissas[33]=.31439631684279107971192;
			m_abscissas[34]=.278464088627850614409527;
			m_abscissas[35]=.242137725629123458809862;
			m_abscissas[36]=.205468643752152383257433;
			m_abscissas[37]=.168508743982650125544796;
			m_abscissas[38]=.131310338926473856135606;
			m_abscissas[39]=.093926078766995225441772;
			m_abscissas[40]=.056408876744665583253423;
			m_abscissas[41]=.018811834264252150374245;
			m_weights[0]=.000286861732644865174986;
			m_weights[1]=.001767782471840435397837;
			m_weights[2]=.003181148545303104174252;
			m_weights[3]=.004589269355179933306019;
			m_weights[4]=.005990792702162181263491;
			m_weights[5]=.007383811435236258796079;
			m_weights[6]=.008766370608787283049575;
			m_weights[7]=.010136518411256585432302;
			m_weights[8]=.011492317424473048764302;
			m_weights[9]=.012831849469581517942544;
			m_weights[10]=.014153218972016109157570;
			m_weights[11]=.0154545558797005381519;
			m_weights[12]=.016734018405800827738020;
			m_weights[13]=.017989795678500229984468;
			m_weights[14]=.019220110324810287313262;
			m_weights[15]=.020423220996890563143787;
			m_weights[16]=.021597424842495305205433;
			m_weights[17]=.022741059918432623263395;
			m_weights[18]=.023852507544778188608107;
			m_weights[19]=.024930194597103006417174;
			m_weights[20]=.025972595733786564676920;
			m_weights[21]=.026978235555437807565699;
			m_weights[22]=.027945690693467789562310;
			m_weights[23]=.028873591824917085284268;
			m_weights[24]=.029760625610722112350480;
			m_weights[25]=.030605536554699290919875;
			m_weights[26]=.031407128780630036300248;
			m_weights[27]=.032164267724940509652888;
			m_weights[28]=.032875881742586393661001;
			m_weights[29]=.033540963623873814144077;
			m_weights[30]=.034158572020072292305436;
			m_weights[31]=.034727832775803850476939;
			m_weights[32]=.035247940166323767414944;
			m_weights[33]=.035718158037942703171593;
			m_weights[34]=.036137820849976736120308;
			m_weights[35]=.036506334616751042182002;
			m_weights[36]=.036823177748324275548288;
			m_weights[37]=.037087901788743963890191;
			m_weights[38]=.037300132050788194793685;
			m_weights[39]=.037459568146295330880160;
			m_weights[40]=.037565984411331235977469;
			m_weights[41]=.037619230225592310836719;
			break;
		case 85:
			m_abscissas[1]=.998972026960157594440641;
			m_abscissas[2]=.996555308801970311092125;
			m_abscissas[3]=.992760889856614496116184;
			m_abscissas[4]=.987594308422085978795572;
			m_abscissas[5]=.981062757128789454657754;
			m_abscissas[6]=.973175277991109608513542;
			m_abscissas[7]=.963942777716858125756335;
			m_abscissas[8]=.953378018871155815477814;
			m_abscissas[9]=.941495604099207804585307;
			m_abscissas[10]=.928311956615371338485956;
			m_abscissas[11]=.913845297781995886183542;
			m_abscissas[12]=.898115622052740982557051;
			m_abscissas[13]=.881144669401412831372322;
			m_abscissas[14]=.862955895308730691038971;
			m_abscissas[15]=.843574438363404933105462;
			m_abscissas[16]=.823027085529270621900637;
			m_abscissas[17]=.801342235129942298488726;
			m_abscissas[18]=.778549857603960513601447;
			m_abscissas[19]=.754681454085661735378072;
			m_abscissas[20]=.729770012869580791867364;
			m_abscissas[21]=.703849963818888038175659;
			m_abscissas[22]=.676957130781074315326282;
			m_abscissas[23]=.649128682076775004051668;
			m_abscissas[24]=.620403079130242170864734;
			m_abscissas[25]=.590820023312514360062239;
			m_abscissas[26]=.560420401070786309400088;
			m_abscissas[27]=.529246227419838278589245;
			m_abscissas[28]=.497340587873641129981123;
			m_abscissas[29]=.464747578897404241181001;
			m_abscissas[30]=.431512246962374912923147;
			m_abscissas[31]=.397680526287626795730756;
			m_abscissas[32]=.363299175354887983803413;
			m_abscissas[33]=.32841571228415408492234;
			m_abscissas[34]=.293078349159405282268331;
			m_abscissas[35]=.257335925395196885445903;
			m_abscissas[36]=.221237840236218053306325;
			m_abscissas[37]=.184833984483111384010368;
			m_abscissas[38]=.148174671538915221490802;
			m_abscissas[39]=.111310567871429322878191;
			m_abscissas[40]=.074292622987611655958548;
			m_abscissas[41]=.037171999016788423210591;
			m_abscissas[42]=0.;
			m_weights[0]=.000280112044817927170868;
			m_weights[1]=.001726201846626938835181;
			m_weights[2]=.003106384022699486771074;
			m_weights[3]=.004481547676585441825457;
			m_weights[4]=.005850417962709958969095;
			m_weights[5]=.007211177725200376026866;
			m_weights[6]=.008561962594443944757483;
			m_weights[7]=.009900910591638705233298;
			m_weights[8]=.011226172998875325503467;
			m_weights[9]=.012535918968397497170378;
			m_weights[10]=.013828338687682475061185;
			m_weights[11]=.015101646110443698163300;
			m_weights[12]=.016354081518747354759061;
			m_weights[13]=.017583913997018648771932;
			m_weights[14]=.018789443844557541605651;
			m_weights[15]=.019969004935073187348877;
			m_weights[16]=.021120967025047393965627;
			m_weights[17]=.022243738010062724635279;
			m_weights[18]=.023335766127107425710015;
			m_weights[19]=.024395542100391090455646;
			m_weights[20]=.02542160122801357787177;
			m_weights[21]=.026412525406774042574774;
			m_weights[22]=.027366945092419508096181;
			m_weights[23]=.028283541192681556531282;
			m_weights[24]=.029161046890519855941143;
			m_weights[25]=.029998249395074432120749;
			m_weights[26]=.030793991617920555072903;
			m_weights[27]=.031547173772318476936084;
			m_weights[28]=.032256754893253635998314;
			m_weights[29]=.032921754276170469067375;
			m_weights[30]=.033541252832414081844594;
			m_weights[31]=.034114394359508308101530;
			m_weights[32]=.034640386724515831234594;
			m_weights[33]=.035118502958845788315116;
			m_weights[34]=.035548082262996400825779;
			m_weights[35]=.035928530919844470032539;
			m_weights[36]=.036259323115219841100480;
			m_weights[37]=.036540001664630987227442;
			m_weights[38]=.036770178645137504934161;
			m_weights[39]=.036949535931496357327218;
			m_weights[40]=.037077825635840967201076;
			m_weights[41]=.037154870450285559732434;
			m_weights[42]=.037180563891981298345447;
			break;
		case 86:
			m_abscissas[1]=.998995929308718902614052;
			m_abscissas[2]=.996635372948596527214277;
			m_abscissas[3]=.992929042580881674793743;
			m_abscissas[4]=.987882228803990546223533;
			m_abscissas[5]=.981501795319843123232066;
			m_abscissas[6]=.973796369891473170968597;
			m_abscissas[7]=.964776359855114388571625;
			m_abscissas[8]=.954453944163468617013227;
			m_abscissas[9]=.942843058763103838833216;
			m_abscissas[10]=.929959378437169090875043;
			m_abscissas[11]=.91582029591590576396227;
			m_abscissas[12]=.900444898521195580230873;
			m_abscissas[13]=.883853942461070793478104;
			m_abscissas[14]=.866069824842393191902521;
			m_abscissas[15]=.847116553454036775098147;
			m_abscissas[16]=.827019714368168428990093;
			m_abscissas[17]=.805806437406744234343742;
			m_abscissas[18]=.783505359521609675555206;
			m_abscissas[19]=.760146586138607408295274;
			m_abscissas[20]=.735761650518427215679702;
			m_abscissas[21]=.710383471189383395772343;
			m_abscissas[22]=.684046307509780035377199;
			m_abscissas[23]=.656785713419975444389916;
			m_abscissas[24]=.62863848944665650870919;
			m_abscissas[25]=.599642633024165954010594;
			m_abscissas[26]=.569837287199980333956612;
			m_abscissas[27]=.539262687793606925857717;
			m_abscissas[28]=.507960109080248370096921;
			m_abscissas[29]=.475971808072570621437731;
			m_abscissas[30]=.443340967475799055003505;
			m_abscissas[31]=.410111637393156304604089;
			m_abscissas[32]=.376328675860340853588373;
			m_abscissas[33]=.342037688289325055084501;
			m_abscissas[34]=.307284965903222853820876;
			m_abscissas[35]=.272117423245338933932576;
			m_abscissas[36]=.236582534846760424179413;
			m_abscissas[37]=.200728271137987923395103;
			m_abscissas[38]=.164603033691122901480239;
			m_abscissas[39]=.128255589880032086440917;
			m_abscissas[40]=.091735007046695031317978;
			m_abscissas[41]=.055090586262607595415365;
			m_abscissas[42]=.018371795774660664856580;
			m_weights[0]=.000273597811217510259918;
			m_weights[1]=.001686071056896356631851;
			m_weights[2]=.003034223634601489065505;
			m_weights[3]=.004377571806013967497882;
			m_weights[4]=.005714913310933687814442;
			m_weights[5]=.007044515807472438325587;
			m_weights[6]=.008364600337361943097553;
			m_weights[7]=.009673389661042158647904;
			m_weights[8]=.010969118760398567103136;
			m_weights[9]=.012250039227787949297169;
			m_weights[10]=.013514422246205217048056;
			m_weights[11]=.014760561147430673160658;
			m_weights[12]=.015986773807408025794958;
			m_weights[13]=.017191404957972418506682;
			m_weights[14]=.018372828441146881724131;
			m_weights[15]=.019529449414531459917353;
			m_weights[16]=.020659706509760991010398;
			m_weights[17]=.021762073943389429460237;
			m_weights[18]=.022835063578455358141897;
			m_weights[19]=.023877226934509871071528;
			m_weights[20]=.024887157143694409200222;
			m_weights[21]=.025863490850394761179850;
			m_weights[22]=.026804910052002394837114;
			m_weights[23]=.027710143878354636885586;
			m_weights[24]=.028577970307485786747593;
			m_weights[25]=.029407217815394263135297;
			m_weights[26]=.030196766957612194210862;
			m_weights[27]=.030945551880451167756473;
			m_weights[28]=.031652561759889813069035;
			m_weights[29]=.032316842166164665087426;
			m_weights[30]=.032937496352224822724777;
			m_weights[31]=.033513686464312877294215;
			m_weights[32]=.034044634673039165046725;
			m_weights[33]=.034529624223423355305238;
			m_weights[34]=.034968000402486518784185;
			m_weights[35]=.035359171423087943964953;
			m_weights[36]=.035702609222813907948895;
			m_weights[37]=.035997850176840193088675;
			m_weights[38]=.036244495723806206162574;
			m_weights[39]=.036442212903855938702327;
			m_weights[40]=.036590734808119541669831;
			m_weights[41]=.036689860939028811360539;
			m_weights[42]=.036739457480980232260765;
			break;
		case 87:
			m_abscissas[1]=.999019007619139196532882;
			m_abscissas[2]=.996712678293160570243622;
			m_abscissas[3]=.993091405907932037753287;
			m_abscissas[4]=.988160247266230916237708;
			m_abscissas[5]=.981925755764266718870927;
			m_abscissas[6]=.974396168366924696656241;
			m_abscissas[7]=.965581421278171166961191;
			m_abscissas[8]=.955493142788405020933261;
			m_abscissas[9]=.944144639714160969712937;
			m_abscissas[10]=.93155088048790976418224;
			m_abscissas[11]=.917728475680321324775628;
			m_abscissas[12]=.902695656213240153663881;
			m_abscissas[13]=.886472249374561211579237;
			m_abscissas[14]=.86907965269937404498265;
			m_abscissas[15]=.850540805766041690181903;
			m_abscissas[16]=.830880159951059273021697;
			m_abscissas[17]=.810123646185886187688016;
			m_abscissas[18]=.788298640760006923908308;
			m_abscissas[19]=.765433929216269833876552;
			m_abscissas[20]=.741559668386661389851939;
			m_abscissas[21]=.716707346618904654312205;
			m_abscissas[22]=.690909742246532205069432;
			m_abscissas[23]=.66420088035732804802666;
			m_abscissas[24]=.636615987917234061548015;
			m_abscissas[25]=.608191447308958963838459;
			m_abscissas[26]=.578964748346601999456552;
			m_abscissas[27]=.548974438829602864853874;
			m_abscissas[28]=.518260073701248918269853;
			m_abscissas[29]=.486862162878806550458589;
			m_abscissas[30]=.45482211782409247437224;
			m_abscissas[31]=.422182196924959797549867;
			m_abscissas[32]=.388985449759740552257781;
			m_abscissas[33]=.35527566031815859577965;
			m_abscissas[34]=.321097289253602375134204;
			m_abscissas[35]=.286495415242924069632145;
			m_abscissas[36]=.25151567553110833408293;
			m_abscissas[37]=.21620420573922867033676;
			m_abscissas[38]=.180607579015080908066154;
			m_abscissas[39]=.14477274460675007538179;
			m_abscissas[40]=.108746965940127928312204;
			m_abscissas[41]=.072577758282052571424990;
			m_abscissas[42]=.036312826071288069870283;
			m_abscissas[43]=0.;
			m_weights[0]=.000267308206361935311414;
			m_weights[1]=.001647323480333473513971;
			m_weights[2]=.002964547872568327427706;
			m_weights[3]=.004277170195334306280304;
			m_weights[4]=.005584056340853757060844;
			m_weights[5]=.006883553952252381246459;
			m_weights[6]=.008173964666509673247472;
			m_weights[7]=.009453591231561440192964;
			m_weights[8]=.010720747661948050123114;
			m_weights[9]=.011973763422655820360923;
			m_weights[10]=.013210986238049869986719;
			m_weights[11]=.014430784490257452920520;
			m_weights[12]=.015631549460495840047774;
			m_weights[13]=.016811697490845454830382;
			m_weights[14]=.017969672092280431798486;
			m_weights[15]=.019103946007477180782616;
			m_weights[16]=.020213023230519237582669;
			m_weights[17]=.021295440983053879280094;
			m_weights[18]=.022349771645372863213008;
			m_weights[19]=.023374624640421513351239;
			m_weights[20]=.024368648268545360552905;
			m_weights[21]=.025330531490717414053687;
			m_weights[22]=.026259005657987520733269;
			m_weights[23]=.027152846184927951830896;
			m_weights[24]=.028010874164901522751094;
			m_weights[25]=.028831957925042618731803;
			m_weights[26]=.029615014518913429882244;
			m_weights[27]=.030359011154875253105183;
			m_weights[28]=.031062966558296595829014;
			m_weights[29]=.031725952265805223500157;
			m_weights[30]=.032347093849879714759156;
			m_weights[31]=.032925572072167167735976;
			m_weights[32]=.033460623964007170721706;
			m_weights[33]=.033951543832737795794619;
			m_weights[34]=.034397684192457010718517;
			m_weights[35]=.034798456618012366563241;
			m_weights[36]=.035153332521092950525184;
			m_weights[37]=.035461843847400245871082;
			m_weights[38]=.035723583693978567605265;
			m_weights[39]=.035938206845890998828942;
			m_weights[40]=.036105430231533094943446;
			m_weights[41]=.036225033295983907064429;
			m_weights[42]=.036296858291901958261244;
			m_weights[43]=.036320810487582542156526;
			break;
		case 88:
			m_abscissas[1]=.999041299337602693528652;
			m_abscissas[2]=.996787350122384118035837;
			m_abscissas[3]=.993248242491367032156767;
			m_abscissas[4]=.98842881242479468275709;
			m_abscissas[5]=.982335320384428979369315;
			m_abscissas[6]=.974975634415231832955056;
			m_abscissas[7]=.966359245934661481567739;
			m_abscissas[8]=.956497263315689390378363;
			m_abscissas[9]=.945402399304722256053177;
			m_abscissas[10]=.933088955261363117771717;
			m_abscissas[11]=.919572802983094466871559;
			m_abscissas[12]=.904871364365576675347847;
			m_abscissas[13]=.889003589005353823828197;
			m_abscissas[14]=.871989929805823471489224;
			m_abscissas[15]=.853852316631806778809779;
			m_abscissas[16]=.834614128053171509043017;
			m_abscissas[17]=.814300161217156508383109;
			m_abscissas[18]=.79293659988992011603114;
			m_abscissas[19]=.770550980709429942332747;
			m_abscissas[20]=.747172157693718380109239;
			m_abscissas[21]=.722830265050560570440933;
			m_abscissas[22]=.697556678336699134290293;
			m_abscissas[23]=.671383974016796560640193;
			m_abscissas[24]=.644345887474316497650166;
			m_abscissas[25]=.616477269528504622315385;
			m_abscissas[26]=.587814041513549031182361;
			m_abscissas[27]=.558393148977842925517846;
			m_abscissas[28]=.528252514063044155338805;
			m_abscissas[29]=.497430986624323321100815;
			m_abscissas[30]=.465968294154811607277889;
			m_abscissas[31]=.433904990578798743142773;
			m_abscissas[32]=.401282403979688146280311;
			m_abscissas[33]=.368142583330088317939601;
			m_abscissas[34]=.334528244292705021233381;
			m_abscissas[35]=.300482714161895935666181;
			m_abscissas[36]=.26604987601685674016066;
			m_abscissas[37]=.231274112158423462619544;
			m_abscissas[38]=.196200246902399108794068;
			m_abscissas[39]=.160873488803141836749095;
			m_abscissas[40]=.125339372381886192468825;
			m_abscissas[41]=.089643699434907210670784;
			m_abscissas[42]=.053832479997178682812669;
			m_abscissas[43]=.017951873037620898532463;
			m_weights[0]=.000261233019853709508882;
			m_weights[1]=.001609896277516938935145;
			m_weights[2]=.002897244003549063682379;
			m_weights[3]=.004180180998415566455415;
			m_weights[4]=.005457637182429821404764;
			m_weights[5]=.006728035679401557555528;
			m_weights[6]=.007989754226935509855776;
			m_weights[7]=.009241171124094942737486;
			m_weights[8]=.01048067505871653982654;
			m_weights[9]=.011706669099867393417652;
			m_weights[10]=.012917573348050568824461;
			m_weights[11]=.014111827185786502187995;
			m_weights[12]=.015287891376486787985024;
			m_weights[13]=.016444250087540557295949;
			m_weights[14]=.017579412863016516288687;
			m_weights[15]=.018691916554481420742016;
			m_weights[16]=.019780327212174948261281;
			m_weights[17]=.020843241936272049922065;
			m_weights[18]=.021879290686900743877619;
			m_weights[19]=.022887138051120916713514;
			m_weights[20]=.023865484964873936370221;
			m_weights[21]=.024813070387842810020471;
			m_weights[22]=.025728672929155333765296;
			m_weights[23]=.026611112421888669853398;
			m_weights[24]=.027459251444378576849843;
			m_weights[25]=.028271996786392709564992;
			m_weights[26]=.029048300858291063333756;
			m_weights[27]=.029787163041365582881434;
			m_weights[28]=.030487630977623930431848;
			m_weights[29]=.031148801797358646719336;
			m_weights[30]=.031769823282921951060813;
			m_weights[31]=.032349894967207875506163;
			m_weights[32]=.032888269165427064595979;
			m_weights[33]=.033384251938845199178020;
			m_weights[34]=.033837203989243453096551;
			m_weights[35]=.034246541482948516257363;
			m_weights[36]=.034611736803370377955394;
			m_weights[37]=.034932319231078127855017;
			m_weights[38]=.035207875550537369426939;
			m_weights[39]=.035438050582727324501155;
			m_weights[40]=.035622547642951211180316;
			m_weights[41]=.035761128923249874138549;
			m_weights[42]=.035853615798925809628600;
			m_weights[43]=.035899889058782530350370;
			break;
		case 89:
			m_abscissas[1]=.999062839807244372719503;
			m_abscissas[2]=.996859506691920422083431;
			m_abscissas[3]=.993399800262138491909225;
			m_abscissas[4]=.988688347789927236669212;
			m_abscissas[5]=.982731133019385751094162;
			m_abscissas[6]=.975535675500774341653081;
			m_abscissas[7]=.967111046463640058584448;
			m_abscissas[8]=.957467863073278108300932;
			m_abscissas[9]=.946618276750204517680211;
			m_abscissas[10]=.934575958471262956784827;
			m_abscissas[11]=.921356081796944730270561;
			m_abscissas[12]=.906975303868458835027277;
			m_abscissas[13]=.891451744477256647159206;
			m_abscissas[14]=.874804963264650425500675;
			m_abscissas[15]=.85705593509383446313128;
			m_abscissas[16]=.838227023631687998978773;
			m_abscissas[17]=.818341953176801239777345;
			m_abscissas[18]=.797425778770872102453398;
			m_abscissas[19]=.775504854632036680593867;
			m_abscissas[20]=.752606800950421214127073;
			m_abscissas[21]=.728760469088055420187327;
			m_abscissas[22]=.703995905227178490372687;
			m_abscissas[23]=.678344312512854895042143;
			m_abscissas[24]=.651838011737672930043219;
			m_abscissas[25]=.624510400618110530886086;
			m_abscissas[26]=.596395911713911435955379;
			m_abscissas[27]=.56752996904351436090596;
			m_abscissas[28]=.537948943450214131553026;
			m_abscissas[29]=.507690106775303347746193;
			m_abscissas[30]=.476791584895943341827359;
			m_abscissas[31]=.445292309686941565605517;
			m_abscissas[32]=.413231969966966974397096;
			m_abscissas[33]=.380650961491013565485537;
			m_abscissas[34]=.347590336052123226870233;
			m_abscissas[35]=.314091749756500858349674;
			m_abscissas[36]=.280197410537195867463016;
			m_abscissas[37]=.24595002497248326487803;
			m_abscissas[38]=.211392744475953450731962;
			m_abscissas[39]=.176569110926111271908193;
			m_abscissas[40]=.14152300180399102911241;
			m_abscissas[41]=.106298574907913922363938;
			m_abscissas[42]=.07094021271504715615439;
			m_abscissas[43]=.035492466459868904782124;
			m_abscissas[44]=0.;
			m_weights[0]=.000255362614913176710930;
			m_weights[1]=.001573730137083271206940;
			m_weights[2]=.002832205614263301929644;
			m_weights[3]=.004086451420539364396458;
			m_weights[4]=.005335457664217781768632;
			m_weights[5]=.006577718744988413650341;
			m_weights[6]=.007811684303221055094143;
			m_weights[7]=.009035804044022397576382;
			m_weights[8]=.010248537253607570686458;
			m_weights[9]=.011448356613041191542240;
			m_weights[10]=.012633750701162080271448;
			m_weights[11]=.013803226108303797324693;
			m_weights[12]=.014955309403322427790156;
			m_weights[13]=.016088549028309278776120;
			m_weights[14]=.017201517145981301349689;
			m_weights[15]=.018292811448216319466314;
			m_weights[16]=.019361056928076479326152;
			m_weights[17]=.020404907615207015306080;
			m_weights[18]=.021423048273454123106140;
			m_weights[19]=.022414196059089396134089;
			m_weights[20]=.023377102137832344776550;
			m_weights[21]=.024310553258789211594450;
			m_weights[22]=.025213373283414132456074;
			m_weights[23]=.026084424667618836854777;
			m_weights[24]=.026922609895195389376607;
			m_weights[25]=.027726872860765724317444;
			m_weights[26]=.028496200200528079063709;
			m_weights[27]=.029229622569131768888354;
			m_weights[28]=.029926215861076845191616;
			m_weights[29]=.030585102375103303923106;
			m_weights[30]=.031205451920105191172468;
			m_weights[31]=.031786482861177869696465;
			m_weights[32]=.032327463104481631104753;
			m_weights[33]=.032827711019681581746606;
			m_weights[34]=.033286596298802144012147;
			m_weights[35]=.033703540750414464414686;
			m_weights[36]=.034078019028156381507868;
			m_weights[37]=.034409559292667262537259;
			m_weights[38]=.034697743806103853285456;
			m_weights[39]=.034942209458488188216442;
			m_weights[40]=.035142648225224465973315;
			m_weights[41]=.035298807555208497005666;
			m_weights[42]=.035410490689040763993252;
			m_weights[43]=.035477556906942189933482;
			m_weights[44]=.035499921706060271088801;
			break;
		case 90:
			m_abscissas[1]=.999083662408302103439943;
			m_abscissas[2]=.996929259694517965048551;
			m_abscissas[3]=.993546313407523700814027;
			m_abscissas[4]=.988939253432408088452721;
			m_abscissas[5]=.983113801946445082433017;
			m_abscissas[6]=.976077149088712930117359;
			m_abscissas[7]=.967837968885887420756275;
			m_abscissas[8]=.958406414125511082663181;
			m_abscissas[9]=.947794105507042111942316;
			m_abscissas[10]=.936014117923373020863843;
			m_abscissas[11]=.923080964599626307114438;
			m_abscissas[12]=.909010579325975634239018;
			m_abscissas[13]=.893820296883389128381933;
			m_abscissas[14]=.877528831716980722260985;
			m_abscissas[15]=.860156254896521787494046;
			m_abscissas[16]=.841723969398704128028238;
			m_abscissas[17]=.822254683744690742139893;
			m_abscissas[18]=.801772384027047073270117;
			m_abscissas[19]=.780302304361398875378568;
			m_abscissas[20]=.757870895799722668554768;
			m_abscissas[21]=.734505793743863237047672;
			m_abscissas[22]=.710235783899603951239372;
			m_abscissas[23]=.685090766813346097319507;
			m_abscissas[24]=.659101721035159180886942;
			m_abscissas[25]=.632300664953631669700715;
			m_abscissas[26]=.604720617349572402779379;
			m_abscissas[27]=.576395556717181277295247;
			m_abscissas[28]=.547360379402819769748652;
			m_abscissas[29]=.517650856612964217273451;
			m_abscissas[30]=.487303590344315010752869;
			m_abscissas[31]=.456355968290360735595219;
			m_abscissas[32]=.424846117779955896899692;
			m_abscissas[33]=.392812858804662432552195;
			m_abscissas[34]=.360295656192727152046519;
			m_abscissas[35]=.327334570988618071490801;
			m_abscissas[36]=.293970211098020993162412;
			m_abscissas[37]=.260243681259102355776081;
			m_abscissas[38]=.226196532401674217150872;
			m_abscissas[39]=.191870710456651182564219;
			m_abscissas[40]=.157308504678866217309583;
			m_abscissas[41]=.122552495546911736956264;
			m_abscissas[42]=.087645502304193408115092;
			m_abscissas[43]=.052630530205826069271167;
			m_abscissas[44]=.017550717536363547392349;
			m_weights[0]=.000249687890137328339576;
			m_weights[1]=.001538769040694787901293;
			m_weights[2]=.002769332190918619777063;
			m_weights[3]=.003995837118109549139632;
			m_weights[4]=.005217330540475201378895;
			m_weights[5]=.006432374205385298898976;
			m_weights[6]=.007639485732308681620053;
			m_weights[7]=.008837182356578174567774;
			m_weights[8]=.010023990150963538754322;
			m_weights[9]=.011198447672339711750176;
			m_weights[10]=.012359108327817702627666;
			m_weights[11]=.013504542361621505248256;
			m_weights[12]=.014633338699983519210650;
			m_weights[13]=.015744106726931328455625;
			m_weights[14]=.016835478015548680809203;
			m_weights[15]=.017906108023132139124575;
			m_weights[16]=.018954677752674151880347;
			m_weights[17]=.019979895380698184258728;
			m_weights[18]=.020980497850447824835525;
			m_weights[19]=.021955252428981760459287;
			m_weights[20]=.022902958226530883646062;
			m_weights[21]=.023822447676397862436795;
			m_weights[22]=.024712587973663149079983;
			m_weights[23]=.025572282470976441630523;
			m_weights[24]=.026400472029745222368272;
			m_weights[25]=.027196136325075122801357;
			m_weights[26]=.027958295102866762356335;
			m_weights[27]=.028686009387528311781722;
			m_weights[28]=.029378382638821157272756;
			m_weights[29]=.030034561856416975399868;
			m_weights[30]=.030653738630807821652585;
			m_weights[31]=.031235150139276184003635;
			m_weights[32]=.031778080085699139507019;
			m_weights[33]=.032281859583029608344288;
			m_weights[34]=.032745867977368087015628;
			m_weights[35]=.033169533612610039049271;
			m_weights[36]=.033552334534727216531243;
			m_weights[37]=.033893799134815473871430;
			m_weights[38]=.034193506730117014933081;
			m_weights[39]=.034451088082301386260196;
			m_weights[40]=.034666225852366793855051;
			m_weights[41]=.034838654991601380330371;
			m_weights[42]=.034968163068122854817378;
			m_weights[43]=.035054590528594221082459;
			m_weights[44]=.035097830894793200935069;
			break;
		case 91:
			m_abscissas[1]=.999103798687492262625765;
			m_abscissas[2]=.99699671469222115516302;
			m_abscissas[3]=.99368800327501366062002;
			m_abscissas[4]=.989181907522294422074361;
			m_abscissas[5]=.983483902210260281695656;
			m_abscissas[6]=.976600865910292186241353;
			m_abscissas[7]=.968541096939741031862795;
			m_abscissas[8]=.959314308800405312043253;
			m_abscissas[9]=.948931620097228569369258;
			m_abscissas[10]=.93740554172623778096181;
			m_abscissas[11]=.924749962041447334067011;
			m_abscissas[12]=.910980130231028268258223;
			m_abscissas[13]=.896112637998085166377653;
			m_abscissas[14]=.88016539959801106663529;
			m_abscissas[15]=.86315763026945911296922;
			m_abscissas[16]=.845109823090989618760541;
			m_abscissas[17]=.826043724294294704777104;
			m_abscissas[18]=.805982307065324820235562;
			m_abscissas[19]=.784949743865748667789941;
			m_abscissas[20]=.762971377308588462740652;
			m_abscissas[21]=.740073689623412343980741;
			m_abscissas[22]=.716284270748051415465624;
			m_abscissas[23]=.691631785085397693066621;
			m_abscissas[24]=.666145936965408163066822;
			m_abscissas[25]=.63985743485397602783592;
			m_abscissas[26]=.612797954351824912885333;
			m_abscissas[27]=.58500010002802997312032;
			m_abscissas[28]=.556497366134167637571245;
			m_abscissas[29]=.527324096246440241371483;
			m_abscissas[30]=.497515441884410629896442;
			m_abscissas[31]=.467107320156212955215169;
			m_abscissas[32]=.436136370481277534584848;
			m_abscissas[33]=.404639910442718198340811;
			m_abscissas[34]=.372655890822578554475774;
			m_abscissas[35]=.340222849874117691807501;
			m_abscissas[36]=.30737986688623478989058;
			m_abscissas[37]=.274166515095984755375312;
			m_abscissas[38]=.240622814005922306960271;
			m_abscissas[39]=.206789181163728919330444;
			m_abscissas[40]=.172706383462224833301696;
			m_abscissas[41]=.138415488018446154799268;
			m_abscissas[42]=.103957812690974196286221;
			m_abscissas[43]=.069374876295140044900431;
			m_abscissas[44]=.034708348576091343205069;
			m_abscissas[45]=0.;
			m_weights[0]=.000244200244200244200244;
			m_weights[1]=.001504960046072077155211;
			m_weights[2]=.002708528731167958856943;
			m_weights[3]=.003908201644267197620730;
			m_weights[4]=.005103078777115176331514;
			m_weights[5]=.006291785552222116799433;
			m_weights[6]=.007472903898052906307034;
			m_weights[7]=.008645014953475471532275;
			m_weights[8]=.009806708009542905601132;
			m_weights[9]=.010956583999801821540899;
			m_weights[10]=.012093257739247319372312;
			m_weights[11]=.013215359793536616249411;
			m_weights[12]=.014321538210655089456164;
			m_weights[13]=.015410460186440101208123;
			m_weights[14]=.016480813688136995403825;
			m_weights[15]=.017531309044352929403047;
			m_weights[16]=.018560680503912370380742;
			m_weights[17]=.019567687763762851125748;
			m_weights[18]=.020551117465075161605269;
			m_weights[19]=.021509784656238629744444;
			m_weights[20]=.022442534221257232047793;
			m_weights[21]=.023348242271974237597152;
			m_weights[22]=.024225817502533141537811;
			m_weights[23]=.025074202504493222560855;
			m_weights[24]=.025892375041045653930083;
			m_weights[25]=.026679349278813814914803;
			m_weights[26]=.027434176975765633633574;
			m_weights[27]=.028155948623814430855695;
			m_weights[28]=.028843794544736675075661;
			m_weights[29]=.029496885938089613378587;
			m_weights[30]=.030114435879868477446417;
			m_weights[31]=.030695700270701591311138;
			m_weights[32]=.031239978732442022201202;
			m_weights[33]=.031746615452076261816953;
			m_weights[34]=.032214999971933675210469;
			m_weights[35]=.032644567925244997924139;
			m_weights[36]=.033034801716163899205445;
			m_weights[37]=.033385231143432466075663;
			m_weights[38]=.033695433966939309414086;
			m_weights[39]=.033965036416487760598157;
			m_weights[40]=.034193713642161228166057;
			m_weights[41]=.034381190105743131388706;
			m_weights[42]=.034527239912719834545929;
			m_weights[43]=.034631687084466584939655;
			m_weights[44]=.034694405770288521841303;
			m_weights[45]=.034715320399061284976331;
			break;
		case 92:
			m_abscissas[1]=.999123278477545503658115;
			m_abscissas[2]=.997061971515730498739737;
			m_abscissas[3]=.993825079207622798835164;
			m_abscissas[4]=.98941666775115997278895;
			m_abscissas[5]=.983841977776044161690288;
			m_abscissas[6]=.977107592982370172641267;
			m_abscissas[7]=.969221456091847163715917;
			m_abscissas[8]=.960192864804434424479451;
			m_abscissas[9]=.95003246242622781260286;
			m_abscissas[10]=.938752225896106261307675;
			m_abscissas[11]=.926365451904429805161268;
			m_abscissas[12]=.912886741327953975553979;
			m_abscissas[13]=.898331982072979443837671;
			m_abscissas[14]=.882718330376205989939954;
			m_abscissas[15]=.866064190598034306339979;
			m_abscissas[16]=.848389193538071213415969;
			m_abscissas[17]=.82971417330134889032505;
			m_abscissas[18]=.810061142744070836928918;
			m_abscissas[19]=.78945326752867248396935;
			m_abscissas[20]=.767914838819259270197589;
			m_abscissas[21]=.745471244649889832621645;
			m_abscissas[22]=.722148939999624989873761;
			m_abscissas[23]=.697975415609722885898401;
			m_abscissas[24]=.672979165579804656119851;
			m_abscissas[25]=.647189653781230518707682;
			m_abscissas[26]=.6206372791273054897515;
			m_abscissas[27]=.593353339741271861680644;
			m_abscissas[28]=.565369996064338555385698;
			m_abscissas[29]=.536720232947242715803284;
			m_abscissas[30]=.50743782077003425571631;
			m_abscissas[31]=.477557275635917589984797;
			m_abscissas[32]=.447113818686074899706789;
			m_abscissas[33]=.416143334583430447692588;
			m_abscissas[34]=.384682329214294388123213;
			m_abscissas[35]=.352767886657745937433072;
			m_abscissas[36]=.320437625473478558192526;
			m_abscissas[37]=.287729654359632897828397;
			m_abscissas[38]=.254682527232885646353213;
			m_abscissas[39]=.221335197783743338300156;
			m_abscissas[40]=.187726973560608608206312;
			m_abscissas[41]=.153897469636741777958498;
			m_abscissas[42]=.119886561914732246839516;
			m_abscissas[43]=.085734340123521386943267;
			m_abscissas[44]=.051481060563381010711802;
			m_abscissas[45]=.017167098654548544059095;
			m_weights[0]=.000238891543239369326326;
			m_weights[1]=.001472253086521608712556;
			m_weights[2]=.002649705385512103162789;
			m_weights[3]=.003823415936445150325708;
			m_weights[4]=.004992534891449485099585;
			m_weights[5]=.006155747912144261178862;
			m_weights[6]=.007311697800571574425377;
			m_weights[7]=.008459026203128649710401;
			m_weights[8]=.009596382286527388075381;
			m_weights[9]=.010722426081610846472777;
			m_weights[10]=.011835830603961948012640;
			m_weights[11]=.012935283613875230550694;
			m_weights[12]=.014019489242961033424609;
			m_weights[13]=.015087169557356852262572;
			m_weights[14]=.016137066081314347150797;
			m_weights[15]=.017167941289463172149015;
			m_weights[16]=.018178580070315932536087;
			m_weights[17]=.019167791161271894065389;
			m_weights[18]=.020134408554391654851678;
			m_weights[19]=.021077292871778102313049;
			m_weights[20]=.021995332709205122958321;
			m_weights[21]=.022887445946555846141932;
			m_weights[22]=.023752581023609181183374;
			m_weights[23]=.024589718179720091113263;
			m_weights[24]=.025397870655962225765823;
			m_weights[25]=.026176085858334473213256;
			m_weights[26]=.026923446480672111590722;
			m_weights[27]=.027639071585946590651800;
			m_weights[28]=.028322117644684423145699;
			m_weights[29]=.028971779529284553040278;
			m_weights[30]=.029587291463064474056486;
			m_weights[31]=.030167927922918022889530;
			m_weights[32]=.030713004494521973447426;
			m_weights[33]=.031221878679084165819404;
			m_weights[34]=.031693950650682799227064;
			m_weights[35]=.032128663963304601592195;
			m_weights[36]=.032525506206748769414003;
			m_weights[37]=.032884009610623766270915;
			m_weights[38]=.033203751595725195872321;
			m_weights[39]=.033484355272144947600140;
			m_weights[40]=.033725489883524571354074;
			m_weights[41]=.033926871196929297156094;
			m_weights[42]=.034088261837883196388554;
			m_weights[43]=.034209471570170608643164;
			m_weights[44]=.034290357520074053643218;
			m_weights[45]=.034330824344784334014654;
			break;
		case 93:
			m_abscissas[1]=.999142130007748739404495;
			m_abscissas[2]=.997125124633743024944165;
			m_abscissas[3]=.993957739316500746693073;
			m_abscissas[4]=.989643872647796604121797;
			m_abscissas[5]=.984188543521907037018403;
			m_abscissas[6]=.977598056402062734380402;
			m_abscissas[7]=.969880017250340265855327;
			m_abscissas[8]=.961043329959517936695617;
			m_abscissas[9]=.951098187636642115777117;
			m_abscissas[10]=.940056061406358584920923;
			m_abscissas[11]=.927929687411480081826089;
			m_abscissas[12]=.914733052229110938703273;
			m_abscissas[13]=.900481376791234361609169;
			m_abscissas[14]=.885191098856923567035378;
			m_abscissas[15]=.86887985406882142798911;
			m_abscissas[16]=.851566455621547033912238;
			m_abscissas[17]=.833270872568366056145263;
			m_abscissas[18]=.814014206792656542652162;
			m_abscissas[19]=.793818668671557704823179;
			m_abscissas[20]=.772707551460341200063613;
			m_abscissas[21]=.75070520442732638479798;
			m_abscissas[22]=.727837004770493197985598;
			m_abscissas[23]=.704129328348288218030455;
			m_abscissas[24]=.679609519258449288518272;
			m_abscissas[25]=.654305858299979575337083;
			m_abscissas[26]=.628247530354675504734423;
			m_abscissas[27]=.601464590725849859543783;
			m_abscissas[28]=.573987930473087960010659;
			m_abscissas[29]=.545849240783028754173751;
			m_abscissas[30]=.51708097641727173116508;
			m_abscissas[31]=.487716318279573115694263;
			m_abscissas[32]=.457789135145509289509977;
			m_abscissas[33]=.427333944598750459523775;
			m_abscissas[34]=.39638587321900201734622;
			m_abscissas[35]=.364980616067533675300578;
			m_abscissas[36]=.333154395517026269797567;
			m_abscissas[37]=.300943919473222122835998;
			m_abscissas[38]=.268386339036566148519324;
			m_abscissas[39]=.235519205652670656916804;
			m_abscissas[40]=.202380427801026284786539;
			m_abscissas[41]=.169008227271913982645827;
			m_abscissas[42]=.135441095081947889731601;
			m_abscissas[43]=.101717747079095680092609;
			m_abscissas[44]=.067877079288381080311012;
			m_abscissas[45]=.033958123049772326645556;
			m_abscissas[46]=0.;
			m_weights[0]=.000233754090696587190276;
			m_weights[1]=.001440600785540810001860;
			m_weights[2]=.002592777125621935757523;
			m_weights[3]=.003741357842277387674976;
			m_weights[4]=.004885540341145310109954;
			m_weights[5]=.006024067305896534215952;
			m_weights[6]=.007155639194123707262291;
			m_weights[7]=.008278954977531928860142;
			m_weights[8]=.009392720565092307514600;
			m_weights[9]=.010495652009466264259517;
			m_weights[10]=.011586477517382044767539;
			m_weights[11]=.012663939108097027675012;
			m_weights[12]=.013726794142439656168846;
			m_weights[13]=.014773816791011345095293;
			m_weights[14]=.015803799464914359606419;
			m_weights[15]=.016815554217234971761526;
			m_weights[16]=.017807914117895531013133;
			m_weights[17]=.018779734602229657921485;
			m_weights[18]=.019729894792668035766893;
			m_weights[19]=.020657298792492111784704;
			m_weights[20]=.021560876950420510658404;
			m_weights[21]=.022439587094712017672890;
			m_weights[22]=.023292415735443345576691;
			m_weights[23]=.024118379233623182741607;
			m_weights[24]=.024916524935823309512315;
			m_weights[25]=.025685932273036275826183;
			m_weights[26]=.026425713822503768152190;
			m_weights[27]=.027135016331298442558754;
			m_weights[28]=.027813021700483572121607;
			m_weights[29]=.028458947928718703688309;
			m_weights[30]=.029072050014225239368177;
			m_weights[31]=.029651620814073188898065;
			m_weights[32]=.030196991859799097136635;
			m_weights[33]=.030707534128415200165760;
			m_weights[34]=.031182658767921095500768;
			m_weights[35]=.031621817776481535108735;
			m_weights[36]=.032024504634487282730428;
			m_weights[37]=.032390254888770243840324;
			m_weights[38]=.032718646688299205457330;
			m_weights[39]=.03300930127073844396458;
			m_weights[40]=.033261883399308103226392;
			m_weights[41]=.033476101749442544226026;
			m_weights[42]=.033651709244800752962682;
			m_weights[43]=.033788503342241297155825;
			m_weights[44]=.033886326265433175953685;
			m_weights[45]=.03394506518683314165096;
			m_weights[46]=.033964652357819619473476;
			break;
		case 94:
			m_abscissas[1]=.999160380006258435521938;
			m_abscissas[2]=.997186263494824104050373;
			m_abscissas[3]=.99408617119616572372413;
			m_abscissas[4]=.98986384279639733514163;
			m_abscissas[5]=.984524087083897864179258;
			m_abscissas[6]=.978072943935417123437997;
			m_abscissas[7]=.970517700205370609329456;
			m_abscissas[8]=.961866886593698583390219;
			m_abscissas[9]=.952130269536089654363742;
			m_abscissas[10]=.941318840727131198424255;
			m_abscissas[11]=.929444804938616761765658;
			m_abscissas[12]=.916521566345757369290254;
			m_abscissas[13]=.902563713447280748282871;
			m_abscissas[14]=.887587002624446712363144;
			m_abscissas[15]=.871608340369701325723458;
			m_abscissas[16]=.854645764210716246032661;
			m_abscissas[17]=.836718422354172199446276;
			m_abscissas[18]=.817846552073744242015473;
			m_abscissas[19]=.798051456867494605591978;
			m_abscissas[20]=.777355482410919216626267;
			m_abscissas[21]=.755781991333064107378239;
			m_abscissas[22]=.733355336844349924693751;
			m_abscissas[23]=.710100835245977053538216;
			m_abscissas[24]=.686044737352009368615698;
			m_abscissas[25]=.661214198856439334040594;
			m_abscissas[26]=.635637249678713867231015;
			m_abscissas[27]=.609342762322344403303004;
			m_abscissas[28]=.582360419282332698832586;
			m_abscissas[29]=.554720679538213681553641;
			m_abscissas[30]=.52645474417054616481128;
			m_abscissas[31]=.497594521139669895774484;
			m_abscissas[32]=.46817258926649179745688;
			m_abscissas[33]=.438222161455964148250102;
			m_abscissas[34]=.407777047204771683636663;
			m_abscissas[35]=.376871614435552159214134;
			m_abscissas[36]=.34554075070073481690679;
			m_abscissas[37]=.313819823799792552828955;
			m_abscissas[38]=.281744641854365567374253;
			m_abscissas[39]=.249351412886326120777144;
			m_abscissas[40]=.216676703944415017470476;
			m_abscissas[41]=.183757399825589957872356;
			m_abscissas[42]=.150630661437683345228867;
			m_abscissas[43]=.117333883850371996810558;
			m_abscissas[44]=.083904654081813022730050;
			m_abscissas[45]=.050380708668598502330281;
			m_abscissas[46]=.016799891066926168905565;
			m_weights[0]=.000228780599405170441547;
			m_weights[1]=.001409958285218535289252;
			m_weights[2]=.002537663437296154243695;
			m_weights[3]=.003661911680619358983019;
			m_weights[4]=.004781944958252564190251;
			m_weights[5]=.005896559961769222107909;
			m_weights[6]=.007004511787824695038547;
			m_weights[7]=.008104553749491364342839;
			m_weights[8]=.009195445558750452745734;
			m_weights[9]=.010275956403924390508744;
			m_weights[10]=.011344866857284808693799;
			m_weights[11]=.012400970439111776890440;
			m_weights[12]=.013443075056065899809263;
			m_weights[13]=.014470004381094858490380;
			m_weights[14]=.015480599197842042480967;
			m_weights[15]=.016473718717708508516014;
			m_weights[16]=.017448241872221374207072;
			m_weights[17]=.018403068581148312557369;
			m_weights[18]=.019337120995849413702316;
			m_weights[19]=.020249344716934295748911;
			m_weights[20]=.021138709985101422809824;
			m_weights[21]=.022004212843954721816622;
			m_weights[22]=.022844876273564719394112;
			m_weights[23]=.023659751293541736169063;
			m_weights[24]=.024447918034404536326443;
			m_weights[25]=.025208486776052790655190;
			m_weights[26]=.025940598952182362352808;
			m_weights[27]=.026643428119516896897547;
			m_weights[28]=.027316180890766427225594;
			m_weights[29]=.027958097830263064765724;
			m_weights[30]=.028568454311264947382739;
			m_weights[31]=.029146561333962204685322;
			m_weights[32]=.029691766303262605619797;
			m_weights[33]=.030203453765479644990184;
			m_weights[34]=.030681046103092004099062;
			m_weights[35]=.031124004186790502130111;
			m_weights[36]=.031531827984076765160206;
			m_weights[37]=.031904057123726810746470;
			m_weights[38]=.032240271415482513391277;
			m_weights[39]=.032540091324384417417684;
			m_weights[40]=.032803178399210537591385;
			m_weights[41]=.033029235654537573457598;
			m_weights[42]=.033218007905993300273829;
			m_weights[43]=.033369282058321727047722;
			m_weights[44]=.033482887345935869833066;
			m_weights[45]=.033558695525686615200498;
			m_weights[46]=.033596621021630083572051;
			break;
		case 95:
			m_abscissas[1]=.999178053794877861741512;
			m_abscissas[2]=.997245472844120499647899;
			m_abscissas[3]=.994210552587176812550878;
			m_abscissas[4]=.990076881965390151868922;
			m_abscissas[5]=.984849070566053041889784;
			m_abscissas[6]=.978532907417260818791358;
			m_abscissas[7]=.971135376819581779403558;
			m_abscissas[8]=.962664655614073321041556;
			m_abscissas[9]=.953130105634210306616888;
			m_abscissas[10]=.942542263896669875257576;
			m_abscissas[11]=.930912831178509915630839;
			m_abscissas[12]=.91825465918816517322275;
			m_abscissas[13]=.904581736413498803451469;
			m_abscissas[14]=.889909172689963003815287;
			m_abscissas[15]=.874253182517828762382495;
			m_abscissas[16]=.857631067152485813342735;
			m_abscissas[17]=.840061195490369959604257;
			m_abscissas[18]=.821562983773087773909714;
			m_abscissas[19]=.802156874132959362743306;
			m_abscissas[20]=.781864312004138720660467;
			m_abscissas[21]=.760707722424539448264622;
			m_abscissas[22]=.738710485254915027462752;
			m_abscissas[23]=.71589690934257886055923;
			m_abscissas[24]=.692292205658379298075557;
			m_abscissas[25]=.667922459436657288766854;
			m_abscissas[26]=.642814601349002117751755;
			m_abscissas[27]=.616996377743679486081409;
			m_abscissas[28]=.590496319983632848655371;
			m_abscissas[29]=.563343712916951190806983;
			m_abscissas[30]=.53556856251465249273651;
			m_abscissas[31]=.50720156271155051369207;
			m_abscissas[32]=.478274061486851933312249;
			m_abscissas[33]=.448818026221970165719148;
			m_abscissas[34]=.418866008373840268375003;
			m_abscissas[35]=.388451107502775342603991;
			m_abscissas[36]=.357606934694617778136876;
			m_abscissas[37]=.326367575417607806337869;
			m_abscissas[38]=.294767551855016331720788;
			m_abscissas[39]=.262841784755168201513549;
			m_abscissas[40]=.230625554841015296112354;
			m_abscissas[41]=.198154463821905481316156;
			m_abscissas[42]=.165464395050633012193788;
			m_abscissas[43]=.13259147386924792804405;
			m_abscissas[44]=.099572027687445891730469;
			m_abscissas[45]=.066442545837655422247464;
			m_abscissas[46]=.033239639251186218399700;
			m_abscissas[47]=0.;
			m_weights[0]=.000223964165733482642777;
			m_weights[1]=.001380283087270199758311;
			m_weights[2]=.002484288035985838404963;
			m_weights[3]=.003584967834740586021025;
			m_weights[4]=.004681606424546652003199;
			m_weights[5]=.005773051678903967054569;
			m_weights[6]=.006858110504033792954914;
			m_weights[7]=.007935587754481437441035;
			m_weights[8]=.009004294186289966957998;
			m_weights[9]=.010063049413203391560121;
			m_weights[10]=.011110683718378111758090;
			m_weights[11]=.012146039530575962871432;
			m_weights[12]=.013167972778206395743604;
			m_weights[13]=.014175354187115519764911;
			m_weights[14]=.015167070544696044777705;
			m_weights[15]=.016142025938386090901137;
			m_weights[16]=.017099142971240733528568;
			m_weights[17]=.018037363955091447998534;
			m_weights[18]=.018955652080878197438016;
			m_weights[19]=.019852992565322322879771;
			m_weights[20]=.020728393772919255169409;
			m_weights[21]=.021580888312147571171597;
			m_weights[22]=.022409534104761163845782;
			m_weights[23]=.023213415427029023931453;
			m_weights[24]=.023991643921799967997472;
			m_weights[25]=.024743359580291288559375;
			m_weights[26]=.025467731692527413162063;
			m_weights[27]=.026163959765385414600260;
			m_weights[28]=.026831274407237577264813;
			m_weights[29]=.027468938178216584019625;
			m_weights[30]=.028076246405165868459332;
			m_weights[31]=.028652527960376036481605;
			m_weights[32]=.029197146003247833336881;
			m_weights[33]=.029709498684062797373018;
			m_weights[34]=.030189019809084406168102;
			m_weights[35]=.030635179466255107439861;
			m_weights[36]=.031047484610798068643269;
			m_weights[37]=.031425479610076714063487;
			m_weights[38]=.031768746747110088583035;
			m_weights[39]=.032076906682187737339058;
			m_weights[40]=.032349618872074065544482;
			m_weights[41]=.032586581946338988774751;
			m_weights[42]=.032787534040399047206885;
			m_weights[43]=.032952253084900983849491;
			m_weights[44]=.033080557051128022828697;
			m_weights[45]=.033172304152157675235594;
			m_weights[46]=.033227392999548792641218;
			m_weights[47]=.033245762715384727697410;
			break;
		case 96:
			m_abscissas[1]=.999195175376926043326695;
			m_abscissas[2]=.997302833017008286457881;
			m_abscissas[3]=.994331051990612280799635;
			m_abscissas[4]=.990283278154330438461231;
			m_abscissas[5]=.985163932126614011137175;
			m_abscissas[6]=.978978564977784333326435;
			m_abscissas[7]=.9717338739890572962493;
			m_abscissas[8]=.963437700287926497097732;
			m_abscissas[9]=.954099021820540160097517;
			m_abscissas[10]=.943727944162188669320199;
			m_abscissas[11]=.932335689799266255828747;
			m_abscissas[12]=.919934586085041681937503;
			m_abscissas[13]=.906538051949890111529772;
			m_abscissas[14]=.89216058340720878923969;
			m_abscissas[15]=.876817737882357948402463;
			m_abscissas[16]=.860526117387032981787528;
			m_abscissas[17]=.843303350559981906211962;
			m_abscissas[18]=.825168073594918295829925;
			m_abscissas[19]=.806139910077042363585527;
			m_abscissas[20]=.786239449750429480180639;
			m_abscissas[21]=.765488226239520781954641;
			m_abscissas[22]=.743908693748980098704968;
			m_abscissas[23]=.72152420276722755296044;
			m_abscissas[24]=.69835897480000278398098;
			m_abscissas[25]=.674438076161338555046375;
			m_abscissas[26]=.649787390850331782201685;
			m_abscissas[27]=.624433592543079489309324;
			m_abscissas[28]=.598404115730098918316191;
			m_abscissas[29]=.571727126030471857380479;
			m_abscissas[30]=.544431489714841513284291;
			m_abscissas[31]=.516546742470244561310207;
			m_abscissas[32]=.48810305744058016067751;
			m_abscissas[33]=.459131212577300665766858;
			m_abscissas[34]=.429662557335654531033583;
			m_abscissas[35]=.399728978752519617573962;
			m_abscissas[36]=.36936286694253394459924;
			m_abscissas[37]=.33859708004986013077193;
			m_abscissas[38]=.307464908693508633376016;
			m_abscissas[39]=.276000039944692763489444;
			m_abscissas[40]=.244236520875194727906564;
			m_abscissas[41]=.212208721716186067171607;
			m_abscissas[42]=.179951298667367315395221;
			m_abscissas[43]=.147499156396670041050966;
			m_abscissas[44]=.1148874102710992260705;
			m_abscissas[45]=.082151348359584838633188;
			m_abscissas[46]=.049326393248958136572065;
			m_abscissas[47]=.016448063714370435100636;
			m_weights[0]=.000219298245614035087719;
			m_weights[1]=.001351534905655567238901;
			m_weights[2]=.002432578603010584801390;
			m_weights[3]=.003510422375024517778209;
			m_weights[4]=.004584389784779426443261;
			m_weights[5]=.005653377236371456058608;
			m_weights[6]=.006716240789720192340943;
			m_weights[7]=.007771834211914708955037;
			m_weights[8]=.008819016711677060969120;
			m_weights[9]=.009856655781517411407194;
			m_weights[10]=.010883628919882028225457;
			m_weights[11]=.011898825025498477896266;
			m_weights[12]=.012901145672924584821447;
			m_weights[13]=.013889506333909183652582;
			m_weights[14]=.014862837566750324613446;
			m_weights[15]=.015820086181631326706956;
			m_weights[16]=.016760216384641604498534;
			m_weights[17]=.017682210901064016654511;
			m_weights[18]=.018585072077597675816903;
			m_weights[19]=.019467822962775370458610;
			m_weights[20]=.020329508364647557709412;
			m_weights[21]=.021169195884722009130520;
			m_weights[22]=.021985976927116856359690;
			m_weights[23]=.022778967681880256853320;
			m_weights[24]=.023547310081440070281895;
			m_weights[25]=.024290172729165633233685;
			m_weights[26]=.025006751799047682936819;
			m_weights[27]=.025696271905529925155224;
			m_weights[28]=.026357986942555655974069;
			m_weights[29]=.026991180890924637439966;
			m_weights[30]=.027595168593088729109472;
			m_weights[31]=.028169296494549358149209;
			m_weights[32]=.028712943351055611780237;
			m_weights[33]=.029225520900838443954447;
			m_weights[34]=.029706474501154116021562;
			m_weights[35]=.030155283728448467571411;
			m_weights[36]=.030571462941492877080998;
			m_weights[37]=.030954561806882766808066;
			m_weights[38]=.031304165786330180452350;
			m_weights[39]=.031619896585223265455023;
			m_weights[40]=.031901412561967375625186;
			m_weights[41]=.032148409097664925898010;
			m_weights[42]=.032360618925734031545791;
			m_weights[43]=.032537812421109301242233;
			m_weights[44]=.032679797848711879112428;
			m_weights[45]=.032786421570919897236022;
			m_weights[46]=.032857568213814858785464;
			m_weights[47]=.032893160792024074672394;
			break;
		case 97:
			m_abscissas[1]=.999211767518767937292512;
			m_abscissas[2]=.997358420211575308380833;
			m_abscissas[3]=.994447829238317218533802;
			m_abscissas[4]=.990483304565576382777909;
			m_abscissas[5]=.985469087450548158033555;
			m_abscissas[6]=.979410503109991065929354;
			m_abscissas[7]=.97231397639338394986252;
			m_abscissas[8]=.964187029755659609253356;
			m_abscissas[9]=.955038276712134050045025;
			m_abscissas[10]=.944877413224633009627481;
			m_abscissas[11]=.93371520763849810980598;
			m_abscissas[12]=.921563489367936527387895;
			m_abscissas[13]=.908435136407928054815107;
			m_abscissas[14]=.894344061712211572323574;
			m_abscissas[15]=.879305198463203883178571;
			m_abscissas[16]=.86333448425479747382835;
			m_abscissas[17]=.8464488442074511804343;
			m_abscissas[18]=.828666173034855392142298;
			m_abscissas[19]=.810005316081936147327894;
			m_abscissas[20]=.790486049354725181792563;
			m_abscissas[21]=.770129058563513614050128;
			m_abscissas[22]=.748955917201652598454668;
			m_abscissas[23]=.726989063683328133299878;
			m_abscissas[24]=.704251777564599720681263;
			m_abscissas[25]=.680768154872942778294622;
			m_abscissas[26]=.656563082571466092697812;
			m_abscissas[27]=.631662212184884059708887;
			m_abscissas[28]=.606091932615206187760061;
			m_abscissas[29]=.579879342175961313997477;
			m_abscissas[30]=.553052219874599625586179;
			m_abscissas[31]=.525638995973510597989378;
			m_abscissas[32]=.497668721860858250637271;
			m_abscissas[33]=.469171039263165698537078;
			m_abscissas[34]=.440176148832277959662886;
			m_abscissas[35]=.410714778139994549176929;
			m_abscissas[36]=.380818149114290800198177;
			m_abscissas[37]=.350517944951638397683906;
			m_abscissas[38]=.319846276540490646377714;
			m_abscissas[39]=.288835648431515910902827;
			m_abscissas[40]=.257518924390642905356213;
			m_abscissas[41]=.22592929257142355389587;
			m_abscissas[42]=.194100230343622518343615;
			m_abscissas[43]=.162065468815306761916132;
			m_abscissas[44]=.129858957086033300630777;
			m_abscissas[45]=.097514826269018239190312;
			m_abscissas[46]=.065067353320414992547590;
			m_abscissas[47]=.032550924714033997196955;
			m_abscissas[48]=0.;
			m_weights[0]=.000214776632302405498282;
			m_weights[1]=.001323675529824700568775;
			m_weights[2]=.002382466540764382539846;
			m_weights[3]=.003438176708756261609713;
			m_weights[4]=.004490166994743619283068;
			m_weights[5]=.005537379844305467094922;
			m_weights[6]=.006578717976538097355842;
			m_weights[7]=.007613081601079004241470;
			m_weights[8]=.008639375943791127812347;
			m_weights[9]=.009656513981519797512856;
			m_weights[10]=.010663418080519123092909;
			m_weights[11]=.011659021314495089149568;
			m_weights[12]=.012642268667039216159120;
			m_weights[13]=.013612118181804500923058;
			m_weights[14]=.014567542082226554527894;
			m_weights[15]=.015507527868679228554510;
			m_weights[16]=.016431079395787912577787;
			m_weights[17]=.017337217930540849428266;
			m_weights[18]=.018224983190943246588992;
			m_weights[19]=.019093434364555941906245;
			m_weights[20]=.019941651106075271609355;
			m_weights[21]=.020768734513027747161153;
			m_weights[22]=.021573808078620492685715;
			m_weights[23]=.022356018620781905687245;
			m_weights[24]=.023114537186434821642947;
			m_weights[25]=.023848559930060547675164;
			m_weights[26]=.024557308965633291303443;
			m_weights[27]=.025240033191029002580870;
			m_weights[28]=.025896009084039484519060;
			m_weights[29]=.026524541469151228747718;
			m_weights[30]=.02712496425427844790121;
			m_weights[31]=.027696641136670978255466;
			m_weights[32]=.028238966277249964563394;
			m_weights[33]=.028751364942657406263694;
			m_weights[34]=.02923329411433965865294;
			m_weights[35]=.029684243064019778295982;
			m_weights[36]=.030103733894949122541514;
			m_weights[37]=.030491322048364807972964;
			m_weights[38]=.030846596774616454962789;
			m_weights[39]=.031169181568463050327708;
			m_weights[40]=.031458734568078703741657;
			m_weights[41]=.031714948917344513050972;
			m_weights[42]=.031937553091042646259305;
			m_weights[43]=.032126311182608051104894;
			m_weights[44]=.032281023154132874195703;
			m_weights[45]=.032401525048358667838395;
			m_weights[46]=.032487689162431741082562;
			m_weights[47]=.032539424183237529025511;
			m_weights[48]=.032556675284170567850399;
			break;
		case 98:
			m_abscissas[1]=.999227851825523078008899;
			m_abscissas[2]=.997412306741663653040419;
			m_abscissas[3]=.99456103602252093942368;
			m_abscissas[4]=.990677220506857548728504;
			m_abscissas[5]=.985764931117584660081107;
			m_abscissas[6]=.979829278590865039523908;
			m_abscissas[7]=.972876429051795194306959;
			m_abscissas[8]=.964913602296992344422626;
			m_abscissas[9]=.955949065697237083519309;
			m_abscissas[10]=.945992126118696333101242;
			m_abscissas[11]=.935053120469195363858474;
			m_abscissas[12]=.923143405062335733348823;
			m_abscissas[13]=.910275343875370061286423;
			m_abscissas[14]=.896462295738773661813925;
			m_abscissas[15]=.881718600481995321476297;
			m_abscissas[16]=.866059564054996802862907;
			m_abscissas[17]=.849501442643624172605682;
			m_abscissas[18]=.832061425796660589853874;
			m_abscissas[19]=.81375761858282070969824;
			m_abscissas[20]=.794609022796633182587506;
			m_abscissas[21]=.774635517232971001609571;
			m_abscissas[22]=.753857837050858332602759;
			m_abscissas[23]=.732297552248070981770094;
			m_abscissas[24]=.709977045268936732893656;
			m_abscissas[25]=.686919487768620762462385;
			m_abscissas[26]=.663148816558043915493017;
			m_abscissas[27]=.638689708754423834147642;
			m_abscissas[28]=.61356755616324810858928;
			m_abscissas[29]=.587808438918282842217639;
			m_abscissas[30]=.561439098406987811655548;
			m_abscissas[31]=.534486909509449550033951;
			m_abscissas[32]=.506979852179655162541486;
			m_abscissas[33]=.478946482398611589821117;
			m_abscissas[34]=.450415902529466549902481;
			m_abscissas[35]=.421417731105407763536769;
			m_abscissas[36]=.391982072081705607745178;
			m_abscissas[37]=.362139483583820403462017;
			m_abscissas[38]=.331920946184018524821047;
			m_abscissas[39]=.301357830739430860716067;
			m_abscissas[40]=.270481865824942344281022;
			m_abscissas[41]=.239325104794721812153863;
			m_abscissas[42]=.207919892506586920342248;
			m_abscissas[43]=.176298831743748822765851;
			m_abscissas[44]=.144494749368795445668064;
			m_abscissas[45]=.112540662245050137675425;
			m_abscissas[46]=.080469742960683951104414;
			m_abscissas[47]=.048315285391164563244866;
			m_abscissas[48]=.016110670135792663313041;
			m_weights[0]=.000210393435724805386072;
			m_weights[1]=.001296668697725162623213;
			m_weights[2]=.002333886745365223141490;
			m_weights[3]=.003368137254799577326209;
			m_weights[4]=.004398816501337053643816;
			m_weights[5]=.005424910633713067226977;
			m_weights[6]=.006445366685723361655670;
			m_weights[7]=.007459128987416986620027;
			m_weights[8]=.008465146491308710029270;
			m_weights[9]=.009462375405901914139949;
			m_weights[10]=.010449780755786946373977;
			m_weights[11]=.011426337628496741867824;
			m_weights[12]=.012391032308784479464789;
			m_weights[13]=.013342863362451257102746;
			m_weights[14]=.014280842691149918796703;
			m_weights[15]=.015203996565960370670331;
			m_weights[16]=.016111366642469823674073;
			m_weights[17]=.017002010958049751838700;
			m_weights[18]=.017875004911142652242194;
			m_weights[19]=.018729442221975396344494;
			m_weights[20]=.019564435873933041211155;
			m_weights[21]=.020379119034743948429220;
			m_weights[22]=.021172645956593313200280;
			m_weights[23]=.021944192854274011863795;
			m_weights[24]=.022692958760489418566559;
			m_weights[25]=.023418166357436606600504;
			m_weights[26]=.024119062783817014240219;
			m_weights[27]=.024794920416443512386297;
			m_weights[28]=.025445037625636897711035;
			m_weights[29]=.026068739503630586172390;
			m_weights[30]=.026665378565229354968032;
			m_weights[31]=.027234335419996158889577;
			m_weights[32]=.027775019415270184753658;
			m_weights[33]=.028286869249349302963237;
			m_weights[34]=.028769353554200851640450;
			m_weights[35]=.029221971447096184964401;
			m_weights[36]=.029644253050596581558753;
			m_weights[37]=.030035759980350894987628;
			m_weights[38]=.030396085800198694136934;
			m_weights[39]=.030724856444106546104501;
			m_weights[40]=.031021730604499499060397;
			m_weights[41]=.031286400086584688919688;
			m_weights[42]=.031518590128299283482363;
			m_weights[43]=.031718059685550652947014;
			m_weights[44]=.031884601682452678389901;
			m_weights[45]=.032018043226298441821067;
			m_weights[46]=.032118245787046144553638;
			m_weights[47]=.032185105341131936472386;
			m_weights[48]=.032218552479460368836395;
			break;
		case 99:
			m_abscissas[1]=.999243448811422935102175;
			m_abscissas[2]=.997464561272040987918646;
			m_abscissas[3]=.994670816388098296096105;
			m_abscissas[4]=.99086527223029431289497;
			m_abscissas[5]=.986051837874146414782303;
			m_abscissas[6]=.980235420267952951361189;
			m_abscissas[7]=.973421939700840749058726;
			m_abscissas[8]=.96561832836999884350219;
			m_abscissas[9]=.956832524698980156095877;
			m_abscissas[10]=.947073465756686868249713;
			m_abscissas[11]=.936351078370739139823785;
			m_abscissas[12]=.924676269123547870754402;
			m_abscissas[13]=.912060913304824077263983;
			m_abscissas[14]=.898517842856990912872717;
			m_abscissas[15]=.884060833336716724144563;
			m_abscissas[16]=.868704589910952525014495;
			m_abscissas[17]=.852464732404262838953796;
			m_abscissas[18]=.835357779413990925522461;
			m_abscissas[19]=.817401131510144556965638;
			m_abscissas[20]=.798613053537505277939465;
			m_abscissas[21]=.779012656038206688075746;
			m_abscissas[22]=.758619875813825917484511;
			m_abscissas[23]=.737455455646851944080375;
			m_abscissas[24]=.715540923202216106744265;
			m_abscissas[25]=.6928985691303837254369;
			m_abscissas[26]=.66955142439430504217776;
			m_abscissas[27]=.645523236843304926595547;
			m_abscissas[28]=.620838447057751395772796;
			m_abscissas[29]=.595522163489081156226144;
			m_abscissas[30]=.569600136920474693244881;
			m_abscissas[31]=.543098734274162782667543;
			m_abscissas[32]=.51604491179200972726581;
			m_abscissas[33]=.488466187616655285388063;
			m_abscissas[34]=.460390613801106408235068;
			m_abscissas[35]=.431846747775250842922562;
			m_abscissas[36]=.402863623298316752146729;
			m_abscissas[37]=.373470720926825153267393;
			m_abscissas[38]=.343697938028074635547776;
			m_abscissas[39]=.313575558369659957100121;
			m_abscissas[40]=.283134221315957271460996;
			m_abscissas[41]=.25240489066290844109083;
			m_abscissas[42]=.221418823142804749122037;
			m_abscissas[43]=.190207536631105943016392;
			m_abscissas[44]=.158802778087633589995297;
			m_abscissas[45]=.127236491264747883681101;
			m_abscissas[46]=.095540784215354037878298;
			m_abscissas[47]=.063747896633787994538151;
			m_abscissas[48]=.031890167062801150697225;
			m_abscissas[49]=0.;
			m_weights[0]=.000206143063285920428778;
			m_weights[1]=.001270479977782297294289;
			m_weights[2]=.002286777395342568411674;
			m_weights[3]=.003300215141163138841779;
			m_weights[4]=.004310222852065934660779;
			m_weights[5]=.005315828181882797191959;
			m_weights[6]=.006316020274268851381576;
			m_weights[7]=.007309785395202577542041;
			m_weights[8]=.008296114068458742968312;
			m_weights[9]=.009274003613616795007807;
			m_weights[10]=.010242459632815281067300;
			m_weights[11]=.011200497191140964870512;
			m_weights[12]=.012147141887327426552489;
			m_weights[13]=.013081430875699181584178;
			m_weights[14]=.014002413860416873931438;
			m_weights[15]=.014909154069722674305596;
			m_weights[16]=.015800729212924282198061;
			m_weights[17]=.016676232420854347752424;
			m_weights[18]=.017534773169679916309994;
			m_weights[19]=.018375478187546865337847;
			m_weights[20]=.019197492343363656639919;
			m_weights[21]=.019999979516945885665131;
			m_weights[22]=.020782123449708474057792;
			m_weights[23]=.021543128575082678362760;
			m_weights[24]=.022282220827838999375873;
			m_weights[25]=.022998648431508780367622;
			m_weights[26]=.023691682663113720855243;
			m_weights[27]=.024360618594432036277336;
			m_weights[28]=.025004775809051620535450;
			m_weights[29]=.025623499094483767718054;
			m_weights[30]=.026216159108635448221410;
			m_weights[31]=.026782153019963602664794;
			m_weights[32]=.027320905120661275305810;
			m_weights[33]=.027831867412252559826872;
			m_weights[34]=.028314520163001202184522;
			m_weights[35]=.028768372436566240677427;
			m_weights[36]=.029192962591367214772217;
			m_weights[37]=.029587858750151199508043;
			m_weights[38]=.029952659239284182957279;
			m_weights[39]=.030286992997320063849070;
			m_weights[40]=.030590519952431769919567;
			m_weights[41]=.030862931368320650402510;
			m_weights[42]=.031103950158252344244680;
			m_weights[43]=.031313331166899735196276;
			m_weights[44]=.031490861419706341971126;
			m_weights[45]=.031636360339516522210708;
			m_weights[46]=.031749679930252158921085;
			m_weights[47]=.031830704927449013151674;
			m_weights[48]=.031879352915499632581704;
			m_weights[49]=.031895574411483567878429;
			break;
		case 100:
			m_abscissas[1]=.99925857796524492280619;
			m_abscissas[2]=.997515249037128610186208;
			m_abscissas[3]=.994777307190455282175161;
			m_abscissas[4]=.991047693712928295458446;
			m_abscissas[5]=.986330163816810990058285;
			m_abscissas[6]=.980629430722016494007241;
			m_abscissas[7]=.973951181007660389330198;
			m_abscissas[8]=.966302073440818998655945;
			m_abscissas[9]=.957689733680974119465208;
			m_abscissas[10]=.948122747162353244659818;
			m_abscissas[11]=.937610650735518652152924;
			m_abscissas[12]=.926163923252228889982543;
			m_abscissas[13]=.913793975165241761723576;
			m_abscissas[14]=.900513137178149364089542;
			m_abscissas[15]=.886334647967296543063887;
			m_abscissas[16]=.871272640993040267451333;
			m_abscissas[17]=.855342130415990328392342;
			m_abscissas[18]=.83855899613357534772703;
			m_abscissas[19]=.820939967952564038711456;
			m_abscissas[20]=.8025026089137250465782;
			m_abscissas[21]=.783265297785486709829295;
			m_abscissas[22]=.763247210744192374678835;
			m_abscissas[23]=.742468302259303058938357;
			m_abscissas[24]=.720949285202659112480663;
			m_abscissas[25]=.698711610201665954681614;
			m_abscissas[26]=.675777444257010061105699;
			m_abscissas[27]=.652169648646236438127448;
			m_abscissas[28]=.62791175613522542155997;
			m_abscissas[29]=.603027947520293029996592;
			m_abscissas[30]=.577543027524303982147443;
			m_abscissas[31]=.551482400070828795047148;
			m_abscissas[32]=.524872042960995227407734;
			m_abscissas[33]=.497738481978278941989572;
			m_abscissas[34]=.470108764447047934545557;
			m_abscissas[35]=.442010432271219376667596;
			m_abscissas[36]=.41347149447990545606048;
			m_abscissas[37]=.384520399307416021788461;
			m_abscissas[38]=.355186005835449841890691;
			m_abscissas[39]=.325497555225742578310693;
			m_abscissas[40]=.29548464157184773339934;
			m_abscissas[41]=.265177182399106408618744;
			m_abscissas[42]=.234605388842212355460444;
			m_abscissas[43]=.203799735530100137367179;
			m_abscissas[44]=.172790930208175936296304;
			m_abscissas[45]=.14160988312817233549013;
			m_abscissas[46]=.110287676236140028426393;
			m_abscissas[47]=.078855532189290610718068;
			m_abscissas[48]=.047344783232575205450427;
			m_abscissas[49]=.015786839966023482370336;
			m_weights[0]=.000202020202020202020202;
			m_weights[1]=.001245076659135294289299;
			m_weights[2]=.002241079755082963045034;
			m_weights[3]=.003234325923634960127431;
			m_weights[4]=.004224276331654507434773;
			m_weights[5]=.005209998070585079866662;
			m_weights[6]=.006190520319146799369175;
			m_weights[7]=.007164869223011786038752;
			m_weights[8]=.008132074847738999053753;
			m_weights[9]=.009091173626580326380516;
			m_weights[10]=.010041209778500831940692;
			m_weights[11]=.010981236426462550780560;
			m_weights[12]=.011910316608771469478412;
			m_weights[13]=.012827524243262733436393;
			m_weights[14]=.013731945065015462697624;
			m_weights[15]=.014622677545198368306161;
			m_weights[16]=.015498833793784865806705;
			m_weights[17]=.016359540446913776029939;
			m_weights[18]=.017203939538825604536246;
			m_weights[19]=.018031189357921348572402;
			m_weights[20]=.018840465286312474029789;
			m_weights[21]=.019630960622148179417898;
			m_weights[22]=.020401887383970712731127;
			m_weights[23]=.021152477096338553970581;
			m_weights[24]=.021881981555959572786226;
			m_weights[25]=.022589673577586142377554;
			m_weights[26]=.023274847718938643155256;
			m_weights[27]=.023936820983941188731127;
			m_weights[28]=.024574933503572831283953;
			m_weights[29]=.025188549193658416298753;
			m_weights[30]=.025777056388945333874325;
			m_weights[31]=.026339868452835451553509;
			m_weights[32]=.026876424362165378011579;
			m_weights[33]=.027386189266452805111813;
			m_weights[34]=.027868655021051939596589;
			m_weights[35]=.028323340693686911759267;
			m_weights[36]=.028749793043858492720559;
			m_weights[37]=.029147586974646425933211;
			m_weights[38]=.029516325956457147251446;
			m_weights[39]=.029855642422294598437121;
			m_weights[40]=.030165198134160199711896;
			m_weights[41]=.030444684520216807011992;
			m_weights[42]=.030693822982380608453520;
			m_weights[43]=.030912365174034381835092;
			m_weights[44]=.031100093247585310487159;
			m_weights[45]=.031256820071620608072336;
			m_weights[46]=.031382389417444503600751;
			m_weights[47]=.031476676114810655384057;
			m_weights[48]=.031539586176694766221355;
			m_weights[49]=.031571056892983030979428;
			break;
		default:
			m_n=0;
			m_abscissas.resize(0);
			m_weights.resize(0);
			throw ErrN(m_n);
			return;
		}

		if(m_n>1) m_abscissas[0]=1.;

		for(i=0,j=m_n-1;j>i;i++,j--)
		{
			m_abscissas[j]=-m_abscissas[i];
			m_weights[j]=m_weights[i];
		}
		break;

	// GaussChebyshev **************************************************
	case GaussChebyshev:
	{
		const long double pin=3.1415926535897932385L/m_n;
		for(i=0;i<m_n;i++)
		{
			m_abscissas[i]=cos((i+.5)*pin);
			m_weights[i]=sqrt(1.-m_abscissas[i]*m_abscissas[i]);
		}
		m_weights*=double(pin);
		break;
	}

	// Unspecified *********************************************************
	default:
		m_abscissas.resize(0);
		m_weights.resize(0);
		throw ErrRule(m_rule);
		m_rule=Unspecified;
		return;
	}

	if(m_lbound!=-1. || m_ubound!=1.) FitOrd2Bounds();
}




// #####################################################################
// Points2D ############################################################
// #####################################################################

/**
 *	\param	n1	Number of integration points in direction 1	(unsigned int)
 *	\param	r1	Integration rule in direction 1	(Numegration::Rule)
 *	\param	n2	Number of integration points in direction 2	(unsigned int)\n
 *		n2==0 or not specified:	n2 is set to n1
 *	\param	r2	Integration rule in direction 2	(Numegration::Rule)\n
 *		r2==Unspecified or not specified:	r2 is set to r1
 *	\param	lbound1	Lower bound of integration region in direction 1	(double)\n
 *		lbound1 not specified:	lbound1 is set to -1
 *	\param	ubound1	Upper bound of integration region in direction 1	(double)\n
 *		ubound1 not specified:	ubound1 is set to 1
 *	\param	lbound2	Lower bound of integration region in direction 2	(double)\n
 *		lbound2 not specified:	lbound2 is set to -1
 *	\param	ubound2	Upper bound of integration region in direction 2	(double)\n
 *		ubound2 not specified:	ubound2 is set to 1
 */
Numegration::Points2D::Points2D(unsigned int n1, Rule r1
				, unsigned int n2, Rule r2
				, double lbound1, double ubound1
				, double lbound2, double ubound2)
{
	SetParameters(n1,r1,n2,r2,lbound1,ubound1,lbound2,ubound2);
}

/**
 *	\param	n1	Number of integration points in direction 1	(unsigned int)
 *	\param	n2	Number of integration points in direction 2	(unsigned int)\n
 *		n2==0 or not specified:	n2 is set to n1
 */
void Numegration::Points2D::Setn(unsigned int n1,unsigned int n2)
{
	if(!n2) n2=n1;

	if(n1 && n2 && (n1!=m_n1 || n2!=m_n2))
	{
		m_n1=n1;
		m_n2=n2;
		Initialize();
	}
}

/**
 *	\param	r1	Integration rule in direction 1	(Numegration::Rule)
 *	\param	r2	Integration rule in direction 2	(Numegration::Rule)\n
 *			r2==Unspecified or not specified:	r2 is set to r1
 */
void Numegration::Points2D::SetRule(Rule r1,Rule r2)
{
	if(r2==Unspecified) r2=r1;

	if(r1!=m_rule1 || r2!=m_rule2)
	{
		m_rule1=r1;
		m_rule2=r2;
		Initialize();
	}
}

/**
 *	\param	lbound1	Lower bound of integration region in direction 1	(double)
 *	\param	ubound1	Upper bound of integration region in direction 1	(double)
 *	\param	lbound2	Lower bound of integration region in direction 2	(double)
 *	\param	ubound2	Upper bound of integration region in direction 2	(double)
 */
void Numegration::Points2D::SetBounds(double lbound1, double ubound1
				     , double lbound2, double ubound2)
{
	if(lbound1!=m_lbound1 || ubound1!=m_ubound1 || lbound2!=m_lbound2 || ubound2!=m_ubound2)
	{
		m_lbound1=lbound1;
		m_ubound1=ubound1;
		m_lbound2=lbound2;
		m_ubound2=ubound2;
		Initialize();
	}
}

/**
 *	\param	n1	Number of integration points in direction 1	(unsigned int)
 *	\param	r1	Integration rule in direction 1	(Numegration::Rule)
 *	\param	n2	Number of integration points in direction 2	(unsigned int)\n
 *		n2==0 or not specified:	n2 is set to n1
 *	\param	r2	Integration rule in direction 2	(Numegration::Rule)\n
 *		r2==Unspecified or not specified:	r2 is set to r1
 *	\param	lbound1	Lower bound of integration region in direction 1	(double)\n
 *		lbound1 not specified:	lbound1 is set to -1
 *	\param	ubound1	Upper bound of integration region in direction 1	(double)\n
 *		ubound1 not specified:	ubound1 is set to 1
 *	\param	lbound2	Lower bound of integration region in direction 2	(double)\n
 *		lbound2 not specified:	lbound2 is set to -1
 *	\param	ubound2	Upper bound of integration region in direction 2	(double)\n
 *		ubound2 not specified:	ubound2 is set to 1
 */
void Numegration::Points2D::SetParameters(unsigned int n1, Rule r1, unsigned int n2, Rule r2, double lbound1, double ubound1, double lbound2, double ubound2)
{
	m_n1=n1;
	m_n2=n2;
	if(!m_n2) m_n2=m_n1;

	m_rule1=r1;
	m_rule2=r2;
	if(m_rule2==Unspecified) m_rule2=m_rule1;

	m_lbound1=lbound1;
	m_ubound1=ubound1;
	m_lbound2=lbound2;
	m_ubound2=ubound2;

	Initialize();
}


/**
 *	\param	funcval	Ordinates of the function at integration points	(vector<valarray<double> >)
 */
double Numegration::Points2D::Integrate(const vector<valarray<double> >& funcval) const
{
	unsigned int size1=m_weights.size();
	double result=0.;

	if(size1 && funcval.size()==size1)
	{
		unsigned int size2=m_weights[0].size();
		if(size2 && funcval[0].size()==size2)
			for(unsigned int i=0;i<size1;i++)
				for(unsigned int j=0;j<size2;j++)
					result+=m_weights[i][j]*funcval[i][j];
	}
	return result;
}

/**
 *	\param	pf	pointer to a C-function	(double* (double,double))
 */
double Numegration::Points2D::Integrate(double (*pf)(double,double)) const
{
	double result=0.;
	double x1,x2;
	for(unsigned int i=0;i<m_n1;i++)
		for(unsigned int j=0;j<m_n2;j++)
		{
			x1=m_abscissas1[i];
			x2=m_abscissas2[j];
			result+=m_weights[i][j]*(*pf)(x1,x2);
		}
	return result;
}

/**
 *	\param	f	function object 	(Function2D)
 */
double Numegration::Points2D::Integrate(Function2D& f) const
{
	double result=0.;
	double x1,x2;
	for(unsigned int i=0;i<m_n1;i++)
		for(unsigned int j=0;j<m_n2;j++)
		{
			x1=m_abscissas1[i];
			x2=m_abscissas2[j];
			result+=m_weights[i][j]*f(x1,x2);
		}
	return result;
}


/**
 *	Integration funcval1*funcval2
 *	\param	funcval1	Ordinates of the function at integration points	(valarray<double>)
 *	\param	funcval2	Ordinates of the function at integration points	(valarray<double>)
 */
double Numegration::Points2D::IntegrateProduct(const valarray<double>& funcval1,const valarray<double>& funcval2) const
{
   Points1D ip1(m_n1,m_rule1,m_lbound1,m_ubound1);
   Points1D ip2(m_n2,m_rule2,m_lbound2,m_ubound2);
   return ip1.Integrate(funcval1)*ip2.Integrate(funcval2);
}

/**
 *	Integration pf1*pf2
 *	\param	pf1	pointer to a C-function	(double* (double))
 *	\param	pf2	pointer to a C-function	(double* (double))
 */
double Numegration::Points2D::IntegrateProduct(double (*pf1)(double),double (*pf2)(double)) const
{
   Points1D ip1(m_n1,m_rule1,m_lbound1,m_ubound1);
   Points1D ip2(m_n2,m_rule2,m_lbound2,m_ubound2);
   return ip1.Integrate(pf1)*ip2.Integrate(pf2);
}

/**
 *	Integration f1*f2
 *	\param	f1	function object 	(Function1D)
 *	\param	f2	function object 	(Function1D)
 */
double Numegration::Points2D::IntegrateProduct(Function1D& f1,Function1D& f2) const
{
   Points1D ip1(m_n1,m_rule1,m_lbound1,m_ubound1);
   Points1D ip2(m_n2,m_rule2,m_lbound2,m_ubound2);
   return ip1.Integrate(f1)*ip2.Integrate(f2);
}


// private -------------------------------------------------------------			
			
void Numegration::Points2D::Initialize()
{
	Points1D p1(m_n1,m_rule1,m_lbound1,m_ubound1);
	Points1D p2(m_n2,m_rule2,m_lbound2,m_ubound2);

	m_n1=p1.Getn();
	m_n2=p2.Getn();

	//m_abscissas1.resize(m_n1);
	m_abscissas1=p1.GetAbscissas();
	//m_abscissas2.resize(m_n2);
	m_abscissas2=p2.GetAbscissas();

	m_weights.resize(m_n1);

	for(unsigned int i=0;i<m_n1;i++)
	{
		m_weights[i].resize(m_n2);
		for(unsigned int j=0;j<m_n2;j++)
			m_weights[i][j]=(p1.GetWeights())[i]*(p2.GetWeights())[j];
	}
}
